/*
	Copyright (C) 2008-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"


#define FREE_PREFS_TEXT(A)	free ( A->format_datetime );\
				free ( A->num_thousands );\
				free ( A->text_prefix );\
				free ( A->text_suffix );


int ced_cell_destroy (
	CedCell		* cell
	)
{
	if ( ! cell )
	{
		return 1;
	}

	ced_cell_prefs_destroy ( cell->prefs );
	free ( cell->text );
	free ( cell );

	return 0;
}

int ced_cmp_cell (
	void		* k1,
	void		* k2
	)
{
	if ( k1 < k2 )
	{
		return -1;
	}

	if ( k1 > k2 )
	{
		return 1;
	}

	return 0;
}

void ced_del_cell (
	mtTreeNode	* node
	)
{
	ced_cell_destroy ( (CedCell *)node->data );
}


static CedCellPrefs default_cell_prefs = { 0, 16777215, 0 };


CedCell * ced_cell_new ( void )
{
	return calloc ( sizeof ( CedCell ), 1 );
}

CedCellPrefs * ced_cell_prefs_new ( void )
{
	CedCellPrefs	* prefs;


	prefs = calloc ( sizeof ( CedCellPrefs ), 1 );
	if ( ! prefs )
	{
		return NULL;
	}

	prefs->color_background = 16777215;

	return prefs;
}

int ced_cell_prefs_destroy (
	CedCellPrefs	* prefs
	)
{
	if ( ! prefs )
	{
		return 1;
	}

	FREE_PREFS_TEXT ( prefs )
	free ( prefs );

	return 0;
}


CedCellPrefs * ced_cell_prefs_default ( void )
{
	return &default_cell_prefs;
}

#define CELL_STRDUP( A ) \
	if ( A && A[0] ) \
	{ \
		A = strdup ( A ); \
		if ( ! A ) \
		{ \
			res = -1; \
		} \
	} \
	else \
	{ \
		A = NULL; \
	}


int ced_cell_set_prefs (
	CedCell		* cell,
	CedCellPrefs	* prefs
	)
{
	int	res = 0;


	// Empty old prefs strings
	if ( cell->prefs )
	{
		FREE_PREFS_TEXT ( cell->prefs )
		// Dangling pointers are safe as the container either get
		// released or set next
	}

	if ( ! prefs )
	{
		ced_cell_prefs_destroy ( cell->prefs );
		cell->prefs = NULL;

		return 0;
	}

	if ( ! cell->prefs )
	{
		cell->prefs = ced_cell_prefs_new ();
		if ( ! cell->prefs )
		{
			return 1;
		}
	}

	memcpy ( cell->prefs, prefs, sizeof ( CedCellPrefs ) );

	// Duplicate new prefs strings or NULL out dangling refs
	CELL_STRDUP ( cell->prefs->format_datetime )
	CELL_STRDUP ( cell->prefs->num_thousands )
	CELL_STRDUP ( cell->prefs->text_prefix )
	CELL_STRDUP ( cell->prefs->text_suffix )

	return res;
}





int ced_strtocellref (
	char		* input,
	CedCellRef	* result,
	char		** next,
	int		strict
	)
{
	int		ref,
			i,
			res[2][2];
	char		* s = input,
			c[] = "rRcC";


	if ( ! input || ! result )
	{
		return 1;
	}

	// Skip all leading spaces
	while ( isspace ( s[0] ) )
	{
		s++;
	}

	for ( i = 0; i < 2; i++ )
	{
		/* Must have input string to parse and begin with rR or cC */
		if (	! s ||
			( s[0] != c[2 * i] && s[0] != c[2 * i + 1] )
			)
		{
			return 1;
		}

		s++;

		if ( s[0] == '_' )
		{
			/* Check for maximum row/col shorthand - absolute ref */
			res[i][0] = 0;
			res[i][1] = (i == 0) ? CED_MAX_ROW : CED_MAX_COLUMN;
			s++;
		}
		else if ( s[0] == '[' )
		{
			/* Check for relative open square bracket */
			s++;

			/* Get relative reference */
			if ( mtkit_strtoi ( s, &ref, &s, 0 ) )
			{
				return 1;
			}

			/* Get closing square bracket */
			if ( s[0] != ']' )
			{
				return 1;
			}

			s++;

			res[i][0] = 1;
			res[i][1] = ref;
		}
		else if ( s[0] >= '0' &&
			s[0] <= '9' &&
			! mtkit_strtoi ( s, &ref, &s, 0 ) )
		{
			/* Absolute cell reference */
			res[i][0] = 0;
			res[i][1] = ref;

			if (	( i == 0 && ref > CED_MAX_ROW ) ||
				( i == 1 && ref > CED_MAX_COLUMN )
				)
			{
			// Row/Column is out of range for an absolute reference
				return 1;
			}
		}
		else
		{
			/* No reference found so must be this row/column */
			res[i][0] = 1;
			res[i][1] = 0;
		}
	}

	if ( strict && mtkit_strnonspaces ( s ) )
	{
		return 1;
	}

	result->row_m = res[0][0];
	result->row_d = res[0][1];
	result->col_m = res[1][0];
	result->col_d = res[1][1];

	if ( next )
	{
		next[0] = s;
	}

/*
printf ( "cell ref parsed = %i,%i %i,%i\n", result->row_m, result->col_m,
	result->row_d, result->col_d );
*/

	return 0;
}

int ced_strtocellrange (
	char		* input,
	CedCellRef	* r1,
	CedCellRef	* r2,
	char		** next,
	int		strict
	)
{
	char		* col = NULL;


	if ( ced_strtocellref ( input, r1, &col, 0 ) || ! col )
	{
		return 1;
	}

	for ( ;; col++ )
	{
		if ( isspace ( col[0] ) )
		{
			continue;
		}

		if ( col[0] == ':' )
		{
			break;
		}

		return 1;	// Invalid char, including NUL
	}

	if ( ced_strtocellref ( col + 1, r2, NULL, strict ) )
	{
		return 1;
	}

	return 0;
}

int ced_cellreftostr (
	char		* output,
	CedCellRef	* cellref
	)
{
	char		rt[32] = {0},
			ct[32] = {0};


	if ( ! output || ! cellref )
	{
		return 1;
	}

	if ( cellref->row_m )
	{
		if ( cellref->row_d )
		{
			sprintf ( rt, "[%i]", cellref->row_d );	/* Relative */
		}
	}
	else
	{
		sprintf ( rt, "%i", cellref->row_d );		/* Absolute */
	}

	if ( cellref->col_m )
	{
		if ( cellref->col_d )
		{
			sprintf ( ct, "[%i]", cellref->col_d );	/* Relative */
		}
	}
	else
	{
		sprintf ( ct, "%i", cellref->col_d );		/* Absolute */
	}

	snprintf ( output, 128, "R%sC%s", rt, ct );

	return 0;
}

int ced_strtosheetref (
	char		* input,
	CedSheet	** result,
	char		** next,
	CedBook		* book
	)
{
	char		* ds,
			* sd,
			* ss;
	mtTreeNode	* node;


	if ( ! input || input[0] != '\'' || ! result || ! book )
	{
		return 1;
	}

	ds = strdup ( input + 1 );
	if ( ! ds )
	{
		return 1;
	}

	// Filter out occurences of '' and replace with ' + remove trailing '
	for (	sd = ds, ss = input + 1;
		ss[0] != 0;
		ss++, sd++ )
	{
		if ( ss[0] == '\'' )
		{
			if ( ss[1] == '\'' )
			{
				// ''
				ss++;
			}
			else
			{
				// Final trailing ' at end
				// Ensure next[0] points to next unparsed char
				ss++;
				break;
			}
		}

		sd[0] = ss[0];
	}

	sd[0] = 0;

	if ( next )
	{
		next[0] = ss;
	}

//printf ( "in = %s out = %s next = %s\n", input, ds, next[0] );

	node = mtkit_tree_node_find ( book->sheets, ds );
	free ( ds );

	if ( ! node )
	{
		return 1;
	}

	result[0] = (CedSheet *)node->data;

	return 0;
}

char * ced_sheetreftostr (
	char		* input
	)
{
	char		* s,
			* t,
			* output;
	int		newlen;


	if ( ! input )
	{
		return NULL;
	}

	// Old string + outer 'x2 + terminator
	newlen = strlen ( input ) + 2 + 1;

	// Add in occurences of ' character
	for (	s = strchr ( input, '\'' );
		s;
		s = strchr ( s + 1, '\'' )
		)
	{
		newlen++;
	}

	output = calloc ( 1, newlen );
	if ( ! output )
	{
		return NULL;
	}

	output[0] = '\'' ;
	for (	s = input, t = output + 1;
		s[0];
		s++, t++
		)
	{
		t[0] = s[0];
		if ( t[0] == '\'' )
		{
			t++;
			t[0] = '\'' ;
		}
	}
	t[0] = '\'' ;

	return output;
}



int ced_cell_paste (
	CedCell		* dest,
	CedCell		* src
	)
{
	if (	! dest ||
		! src ||
		mtkit_strfreeswap ( &dest->text, src->text ) ||
		ced_cell_set_prefs ( dest, src->prefs )
		)
	{
		return 1;	// Fail
	}

	dest->type = src->type;
	dest->value = src->value;

	return 0;		// Success
}

CedCell * ced_cell_duplicate (
	CedCell		* cell
	)
{
	CedCell		* newcell;


	if ( ! cell )
	{
		return NULL;
	}

	newcell = ced_cell_new ();
	if ( ! newcell )
	{
		return NULL;
	}

	if ( ced_cell_paste ( newcell, cell ) )
	{
		ced_cell_destroy ( newcell );
		return NULL;
	}

	return newcell;
}


static int col_dup_func (
	mtTreeNode	* old,
	mtTreeNode	* new
	)
{
	new->key = old->key;
	new->data = ced_cell_duplicate ( old->data );

	if ( ! new->data )
	{
		return 1;	// Fail
	}

	return 0;		// Success
}

static int row_dup_func (
	mtTreeNode	* old,
	mtTreeNode	* new
	)
{
	new->key = old->key;
	new->data = mtkit_tree_duplicate ( old->data, col_dup_func );

	if ( ! new->data )
	{
		return 1;	// Fail
	}

	return 0;		// Success
}

CedSheet * ced_sheet_duplicate (
	CedSheet	* sheet
	)
{
	CedSheet	* newsheet;


	if ( ! sheet )
	{
		return NULL;
	}

	newsheet = ced_sheet_new ();
	if ( ! newsheet )
	{
		return NULL;
	}

	newsheet->prefs = sheet->prefs;

	if ( sheet->rows )
	{
		// Remove empty tree
		mtkit_tree_destroy ( newsheet->rows );
		newsheet->rows = mtkit_tree_duplicate ( sheet->rows,
			row_dup_func );

		if ( ! newsheet->rows )
		{
			ced_sheet_destroy ( newsheet );

			return NULL;
		}
	}

	return newsheet;
}




typedef struct
{
	unsigned int	row,		// Current row
			min_row,
			min_col,
			max_row,
			max_col
			;
	CedFuncScanArea	callback;
	void		* user_data;
	CedSheet	* sheet;
} scanSTATE;



static int recurse_scan_col_back (
	scanSTATE	* state,
	mtTreeNode	* col_node
	)
{
	unsigned int	col = (uintptr_t)col_node->key;


	if (	col_node->right &&
		col < state->max_col &&
		recurse_scan_col_back ( state, col_node->right ) )
	{
		return 2;
	}

	if ( col >= state->min_col && col <= state->max_col )
	{
		if ( state->callback ( state->sheet, (CedCell *)col_node->data,
			state->row, col, state->user_data ) )
		{
			return 2;
		}
	}

	if (	col_node->left &&
		col > state->min_col &&
		recurse_scan_col_back ( state, col_node->left ) )
	{
		return 2;
	}

	return 0;
}

static int recurse_scan_row_back (
	scanSTATE	* state,
	mtTreeNode	* row_node
	)
{
	unsigned int	row = (uintptr_t)row_node->key;


	if (	row_node->right &&
		row < state->max_row &&
		recurse_scan_row_back ( state, row_node->right )
		)
	{
		return 2;
	}

	if ( row >= state->min_row && row <= state->max_row )
	{
		mtTreeNode	* cnode;


		cnode = ( (mtTree *) row_node->data )->root;
		state->row = row;

		if ( cnode && recurse_scan_col_back ( state, cnode ) )
		{
			return 2;
		}
	}

	if (	row_node->left &&
		row > state->min_row &&
		recurse_scan_row_back ( state, row_node->left )
		)
	{
		return 2;
	}

	return 0;
}

int ced_sheet_scan_area_backwards (
	CedSheet	* sheet,
	unsigned int	row,
	unsigned int	col,
	unsigned int	rowtot,
	unsigned int	coltot,
	CedFuncScanArea	callback,
	void		* user_data
	)
{
	scanSTATE	state = { 0, 0, 0, row, col, callback, user_data,
				sheet };


	if (	! sheet ||
		! callback ||
		row > CED_MAX_ROW ||
		col > CED_MAX_COLUMN
		)
	{
		return 1;
	}

	if ( ! sheet->rows || ! sheet->rows->root )
	{
		return 0;
	}

	if ( ! row || ! rowtot || rowtot > row )
	{
		state.min_row = 0;
	}
	else
	{
		state.min_row = row - rowtot + 1;
	}

	if ( ! col || ! coltot || coltot > col )
	{
		state.min_col = 0;
	}
	else
	{
		state.min_col = col - coltot + 1;
	}

	return recurse_scan_row_back ( &state, sheet->rows->root );
}


static int recurse_scan_col (
	scanSTATE	* state,
	mtTreeNode	* col_node )
{
	unsigned int	col = (uintptr_t)col_node->key;


	if (	col_node->left &&
		col > state->min_col &&
		recurse_scan_col ( state, col_node->left )
		)
	{
		return 2;
	}

	if ( col >= state->min_col && col <= state->max_col )
	{
		if ( state->callback ( state->sheet, (CedCell *)col_node->data,
			state->row, col, state->user_data ) )
		{
			return 2;
		}
	}

	if ( col_node->right && col < state->max_col &&
		recurse_scan_col ( state, col_node->right ) )
	{
		return 2;
	}

	return 0;
}

static int recurse_scan_row (
	scanSTATE	* state,
	mtTreeNode	* row_node )
{
	unsigned int	row = (uintptr_t)row_node->key;


	if (	row_node->left &&
		row > state->min_row &&
		recurse_scan_row ( state, row_node->left )
		)
	{
		return 2;
	}

	if ( row >= state->min_row && row <= state->max_row )
	{
		mtTreeNode	* cnode;


		cnode = ( (mtTree *) row_node->data )->root;
		state->row = row;

		if ( cnode && recurse_scan_col ( state, cnode ) )
		{
			return 2;
		}
	}

	if (	row_node->right &&
		row < state->max_row &&
		recurse_scan_row ( state, row_node->right )
		)
	{
		return 2;
	}

	return 0;
}

int ced_sheet_scan_area (
	CedSheet	* sheet,
	unsigned int	row,
	unsigned int	col,
	unsigned int	rowtot,
	unsigned int	coltot,
	CedFuncScanArea	callback,
	void		* user_data
	)
{
	scanSTATE	state = { 0, row, col, 0, 0, callback, user_data,
				sheet };


	if (	! sheet ||
		! callback ||
		row > CED_MAX_ROW ||
		col > CED_MAX_COLUMN
		)
	{
		return 1;
	}

	if ( ! sheet->rows || ! sheet->rows->root )
	{
		return 0;
	}

	if ( ! rowtot )
	{
		state.max_row = CED_MAX_ROW;
	}
	else
	{
		state.max_row = row + rowtot - 1;
	}

	if ( ! coltot )
	{
		state.max_col = CED_MAX_COLUMN;
	}
	else
	{
		state.max_col = col + coltot - 1;
	}

	return recurse_scan_row ( &state, sheet->rows->root );
}
