/*
	Copyright (C) 2008-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



enum
{
	CELL_TYPE_NONE,

	CELL_TYPE_TEXT,
	CELL_TYPE_FIXED_DECIMAL,
	CELL_TYPE_HEX,
	CELL_TYPE_BINARY,
	CELL_TYPE_SCIENTIFIC,
	CELL_TYPE_PERCENTAGE,
	CELL_TYPE_DATETIME,

	CELL_TYPE_GENERAL_NUMBER,
	CELL_TYPE_ERROR,

	CELL_TYPE_TOTAL
};



#define GUI_PRINTF_NUM		"%.8g"
#define CELL_TXT_SIZE		2048



static int get_cell_type (
	CedCell		* cell
	)
{
	// Errors and text cannot be manually overidden
	if ( cell->type == CED_CELL_TYPE_ERROR )
	{
		return CELL_TYPE_ERROR;
	}

	if (	cell->type == CED_CELL_TYPE_TEXT ||
		cell->type == CED_CELL_TYPE_TEXT_EXPLICIT
		)
	{
		return CELL_TYPE_TEXT;
	}

	if ( cell->prefs && cell->prefs->format )
	{
		return cell->prefs->format;
	}

	// General cell type (based on internal cell types)
	switch ( cell->type )
	{
	case CED_CELL_TYPE_DATE:
		return CELL_TYPE_DATETIME;

	default:
		break;
	}

/*
	CED_CELL_TYPE_VALUE:
	CED_CELL_TYPE_FORMULA:
	CED_CELL_TYPE_FORMULA_EVAL:
	unknown
*/
	return CELL_TYPE_GENERAL_NUMBER;
}

char * ced_cell_create_output (
	CedCell		* cell,
	int		* hjustify
	)
{
	char		txt[CELL_TXT_SIZE],
			* txt_list[10] = { NULL },
			* newtxt = NULL,
			* fmt_ddt = NULL,
			* thou = NULL;
	int		ctype,
			justify = 0,
			txt_items = 0,
			dp = 0,
			nz = 0;


	if ( ! cell || ! cell->text )
	{
		return NULL;
	}

	ctype = get_cell_type ( cell );

	if ( ctype == CELL_TYPE_ERROR )
	{
		int e = cell->value;


		justify = CED_CELL_JUSTIFY_LEFT;
		sprintf ( txt, "Err:%i,%i", e%1000, e/1000 );
		txt_list[0] = txt;

		goto finish;
	}

	// PREFIX
	if (	cell->prefs &&
		cell->prefs->text_prefix &&
		cell->prefs->text_prefix[0]
		)
	{
		txt_list[ txt_items ] = cell->prefs->text_prefix;
		txt_items ++;
	}

	justify = CED_CELL_JUSTIFY_RIGHT;		// Default

	if ( cell->prefs )
	{
		dp = cell->prefs->num_decimal_places;
		nz = cell->prefs->num_zeros;
		fmt_ddt = cell->prefs->format_datetime;
		thou = cell->prefs->num_thousands;
	}

	switch ( ctype )
	{
	case CELL_TYPE_TEXT:
		justify = CED_CELL_JUSTIFY_LEFT;
		txt_list[ txt_items ] = cell->text;
		txt_items ++;
		break;

	case CELL_TYPE_FIXED_DECIMAL:
	case CELL_TYPE_PERCENTAGE:

		if ( fabs ( cell->value ) > 1000000000000000.0 )
		{
			// Enforce sanity for very large numbers
			sprintf ( txt, GUI_PRINTF_NUM,
				ctype == CELL_TYPE_PERCENTAGE ?
				cell->value * 100 : cell->value );
		}
		else
		{
			sprintf ( txt + 128, "%%.%if", dp );
			sprintf ( txt, txt + 128, ctype == CELL_TYPE_PERCENTAGE
				? cell->value * 100 : cell->value );
		}

		if ( thou )
		{
			mtkit_strtothou ( txt, txt, 64, thou[0], '-', '.', 3, 0
				);
		}

		txt_list[ txt_items ] = txt;
		txt_items ++;

		if ( ctype == CELL_TYPE_PERCENTAGE )
		{
			txt_list[ txt_items ] = "%";
			txt_items ++;
		}

		break;

	case CELL_TYPE_HEX:
		sprintf ( txt + 128, "%%.%iX", nz );
		sprintf ( txt, txt + 128, (unsigned int)cell->value );
		txt_list[ txt_items ] = txt;
		txt_items ++;
		break;

	case CELL_TYPE_BINARY:
		{
			int		i;
			unsigned int	b, v = (unsigned int)cell->value;
			char		* cp = txt;


			if ( v == 0 )
			{
				cp[0] = '0';
				cp++;
			}
			else for ( i = 31; i >= 0; i-- )
			{
				b = 1 << i;

				if ( b & v )
				{
					cp[0] = '1';
					cp++;
				}
				else
				{
					if ( cp != txt )
					{
						cp[0] = '0';
						cp++;
					}
				}
			}

			cp[0] = '\0';
			cp++;

			i = strlen ( txt );
			if ( nz > i && i <= 32 && i > 0 )
			{
				b = nz - i;	// Leading zeros required
				txt[nz] = 0;

//printf ( "txt = '%s' nz = %i b = %i\n", txt, nz, b );

				// Shift old string to the right
				for ( i = nz - 1; i >= b; i-- )
				{
//printf ( "i = %i\n", i );
					txt[i] = txt[i - b];
				}

				// Put in leading zeros
				for ( ; i >= 0; i-- )
				{
					txt[i] = '0';
				}
			}
		}

		txt_list[ txt_items ] = txt;
		txt_items ++;
		break;

	case CELL_TYPE_SCIENTIFIC:
		sprintf ( txt + 128, "%%.%ie", dp );
		sprintf ( txt, txt + 128, cell->value );

		txt_list[ txt_items ] = txt;
		txt_items ++;
		break;

	case CELL_TYPE_DATETIME:
		justify = CED_CELL_JUSTIFY_LEFT;

		{
			int		val,
					chsq,
					day = 0,
					month = 0,
					year = 0,
					hour = 0,
					min = 0,
					sec = 0,
					backslash = 0;
			char		* dest,
					* src;
			struct tm	tm = { 0 };


			mtkit_ddttoi ( cell->value, &day, &month, &year, &hour,
				&min, &sec );

/*
printf ("v = %f %i %i %i %i %i %i\n", cell->value, day, month, year, hour, min
, sec );
*/

			if ( fmt_ddt && fmt_ddt[0] )
			{
				dest = txt;
				src = fmt_ddt;

				dest[0] = 0;

				for ( ; src[0] && (dest - txt) <
					(CELL_TXT_SIZE - 35); src++ )
				{
					val = -1;

					for ( chsq = 0; src[0] == src[chsq + 1];
						chsq++ )
					{
					}

		// Skip the leading backslash, record its existence
					if ( src[0] == '\\' && ! backslash )
					{
						backslash = 1;
						continue;
					}

					if ( ! backslash ) switch ( src[0] )
					{
					case 'd':
						if ( chsq >= 2 )
						{
						tm.tm_wday = mtkit_ddt_weekday (
							cell->value );

						if ( chsq == 2 )
						{
							strftime ( dest, 32,
								"%a", &tm );
						}
						else
						{
							strftime ( dest, 32,
								"%A", &tm );
						}

						dest[32] = 0;

						if ( chsq > 3 )
						{
							chsq = 3;
						}

						}
						val = day;
						break;

					case 'm':
						if ( chsq >= 2 )
						{
						tm.tm_mon = month - 1;
						if ( chsq == 2 )
						{
							strftime ( dest, 32,
								"%b", &tm );
						}
						else
						{
							strftime ( dest, 32,
								"%B", &tm );
						}

						dest[32] = 0;

						if ( chsq > 3 )
						{
							chsq = 3;
						}

						}
						val = month;
						break;

					case 'y':
						val = year;
						if ( chsq > 1 )
						{
							chsq = 1;
						}
						break;

					case 'H':
						val = hour;
						if ( chsq > 1 )
						{
							chsq = 1;
						}
						break;

					case 'M':
						val = min;
						if ( chsq > 1 )
						{
							chsq = 1;
						}
						break;

					case 'S':
						val = sec;
						if ( chsq > 1 )
						{
							chsq = 1;
						}
						break;
					}

					backslash = 0;

					if ( val < 0 )
					{
						dest[0] = src[0];
						dest++;

						dest[0] = 0;
					}
					else	// Numerical substitution
					{
						if ( chsq == 1 )
						{
							sprintf ( dest, "%02i",
								val % 100 );
						}
						else if ( chsq == 0 )
						{
							sprintf ( dest, "%i",
								val );
						}

						src += chsq;
					}

					dest += strlen ( dest );
				}
			}
			else
			{
				sprintf ( txt, "%i-%i-%i", year, month, day );
			}
		}

		txt_list[ txt_items ] = txt;
		txt_items ++;
		break;

	case CELL_TYPE_GENERAL_NUMBER:
	default:
		sprintf ( txt, GUI_PRINTF_NUM, cell->value );
		txt_list[ txt_items ] = txt;
		txt_items ++;

		if ( thou )
		{
			mtkit_strtothou ( txt, txt, 64, thou[0], '-', '.', 3, 0
				);
		}
		break;
	}

	// SUFFIX
	if (	cell->prefs &&
		cell->prefs->text_suffix &&
		cell->prefs->text_suffix[0]
		)
	{
		txt_list[ txt_items ] = cell->prefs->text_suffix;
		txt_items ++;
	}

finish:

	if ( ! txt_list[0] )
	{
		return NULL;		// Nothing to do
	}

	newtxt = mtkit_strjoinv ( NULL, txt_list );

	if ( hjustify )
	{
		if ( cell->prefs && cell->prefs->align_horizontal )
		{
			hjustify[0] = cell->prefs->align_horizontal;
		}
		else
		{
			hjustify[0] = justify;
		}
	}

	return newtxt;
}
