/*
	Copyright (C) 2012-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



static char * get_temp_clip_filename ( void )
{
	static char	temp_clip_filename[MTKIT_PATHBUF_MAX];


	if ( ! temp_clip_filename[0] )
	{
		// First time usage

		snprintf ( temp_clip_filename, MTKIT_PATHBUF_MAX, "%s%c%s",
			mtkit_file_home (), MTKIT_DIR_SEP,
			".mtcelledit_clip0" );
		temp_clip_filename[MTKIT_PATHBUF_MAX - 1] = 0;
	}

	return temp_clip_filename;
}


int cuf_clip_set_timestamp (
	CedUI_clip	* clipboard,
	char		* txt
	)
{
	if ( ! clipboard )
	{
		return 1;
	}


	memset ( clipboard->timestamp, 0, CEDUI_CLIPBOARD_TIMESTAMP_SIZE );

	if ( txt )
	{
		mtkit_strnncat ( clipboard->timestamp, txt,
			CEDUI_CLIPBOARD_TIMESTAMP_SIZE );
	}
	else
	{
		snprintf ( clipboard->timestamp, CEDUI_CLIPBOARD_TIMESTAMP_SIZE,
			"%u", (unsigned int)time ( NULL ) );
	}

	clipboard->timestamp[CEDUI_CLIPBOARD_TIMESTAMP_SIZE - 1] = 0;

	return 0;
}

CedUI_clip * cuf_clip_new ( void )
{
	CedUI_clip	* clipboard;


	clipboard = calloc ( sizeof ( CedUI_clip ), 1 );
	if ( ! clipboard )
	{
		return NULL;
	}

	return clipboard;
}

int cuf_clip_free (
	CedUI_clip	* clipboard
	)
{
	cuf_clip_flush ( clipboard );
	free ( clipboard );

	return 0;
}

int cuf_clip_flush (
	CedUI_clip	* clipboard
	)
{
	if ( ! clipboard )
	{
		return 1;
	}

	if ( clipboard->temp )
	{
		// Remove the temp file

		remove ( get_temp_clip_filename () );
		clipboard->temp = 0;
	}

	if ( clipboard->sheet )
	{
		ced_sheet_destroy ( clipboard->sheet );
		clipboard->sheet = NULL;
	}

	free ( clipboard->tsv );
	clipboard->tsv = NULL;

	free ( clipboard->ced );
	clipboard->ced = NULL;

	return 0;
}

int cuf_clip_import_text (
	CedUI_clip	* clipboard,
	char		* text
	)
{
	unsigned int	rtot,
			ctot,
			txt_len;
	CedSheet	* newsheet;


	if ( ! clipboard || ! text )
	{
		return 1;
	}

	txt_len = strlen ( text );
	newsheet = ced_sheet_load_mem ( text, txt_len + 1, "ISO-8859-1", NULL );
	if ( newsheet )
	{
		rtot = ctot = 0;
		if ( ced_sheet_tsvmem_geometry ( text, txt_len + 1, &rtot,
			&ctot ) )
		{
			ced_sheet_get_geometry ( newsheet, &rtot, &ctot );
		}

		if ( rtot < 1 )
		{
			rtot = 1;
		}

		if ( ctot < 1 )
		{
			ctot = 1;
		}

		if ( cuf_clip_flush ( clipboard ) || clipboard->sheet )
		{
			// Should never happen

			return 1;
		}

		clipboard->sheet = newsheet;
		clipboard->rows = rtot;
		clipboard->cols = ctot;
	}

	return 0;
}

int cuf_clip_export_text (
	CedUI_clip	* clipboard
	)
{
	if ( ! clipboard || ! clipboard->sheet )
	{
		return 1;
	}

	if ( ! clipboard->tsv )
	{
		mtFile		* mtfp;
		void		* buf;


		mtfp = ced_sheet_save_mem ( clipboard->sheet,
			CED_FILE_TYPE_TSV_CONTENT );
		if ( ! mtfp )
		{
			return 1;
		}

		if (	mtkit_file_write ( mtfp, "", 1 )	||
			mtkit_file_get_mem ( mtfp, &buf, NULL )
			)
		{
			mtkit_file_close ( mtfp );

			return 1;
		}

		clipboard->tsv = strdup ( buf );
		mtkit_file_close ( mtfp );
	}

	return 0;
}

int cuf_clip_save_temp_file (
	CedUI_clip	* clipboard,
	char		* filename
	)
{
	CedBook		* book;


	if (	! clipboard ||
		! clipboard->sheet ||
		! filename
		)
	{
		return 1;
	}

	// Save the current clipboard as a CED ZIP file with the cursor position
	// denoting the geometry of the paste.
	book = ced_book_new ();
	if ( ! book )
	{
		return 1;
	}

	if ( ced_book_add_sheet ( book, clipboard->sheet,
		CEDUI_CLIPBOARD_SHEET_NAME ) )
	{
		ced_book_destroy ( book );

		return 1;
	}

	clipboard->sheet->prefs.cursor_r1 = clipboard->rows;
	clipboard->sheet->prefs.cursor_c1 = clipboard->cols;

	if ( ced_book_save ( book, filename, CED_FILE_TYPE_LEDGER_BOOK ) )
	{
		ced_book_detach_sheet ( clipboard->sheet );
		ced_book_destroy ( book );

		return 1;
	}

	ced_book_detach_sheet ( clipboard->sheet );
	ced_book_destroy ( book );

	return 0;
}

int cuf_clip_save_temp (
	CedUI_clip	* clipboard
	)
{
	if ( ! clipboard )
	{
		return 1;
	}

	// Only save to temp if we haven't done so already
	if ( clipboard->temp == 0 )
	{
		if ( cuf_clip_save_temp_file ( clipboard,
			get_temp_clip_filename () ) )
		{
			return 1;
		}

		clipboard->temp = 1;
		cuf_clip_set_timestamp ( clipboard, NULL );
	}

	return 0;
}

int cuf_clip_load_temp_file (
	CedUI_clip	* clipboard,
	char		* filename
	)
{
	CedBook		* book;
	CedSheet	* sheet;


	if ( ! clipboard || ! filename )
	{
		return 1;
	}

	book = ced_book_load ( filename, NULL, "ISO-8859-1" );
	if ( ! book )
	{
		return 1;		// No file found - error!
	}

	// Extract "clipboard" sheet and keep as new clipboard
	sheet = ced_book_get_sheet ( book, CEDUI_CLIPBOARD_SHEET_NAME );
	if ( ! sheet || ced_book_detach_sheet ( sheet ) )
	{
		ced_book_destroy ( book );

		return 1;		// No clipboard sheet found/extracted
	}

	ced_book_destroy ( book );

	ced_sheet_destroy ( clipboard->sheet );
	clipboard->sheet = sheet;
	clipboard->rows = sheet->prefs.cursor_r1;
	clipboard->cols = sheet->prefs.cursor_c1;

	return 0;
}

int cuf_clip_load_temp (
	CedUI_clip	* clipboard
	)
{
	return cuf_clip_load_temp_file ( clipboard, get_temp_clip_filename () );
}

int cuf_clip_paste (
	CedUI_file	* file,
	CedUI_clip	* clipboard,
	int		paste_mode
	)
{
	int		r1,
			c1,
			r2,
			c2,
			rtot,
			ctot;
	CedSheet	* sheet;


	sheet = cuf_file_get_sheet ( file );

	if (	! sheet ||
		! clipboard ||
		! clipboard->sheet
		)
	{
		return 1;
	}

	ced_sheet_cursor_max_min ( sheet, &r1, &c1, &r2, &c2 );

	rtot = r2 - r1 + 1;
	ctot = c2 - c1 + 1;

/*
If selected area rows/cols are less than the rows/cols of the paste then
the selection is exapnded to use the full clipboard width/height.
*/
	if ( rtot < clipboard->rows )
	{
		rtot = clipboard->rows;
	}

	if ( ctot < clipboard->cols )
	{
		ctot = clipboard->cols;
	}

	return cuf_sheet_paste_area ( file->cubook, sheet, clipboard->sheet,
		r1, c1, rtot, ctot, clipboard->rows, clipboard->cols,
		paste_mode );
}

int cuf_clip_copy (
	CedUI_file	* file,
	CedUI_clip	* clipboard
	)
{
	CedSheet	* sheet,
			* new_clipsheet;
	int		rtot,
			ctot;


	if ( ! file || ! clipboard )
	{
		return 1;
	}

	sheet = cuf_file_get_sheet ( file );
	if ( ! sheet )
	{
		return 1;
	}

	new_clipsheet = ced_sheet_copy_selection ( sheet, &rtot, &ctot );
	if ( ! new_clipsheet )
	{
		return 1;
	}

	if ( cuf_clip_flush ( clipboard ) )
	{
		ced_sheet_destroy ( new_clipsheet );

		return 1;
	}

	clipboard->sheet = new_clipsheet;
	clipboard->rows = rtot;
	clipboard->cols = ctot;

	return 0;		// Success
}
