/*
	Copyright (C) 2010-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



#define FIRST_DUPLICATE_ID	2



typedef int (* NameCB) (
	CuBook		* cubook,
	char		* old_name,
	char		* new_name,
	void		* user_data
	);



static int duplicate_name (
	CuBook		* cubook,
	char		* name,
	NameCB		callback,
	void		* user_data
	)
{
	char		* new_name = NULL,
			* cp;
	int		size,
			next_id,
			res = 0;


	if ( ! name || ! callback )
	{
		return CU_ERROR_NO_CHANGES;
	}

	// Create space for a large number + brackets/space
	size = strlen ( name );
	new_name = calloc ( size + 16, 1 );

	if ( ! new_name )
	{
		return CU_ERROR_NO_CHANGES;
	}

	strcpy ( new_name, (char *)name );

	next_id = FIRST_DUPLICATE_ID;

	// Find out if this name is already a duplicate - if so use its number
	if ( size > 3 && new_name[size - 1] == ')' )
	{
		// Lose the final bracket so we can test for a valid integer
		new_name[size - 1] = 0;

		// Chop off the last ' (%i)'
		cp = strrchr ( new_name, '(' );
		if (	cp > new_name &&
			cp[-1] == ' ' &&
			! mtkit_strtoi ( cp + 1, &next_id, NULL, 1 ) &&
			next_id >= FIRST_DUPLICATE_ID
			)
		{
			cp[-1] = 0;
			size = cp - 1 - new_name;
		}
		else
		{
			// Stop negative index being valid
			mtMAX ( next_id, next_id, FIRST_DUPLICATE_ID )

		// Reinstate the bracket as its needed in the original name
			new_name[size - 1] = ')';
		}
	}

	// Keep looping until we get a valid number we can use
	for ( ; next_id < CEDUI_SHEET_MAX_NAME; next_id ++ )
	{
		sprintf ( new_name + size, " (%i)", next_id );
		res = callback ( cubook, name, new_name, user_data );

		if ( res )
		{
		// Keep looping until we find a valid new name or get error

			if ( res == 1 )
			{
				res = 0;
			}

			break;
		}
	}

	if ( ! res && next_id == CEDUI_SHEET_MAX_NAME )
	{
		// No spare names found
		res = CU_ERROR_NO_CHANGES;
	}

	free ( new_name );

	return res;
}

static int dup_sheet_cb (
	CuBook		* cubook,
	char		* old_name,
	char		* new_name,
	void		* user_data
	)
{
	if ( ! ced_book_get_sheet ( cubook->book, new_name ) )
	{
		int	res;


		res = cuf_book_add_sheet ( cubook, (CedSheet *)user_data,
			new_name );

		if ( res < 0 )
		{
			return res;	// Stop - error occurred
		}

		return 1;		// Use this new_name
	}

	return 0;			// Keep looking - this name is in use
}

int cuf_book_duplicate_sheet (
	CuBook		* cubook,
	CedSheet	* sheet,
	CedSheet	** new_sh
	)
{
	int		res;
	CedSheet	* newsheet = NULL;


	if (	! cubook ||
		! sheet ||
		! sheet->book_tnode ||
		! sheet->book_tnode->key
		)
	{
		return CU_ERROR_NO_CHANGES;
	}

	newsheet = ced_sheet_duplicate ( sheet );
	if ( ! newsheet )
	{
		return CU_ERROR_NO_CHANGES;
	}

	res = duplicate_name ( cubook, (char *)sheet->book_tnode->key,
		dup_sheet_cb, newsheet );

	if ( res )
	{
		return res;
	}

	if ( new_sh )
	{
		new_sh[0] = newsheet;
	}

	return 0;
}

static int check_graph_name_available (
	CuBook		* cubook,
	char		* old_name,
	char		* new_name,
	void		* user_data
	)
{
	CedBook		* book = cubook->book;


	if ( ! ced_graph_get ( book, new_name ) )
	{
		CedBookFile	* old,
				* new;
		char		* newmem = NULL,
				** nameret = user_data;


		old = ced_graph_get ( book, old_name );
		if ( ! old )
		{
			goto error;
		}

		if ( old->size > 0 )
		{
			newmem = malloc ( old->size );
			if ( ! newmem )
			{
				goto error;
			}

			memcpy ( newmem, old->mem, old->size );
		}

		new = ced_graph_new ( book, newmem, old->size, new_name );
		if ( ! new )
		{
			goto error;
		}

		if ( nameret )
		{
// Note: we can't return 'newname' because it is free'd by duplicate_name ()
			nameret[0] = strdup ( new_name );
		}

		return 1;		// Success: Use this name

error:
		free ( newmem );
		return -1;		// Bail out with error
	}

	return 0;		// Do not use, keep looking for next number
}

int ced_graph_duplicate (
	CuBook		* cubook,
	char		* graph_name,
	char		** new_name )
{
	if (	! cubook ||
		! graph_name ||
		duplicate_name ( cubook, graph_name,
			check_graph_name_available, new_name )
		)
	{
		return 1;
	}

	return 0;
}
