/*
	Copyright (C) 2012-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



CedUI_file * cuf_file_new ( void )
{
	CedUI_file	* file;


	file = calloc ( sizeof ( CedUI_file ), 1 );
	if ( ! file )
	{
		return NULL;
	}

	file->cubook = cuf_book_new ();
	if ( ! file->cubook )
	{
		cuf_file_free ( file );

		return NULL;
	}

	file->type = CED_FILE_TYPE_TSV_VAL_BOOK;

	return file;
}

int cuf_file_free (
	CedUI_file	* file
	)
{
	if ( ! file )
	{
		return 1;
	}

	cuf_book_destroy ( file->cubook );
	free ( file->name );
	free ( file );

	return 0;
}

int cuf_file_load (
	CedUI_file	* file,
	char		* filename
	)
{
	char		link_filename[MTKIT_PATHBUF_MAX],
			real_filename[MTKIT_PATHBUF_MAX],
			* tcp;
	int		cedfiletype,
			l;
	ssize_t		link_res;
	CuBook		* new_cubook = NULL;
	CedBook		* newbook = NULL;
	CedFileType	ftype;


	if ( ! file || ! filename || ! filename[0] )
	{
		return 1;
	}

	l = 0;
	if ( filename[0] != MTKIT_DIR_SEP )
	{
		// Expand the filename to include the full path
		if ( NULL == getcwd ( real_filename, MTKIT_PATHBUF_MAX - 2 ) )
		{
			goto fail;
		}

		l = strlen ( real_filename );
		real_filename[ l++ ] = MTKIT_DIR_SEP;
	}
	real_filename[ l ] = 0;
	mtkit_strnncat ( real_filename, filename, MTKIT_PATHBUF_MAX );

/*
Check to see if this is a symlink.  We do this to avoid problems when saving
as we rename onto this input filename which destroys the symlink and zero's
the linked file.
*/
	for ( l = 0; ; l++ )
	{
		if ( l > 5 )
		{
			goto fail;	// More than 5 chained links so give up
		}

		link_res = readlink ( real_filename, link_filename,
			MTKIT_PATHBUF_MAX - 1 );
		link_filename[MTKIT_PATHBUF_MAX - 1] = 0;
		if ( link_res < 0 )
		{
			if ( errno == EINVAL )
			{
				// This is not a symlink so load as normal file
				break;
			}

			// Error
			goto fail;
		}

		// Symlink found, so loop and check again
		link_filename[link_res] = 0;

		if ( link_filename[0] != MTKIT_DIR_SEP )
		{
			// Relative symlink

			tcp = strrchr ( real_filename, MTKIT_DIR_SEP );
			if ( ! tcp )
			{
				goto fail;
			}

			tcp[1] = 0;
		}
		else
		{
			// Full symlink
			real_filename[0] = 0;
		}

		mtkit_strnncat ( real_filename, link_filename,
			MTKIT_PATHBUF_MAX );
	}

	cedfiletype = ced_file_type_detect ( real_filename );
	if ( cedfiletype < 1 || cedfiletype > 3 )
	{
		// File error / Unknown file type

		goto fail;
	}

	if ( cedfiletype == 1 )
	{
		newbook = ced_book_load ( real_filename, &ftype, "ISO-8859-1" );
		if ( ! newbook )
		{
			goto fail;
		}
	}
	else
	{
		char		* sheet_name,
				* free_sheet_name = NULL;
		CedSheet	* newsheet = NULL;


		// TSV or CSV
		if ( cedfiletype == 2 )
		{
			newsheet = ced_sheet_load ( real_filename,
				"ISO-8859-1", &ftype );
		}
		else if ( cedfiletype == 3 )
		{
			newsheet = ced_sheet_load_csv ( real_filename,
				"ISO-8859-1" );
			ftype = CED_FILE_TYPE_CSV_CONTENT;
		}

		if ( ! newsheet )
		{
			goto fail;
		}

		newbook = ced_book_new ();
		if ( ! newbook )
		{
			ced_sheet_destroy ( newsheet );

			goto fail;
		}

		sheet_name = strrchr ( filename, MTKIT_DIR_SEP );
		if ( sheet_name )
		{
			sheet_name ++;
		}
		else
		{
			sheet_name = filename;
		}

		if ( ! mtkit_utf8_string_legal ( (unsigned char *)sheet_name,
			0 ) )
		{
			if ( mtkit_string_encoding_conversion ( sheet_name,
				"ISO-8859-1", &free_sheet_name, "UTF-8" ) )
			{
				goto fail;
			}

			sheet_name = free_sheet_name;
		}

		if ( ced_book_add_sheet ( newbook, newsheet, sheet_name ) )
		{
			ced_sheet_destroy ( newsheet );
			free ( free_sheet_name );

			goto fail;
		}

		free ( free_sheet_name );
	}

	new_cubook = cuf_book_new ();
	if ( ! new_cubook )
	{
		goto fail;
	}

	// Everything has worked so lets commit to the file structure
	new_cubook->book = newbook;

	cuf_book_destroy ( file->cubook );
	file->cubook = new_cubook;
	file->type = ftype;
	mtkit_strfreeswap ( &file->name, real_filename );

	// If sheet was loaded populate the active sheet name
	if ( cedfiletype == 1 )
	{
		cuf_file_get_sheet ( file );
	}

	return 0;

fail:
	ced_book_destroy ( newbook );

	return 1;
}

int cuf_sheet_check_geometry (
	CedSheet	* sheet
	)
{
	unsigned int	r,
			c;


	if ( ced_sheet_get_geometry ( sheet, &r, &c ) )
	{
		return 2;
	}

	if (	r > CEDUI_SHEET_MAX_ROW ||
		c > CEDUI_SHEET_MAX_COL ||
		(r * c) > CEDUI_SHEET_MAX_AREA
		)
	{
		return 1;
	}

	return 0;
}

static int tsv_sheets_cb (
	CedSheet	* sheet,
	char		* name,
	void		* user_data
	)
{
	return cuf_sheet_check_geometry ( sheet );
}

#define MAX_TEMP_SUFFIX 1000

int cuf_file_save (
	CedUI_file	* file,
	char		* filename,
	CedFileType	filetype
	)
{
	CedSheet	* sheet = NULL;
	CedBook		* book = NULL;
	char		* tempfile,
			* suffix = "",
			* s;
	int		i;
	FILE		* fp;


	if ( ! file || ! filename || ! filename[0] )
	{
		return 1;
	}

/*
Check the sheet geometry now as we are saving to a TSV format which may
struggle with very large sheets.
*/

	if (	filetype < CED_FILE_TYPE_NONE ||
		filetype > CED_FILE_TYPE_TOTAL )
	{
		// Type has been corrupted beyond defined boundaries
		return 1;
	}

	i = 0;
	switch ( filetype )
	{
	case CED_FILE_TYPE_TSV_CONTENT:
	case CED_FILE_TYPE_TSV_CONTENT_GZ:
	case CED_FILE_TYPE_TSV_CONTENT_NOQ:
	case CED_FILE_TYPE_TSV_VALUE:
	case CED_FILE_TYPE_TSV_VALUE_GZ:
	case CED_FILE_TYPE_TSV_VALUE_NOQ:
	case CED_FILE_TYPE_CSV_CONTENT:
	case CED_FILE_TYPE_CSV_CONTENT_NOQ:
	case CED_FILE_TYPE_CSV_VALUE:
	case CED_FILE_TYPE_CSV_VALUE_NOQ:
		sheet = cuf_file_get_sheet ( file );
		if ( ! sheet )
		{
			return 1;
		}

		i = cuf_sheet_check_geometry ( sheet );
		break;

	case CED_FILE_TYPE_LEDGER:
	case CED_FILE_TYPE_LEDGER_GZ:
	case CED_FILE_TYPE_LEDGER_VAL:
	case CED_FILE_TYPE_LEDGER_VAL_GZ:
		sheet = cuf_file_get_sheet ( file );
		if ( ! sheet )
		{
			return 1;
		}
		break;

	case CED_FILE_TYPE_TSV_BOOK:
	case CED_FILE_TYPE_TSV_VAL_BOOK:
		book = file->cubook->book;
		if ( ! book )
		{
			return 1;
		}
		i = ced_book_scan ( book, tsv_sheets_cb, NULL );
		break;

	case CED_FILE_TYPE_LEDGER_BOOK:
	case CED_FILE_TYPE_LEDGER_VAL_BOOK:
		book = file->cubook->book;
		if ( ! book )
		{
			return 1;
		}
		break;

	case CED_FILE_TYPE_NONE:
	case CED_FILE_TYPE_TOTAL:
	case CED_FILE_TYPE_OUTPUT_TSV:
	case CED_FILE_TYPE_OUTPUT_TSV_QUOTED:
	case CED_FILE_TYPE_OUTPUT_HTML:
		// Bogus types
		return 1;
	}

	if ( i )
	{
		// Sheet too large
		return 1;
	}

	tempfile = calloc ( strlen ( filename ) + 16, 1 );
	if ( ! tempfile )
	{
		goto error;
	}

	if ( book )
	{
	// Add a .zip suffix if no file type specified by user
	// This trick ensures that the internal zip filename isn't filename_000
		s = strrchr ( filename, MTKIT_DIR_SEP );
		if ( ! s )
		{
			s = filename;
		}
		else
		{
			s++;
		}

		if ( ! strrchr ( s, '.' ) )
		{
			suffix = ".zip";
		}
	}

	// Create new temp filename that is available in form "filename_000"
	for ( i = 0; i < MAX_TEMP_SUFFIX; i++ )
	{
		sprintf ( tempfile, "%s%s_%03i", filename, suffix, i );
		if ( mtkit_file_valid ( tempfile ) )
		{
			break;
		}
	}

	// No free filename found
	if ( i >= MAX_TEMP_SUFFIX )
	{
		goto error;
	}

	if ( book )
	{
		if ( ced_book_save ( book, tempfile, filetype ) )
		{
			goto error;
		}
	}
	else if ( sheet )
	{
		if ( ced_sheet_save ( sheet, tempfile, filetype ) )
		{
			goto error;
		}
	}

	// This step is required to avoid removing a write protected file
	fp = fopen ( filename, "w" );
	if ( ! fp )
	{
		goto error;
	}
	fclose ( fp );

	if ( rename ( tempfile, filename ) )
	{
		goto error;
	}

	free ( tempfile );
	tempfile = NULL;

	// Everything worked so change state values
	file->type = filetype;
	if ( filename != file->name )
	{
		mtkit_strfreeswap ( &file->name, filename );
	}

	return 0;			// Success

error:
	if ( tempfile )
	{
		remove ( tempfile );
		free ( tempfile );
		tempfile = NULL;
	}

	return -1;
}

int cuf_file_sheet_add (
	CedUI_file	* file
	)
{
	CedSheet	* newsheet = NULL;
	char		page_name[256];
	int		i,
			res = 1;


	if (	! file ||
		! file->cubook ||
		! file->cubook->book
		)
	{
		goto error;
	}

	for ( i = 1; i < CEDUI_SHEET_MAX_NAME; i++ )
	{
		sprintf ( page_name, "Sheet %i", i );
		if ( ! ced_book_get_sheet ( file->cubook->book, page_name ) )
		{
			break;
		}

		// Keep looping until we find a valid new name
	}

	if (	i == CEDUI_SHEET_MAX_NAME ||
		! ( newsheet = ced_sheet_new () ) )
	{
		goto error;
	}

	res = cuf_book_add_sheet ( file->cubook, newsheet, page_name );
	if (	res == CU_ERROR_UNDO_OP ||
		res == CU_ERROR_NO_CHANGES
		)
	{
	// We only report an error if the operation was not carried out.
		goto error;
	}

	mtkit_strfreeswap ( &file->cubook->book->prefs.active_sheet,
		page_name );

	return 0;

error:
	ced_sheet_destroy ( newsheet );

	return res;
}

static int cuf_file_sheet_new (
	CedUI_file	* file,
	char		* sheet_name
	)
{
	CedSheet	* sheet;


	if ( ced_book_get_sheet ( file->cubook->book, sheet_name ) )
	{
		fprintf ( stderr, "cuf_file_sheet_new: Sheet already exists.\n"
			);

		goto fail;
	}

	sheet = ced_sheet_new ();
	if ( ! sheet )
	{
		goto fail;
	}

	if ( ced_book_add_sheet ( file->cubook->book, sheet, sheet_name ) )
	{
		ced_sheet_destroy ( sheet );

		goto fail;
	}

	mtkit_strfreeswap ( &file->cubook->book->prefs.active_sheet,
		sheet_name );

	return 0;

fail:
	fprintf ( stderr,
		"cuf_file_sheet_new: Unable to create new sheet '%s'.\n",
		sheet_name );

	return 1;
}

int cuf_file_book_new (
	CedUI_file	* file
	)
{
	if ( ! file )
	{
		return 1;
	}

	cuf_book_destroy ( file->cubook );
	file->cubook = cuf_book_new ();
	if ( ! file->cubook )
	{
		goto fail;
	}

	file->cubook->book = ced_book_new ();
	if ( ! file->cubook->book )
	{
		goto fail;
	}

	if ( cuf_file_sheet_new ( file, "Sheet 1" ) )
	{
		goto fail;
	}

	mtkit_strfreeswap ( &file->name, NULL );
	file->type = CED_FILE_TYPE_TSV_VAL_BOOK;

	return 0;

fail:
	fprintf ( stderr, "cuf_file_book_new: Unable to create new book.\n" );

	return 1;
}

CedSheet * cuf_file_get_sheet (
	CedUI_file	* file
	)
{
	CedBook		* book;
	CedSheet	* sheet;


	if ( ! file )
	{
		return NULL;
	}

	book = file->cubook->book;
	sheet = ced_book_get_sheet ( book, book->prefs.active_sheet );

	if ( ! sheet )
	{
		if ( book->prefs.active_sheet )
		{
// No valid sheet found so flush current dangling reference
			mtkit_strfreeswap ( &book->prefs.active_sheet, NULL );
		}

		if ( book->sheets->root )
		{
// Current name is empty, but a sheet exists so pick root sheet name
			mtkit_strfreeswap ( &book->prefs.active_sheet,
				book->sheets->root->key );
			sheet = (CedSheet *)book->sheets->root->data;
		}
	}

	return sheet;
}

static int find_graph (
	CedBook		* book,
	char		* graph_name,
	CedBookFile	* bookfile,
	void		* user_data
	)
{
	char		** name = user_data;


	name[0] = graph_name;

	return 1;		// Stop
}

CedBookFile * cuf_file_get_graph (
	CedUI_file	* file
	)
{
	CedBook		* book;
	CedBookFile	* graph;


	if ( ! file )
	{
		return NULL;
	}

	book = file->cubook->book;
	graph = ced_graph_get ( book, book->prefs.active_graph );

	if ( ! graph )
	{
		char		* name = NULL;


		if ( book->prefs.active_graph )
		{
	// No valid graph found so flush current dangling reference
			mtkit_strfreeswap ( &book->prefs.active_graph, NULL );
		}

		// Use the first graph (if there is one)
		ced_graph_scan ( book, find_graph, &name );
		graph = ced_graph_get ( book, name );
		mtkit_strfreeswap ( &book->prefs.active_graph, name );
	}

	return graph;
}

static int scan_2dyear (
	CedSheet	* sheet,
	CedCell		* cell,
	unsigned int	row,
	unsigned int	col,
	void		* user_data
	)
{
	ced_cell_set_2dyear ( cell, ( (int *)user_data )[0] );

	return 0;			// Continue
}

int cuf_sheet_2dyear (
	CedUI_file	* file,
	int		year_start
	)
{
	int		res;
	CedUI_clip	* clip;


	if (	! file ||
		year_start < MTKIT_DDT_MIN_DATE_YEAR ||
		year_start > MTKIT_DDT_MAX_DATE_YEAR )
	{
		return 1;
	}

	clip = cuf_clip_new ();
	if ( ! clip )
	{
		return CU_ERROR_NO_CHANGES;
	}

	if ( cuf_clip_copy ( file, clip ) )
	{
		cuf_clip_free ( clip );

		return CU_ERROR_NO_CHANGES;
	}

	ced_sheet_scan_area ( clip->sheet, 1, 1, 0, 0, scan_2dyear,
		(void *)&year_start );

	res = cuf_clip_paste ( file, clip, 0 );
	cuf_clip_free ( clip );

	return res;
}

