/*
	Copyright (C) 2008-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "gui.h"



/*
Summary of clipboard actions

When copying data we initially just put it into (CedSheet *) global.clip_sheet;
When this is pasted by this instance of mtCellEdit, we recognise it as our
data by the owner object being the main window.

When another app pastes this data we create a TSV file and pass over the data
- this data is kept in memory just in case it is pasted again later.  This is
pure text and is in (char *)global.clip_tsv;

When another instance of mtCellEdit pastes this data a CED ZIP file is created
and a timestamp is created to identify it between processes.

When another program, e.g. Gnumeric, copies data to the system clipboard this
is always read in on demand as a TSV memory chunk.
*/



#define MTCELLEDIT_CLIP_CODE		123456
#define MTCELLEDIT_CLIP_NAME		"application/x-mtcelledit-clipboard"



///	COPYING + XFERING DATA TO/FROM SYSTEM CLIPBOARD

static void flush_internal_clipboard ( void )
{
	GtkClipboard	* clip;
	GObject		* owner = NULL;


	clip = gtk_clipboard_get ( GDK_SELECTION_CLIPBOARD );
	if ( clip )
	{
		owner = gtk_clipboard_get_owner ( clip );
	}

	if ( owner && owner == (GObject *)(global.main_window) )
	{
		// We own the system clipboard so destroy it

		gtk_clipboard_clear ( clip );
	}
	else
	{
		// We don't own the clipboard so only flush our copy of it

		cuf_clip_flush ( global.clipboard );
	}
}

static void get_clip_func (
	GtkClipboard	* clipboard,
	GtkSelectionData * selection_data,
	guint		info,
	gpointer	user_data_or_owner
	)
{
	GdkAtom		ced_clip_atom;


	ced_clip_atom = gdk_atom_intern ( MTCELLEDIT_CLIP_NAME, FALSE );

	if ( selection_data->target == ced_clip_atom )
	{

		if ( cuf_clip_save_temp ( global.clipboard ) )
		{
			goto error_temp;
		}

		// Pass the timestamp
		gtk_selection_data_set_text ( selection_data,
			global.clipboard->timestamp,
			CEDUI_CLIPBOARD_TIMESTAMP_SIZE );

		gtk_selection_data_set ( selection_data, ced_clip_atom, 8,
			(guchar *)global.clipboard->timestamp,
			CEDUI_CLIPBOARD_TIMESTAMP_SIZE );
	}
	else
	{

		if (	cuf_clip_export_text ( global.clipboard ) ||
			! global.clipboard->tsv )
		{
			goto error;
		}

		gtk_selection_data_set_text ( selection_data,
			global.clipboard->tsv, -1 );
	}

	return;				// Success

error:
	gtk_selection_data_set_text ( selection_data,
		_("Error - get_clip_func: unable to create TSV data\n"), -1 );

	return;

error_temp:
	gtk_selection_data_set_text ( selection_data,
		_("Error - get_clip_func: unable to create temp file\n"), -1 );
}


static void get_clear_func (
	GtkClipboard	* clipboard,
	gpointer	user_data_or_owner
	)
{
	cuf_clip_flush ( global.clipboard );
}

static void set_clip_owner ( void )
{
	GtkTargetList	* list;
	GtkTargetEntry	* targets;
	gint		n_targets;
	GtkClipboard	* clip;


	clip = gtk_clipboard_get ( GDK_SELECTION_CLIPBOARD );
	if ( ! clip )
	{
		return;
	}

	list = gtk_target_list_new ( NULL, 0 );
	gtk_target_list_add ( list, gdk_atom_intern ( MTCELLEDIT_CLIP_NAME,
		FALSE ), 0, MTCELLEDIT_CLIP_CODE );
	gtk_target_list_add_text_targets ( list, 0 );
	targets = gtk_target_table_new_from_list ( list, &n_targets );

	gtk_clipboard_set_with_owner ( clip, targets, n_targets,
		get_clip_func, get_clear_func,
		G_OBJECT ( global.main_window ) );

	gtk_target_table_free ( targets, n_targets );
	gtk_target_list_unref ( list );
}

static int copy_selection_to_clip ( void )
{
	flush_internal_clipboard ();
	if ( cuf_clip_copy ( global.file, global.clipboard ) )
	{
		mtgex_alert_box ( _("Error"),
			_("Unable to copy selection to clipboard."),
			_("OK"), NULL, NULL, global.main_window );

		return 1;
	}

	set_clip_owner ();

	return 0;			// Success
}

void pressed_copy (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	copy_selection_to_clip ();

	ced_view_bell_swap ( global.cedview );
}

static int copy_value_scan (
	CedSheet	* sheet,
	CedCell		* cell,
	unsigned int	row,
	unsigned int	col,
	void		* user_data
	)
{
	char		txt[256];


	switch ( cell->type )
	{
	case CED_CELL_TYPE_FORMULA:
	case CED_CELL_TYPE_FORMULA_EVAL:
		sprintf ( txt, CED_PRINTF_NUM, cell->value );
		if ( ! mtkit_strfreeswap ( &cell->text, txt ) )
		{
			cell->type = CED_CELL_TYPE_VALUE;
		}
		else
		{
			return 1;
		}
		break;

	default:
		break;
	}

	return 0;
}

void pressed_copy_values (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( copy_selection_to_clip () )
	{
		return;
	}

	if ( ced_sheet_scan_area ( global.clipboard->sheet, 1, 1, 0, 0,
		copy_value_scan, NULL ) )
	{
		mtgex_alert_box ( _("Error"), _("Unable to create clipboard."),
			_("OK"), NULL, NULL, global.main_window );

		return;
	}

	ced_view_bell_swap ( global.cedview );
}



static int copy_output_scan (
	CedSheet	* sheet,
	CedCell		* cell,
	unsigned int	row,
	unsigned int	col,
	void		* user_data
	)
{
	char		* txt;


	txt = ced_cell_create_output ( cell, NULL );
	free ( cell->text );

	cell->text = txt;
	cell->type = CED_CELL_TYPE_TEXT_EXPLICIT;
	cell->value = 0;

	return 0;
}

void pressed_copy_output (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( copy_selection_to_clip () )
	{
		return;
	}

	if ( ced_sheet_scan_area ( global.clipboard->sheet, 1, 1, 0, 0,
		copy_output_scan, NULL ) )
	{
		mtgex_alert_box ( _("Error"), _("Unable to create clipboard."),
			_("OK"), NULL, NULL, global.main_window );

		return;
	}

	ced_view_bell_swap ( global.cedview );
}



///	PASTING CUTTING CLEARING



static int request_mtcelledit_content (
	GtkClipboard	* clip
	)
{
	GtkSelectionData * selection;


	selection = gtk_clipboard_wait_for_contents ( clip,
			gdk_atom_intern ( MTCELLEDIT_CLIP_NAME, FALSE ) );

	if (	! selection ||
		selection->length != CEDUI_CLIPBOARD_TIMESTAMP_SIZE
		)
	{
		return 0;		// Not sent by mtCellEdit
	}

	if (	! global.clipboard->sheet ||
		memcmp( global.clipboard->timestamp, selection->data,
			CEDUI_CLIPBOARD_TIMESTAMP_SIZE ) )
	{
		if ( cuf_clip_load_temp ( global.clipboard ) )
		{
			return 0;	// Unable to load clipboard - error
		}

		// Set the new timestamp
		cuf_clip_set_timestamp ( global.clipboard,
			(char *)selection->data );
	}

	gtk_selection_data_free ( selection );

	return 1;			// mtCellEdit clipboard is available
}

static int read_system_clipboard (
	GtkClipboard	* clip
	)
{
	char		* txt;
	GObject		* owner;


	owner = gtk_clipboard_get_owner ( clip );

	if ( owner && owner == (GObject *)(global.main_window) )
	{
		// If we created this clipboard ourselves we can stop here

		return 0;
	}

	if ( request_mtcelledit_content ( clip ) )
	{
		// This clipboard was created by another mtCellEdit instance

		return 0;
	}

// This is not mtCellEdit clipboard data so load it in as a TSV text chunk
	txt = gtk_clipboard_wait_for_text ( clip );
	if ( ! txt )
	{
		return 1;		// No clipboard
	}

	if ( cuf_clip_import_text ( global.clipboard, txt ) )
	{
		g_free ( txt );

		return 1;		// Error importing
	}

	g_free ( txt );

	return 0;			// Success
}

static int obtain_paste ( void )
{
	CedSheet	* sheet;
	GtkClipboard	* clip;


	sheet = global.cedview->ren.sheet;
	if ( ! sheet )
	{
		return 0;
	}

	clip = gtk_clipboard_get ( GDK_SELECTION_CLIPBOARD );
	if ( ! clip )
	{
		return 0;
	}

	if ( read_system_clipboard ( clip ) )
	{
		return 0;		// Nothing read in
	}

	if ( ! global.clipboard->sheet )
	{
		return 0;		// Nothing to paste
	}

	return 1;	// Successfully transferred something to clipboard
}

static int paste_clipboard_at_cursor (
	int		mode
	)
{
	int		res;


	if ( ! obtain_paste () )
	{
		return 1;		// No paste found
	}

	res = cuf_clip_paste ( global.file, global.clipboard, mode );
	if ( res == 1 )
	{
		return 1;
	}

	undo_report_updates ( res );

	if (	res == CU_ERROR_LOCKED_CELL ||
		res == CU_ERROR_LOCKED_SHEET ||
		res == CU_ERROR_NO_CHANGES
		)
	{
		return 1;		// Nothing changed
	}

	update_changes_chores ( 1, 0 );

	return 0;			// Paste committed
}

void pressed_paste (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	paste_clipboard_at_cursor ( 0 );
}

void pressed_paste_content (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	paste_clipboard_at_cursor ( CED_PASTE_CONTENT );
}

void pressed_paste_prefs (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	paste_clipboard_at_cursor ( CED_PASTE_PREFS );
}

static int clear_selection (
	int		mode
	)
{
	CedSheet	* sheet;
	unsigned int	r,
			c,
			rtot,
			ctot;
	int		res;


	sheet = global.cedview->ren.sheet;
	if ( ! sheet )
	{
		return 1;
	}

	r = MIN ( sheet->prefs.cursor_r1, sheet->prefs.cursor_r2 );
	c = MIN ( sheet->prefs.cursor_c1, sheet->prefs.cursor_c2 );
	rtot = 1 + abs ( sheet->prefs.cursor_r1 - sheet->prefs.cursor_r2 );
	ctot = 1 + abs ( sheet->prefs.cursor_c1 - sheet->prefs.cursor_c2 );

	if ( r < 1 || c < 1 )
	{
		return 1;
	}

	res = cuf_sheet_clear_area ( global.file->cubook, sheet, r, c, rtot,
		ctot, mode );
	undo_report_updates ( res );

	if (	res == CU_ERROR_LOCKED_CELL ||
		res == CU_ERROR_NO_CHANGES
		)
	{
		return 1;
	}

	update_changes_chores ( 1, 0 );

	return 0;			// Success
}

void pressed_cut (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( copy_selection_to_clip () )
	{
		return;
	}

	if ( clear_selection ( 0 ) )
	{
		return;
	}
}

void pressed_clear (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( clear_selection ( 0 ) )
	{
		return;
	}
}

void pressed_clear_contents (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( clear_selection ( CED_PASTE_CONTENT ) )
	{
		return;
	}
}

void pressed_clear_prefs (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	if ( clear_selection ( CED_PASTE_PREFS ) )
	{
		return;
	}
}

static int calc_selection_position (
	int		* row,
	int		* col
	)
{
	CedSheet	* sheet = global.cedview->ren.sheet;


	if ( ! sheet )
	{
		return 1;
	}

	if ( row )
	{
		row[0] = sheet->prefs.cursor_r1;
	}

	if ( col )
	{
		col[0] = sheet->prefs.cursor_c1;
	}

	if ( ! obtain_paste () )
	{
		return 1;		// No paste found
	}

	return 0;
}

void pressed_row_insert_paste_height (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	int		row,
			res;


	if ( undo_report_updates ( cuf_check_sheet_lock (
		global.cedview->ren.sheet ) ) )
	{
		return;
	}

	if (	calc_selection_position ( &row, NULL ) ||
		row < 1 )
	{
		return;
	}

	res = cuf_sheet_insert_row ( global.file->cubook,
		global.cedview->ren.sheet, row, global.clipboard->rows );
	undo_report_updates ( res );

	if (	res == CU_ERROR_LOCKED_CELL ||
		res == CU_ERROR_NO_CHANGES
		)
	{
		return;
	}

	update_changes_chores ( 1, 0 );
}

void pressed_column_insert_paste_width (
	GtkMenuItem	* menu_item,
	gpointer	user_data
	)
{
	int		col,
			res;


	if ( undo_report_updates ( cuf_check_sheet_lock (
		global.cedview->ren.sheet ) ) )
	{
		return;
	}

	if (	calc_selection_position ( NULL, &col ) ||
		col < 1
		)
	{
		return;
	}

	res = cuf_sheet_insert_column ( global.file->cubook,
		global.cedview->ren.sheet, col, global.clipboard->cols );
	undo_report_updates ( res );

	if (	res == CU_ERROR_LOCKED_CELL ||
		res == CU_ERROR_NO_CHANGES
		)
	{
		return;
	}

	update_changes_chores ( 1, 0 );
}

void pressed_transform_clipboard (
	GtkMenuItem	* menu_item,
	gpointer	user_data,
	gint		item
	)
{
	CedSheet	* sheet = NULL;


	if ( ! obtain_paste () )
	{
		return;			// No paste to act on
	}

	switch ( item )
	{
	case 0:
		sheet = ced_sheet_transpose ( global.clipboard->sheet );
		break;

	case 1:
		sheet = ced_sheet_flip_horizontal ( global.clipboard->sheet );
		break;

	case 2:
		sheet = ced_sheet_flip_vertical ( global.clipboard->sheet );
		break;

	case 3:
		sheet = ced_sheet_rotate_clockwise ( global.clipboard->sheet );
		break;

	case 4:
		sheet = ced_sheet_rotate_anticlockwise (
			global.clipboard->sheet );
		break;
	}

	if ( ! sheet )
	{
		goto error;
	}

	// Remove old, install new
	flush_internal_clipboard ();
	if ( global.clipboard->sheet )
	{
		ced_sheet_destroy ( sheet );

		goto error;
	}

	global.clipboard->sheet = sheet;

	if (	item != 1 &&
		item != 2
		)
	{
		int		tmp = global.clipboard->rows;


		// X/Y Geometry has been swapped

		global.clipboard->rows = global.clipboard->cols;
		global.clipboard->cols = tmp;
	}

	set_clip_owner ();		// Claim ownership

	return;				// Success

error:
	mtgex_alert_box ( _("Error"), _("Unable to transform clipboard."),
		_("OK"), NULL, NULL, global.main_window );
}
