/*
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "eleana.h"



static int ptree_cmp (
	void		* k1,
	void		* k2
	)
{
	mapPOLYkey	* key1 = k1,
			* key2 = k2;


	if ( key1->miny < key2->miny )
	{
		return -1;
	}

	if ( key1->miny > key2->miny )
	{
		return 1;
	}

	if ( key1->seat < key2->seat )
	{
		return -1;
	}

	if ( key1->seat > key2->seat )
	{
		return 1;
	}

	return 0;			// identical
}

static void ptree_del (
	mtTreeNode	* node
	)
{
	mapPOLYkey	* key = node->key;
	mapPOLYdata	* data = node->data;


	free ( data->array );
	free ( data );
	free ( key );
}

static int parse_poly (
	mtTree		* tree,
	int		seat,
	char		* text
	)
{
	mapPOLYkey	* key;
	mapPOLYdata	* data;
	int		i,
			nodes,
			literal;
	char		* st;


	nodes = mtkit_strtok_count ( text, "," );
	if ( nodes < 3 )
	{
		// Nothing to do as there aren't enough nodes

		return 0;
	}

	key = calloc ( 1, sizeof ( mapPOLYkey ) );
	if ( ! key )
	{
		return 1;	// Memory error
	}

	data = calloc ( 1, sizeof ( mapPOLYdata ) );
	if ( ! data )
	{
		free ( key );

		return 1;	// Memory error
	}

	data->array = calloc ( nodes, sizeof ( mapPOLYpoint ) );
	if ( ! data->array )
	{
		goto fail;
	}

	data->points = nodes - 1;

	st = text;
	for ( i = 0; i < nodes; i++ )
	{
		while ( isspace ( st[0] ) )
		{
			st ++;
		}

		switch ( st[0] )
		{
		case 'Z':
		case 'z':
			goto finish;

		case 'M':
		case 'm':
		case 'L':
		case 'l':
			literal = 1;
			st ++;
			break;

		default:
			if ( i > 0 )
			{
				literal = 0;
			}
			else
			{
				literal = 1;
			}
		}

		while ( isspace ( st[0] ) )
		{
			st ++;
		}

		if ( mtkit_strtod ( st, &data->array[i].x, &st, 0 ) )
		{
			goto fail;
		}

		st ++;			// ','

		if ( mtkit_strtod ( st, &data->array[i].y, &st, 0 ) )
		{
			goto fail;
		}

		if ( ! literal )
		{
			data->array[i].x = data->array[i - 1].x +
				data->array[i].x;
			data->array[i].y = data->array[i - 1].y +
				data->array[i].y;
		}

		if ( i == 0 )
		{
			key->minx = key->maxx = data->array[i].x;
			key->miny = key->maxy = data->array[i].y;
		}
		else
		{
			key->minx = MIN ( key->minx, data->array[i].x );
			key->maxx = MAX ( key->maxx, data->array[i].x );
			key->miny = MIN ( key->miny, data->array[i].y );
			key->maxy = MAX ( key->maxy, data->array[i].y );
		}
	}

finish:

	key->seat = seat;
	if ( mtkit_tree_node_add ( tree, key, data ) == 0 )
	{
		goto fail;
	}

	return 0;			// Success

fail:
	free ( data->array );
	free ( data );
	free ( key );

	return 1;
}

mtTree * eleana_poly_new (
	CedSheet	* sheet,
	int		seats
	)
{
	mtTree		* tree;
	CedCell		* cell;
	int		i;


	tree = mtkit_tree_new ( ptree_cmp, ptree_del );
	if ( ! tree )
	{
		return NULL;
	}

	for ( i = 0; i < seats; i++ )
	{
		cell = ced_sheet_get_cell ( sheet, ROW_DATA_PARTY1 + i,
			COL_MAP_POLY );

		if ( ! cell )
		{
			continue;
		}

		if ( parse_poly ( tree, i, cell->text ) )
		{
			goto fail;
		}
	}

	if ( ! tree->root )
	{
		// Nothing found

		goto fail;
	}

	return tree;

fail:
	mtkit_tree_destroy ( tree );

	return NULL;
}



typedef struct
{
	mtImage		* image;	// Image to render on
	int		x,		// Offset that the image origin starts
			y;		// at
	mtTree		* tree;		// Polygon tree
	double		* table,	// Table of results
			zoom;		// Zoom factor for the polygons

	elRenderFunc	render_cb;
	void		* render_cb_user_data;

	double		winx1,
			winx2,		// Visible window on polygons
			winy1,
			winy2;		// At the original polygon scale

	mtBrush		* brush[5],
			* brush_fill,
			* brush_outline;

	mapPOLYkey	* key;
	mapPOLYdata	* data;

	mtPolygon	* polygon;

} polySTATE;



static int default_render_cb (
	int		seat,
	mtBrush		** brush_fill,
	mtBrush		** brush_outline,
	void		* user_data
	)
{
	int		win;
	polySTATE	* state = user_data;


	win = EL_TABLE_VALUE ( state->table, seat, EL_TAB_PROJ_WINNER ) - 1;
	if ( win >= 0 && win <= 3 )
	{
		brush_fill[0] = state->brush[win];
	}

	brush_outline[0] = state->brush[4];

	return 0;			// Always render
}

static int poly_recurse (
	polySTATE	* state,
	mtTreeNode	* node
	)
{
	int		res = 0;


	if ( node->left )
	{
		res = poly_recurse ( state, node->left );

		if ( res )
		{
			return res;
		}
	}

	state->key = node->key;
	state->data = node->data;

	state->brush_fill = NULL;
	state->brush_outline = NULL;

	if (	state->key->maxx >= state->winx1 &&
		state->key->minx <= state->winx2 &&
		state->key->maxy >= state->winy1 &&
		state->key->miny <= state->winy2 &&
		(res = state->render_cb ( state->key->seat, &state->brush_fill,
			&state->brush_outline, state->render_cb_user_data ) )
			== 0
		)
	{
		int		i,
				x,
				y;


		mtpixel_polygon_clear ( state->polygon );

		for ( i = 0; i < state->data->points; i++ )
		{
			x = rint ( state->data->array[i].x * state->zoom ) -
				state->x;
			y = rint ( state->data->array[i].y * state->zoom ) -
				state->y;
			mtpixel_polygon_add_point ( state->polygon, x, y );
		}

		if ( state->brush_fill )
		{
			mtpixel_brush_polygon ( state->image, state->brush_fill,
				state->polygon );
		}

		if ( state->brush_outline )
		{
			mtpixel_brush_polygon ( state->image,
				state->brush_outline, state->polygon );
		}
	}

	if ( res == 2 )
	{
		return 2;		// User requested stop
	}

	if ( node->right && state->key->miny <= state->winy2 )
	{
		res = poly_recurse ( state, node->right );
		if ( res )
		{
			return res;
		}
	}

	return 0;
}

static void poly_brush_destroy (
	polySTATE	* state
	)
{
	int		i;

	for ( i = 0; i < 5; i++ )
	{
		mtpixel_brush_destroy ( state->brush[i] );
		state->brush[i] = NULL;
	}
}

static int poly_brush_new (
	polySTATE	* state,
	int		cols[5]
	)
{
	int		i;


	for ( i = 0; i < 5; i++ )
	{
		state->brush[i] = mtpixel_brush_new ();
		if ( ! state->brush[i] )
		{
			poly_brush_destroy ( state );

			return 1;
		}

		mtpixel_brush_set_RGB ( state->brush[i], cols[i], -1 );
	}

	return 0;
}

int eleana_poly_render (
	mtImage		* image,
	int		x,
	int		y,
	mtTree		* tree,
	double		* table,
	double		zoom,
	int		cols[5]
	)
{
	polySTATE	state = { image, x, y, tree, table, zoom };


	state.winx1 = x / zoom - zoom / 2;
	state.winx2 = (x + image->width - 1) / zoom + zoom / 2;
	state.winy1 = y / zoom - zoom / 2;
	state.winy2 = (y + image->height - 1) / zoom + zoom / 2;

	state.render_cb = default_render_cb;
	state.render_cb_user_data = &state;

	state.polygon = mtpixel_polygon_new ();

	if ( ! state.polygon )
	{
		return 1;
	}

	if ( poly_brush_new ( &state, cols ) )
	{
		mtpixel_polygon_destroy ( state.polygon );

		return 1;
	}

	mtpixel_brush_set_size ( state.brush[0], 0 );
	mtpixel_brush_set_size ( state.brush[1], 0 );
	mtpixel_brush_set_size ( state.brush[2], 0 );
	mtpixel_brush_set_size ( state.brush[3], 0 );

	poly_recurse ( &state, tree->root );

	poly_brush_destroy ( &state );
	mtpixel_polygon_destroy ( state.polygon );

	return 0;
}

int eleana_poly_render_custom (
	mtImage		* image,
	int		x,
	int		y,
	mtTree		* tree,
	double		* table,
	double		zoom,
	elRenderFunc	callback,
	void		* user_data
	)
{
	polySTATE	state = { image, x, y, tree, table, zoom, callback,
				user_data };


	state.winx1 = x / zoom - zoom / 2;
	state.winx2 = (x + image->width - 1) / zoom + zoom / 2;
	state.winy1 = y / zoom - zoom / 2;
	state.winy2 = (y + image->height - 1) / zoom + zoom / 2;

	state.polygon = mtpixel_polygon_new ();

	if ( ! state.polygon )
	{
		return 1;
	}

	poly_recurse ( &state, tree->root );

	mtpixel_polygon_destroy ( state.polygon );

	return 0;
}

static int tree_scan_missing_cb (
	mtTreeNode	* node,
	void		* user_data
	)
{
	CedIndexItem	* item;


	item = (CedIndexItem *)node->data;
	if ( ! item )
	{
		return 1;		// Unexpected error
	}

	if ( item->row == 0 )
	{
		// This item hasn't been matched

		printf ( "'%s'\n", (char *)node->key );
	}

	return 0;
}



typedef struct
{
	int		i;
	CedIndex	* idx_seats;
	CedIndexItem	* index_item;
} prepSTATE;



static int result_scan_seats (
	CedSheet	* sheet,
	CedCell		* cell,
	unsigned int	row,
	unsigned int	col,
	void		* user_data
	)
{
	prepSTATE	* state = (prepSTATE *)user_data;


	// Match seat name in cell with index
	if ( ced_index_query ( state->idx_seats, 0, cell->text,
		&state->index_item ) != 1
		)
	{
		return 1;		// Should never happen
	}

	state->index_item->row = 0;
	state->index_item->col = state->i;
	state->i ++;

	return 0;			// Continue
}

static CedIndex * prepare_idx_seats (
	CedSheet	* results
	)
{
	prepSTATE	state = {0};


	state.idx_seats = ced_index_new ( CED_INDEX_TYPE_TEXT );
	if ( ! state.idx_seats )
	{
		return NULL;
	}

	if ( ced_index_add_items ( state.idx_seats, results, FULL_ROW_PARTY1,
		FULL_COL_SEAT_NAME, 0, 1 )
		)
	{
		goto error;
	}

	// For each seat in the index, change row to 0, change col to its index
	state.i = 0;
	if ( ced_sheet_scan_area ( results, FULL_ROW_PARTY1,
		FULL_COL_SEAT_NAME, 0, 1, result_scan_seats, &state )
		)
	{
		goto error;
	}

	return state.idx_seats;

error:
	ced_index_destroy ( state.idx_seats );

	return NULL;
}

mtTree * eleana_polymap_load (
	char		* filename,
	eleanaINDEX	* index,
	int		validate,
	int		tot_seats,
	CedSheet	* results
	)
{
	mtTree		* tree;
	char		* svg,
			* st,
			* d,
			* id,
			* d2,
			* id2;
	int		svg_size,
			tot,
			not_found = 0,
			res;
	CedIndex	* idx_seats = NULL;
	CedIndexItem	* index_item;


	svg = mtkit_file_load ( filename, &svg_size,
		MTKIT_FILE_ZERO | MTKIT_FILE_GUNZIP, NULL );

	if ( ! svg )
	{
		return NULL;
	}

	tree = mtkit_tree_new ( ptree_cmp, ptree_del );
	if ( ! tree )
	{
		free ( svg );

		return NULL;
	}

	st = svg;

	idx_seats = prepare_idx_seats ( results );
	if ( ! idx_seats )
	{
		goto fail;
	}

	for ( tot = 0; ; tot ++ )
	{
		st = mtkit_strcasestr ( st, "<path" );
		if ( ! st )
		{
			break;
		}

		d = mtkit_strcasestr ( st, "d=\"" );
		if ( ! d )
		{
			break;
		}

		d += 3;

		id = mtkit_strcasestr ( st, "id=\"" );
		if ( ! id )
		{
			break;
		}

		id += 4;

		d2 = strchr ( d, '"' );
		if ( ! d2 )
		{
			break;
		}

		d2[0] = 0;

		id2 = strchr ( id, '"' );
		if ( ! id2 )
		{
			break;
		}

		id2[0] = 0;

		if ( d2 > id2 )
		{
			st = d2 + 1;
		}
		else
		{
			st = id2 + 1;
		}

		res = ced_index_query ( idx_seats, 0, id, &index_item );
		if ( res < 0 )
		{
			goto fail;
		}

		if ( res == 0 )
		{
			if ( validate )
			{
				printf ( "ERROR eleana_polymap_load - not in "
					"TSV - '%s'\n", id );
			}

			not_found ++;

			continue;
		}
		else
		{
			// Seat matched so set the found flag
			index_item->row = 1;

			if ( parse_poly ( tree, index_item->col, d ) )
			{
				goto fail;
			}
		}
	}

	if ( validate )
	{
		if ( not_found )
		{
			printf ( "SVG items not_found = %i\n", not_found );
			mtkit_tree_scan ( idx_seats->tree,
				tree_scan_missing_cb, NULL, 0 );
		}
	}

	if ( ! tree->root )
	{
		// Nothing found

		goto fail;
	}

	free ( svg );
	ced_index_destroy ( idx_seats );

	return tree;

fail:
	free ( svg );
	ced_index_destroy ( idx_seats );
	mtkit_tree_destroy ( tree );

	return NULL;
}
