/*
	Copyright (C) 2009 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

/*

gcc -Wall -ggdb _mapgraded1.c poly.o eleana.o -lmtkit -lmtcelledit -lmtpixel

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mtcelledit.h>

#include "eleana.h"


enum
{
	BRUSH_OUTLINE,

	BRUSH_CON_A,
	BRUSH_CON_B,
	BRUSH_CON_C,

	BRUSH_LAB_A,
	BRUSH_LAB_B,
	BRUSH_LAB_C,

	BRUSH_LIB_A,
	BRUSH_LIB_B,
	BRUSH_LIB_C,

	BRUSH_OTH_A,
	BRUSH_OTH_B,
	BRUSH_OTH_C,

	BRUSH_TOTAL
};


static eleanaINDEX	*main_index = NULL;
static CedSheet		*sheet_full = NULL;
static int		record = 0, tot_seats, imw, imh,
			brush_pal[BRUSH_TOTAL] = { 0, 33,112,110,  16,14,68,  29,27,3,  12,10,8 };
static mtTree		*polymap = NULL;
static double		*seat_table = NULL, scale = 1.0;
static mtBrush		*brush[BRUSH_TOTAL] = { NULL };


static void print_version()
{
	printf( "Reads election index, and creates a map based on the record chosen.\n"
		"\n"
		"Usage:\n"
		"\n"
		"./a.out ARG1 ARG2 ARG3 ARG4\n"
		"\n"
		"ARG1 is the input TSV file (index).\n"
		"ARG2 is the record to use from the index.\n"
		"ARG3 is the filename of the output PNG file.\n"
		"ARG4 is the scale factor for the output image.\n"
		"\n"
		);
}

static void create_brushes( mtPalette *palette )
{
	int i;

	if ( !palette ) return;

	for ( i=0; i<BRUSH_TOTAL; i++ )
	{
		brush[i] = mtpixel_brush_new();
		mtpixel_brush_set_index( brush[i], palette, brush_pal[i], brush_pal[i], MTPIXEL_CHANNEL_IMAGE );

		if ( i>0 )
			mtpixel_brush_set_size( brush[i], 0 );	// Fill polygon

//printf("brush %i = %i\n", i, (int)brush[i]);
	}
}

static void destroy_brushes()
{
	int i;

	for ( i=0; i<BRUSH_TOTAL; i++ )
	{
		mtpixel_brush_destroy( brush[i] );
	}
}

static int render_callback( int seat, mtBrush **brush_fill, mtBrush **brush_outline, void *user_data )
{
	int shade = 0;
	double op;

	op = 100 * EL_TABLE_VALUE( seat_table, seat, EL_TAB_O ) / EL_TABLE_VALUE( seat_table, seat, EL_TAB_TURNOUT );

//printf("seat %i op=%f\n", seat, op);

	if ( op < 10 ) shade = 0;
	else if ( op < 20 ) shade = 1;
	else shade = 2;

	*brush_outline = brush[BRUSH_OUTLINE];
	*brush_fill = brush[BRUSH_OTH_A + shade];

	return 0;	// Always render
}

int main( int argc, char **argv ) 
{
	mtImage *image = NULL;


	if ( argc != 5 )
	{
		print_version();
		exit(0);
	}

	main_index = eleana_index_open( argv[1] );
	if ( !main_index )
	{
		fprintf( stderr, "Unable to load TSV index file %s\n", argv[1] );
		exit(1);
	}

	if ( mtkit_strtoi( argv[2], &record, NULL, 0 ) )
	{
		fprintf( stderr, "Unable to get integer from '%s'\n", argv[2] );
		goto fail;
	}

	if ( mtkit_strtod( argv[4], &scale, NULL, 0 ) || scale<0.3 || scale>30 )
	{
		fprintf( stderr, "Unable to get scale from '%s' - should be 0.3 to 30.0\n", argv[2] );
		goto fail;
	}

	imw = POLYMAP_WIDTH * scale + 1;
	imh = POLYMAP_HEIGHT * scale + 1;

	image = mtpixel_image_new_rgb( imw, imh );
	if ( !image )
	{
		fprintf( stderr, "Unable to create image %i x %i\n", imw, imh );
		goto fail;
	}

	if ( eleana_index_record_load( main_index, record, 0, &sheet_full, &polymap, &seat_table, &tot_seats, NULL, NULL ) )
	{
		fprintf( stderr, "Unable to load record %i\n", record );
		goto fail;
	}

	printf( "Index File      %s\n"
		"Record #        %i\n"
		"Record name     %s\n"
		"Seats           %i\n"
		"Scale           %f\n"
		"Image           %i x %i - %s\n",
		argv[1], record, eleana_index_get( main_index, record, ELEANA_INDEX_COL_YEAR ), tot_seats, scale, imw, imh, argv[3]
		);

	create_brushes( &image->palette );
	memset( image->channel[MTPIXEL_CHANNEL_IMAGE], 255, imw*imh*3 );		// White background

	if ( eleana_poly_render_custom(	image, 0, 0, polymap, seat_table, scale, render_callback, NULL ) )
	{
		fprintf( stderr, "Unable to render the map\n" );
		goto fail;
	}

	mtpixel_image_save( image, argv[3], MTPIXEL_FILE_TYPE_PNG, 5 );

	eleana_index_record_close( sheet_full, polymap, seat_table, NULL, NULL );
	eleana_index_close( main_index );
	mtpixel_image_destroy( image );
	destroy_brushes();

	return 0;

fail:
	eleana_index_record_close( sheet_full, polymap, seat_table, NULL, NULL );
	eleana_index_close( main_index );
	mtpixel_image_destroy( image );
	destroy_brushes();

	return 1;
}
