/*
	Copyright (C) 2009-2012 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mtcelledit.h>

#include "eleana.h"




static void print_version()
{
	printf( "places - Load an index and calculate party positions.\n"
		"\n"
		"Usage:\n"
		"\n"
		"./places FILE1 FILE2\n"
		"\n"
		"FILE1 is the input TSV index file.\n"
		"FILE2 is the output CED file.\n"
		"\n"
		);
}


#define TABRES_ROWS 4
#define TABRES_COLS 5

int cmp_vote_table( const void *a, const void *b )
{
	int	*ai = (int *)a, *bi = (int *)b;

	// Sort in ascending order
	if ( bi[1] < ai[1] ) return -1;
	if ( bi[1] > ai[1] ) return 1;

	return 0;
}

int main( int argc, char **argv )
{
	eleanaINDEX	*index;
	char		*filename;
	int		i, r, c, seats, tabres[TABRES_ROWS][TABRES_COLS],
			reslist[8];	// party ID, votes
	CedBook		*book;
	CedSheet	*full, *sheet_new;
	mtTree		*polymap;
	double		*table;

	if ( argc != 3 )
	{
		print_version();
		exit(0);
	}

	book = ced_book_new();

	filename = argv[1];

	index = eleana_index_open( filename );
	if ( !index )
	{
		fprintf( stderr, "Unable to load index file %s\n", filename );
		exit(1);
	}

	printf( "\nIndex loaded at %i.  Path = '%s'\n\n", (int)(intptr_t)index, index->path );

	for ( i=0; i<index->records; i++ )
	{
		if ( eleana_index_record_load( index, i, 1, &full,
			&polymap, &table, &seats, NULL, NULL ) )
		{
			printf( "ERROR on record %i\n", i );
			continue;
		}

		printf("Record %i '%s' = %i %i %i %i\n", i,
			eleana_index_get( index, i, ELEANA_INDEX_COL_YEAR),
			(int)(intptr_t)full, (int)(intptr_t)polymap, (int)(intptr_t)table,
			seats );

		sheet_new = ced_sheet_new();

		// Clear old results
		memset( tabres, 0, sizeof(tabres));

		for ( r=0; r<seats; r++ )
		{
			for ( c=0; c<4; c++ )
			{
				reslist[ 2*c ] = c;
				reslist[ 2*c + 1 ] = EL_TABLE_VALUE( table, r, c + EL_TAB_A );
			}

			// Sort the results to get placings
			qsort( reslist, 4, 2*sizeof(int), cmp_vote_table );

			for ( c=0; c<4; c++ )
			{
				if ( reslist[ 2*c + 1 ] == 0 )
				{
					tabres[ reslist[2*c] ][ 4 ] ++;
				}
				else
				{
					tabres[ reslist[2*c] ][ c ] ++;
				}
			}
		}

		for ( r=0; r<TABRES_ROWS; r++ )
		{
			for ( c=0; c<TABRES_COLS; c++ )
			{
				ced_sheet_set_cell_value( sheet_new,
					2 + r, 2 + c, tabres[r][c] );
			}
		}

		// Create the titles
		ced_sheet_set_cell( sheet_new, 1, 2, "1st" );
		ced_sheet_set_cell( sheet_new, 1, 3, "2nd" );
		ced_sheet_set_cell( sheet_new, 1, 4, "3rd" );
		ced_sheet_set_cell( sheet_new, 1, 5, "4th" );
		ced_sheet_set_cell( sheet_new, 1, 6, "0" );
		ced_sheet_set_cell( sheet_new, 1, 7, "Total" );

		ced_sheet_set_cell( sheet_new, 2, 1, "Con" );
		ced_sheet_set_cell( sheet_new, 3, 1, "Lab" );
		ced_sheet_set_cell( sheet_new, 4, 1, "Lib" );
		ced_sheet_set_cell( sheet_new, 5, 1, "Oth" );
		ced_sheet_set_cell( sheet_new, 6, 1, "Total" );

		ced_sheet_set_cell( sheet_new, 2, 7, "=sum( rc1:rc[-1] )" );
		ced_sheet_set_cell( sheet_new, 3, 7, "=sum( rc1:rc[-1] )" );
		ced_sheet_set_cell( sheet_new, 4, 7, "=sum( rc1:rc[-1] )" );
		ced_sheet_set_cell( sheet_new, 5, 7, "=sum( rc1:rc[-1] )" );
		ced_sheet_set_cell( sheet_new, 6, 7, "=sum( rc1:rc[-1] )" );

		ced_sheet_set_cell( sheet_new, 6, 2, "=sum( r1c:r[-1]c )" );
		ced_sheet_set_cell( sheet_new, 6, 3, "=sum( r1c:r[-1]c )" );
		ced_sheet_set_cell( sheet_new, 6, 4, "=sum( r1c:r[-1]c )" );
		ced_sheet_set_cell( sheet_new, 6, 5, "=sum( r1c:r[-1]c )" );
		ced_sheet_set_cell( sheet_new, 6, 6, "=sum( r1c:r[-1]c )" );

		ced_book_add_sheet( book, sheet_new,
			eleana_index_get( index, i, ELEANA_INDEX_COL_YEAR) );

		eleana_index_record_close( full, polymap, table, NULL, NULL );
	}

	eleana_index_close( index );

	ced_book_save( book, argv[2], CED_FILE_TYPE_TSV_BOOK );
	ced_book_destroy( book );

	return 0;
}

