/*
	Copyright (C) 2010-2012 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <mtcelledit.h>

#include "eleana.h"




static void print_version()
{
	printf( "scrape2010 - Scrape data out from the Electoral Commissions results for 2010.\n"
		"\n"
		"Usage:\n"
		"\n"
		"./scrape2010 FILE\n"
		"\n"
		"2010.tsv is then saved in the current directory\n"
		"\n"
		);
}


static char *get_first_nonspace( char *s )
{
	for ( ; *s; s++ )
	{
		if ( !isspace( *s ) ) break;
	}

	return s;
}

static int read_int( char *s, int *num, char **next )
{
	char	buf[16];
	int	i;


	while ( isspace( *s ) ) s++;		// Skip leading spaces

	// Copy all num characters (skipping any , ) into buf until space hit.
	for ( i=0; *s && i<10; s++ )
	{
		if ( isspace( *s ) ) break;
		if ( *s == ',' || *s == '.' ) continue;
		if ( !isdigit( *s ) ) break;
		buf[ i++ ] = *s;
	}

	if ( i<1 || i>10 ) return 1;

	buf[i] = 0;
	if ( mtkit_strtoi( buf, num, NULL, 0 ) ) return 1;

	*next = s;

printf( "read_int = %i\n", *num );

	return 0;		// Success
}

static int parse_stage3( char *s, CedSheet *sheet, int row )
{
	char	*t;
	int	num;


	s = get_first_nonspace( s );
	if ( *s == '-' ) return -1;	// STAGE 4 discovered

	t = strstr( s, "  " );
	if ( !t ) return 1;
	*t = 0;
	ced_sheet_set_cell_text( sheet, row, FULL_COL_MP_NAME, s );
	s = t + 2;
	s = get_first_nonspace( s );
	if ( !s ) return 1;

	t = strstr( s, "  " );
	if ( !t ) return 1;
	*t = 0;
	ced_sheet_set_cell_text( sheet, row, FULL_COL_PARTY, s );
	s = t + 2;
	s = get_first_nonspace( s );
	if ( !s ) return 1;

	if ( read_int( s, &num, &t ) ) return 1;
	ced_sheet_set_cell_value( sheet, row, FULL_COL_VOTES, (double)num );

	return 0;		// Success
}

/*
We are parsing data from these line structures:

1	consituency [num]
2	  num num name  party  num
3	  name  party  num
4	  ----
5	2005  num

There are 6 stages, with each stage having one line except stage 3 which can have any number of lines.
*/

static void parse_file( FILE *fp, CedSheet *sheet )
{
	int	i=1, j, row = FULL_ROW_PARTY1, num;
	char	*line, *s;


	while (1)
	{
		line = mtkit_file_readline( fp, NULL, NULL );
		if ( !line ) break;		// EOF

printf("LINE1='%s'\n", line);

		s = get_first_nonspace( line );
		if ( *s == 0 )
		{
			free( line );
			continue;	// Empty line
		}

printf("LINE2='%s'\n", s);

// STAGE 1
		s = strrchr( line+2, '[' );
		if ( !s ) break;
		s[-1] = 0;

		ced_sheet_set_cell_text( sheet, row, FULL_COL_SEAT_NAME, line );

		printf( "%4i  '%s'\n", i, line );
		free( line );

// STAGE 2
		line = mtkit_file_readline( fp, NULL, NULL );
		if ( !line ) break;		// Error
		s = get_first_nonspace( line );
		if ( *s == 0 || read_int( s, &num, &s ) )
		{
			free( line );
			break;		// Error
		}

		ced_sheet_set_cell_value( sheet, row, FULL_COL_ELECTORATE, (double)num );

		if ( read_int( s, &num, &s ) )
		{
			free( line );
			break;		// Error
		}

		if ( parse_stage3( s, sheet, row ) )
		{
			free( line );
			break;
		}

		row++;

		free( line );

// STAGE 3
		while (1)
		{
			line = mtkit_file_readline( fp, NULL, NULL );
			if ( !line ) return;		// EOF

			j = parse_stage3( line, sheet, row );
			free( line );
			row++;

			if ( j<0 )
			{
// STAGE 5
				line = mtkit_file_readline( fp, NULL, NULL );
				if ( !line ) return;		// EOF
				free( line );

				break;	// End of constituency
			}
			if ( j>0 ) return;	// Error
		}

		i++;
	}
}

int main( int argc, char **argv )
{
	FILE		*fp;
	CedSheet	*sheet;


	if ( argc != 2 || !(sheet = ced_sheet_new()) )
	{
		print_version();
		return 1;
	}

	fp = fopen( argv[1], "r" );

	if ( fp ) parse_file( fp, sheet );
	else
	{
		fprintf( stderr, "Unable to open file %s\n", argv[1] );
	}

	ced_sheet_save( sheet, "2010.tsv", CED_FILE_TYPE_TSV_CONTENT );
	ced_sheet_destroy( sheet );

	return 0;
}

