/*
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



#define GEXSDRAW_DATA_KEY	"mtGEX.sdraw_data"



static void scrollbar_moved (
	GtkAdjustment	* adjustment,
	gpointer	user_data
	)
{
	int		delta,
			val;
	gexSdraw	* sdraw;


	sdraw = (gexSdraw *)user_data;

	if ( ! sdraw )
	{
		return;
	}

	val = gtk_adjustment_get_value ( GTK_ADJUSTMENT ( adjustment ) );

	if ( adjustment == GTK_ADJUSTMENT ( sdraw->adj_vert ) )
	{
		delta = val - sdraw->y;
		sdraw->y = val;

		if ( delta != 0 )
		{
			if ( abs ( delta ) >=
				sdraw->drawing_area->allocation.height )
			{
				// Full update as scroll was larger than area
				// height

				gtk_widget_queue_draw ( sdraw->drawing_area );
			}
			else
			{
				// Partial update so scroll the area for
				// accelerated blit

				gdk_window_scroll ( sdraw->drawing_area->window,
					0, -delta );
			}
		}
	}
	else
	{
		// Horizontal

		delta = val - sdraw->x;
		sdraw->x = val;

		if ( delta != 0 )
		{
			if ( abs ( delta ) >=
				sdraw->drawing_area->allocation.width )
			{
				// Full update as scroll was larger than area
				// width

				gtk_widget_queue_draw ( sdraw->drawing_area );
			}
			else
			{
				// Partial update so scroll the area for
				// accelerated blit

				gdk_window_scroll ( sdraw->drawing_area->window,
					-delta, 0 );
			}
		}
	}
}

static gboolean canvas_configure (
	GtkWidget	* widget,
	GdkEventConfigure * event,
	gpointer	data
	)
{
	// Calling this we force a rechecking of the page width/height

	mtgex_sdraw_set ( (GtkWidget *)data, -1, -1, -1, -1 );

	return TRUE;
}

static gboolean scroll_wheel (
	GtkWidget	* widget,
	GdkEventScroll	* event,
	gexSdraw	* sdraw
	)
{
	double		val,
			page;
	GtkAdjustment	* adj;


	adj = GTK_ADJUSTMENT ( sdraw->adj_vert );
	val = adj->value;
	page = adj->page_size;

	switch ( event->direction )
	{
	case GDK_SCROLL_DOWN:
		val += page / 8;
		break;

	case GDK_SCROLL_UP:
		val -= page / 8;
		break;

	default:
		return FALSE;
	}

	if ( ( val + page ) > adj->upper )
	{
		val = adj->upper - page;
	}

	gtk_adjustment_set_value ( adj, val );

	return TRUE;
}

static gboolean destroy_sdraw (
	GtkWidget	* widget,
	GdkEvent	* event,
	gpointer	user_data
	)
{
	gexSdraw	* sdraw;


	if ( mtgex_sdraw_get ( widget, &sdraw ) )
	{
		fprintf ( stderr,
			"mtGEX error: destroy_sdraw unable to find sdraw\n" );

		return FALSE;
	}

	if ( ! sdraw )
	{
		return FALSE;
	}

/*
	Stop all callbacks to avoid any dangling references to to destroyed
	sdraw canvas_configure () callback relies on GEXSDRAW_DATA_KEY being
	available so never dangles
*/

	g_signal_handlers_disconnect_by_func ( sdraw->adj_vert,
		G_CALLBACK ( scrollbar_moved ), sdraw );
	g_signal_handlers_disconnect_by_func ( sdraw->adj_hori,
		G_CALLBACK ( scrollbar_moved ), sdraw );

	gtk_object_remove_data ( GTK_OBJECT ( sdraw->table ),
		GEXSDRAW_DATA_KEY );
	free ( sdraw );

	return FALSE;
}

GtkWidget * mtgex_sdraw_new ( void )
{
	gexSdraw	* sdraw;
	GtkWidget	* table,
			* area,
			* scroll;
	GtkObject	* adj;


	sdraw = calloc ( sizeof ( gexSdraw ), 1 );
	if ( ! sdraw )
	{
		return NULL;
	}

	table = gtk_table_new ( 2, 2, FALSE );
	gtk_object_set_data ( GTK_OBJECT ( table ), GEXSDRAW_DATA_KEY, sdraw );
	gtk_widget_show ( table );
	sdraw->table = table;

	gtk_signal_connect ( GTK_OBJECT ( table ), "destroy",
		GTK_SIGNAL_FUNC ( destroy_sdraw ), NULL );

	area = gtk_drawing_area_new ();
	mtgex_fix_darea ( area );
	gtk_widget_show ( area );
	g_signal_connect ( area, "configure_event",
		G_CALLBACK ( canvas_configure ), (gpointer)table );
	g_signal_connect ( area, "scroll_event",
		G_CALLBACK ( scroll_wheel ), sdraw );
	sdraw->drawing_area = area;
	gtk_table_attach ( GTK_TABLE ( table ), area, 0, 0 + 1, 0, 0 + 1,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0 );

	adj = gtk_adjustment_new ( 0, 0, 1, 1, 1, 1 );
	g_signal_connect ( adj, "value-changed",
		G_CALLBACK ( scrollbar_moved ), (gpointer)sdraw );

	scroll = gtk_vscrollbar_new ( GTK_ADJUSTMENT ( adj ) );
	gtk_widget_show ( scroll );
	gtk_table_attach ( GTK_TABLE ( table ), scroll, 1, 1 + 1, 0, 0 + 1,
		GTK_FILL, GTK_FILL, 0, 0 );
	sdraw->adj_vert = adj;
	sdraw->vscroll = scroll;

	adj = gtk_adjustment_new ( 0, 0, 1, 1, 1, 1 );
	g_signal_connect ( adj, "value-changed",
		G_CALLBACK ( scrollbar_moved ), (gpointer)sdraw );

	scroll = gtk_hscrollbar_new ( GTK_ADJUSTMENT ( adj ) );
	gtk_widget_show ( scroll );
	gtk_table_attach ( GTK_TABLE ( table ), scroll, 0, 0 + 1, 1, 1 + 1,
		GTK_FILL, GTK_FILL, 0, 0 );
	sdraw->adj_hori = adj;
	sdraw->hscroll = scroll;

	return table;
}

int mtgex_sdraw_set (
	GtkWidget	* widget,
	int		x,
	int		y,
	int		w,
	int		h
	)
{
	int		nx,
			ny,
			nw,
			nh,
			nhpage,
			nvpage;
	gexSdraw	* sdraw;


	if ( mtgex_sdraw_get ( widget, &sdraw ) || ! sdraw )
	{
		return 1;
	}

	// Block scrollbar callbacks
	g_signal_handlers_block_by_func ( sdraw->adj_vert,
		G_CALLBACK ( scrollbar_moved ), sdraw );
	g_signal_handlers_block_by_func ( sdraw->adj_hori,
		G_CALLBACK ( scrollbar_moved ), sdraw );

	// Set as current values and then change if requested
	nx = sdraw->x;
	ny = sdraw->y;
	nw = sdraw->w;
	nh = sdraw->h;
	nhpage = sdraw->drawing_area->allocation.width;
	nvpage = sdraw->drawing_area->allocation.height;

	if ( x >= 0 ) nx = x;
	if ( y >= 0 ) ny = y;

	if ( w > 0 ) nw = w;
	if ( h > 0 ) nh = h;

	// Ensure x/y/w/h are sensible for current visible state (change if not)
	if ( nw < 1 ) nw = 1;
	if ( nh < 1 ) nw = 1;

	if ( nw < nhpage ) nhpage = nw; // If area is smaller than visible
	if ( nh < nvpage ) nvpage = nh;

	if ( (nx + nhpage) > nw ) nx = nw - nhpage;
	if ( (ny + nvpage) > nh ) ny = nh - nvpage;

	sdraw->x = nx;
	sdraw->y = ny;
	sdraw->w = nw;
	sdraw->h = nh;

	GTK_ADJUSTMENT ( sdraw->adj_hori )->upper = nw;
	GTK_ADJUSTMENT ( sdraw->adj_hori )->page_size = nhpage;
	GTK_ADJUSTMENT ( sdraw->adj_hori )->page_increment = nhpage;
	gtk_adjustment_changed ( GTK_ADJUSTMENT ( sdraw->adj_hori ) );
	gtk_adjustment_set_value ( GTK_ADJUSTMENT ( sdraw->adj_hori ), nx );

	GTK_ADJUSTMENT ( sdraw->adj_vert )->upper = nh;
	GTK_ADJUSTMENT ( sdraw->adj_vert )->page_size = nvpage;
	GTK_ADJUSTMENT ( sdraw->adj_vert )->page_increment = nvpage;
	gtk_adjustment_changed ( GTK_ADJUSTMENT ( sdraw->adj_vert ) );
	gtk_adjustment_set_value ( GTK_ADJUSTMENT ( sdraw->adj_vert ), ny );

// FIXME - this is lazy because we don't always need to redraw everything
// (e.g. x/y stay the same, but w/h are reduced)

	gtk_widget_queue_draw ( sdraw->drawing_area );

	// Unblock scrollbar callbacks
	g_signal_handlers_unblock_by_func ( sdraw->adj_vert,
		G_CALLBACK ( scrollbar_moved ), sdraw );
	g_signal_handlers_unblock_by_func ( sdraw->adj_hori,
		G_CALLBACK ( scrollbar_moved ), sdraw );

	return 0;
}

int mtgex_sdraw_get (
	GtkWidget	* widget,
	gexSdraw	** gexsdraw
	)
{
	if ( ! widget || ! gexsdraw )
	{
		return 1;
	}

	gexsdraw[0] = gtk_object_get_data ( GTK_OBJECT ( widget ),
		GEXSDRAW_DATA_KEY );

	return 0;
}

