/*
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "jigsaw.h"



static void flush_piece_array (
	mtJigsaw	* jigsaw
	)
{
	int		i,
			j;


	if ( ! jigsaw->piece_array )
	{
		return;
	}

	j = ( jigsaw->nodes_hori - 1 ) * ( jigsaw->nodes_vert - 1 );

	for ( i = 0; i < j; i++ )
	{
		mtpixel_image_destroy ( jigsaw->piece_array[i].image );
		jigsaw->piece_array[i].image = NULL;
		jigsaw->piece_array[i].ID = i;
	}
}

static int create_piece_array (
	mtJigsaw	* jigsaw,
	int		x,
	int		y
	)
{
	int		i,
			tot;


	if ( jigsaw->piece_array )
	{
		flush_piece_array ( jigsaw );
		free ( jigsaw->piece_array );
		jigsaw->piece_array = NULL;
	}

	jigsaw->nodes_hori = x;
	jigsaw->nodes_vert = y;
	tot = (x - 1) * (y - 1);

	jigsaw->piece_array = calloc ( sizeof ( mtJigsawPiece ), tot );
	if ( ! jigsaw->piece_array )
	{
		return 1;
	}

	for ( i = 0; i < tot; i++ )
	{
		jigsaw->piece_array[i].ID = i;
	}

	return 0;
}

static int find_piece_ID (
	mtJigsaw	* jigsaw,
	int		id
	)
{
	int		i,
			j;


	if ( jigsaw->piece_array[id].ID == id )
	{
		return id;		// Simple/Quick
	}

	j = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);

	for ( i = 0; i < j; i++ )	// Slow find
	{
		if ( jigsaw->piece_array[i].ID == id )
		{
			return i;
		}
	}

	return id;			// Fallback
}

static int cutup_jigsaw (
	mtJigsaw	* jigsaw,
	int		place_pieces
	)
{
	double		a,
			b,
			c,
			d,
			e,
			f,
			g,
			td;
	int		i,
			j,
			k,
			kk,
			o,
			o1,
			o2,
			o3,
			o4,
			x1,
			y1,
			x2,
			y2,
			x3,
			y3,
			x4,
			y4;
	mtImage		* new_scaled = NULL,
			* ti;
	mtBrush		* brush = NULL;
	mtPolygon	* poly;


	if (	! jigsaw->image					||
		! jigsaw->piece_array				||
		jigsaw->table_width < MTJIGSAW_TABLE_SIZE_MIN	||
		jigsaw->table_height < MTJIGSAW_TABLE_SIZE_MIN
		)
	{
		return 1;		// Structure not ready for cutting up
	}

	// Do calculations to get required size
	a = jigsaw->table_width;
	b = jigsaw->table_height;
	e = jigsaw->image->width;
	f = jigsaw->image->height;

	g = sqrt ( ( a * b ) / ( 3 * e * f ) );
	c = g * e;
	d = g * f;

	// Ensure that the final rectangle contains the whole of scaled
	// rectangle
	if ( c > a )
	{
		g = a / e;
		c = a;
		d = g * f;
	}
	else if ( d > b )
	{
		g = b / f;
		c = g * e;
		d = b;
	}

	// Rescale image
	new_scaled = mtpixel_image_scale ( jigsaw->image,
		MTPIXEL_SCALE_BILINEAR, rint ( c ), rint ( d ), 0, 1 );

	if ( ! new_scaled )
	{
		fprintf ( stderr, "ERROR - Unable to rescla image\n" );

		return 1;
	}

	jigsaw->xo  = (jigsaw->table_width - new_scaled->width) / 2;
	jigsaw->yo  = (jigsaw->table_height - new_scaled->height) / 2;
	jigsaw->vmx = new_scaled->width - 1;
	jigsaw->vmy = new_scaled->height - 1;

	jigsaw->vmx /= MTJIGSAW_NODE_COORD_MAX;
	jigsaw->vmy /= MTJIGSAW_NODE_COORD_MAX;

	mtpixel_image_destroy ( jigsaw->scaled_image );
	jigsaw->scaled_image = new_scaled;
	new_scaled = NULL;

	if ( place_pieces )
	{
		flush_piece_array ( jigsaw );
	}

	brush = mtpixel_brush_new ();

	// Render row lines
	for ( j = 0; j < jigsaw->nodes_vert; j++ )
	{
		for ( i = 0; i < (jigsaw->nodes_hori - 1); i++ )
		{
			o = i + j * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x1 = rint ( td * jigsaw->mesh_x[o] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y1 = rint ( td * jigsaw->mesh_y[o] /
				MTJIGSAW_NODE_COORD_MAX );

			o++;
			td = jigsaw->scaled_image->width - 1;
			x2 = rint ( td * jigsaw->mesh_x[o] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y2 = rint ( td * jigsaw->mesh_y[o] /
				MTJIGSAW_NODE_COORD_MAX );

			mtpixel_brush_line ( jigsaw->scaled_image, brush, x1,
				y1, x2, y2 );
		}
	}

	// Render col lines
	for ( i = 0; i < jigsaw->nodes_hori; i++ )
	{
		for ( j = 0; j < (jigsaw->nodes_vert - 1); j++ )
		{
			o = i + j * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x1 = rint ( td * jigsaw->mesh_x[o] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y1 = rint ( td * jigsaw->mesh_y[o] /
				MTJIGSAW_NODE_COORD_MAX );

			o += jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x2 = rint ( td * jigsaw->mesh_x[o] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y2 = rint ( td * jigsaw->mesh_y[o] /
				MTJIGSAW_NODE_COORD_MAX );

			mtpixel_brush_line ( jigsaw->scaled_image, brush,
				x1, y1, x2, y2 );
		}
	}

	mtpixel_brush_destroy ( brush );
	brush = NULL;

	// Cut up pieces
	poly = mtpixel_polygon_new ();

	kk = 0;
	for ( j = 0; j < (jigsaw->nodes_vert - 1); j++ )
	{
		for ( i = 0; i < (jigsaw->nodes_hori - 1); i++, kk++ )
		{
			mtpixel_polygon_clear ( poly );

			o1 = i + j * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x1 = rint ( td * jigsaw->mesh_x[o1] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y1 = rint ( td * jigsaw->mesh_y[o1] /
				MTJIGSAW_NODE_COORD_MAX );
			mtpixel_polygon_add_point ( poly, x1, y1 );

			o2 = (i + 1) + j * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x2 = rint ( td * jigsaw->mesh_x[o2] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y2 = rint ( td * jigsaw->mesh_y[o2] /
				MTJIGSAW_NODE_COORD_MAX );
			mtpixel_polygon_add_point ( poly, x2, y2 );

			o3 = (i + 1) + (j + 1) * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x3 = rint ( td * jigsaw->mesh_x[o3] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y3 = rint ( td * jigsaw->mesh_y[o3] /
				MTJIGSAW_NODE_COORD_MAX );
			mtpixel_polygon_add_point ( poly, x3, y3 );

			o4 = i + (j + 1) * jigsaw->nodes_hori;
			td = jigsaw->scaled_image->width - 1;
			x4 = rint ( td * jigsaw->mesh_x[o4] /
				MTJIGSAW_NODE_COORD_MAX );
			td = jigsaw->scaled_image->height - 1;
			y4 = rint ( td * jigsaw->mesh_y[o4] /
				MTJIGSAW_NODE_COORD_MAX );
			mtpixel_polygon_add_point ( poly, x4, y4 );

			ti = mtpixel_image_new_from_polygon (
				jigsaw->scaled_image, poly );

			if ( ! ti )
			{
				goto fail;
			}

			k = find_piece_ID ( jigsaw, kk );
			jigsaw->piece_array[k].image = ti;

			if ( place_pieces )
			{
				// Smallest X always on left side
				if ( jigsaw->mesh_x[o1] < jigsaw->mesh_x[o4] )
				{
					jigsaw->piece_array[k].tx =
					jigsaw->piece_array[k].ix =
						jigsaw->mesh_x[o1];
				}
				else
				{
					jigsaw->piece_array[k].tx =
					jigsaw->piece_array[k].ix =
						jigsaw->mesh_x[o4];
				}

				// Smallest Y always on top side
				if ( jigsaw->mesh_y[o1] < jigsaw->mesh_y[o2] )
				{
					jigsaw->piece_array[k].ty =
					jigsaw->piece_array[k].iy =
						jigsaw->mesh_y[o1];
				}
				else
				{
					jigsaw->piece_array[k].ty =
					jigsaw->piece_array[k].iy =
						jigsaw->mesh_y[o2];
				}
			}
		}
	}

	mtpixel_polygon_destroy ( poly );

	return 0;			// Success

fail:
	flush_piece_array ( jigsaw );

	return 1;
}

static mtImage * mtjigsaw_image_load (
	char		* filename
	)
{
	mtImage		* i;


	i = mtpixel_image_load ( filename );
	if ( ! i )
	{
		return NULL;
	}

	mtpixel_channel_destroy ( i, MTPIXEL_CHANNEL_ALPHA );

	if ( i->type != MTPIXEL_IMAGE_RGB )
	{
		mtImage		* j;


		j = mtpixel_image_to_rgb ( i );
		mtpixel_image_destroy ( i );

		if ( ! j )
		{
			return NULL;
		}

		i = j;
	}

	return i;
}

mtJigsaw * mtjigsaw_load (
	char		* filename
	)
{
	int		tot,
			i,
			j,
			k,
			iv[4];
	double		d;
	mtJigsaw	* new_jigsaw;
	CedSheet	* sheet = NULL;
	CedCell		* cell;


	new_jigsaw = calloc ( sizeof ( mtJigsaw ), 1 );
	if ( ! new_jigsaw )
	{
		return NULL;
	}

	if ( filename )
	{
		sheet = ced_sheet_load ( filename, NULL, NULL );
		if ( ! sheet )
		{
			goto fail;
		}

		cell = ced_sheet_get_cell ( sheet, 1, 1 );
		if ( cell )
		{
			new_jigsaw->image = mtjigsaw_image_load ( cell->text );
			if ( new_jigsaw->image )
			{
				new_jigsaw->image_filename =
					strdup ( cell->text );

				if ( ! new_jigsaw->image_filename )
				{
					goto fail;
				}
			}
		}

		cell = ced_sheet_get_cell ( sheet, 2, 1 );

		if (	! cell					||
			cell->value < MTJIGSAW_NODES_MIN	||
			cell->value > MTJIGSAW_NODES_MAX
			)
		{
			goto fail;
		}

		new_jigsaw->nodes_hori = cell->value;

		cell = ced_sheet_get_cell ( sheet, 2, 2 );

		if (	! cell					||
			cell->value < MTJIGSAW_NODES_MIN	||
			cell->value > MTJIGSAW_NODES_MAX
			)
		{
			goto fail;
		}

		new_jigsaw->nodes_vert = cell->value;

		tot = new_jigsaw->nodes_hori * new_jigsaw->nodes_vert;
		new_jigsaw->mesh_x = calloc ( sizeof ( int ), tot );
		new_jigsaw->mesh_y = calloc ( sizeof ( int ), tot );

		if ( create_piece_array ( new_jigsaw, new_jigsaw->nodes_hori,
			new_jigsaw->nodes_vert ) )
		{
			goto fail;
		}

		for ( j = 0; j < new_jigsaw->nodes_vert; j++ )
		{
			for ( i = 0; i < new_jigsaw->nodes_hori; i++ )
			{
				cell = ced_sheet_get_cell ( sheet, 3 + j,
					1 + i );

				if ( ! cell )
				{
					goto fail;
				}

				if ( mtkit_strtok_num ( cell->text, ",", 0, &d )
					)
				{
					goto fail;
				}

				new_jigsaw->mesh_x[i +
					j * new_jigsaw->nodes_hori]
					= d;

				if ( mtkit_strtok_num ( cell->text, ",", 1, &d )
					)
				{
					goto fail;
				}

				new_jigsaw->mesh_y[i +
					j * new_jigsaw->nodes_hori]
					= d;

				if (	new_jigsaw->mesh_x[i + j *
						new_jigsaw->nodes_hori] < 0 ||
					new_jigsaw->mesh_x[i + j *
						new_jigsaw->nodes_hori] >
						MTJIGSAW_NODE_COORD_MAX ||
					new_jigsaw->mesh_y[i + j *
						new_jigsaw->nodes_hori] < 0 ||
					new_jigsaw->mesh_y[i + j *
						new_jigsaw->nodes_hori] >
						MTJIGSAW_NODE_COORD_MAX
					)
				{
					goto fail;
				}
			}
		}

		j += 3;
		tot =	(new_jigsaw->nodes_hori - 1) *
			(new_jigsaw->nodes_vert - 1);

		for ( i = 0; i < tot; i++, j++ )
		{
			for ( k = 0; k < 4; k++ )
			{
				cell = ced_sheet_get_cell ( sheet, j, k + 1 );

				if ( ! cell )
				{
					break;
				}

				iv[k] = cell->value;
			}
			if ( k<4 )
			{
				break;
			}

			new_jigsaw->piece_array[i].tx = iv[0];
			new_jigsaw->piece_array[i].ty = iv[1];
			new_jigsaw->piece_array[i].ix = iv[2];
			new_jigsaw->piece_array[i].iy = iv[3];
		}

		ced_sheet_destroy ( sheet );
	}

	return new_jigsaw;

fail:
	ced_sheet_destroy ( sheet );
	mtjigsaw_destroy ( new_jigsaw );

	return NULL;
}

int mtjigsaw_save (
	mtJigsaw	* jigsaw,
	char		* filename
	)
{
	int		res,
			i,
			j,
			tot;
	char		txt[64];
	CedSheet	* sheet;


	if ( ! jigsaw || ! filename )
	{
		return 1;
	}

	sheet = ced_sheet_new ();
	if ( ! sheet )
	{
		return 1;
	}

	ced_sheet_set_cell_text ( sheet, 1, 1, jigsaw->image_filename );
	ced_sheet_set_cell_value ( sheet, 2, 1, jigsaw->nodes_hori );
	ced_sheet_set_cell_value ( sheet, 2, 2, jigsaw->nodes_vert );

	for ( j = 0; j < jigsaw->nodes_vert; j++ )
	{
		for ( i = 0; i < jigsaw->nodes_hori; i++ )
		{
			sprintf ( txt, "%i,%i",
				jigsaw->mesh_x[i + j * jigsaw->nodes_hori],
				jigsaw->mesh_y[i + j * jigsaw->nodes_hori] );
			ced_sheet_set_cell_text ( sheet, j + 3, i + 1, txt );
		}
	}

	j += 3;
	tot = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);

	if ( jigsaw->piece_array && jigsaw->scaled_image )
	{
		for ( i = 0; i < tot; i++ )
		{
			ced_sheet_set_cell_value ( sheet,
				j + jigsaw->piece_array[i].ID, 1,
				jigsaw->piece_array[i].tx );

			ced_sheet_set_cell_value ( sheet,
				j + jigsaw->piece_array[i].ID, 2,
				jigsaw->piece_array[i].ty );

			ced_sheet_set_cell_value ( sheet,
				j + jigsaw->piece_array[i].ID, 3,
				jigsaw->piece_array[i].ix );

			ced_sheet_set_cell_value ( sheet,
				j + jigsaw->piece_array[i].ID, 4,
				jigsaw->piece_array[i].iy );
		}
	}

	res = ced_sheet_save ( sheet, filename, CED_FILE_TYPE_TSV_CONTENT );
	ced_sheet_destroy ( sheet );

	return res;			// Success
}

static int mtjigsaw_flush (
	mtJigsaw	* jigsaw
	)
{
	if ( ! jigsaw )
	{
		return 1;
	}

	free ( jigsaw->image_filename );
	mtpixel_image_destroy ( jigsaw->image );
	mtpixel_image_destroy ( jigsaw->scaled_image );
	free ( jigsaw->mesh_x );
	free ( jigsaw->mesh_y );
	flush_piece_array ( jigsaw );
	free ( jigsaw->piece_array );

	memset ( jigsaw, 0, sizeof ( mtJigsaw ) );

	return 0;
}

int mtjigsaw_destroy (
	mtJigsaw	* jigsaw
	)
{
	if ( ! jigsaw )
	{
		return 1;
	}

	mtjigsaw_flush ( jigsaw );

	free ( jigsaw );

	return 0;			// Success
}

// Ensure that if any pieces exists they must completely lie within visible area
static void validate_piece_positions (
	mtJigsaw	* jigsaw
	)
{
	int		i,
			j,
			min_xv,
			min_yv,
			max_xv,
			max_yv,
			pxvs,
			pyvs;


	if (	! jigsaw->piece_array ||
		jigsaw->table_width<MTJIGSAW_TABLE_SIZE_MIN
		)
	{
		return;		// Nothing to validate
	}

	// Set up max/min vectors for this table
	min_xv = rint ( ( (double)( -(jigsaw->xo) ) ) / jigsaw->vmx );
	min_yv = rint ( ( (double)( -(jigsaw->yo) ) ) / jigsaw->vmy );
	max_xv = rint ( ( (double)( jigsaw->table_width  - 1 - jigsaw->xo ) ) /
		jigsaw->vmx );
	max_yv = rint ( ( (double)( jigsaw->table_height - 1 - jigsaw->yo ) ) /
		jigsaw->vmy );

	j = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);

	for ( i = 0; i < j; i++ )
	{
		if ( ! jigsaw->piece_array[i].image )
		{
			continue;
		}

		// Get piece vector size
		pxvs = rint ( ( (double)jigsaw->piece_array[i].image->width )  /
			jigsaw->vmx );
		pyvs = rint ( ( (double)jigsaw->piece_array[i].image->height ) /
			jigsaw->vmy );

		// Ensure within limits
		jigsaw->piece_array[i].tx = MIN ( jigsaw->piece_array[i].tx,
			max_xv - pxvs );
		jigsaw->piece_array[i].ty = MIN ( jigsaw->piece_array[i].ty,
			max_yv - pyvs );

		jigsaw->piece_array[i].tx = MAX ( jigsaw->piece_array[i].tx,
			min_xv );
		jigsaw->piece_array[i].ty = MAX ( jigsaw->piece_array[i].ty,
			min_yv );
	}
}

int mtjigsaw_set_image (
	mtJigsaw	* jigsaw,
	char		* filename,
	int		width,
	int		height
	)
{
	if ( ! jigsaw )
	{
		return 1;
	}

	if (	width < MTJIGSAW_TABLE_SIZE_MIN ||
		height < MTJIGSAW_TABLE_SIZE_MIN
		)
	{
		return 1;
	}

	if ( filename )
	{
		mtImage		* new_image = NULL;


		new_image = mtjigsaw_image_load ( filename );
		if ( ! new_image )
		{
			// No structure flush required (no changes made yet)
			return 1;
		}

		mtpixel_image_destroy ( jigsaw->image );
		jigsaw->image = new_image;

		free ( jigsaw->image_filename );
		jigsaw->image_filename = strdup ( filename );

		if ( ! jigsaw->image_filename )
		{
			// Structure flush required as we have altered the image
			goto fail;
		}
	}

	jigsaw->table_width = width;
	jigsaw->table_height = height;

	if ( cutup_jigsaw ( jigsaw, filename ? 1 : 0 ) )
	{
		// Structure flush required as we have altered the table size
		goto fail;
	}

	// Ensure that if any pieces exists they must completely lie within
	// visible area
	validate_piece_positions ( jigsaw );

	return 0;			// Success

fail:

	return 1;
}

int mtjigsaw_set_mesh (
	mtJigsaw	* jigsaw,
	int		x_nodes,
	int		y_nodes
	)
{
	int		i,
			j,
			xvar,
			yvar,
			* new_mesh_x = NULL,
			* new_mesh_y = NULL,
			tot_nodes,
			rx,
			ry,
			x,
			y;


	if ( ! jigsaw )
	{
		return 1;
	}

	if (	x_nodes < MTJIGSAW_NODES_MIN ||
		x_nodes > MTJIGSAW_NODES_MAX ||
		y_nodes < MTJIGSAW_NODES_MIN ||
		y_nodes > MTJIGSAW_NODES_MAX
		)
	{
		return 1;
	}

	tot_nodes = x_nodes * y_nodes;

	new_mesh_x = calloc ( sizeof ( int ), tot_nodes );
	new_mesh_y = calloc ( sizeof ( int ), tot_nodes );

	if ( ! new_mesh_x || ! new_mesh_y )
	{
		free ( new_mesh_x );
		free ( new_mesh_y );

		return 1;
	}

	xvar = MTJIGSAW_NODE_COORD_MAX / (x_nodes - 1) / 3;
	yvar = MTJIGSAW_NODE_COORD_MAX / (y_nodes - 1) / 3;

	for ( j = 0; j < y_nodes; j++ )
	{
		// Set up uniform coordinate
		y = MTJIGSAW_NODE_COORD_MAX * j / (y_nodes - 1);

		for ( i = 0; i < x_nodes; i++ )
		{
			// Set up uniform coordinate
			x = MTJIGSAW_NODE_COORD_MAX * i / (x_nodes - 1);

			// Random variation from uniform coordinate
			rx = 0;
			ry = 0;

			if ( i > 0 && i < (x_nodes - 1) )
			{
				rx = ( rand () % (2 * xvar + 1) ) - xvar;
			}

			if ( j > 0 && j < (y_nodes - 1) )
			{
				ry = ( rand () % (2 * yvar + 1) ) - yvar;
			}

			new_mesh_x[i + j * x_nodes] = x + rx;
			new_mesh_y[i + j * x_nodes] = y + ry;
		}
	}

	// Remove old stuff, put new stuff in
	free ( jigsaw->mesh_x );
	free ( jigsaw->mesh_y );

	jigsaw->mesh_x = new_mesh_x;
	jigsaw->mesh_y = new_mesh_y;

	if ( create_piece_array ( jigsaw, x_nodes, y_nodes ) )
	{
		return 1;
	}

	if ( cutup_jigsaw ( jigsaw, 1 ) )
	{
		return 1;
	}

	return 0;			// Success
}

int mtjigsaw_scatter (
	mtJigsaw	* jigsaw
	)
{
	int		i,
			j,
			x,
			y;


	if (	! jigsaw		||
		! jigsaw->piece_array	||
		! jigsaw->table_width
		)
	{
		return 1;
	}

	j = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);

	for ( i = 0; i < j; i++ )
	{
		if ( ! jigsaw->piece_array[i].image )
		{
			continue;
		}

		x = rand () % ( jigsaw->table_width  -
			jigsaw->piece_array[i].image->width );
		y = rand () % ( jigsaw->table_height -
			jigsaw->piece_array[i].image->height );

		jigsaw->piece_array[i].tx = ( (double)x - jigsaw->xo )  /
			jigsaw->vmx;
		jigsaw->piece_array[i].ty = ( (double)y - jigsaw->yo )  /
			jigsaw->vmy;
	}

	return 0;			// Success
}

int mtjigsaw_join (
	mtJigsaw	* jigsaw
	)
{
	int		i,
			j;


	if (	! jigsaw		||
		! jigsaw->piece_array	||
		! jigsaw->table_width
		)
	{
		return 1;
	}

	j = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);

	for ( i = 0; i < j; i++ )
	{
		jigsaw->piece_array[i].tx = jigsaw->piece_array[i].ix;
		jigsaw->piece_array[i].ty = jigsaw->piece_array[i].iy;
	}

	return 0;			// Success
}

int mtjigsaw_render (
	mtJigsaw	* jigsaw,
	mtImage		* image,
	int		sx,
	int		sy
	)
{
	mtBrush		* brush = NULL;


	if (	! jigsaw		||
		! image			||
		image->type != MTPIXEL_IMAGE_RGB
		)
	{
		return 1;
	}

	memset ( image->channel[MTPIXEL_CHANNEL_IMAGE], 255,
		3 * image->width * image->height );

	if ( jigsaw->scaled_image && jigsaw->piece_array )
	{
		int		x,
				y,
				i,
				j;


		brush = mtpixel_brush_new ();
		if ( ! brush )
		{
			return 0;
		}

		x = jigsaw->xo - sx;
		y = jigsaw->yo - sy;

		mtpixel_brush_rectangle ( image, brush, x, y,
			x + jigsaw->scaled_image->width - 1,
			y + jigsaw->scaled_image->height - 1 );

		// Render pieces
		j = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1);
		for ( i = 0; i < j; i++ )
		{
			x = rint ( jigsaw->piece_array[i].tx * jigsaw->vmx) +
				jigsaw->xo;
			y = rint ( jigsaw->piece_array[i].ty * jigsaw->vmy) +
				jigsaw->yo;

			mtpixel_image_paste ( image,
				jigsaw->piece_array[i].image,
				brush, x - sx, y - sy );
		}

		mtpixel_brush_destroy ( brush );
	}

	return 0;			// Success
}

int mtjigsaw_drag_piece (
	mtJigsaw	* jigsaw,
	int		piece_no,
	int		dx,
	int		dy,
	int		* txu,
	int		* tyu,
	int		* twu,
	int		* thu
	)
{
	int		tx1,
			ty1,
			tx2,
			ty2;


	if ( ! jigsaw )
	{
		return 1;
	}

	tx1 = rint ( jigsaw->piece_array[piece_no].tx * jigsaw->vmx ) +
		jigsaw->xo;
	ty1 = rint ( jigsaw->piece_array[piece_no].ty * jigsaw->vmy ) +
		jigsaw->yo;

	tx2 = tx1 + dx;
	ty2 = ty1 + dy;

	if ( tx2 < 0 )
	{
		tx2 = 0;
	}

	if ( ty2 < 0 )
	{
		ty2 = 0;
	}

	if ( (tx2 + jigsaw->piece_array[piece_no].image->width) >
		jigsaw->table_width
		)
	{
		tx2 = jigsaw->table_width -
			jigsaw->piece_array[piece_no].image->width;
	}

	if ( (ty2 + jigsaw->piece_array[piece_no].image->height) >
		jigsaw->table_height
		)
	{
		ty2 = jigsaw->table_height -
			jigsaw->piece_array[piece_no].image->height;
	}

	// Obtain REAL change after boundary checking
	dx = tx2 - tx1;
	dy = ty2 - ty1;

	if ( dx == 0 && dy == 0 )
	{
		return 1;		// Nothing changed
	}

	// Calculate new vector position for this piece
	jigsaw->piece_array[piece_no].tx = rint ( ( (double)(tx2 - jigsaw->xo) )
		/ jigsaw->vmx );
	jigsaw->piece_array[piece_no].ty = rint ( ( (double)(ty2 - jigsaw->yo) )
		/ jigsaw->vmy );

	// Find update rectangle for the pixel table
	txu[0] = MIN ( tx1, tx2 );
	tyu[0] = MIN ( ty1, ty2 );
	twu[0] = jigsaw->piece_array[piece_no].image->width  +
		abs ( tx2 - tx1 );
	thu[0] = jigsaw->piece_array[piece_no].image->height +
		abs ( ty2 - ty1 );

	return 0;			// Success
}

int mtjigsaw_drop_piece (
	mtJigsaw	* jigsaw,
	int		piece_no,
	int		* txu,
	int		* tyu,
	int		* twu,
	int		* thu
	)
{
	int		snap_dx,
			snap_dy,
			tx1,
			ty1,
			tx2,
			ty2;


	if ( ! jigsaw )
	{
		return 1;
	}

	// Snap distance is 1/4 of the uniform piece width/height
	snap_dx = MTJIGSAW_NODE_COORD_MAX / (jigsaw->nodes_hori - 1);
	snap_dy = MTJIGSAW_NODE_COORD_MAX / (jigsaw->nodes_vert - 1);

	if ( abs ( jigsaw->piece_array[piece_no].tx -
		jigsaw->piece_array[piece_no].ix ) > snap_dx
		)
	{
		return 0;
	}

	if ( abs ( jigsaw->piece_array[piece_no].ty -
		jigsaw->piece_array[piece_no].iy ) > snap_dy
		)
	{
		return 0;
	}

	// Find update rectangle for the pixel table

	tx1 = rint ( jigsaw->piece_array[piece_no].tx * jigsaw->vmx ) +
		jigsaw->xo;
	ty1 = rint ( jigsaw->piece_array[piece_no].ty * jigsaw->vmy ) +
		jigsaw->yo;
	tx2 = rint ( jigsaw->piece_array[piece_no].ix * jigsaw->vmx ) +
		jigsaw->xo;
	ty2 = rint ( jigsaw->piece_array[piece_no].iy * jigsaw->vmy ) +
		jigsaw->yo;

	txu[0] = MIN ( tx1, tx2 );
	tyu[0] = MIN ( ty1, ty2 );
	twu[0] = jigsaw->piece_array[piece_no].image->width  +
		abs ( tx2 - tx1 );
	thu[0] = jigsaw->piece_array[piece_no].image->height +
		abs ( ty2 - ty1 );

	jigsaw->piece_array[piece_no].tx = jigsaw->piece_array[piece_no].ix;
	jigsaw->piece_array[piece_no].ty = jigsaw->piece_array[piece_no].iy;

	return -1;			// Piece snapped into place
}

int mtjigsaw_grab_piece (
	mtJigsaw	* jigsaw,
	int		table_x,
	int		table_y,
	int		* txu,
	int		* tyu,
	int		* twu,
	int		* thu
	)
{
	unsigned char	* src;
	int		i,
			px,
			py,
			ix,
			iy,
			last;


	if (	! jigsaw		||
		! jigsaw->piece_array	||
		jigsaw->nodes_hori < MTJIGSAW_NODES_MIN
		)
	{
		return -1;		// No piece selected
	}

	last = (jigsaw->nodes_hori - 1) * (jigsaw->nodes_vert - 1) - 1;

	for ( i = last; i >= 0; i-- )
	{
		if ( ! jigsaw->piece_array[i].image )
		{
			continue;
		}

		// Does this grab lie within the piece rectangle?
		px = rint ( jigsaw->piece_array[i].tx * jigsaw->vmx ) +
			jigsaw->xo;

		if ( table_x < px )
		{
			continue;
		}

		py = rint ( jigsaw->piece_array[i].ty * jigsaw->vmy ) +
			jigsaw->yo;

		if ( table_y < py )
		{
			continue;
		}

		if ( table_x >= (px + jigsaw->piece_array[i].image->width) )
		{
			continue;
		}

		if ( table_y >= (py + jigsaw->piece_array[i].image->height) )
		{
			continue;
		}

		// Does this grab lie on the correct alpha value?
		src = jigsaw->piece_array[ i ].
			image->channel[ MTPIXEL_CHANNEL_ALPHA ];

		if ( ! src )
		{
			continue;
		}

		ix = table_x - px;
		iy = table_y - py;

		if ( src[ix + iy * jigsaw->piece_array[i].image->width] > 127 )
		{
			mtJigsawPiece	tp;


			txu[0] = rint ( jigsaw->piece_array[i].tx *
				jigsaw->vmx ) + jigsaw->xo;
			tyu[0] = rint ( jigsaw->piece_array[i].ty *
				jigsaw->vmy ) + jigsaw->yo;
			twu[0] = jigsaw->piece_array[i].image->width;
			thu[0] = jigsaw->piece_array[i].image->height;

			if ( i != last )
			{
				int		bxu,
						byu,
						bwu,
						bhu,
						cxu,
						cyu,
						cwu,
						chu;


				// We also need to update the other piece
				bxu = rint ( jigsaw->piece_array[last].tx *
					jigsaw->vmx ) + jigsaw->xo;
				byu = rint ( jigsaw->piece_array[last].ty *
					jigsaw->vmy ) + jigsaw->yo;

				bwu = jigsaw->piece_array[last].image->width;
				bhu = jigsaw->piece_array[last].image->height;

				cxu = MIN ( bxu, txu[0] );
				cyu = MIN ( byu, tyu[0] );

				if ( (bxu + bwu) > (txu[0] + twu[0]) )
				{
					cwu = bxu + bwu - cxu;
				}
				else
				{
					cwu = txu[0] + twu[0] - cxu;
				}

				if ( (byu + bhu) > (tyu[0] + thu[0]) )
				{
					chu = byu + bhu - cyu;
				}
				else
				{
					chu = tyu[0] + thu[0] - cyu;
				}

				txu[0] = cxu;
				tyu[0] = cyu;
				twu[0] = cwu;
				thu[0] = chu;

				// Put this grabbed piece to the top
				memcpy ( &tp, &jigsaw->piece_array[last],
					sizeof ( tp ) );
				memcpy ( &jigsaw->piece_array[last],
					&jigsaw->piece_array[i],
					sizeof ( tp ) );
				memcpy ( &jigsaw->piece_array[i], &tp,
					sizeof ( tp ) );

				i = last;
			}

			break;		// Grab is on this piece
		}
	}

	return i;		// Topmost piece clicked, or -1 if none found
}

