/*
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <mtpixel.h>
#include <mtcelledit.h>



// NOTE - if changed, must also change preferences min/max
#define MTJIGSAW_NODES_MIN	6	// Mesh Nodes per row/column
#define MTJIGSAW_NODES_MAX	100	// Mesh Nodes per row/column
#define MTJIGSAW_NODE_COORD_MAX	1000000
#define MTJIGSAW_TABLE_SIZE_MIN	10	// Minimum table width/height



typedef struct
{
	int		ID,		// Original position in the array
			tx,		// Current Table X,Y of this piece
			ty,		// (for dragging)
			ix,		// Image X,Y - if same as tx,ty then
			iy		// this piece is 'home'
			;
	mtImage		* image
			;
} mtJigsawPiece;

typedef struct
{
	char		* image_filename;

	mtImage		* image,	// Original image
			* scaled_image	// Scaled image for pieces to be cut
					// from
			;

	int		nodes_hori,	// Horizontal mesh nodes
			nodes_vert,	// Vertical mesh nodes
			* mesh_x,	// Node X coords (nodes_hori *
					// nodes_vert) items
			* mesh_y,	// Node Y coords (nodes_hori *
					// nodes_vert) items
			table_width,	// Pixels
			table_height,	// Pixels
			xo,		// Origin of the scaled_image on the
					// table
			yo
			;

	double		vmx,		// Vector multiplier: multiplied with
					// nodex_table[?]
			vmy		// to get a pixel coord, e.g.:
					// xp = rint (xv * vmx) + xo
					// To get a vector coord from a canvas
					// point:
					// xv = rint ( ( (double)(xp - xo) ) /
					// vmx )
			;

	mtJigsawPiece	* piece_array;	// Array of pieces
} mtJigsaw;



// NULL = fail, or 0 = success unless otherwise stated



mtJigsaw * mtjigsaw_load (		// Create a new Jigsaw structure and
					// load a file into it
	char		* filename	// CED/TSV mtJigsaw datafile.
					// NULL = empty
	);

int mtjigsaw_save (			// Save a Jigsaw structure to CED/TSV
					// mtJigsaw datafile
	mtJigsaw	* jigsaw,
	char		* filename
	);

int mtjigsaw_destroy (
	mtJigsaw	* jigsaw
	);

int mtjigsaw_set_image (		// Set image, rescaling as needed, cut
					// up image if mesh already set
	mtJigsaw	* jigsaw,
	char		* filename,	// NULL = use current image
	int		width,		// Geometry of the target 'table' space
	int		height
	);

int mtjigsaw_set_mesh (			// Set up random nodes, cut up scaled
					// image if an image is in place
	mtJigsaw	* jigsaw,
	int		x_nodes,
	int		y_nodes
	);

int mtjigsaw_scatter (			// Randomize positions
	mtJigsaw	* jigsaw
	);

int mtjigsaw_join (			// Reset to start positions
	mtJigsaw	* jigsaw
	);

int mtjigsaw_render (			// Render the Jigsaw onto an image
	mtJigsaw	* jigsaw,
	mtImage		* image,
	int		sx,		// Start X
	int		sy		// Start Y
	);

int mtjigsaw_drag_piece (
	mtJigsaw	* jigsaw,
	int		piece_no,	// Piece to move 0..
	int		dx,		// Change in X coord (pixel/table)
	int		dy,		// Change in Y coord (pixel/table)
	int		* txu,		// Table area to update: (txu, tyu)
	int		* tyu,		// size (twu, thu)
	int		* twu,
	int		* thu
	);

int mtjigsaw_drop_piece (		// Attempt to drop a piece in its
					// position
	mtJigsaw	* jigsaw,
	int		piece_no,	// Piece to drop 0..
	int		* txu,		// Table area to update: (txu, tyu)
	int		* tyu,		// size (twu, thu)
	int		* twu,
	int		* thu
	);
	// -1 = snapped to correct position
	//  0 = Not snapped into position

int mtjigsaw_grab_piece (		// Click from user: brings piece to the
					// top of the pile
	mtJigsaw	* jigsaw,
	int		table_x,
	int		table_y,
	int		* txu,		// Table area to update: (txu, tyu)
	int		* tyu,		// size (twu, thu)
	int		* twu,
	int		* thu
	);
	// -2 = error
	// -1 = no piece grabbed
	// else ID of piece

/*
	Jigsaw file format

	r1c1			Image Filename - absolute path

	r2c1			Horizontal nodes - 6-100
	r2c2			Vertical nodes - 6-100

	r3c1 -> r(2 + r2c1)c(r2c2)

	Table containing node coordinates of each node (cell text = 'X,Y)

	r(3 + r2c1)c1		First jigsaw piece data

	r(2 + (r2c1 - 1) * (r2c2 - 1)

		Last jigsaw piece data :
			c1 = table X
			c2 = table Y
			c3 = image X
			c4 = image Y
*/
