/*
	Copyright (C) 2008-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



#define MTKIT_FILE_MEM_BUFLEN		32768



struct _mtFile
{
	FILE		* fp;		// If NULL we are using memory file
					// below

// ---------	MEMORY FILE ONLY

	void		* buf_start,	// Start of memory buffer
			* buf_next	// Next byte (if buf_left > 0)
			;

	int		buf_left;	// Bytes left remaining unused in the
					// buffer
};



static char * load_gzip (
	char		* filename,
	FILE		* fp,
	int		* bytes,
	int		zero
	)
{
	unsigned char	txt[16];
	char		* buf;
	int		newsize;


	if ( fseek ( fp, -4, SEEK_END ) )
	{
		return NULL;
	}

	if ( fread ( txt, 1, 4, fp ) != 4 )
	{
		return NULL;
	}

	newsize = (txt[3] << 24) + (txt[2] << 16) + (txt[1] << 8) + txt[0];

	if ( newsize > MTKIT_FILESIZE_MAX )
	{
		return NULL;
	}

	if ( newsize < 1 )
	{
		// No bytes to read so just have a 1 byte buffer

		newsize = 0;
		zero = 1;
	}

	buf = calloc ( newsize + zero, 1 );
	if ( ! buf )
	{
		return NULL;
	}

	if ( newsize > 0 )
	{
		int		i;
		gzFile		gzf;


		gzf = gzopen ( filename, "rb" );
		if ( ! gzf )
		{
			goto fail;
		}

		i = gzread ( gzf, (voidp)buf, (unsigned)newsize );
		gzclose ( gzf );

		if ( i != newsize )
		{
			goto fail;
		}
	}

	bytes[0] = newsize;

	return buf;			// Success

fail:
	free ( buf );

	return NULL;
}


char * mtkit_file_load (
	char		* filename,
	int		* bytes,
	int		in_flag,
	int		* out_flag
	)
{
	FILE		* fp;
	char		* buf;
	unsigned char	txt[16];
	int64_t		fsize;
	int		i,
			outi = 0,
			zero = 0,
			gzip = 0;


	if ( ! bytes )
	{
		return NULL;
	}

	fsize = mtkit_file_size ( filename );

	if ( fsize < 0 || fsize > MTKIT_FILESIZE_MAX )
	{
		return NULL;
	}

	if ( in_flag & MTKIT_FILE_ZERO )
	{
		zero = 1;		// Extra 0 byte at end
	}

	fp = fopen ( filename, "rb" );

	if ( ! fp )
	{
		return NULL;
	}

	if ( in_flag & MTKIT_FILE_GUNZIP )
	{
		i = fread ( txt, 1, 2, fp );

		if (	i == 2		&&
			txt[0] == 31	&&
			txt[1] == 139
			)
		{
			gzip = 1;	// GZIP header found
		}

		rewind ( fp );
	}

	bytes[0] = fsize;

	if ( fsize == 0 )
	{
		// No bytes to read so just have a 1 byte buffer

		zero = 1;
	}

	if ( gzip && fsize > 20 )
	{
		buf = load_gzip ( filename, fp, bytes, zero );

		if ( buf )
		{
			// GZIP file successfully unpacked
 			outi |= MTKIT_FILE_GUNZIP;

			goto finish;
		}

		// Failure to successfully open a gzip file leads to us loading
		// the raw file anyway.
	}

	buf = malloc ( fsize + zero );

	if ( buf && fsize > 0 )
	{
		i = fread ( buf, 1, fsize, fp );

		if ( i != fsize )
		{
			free ( buf );
			buf = NULL;
		}
	}

	if ( buf && zero )
	{
		buf[ fsize ] = 0;
	}

finish:
	fclose ( fp );

	if ( out_flag )
	{
		out_flag[0] = outi;
	}

	return buf;
}

int mtkit_file_save (
	char		* filename,
	char		* buf,
	int		buf_size,
	int		out_flag
	)
{
	int		i = 0;


	if (	! filename			||
		! buf				||
		buf_size < 0			||
		buf_size > MTKIT_FILESIZE_MAX
		)
	{
		return 1;
	}

	if ( out_flag & MTKIT_FILE_GUNZIP )
	{
		gzFile		fp;


		fp = gzopen ( filename, "wb" );

		if ( ! fp )
		{
			return 1;
		}

		if (	buf_size > 0 &&
			gzwrite ( fp, buf, buf_size ) != buf_size
			)
		{
			i = 1;
		}

		gzclose ( fp );
	}
	else
	{
		FILE		* fp;


		fp = fopen ( filename, "wb" );
		if ( ! fp )
		{
			return 1;
		}

		if (	buf_size > 0 &&
			fwrite ( buf, 1, buf_size, fp ) != buf_size
			)
		{
			i = 1;
		}

		fclose ( fp );
	}

	return i;
}



static char	* homedir = NULL;



char * mtkit_file_home ( void )
{
	struct passwd	* p;


	if ( homedir )
	{
		return homedir;
	}

	homedir = getenv ( "HOME" );

	if ( ! homedir )
	{
		p = getpwuid ( getuid () );

		if ( p )
		{
			homedir = p->pw_dir;
		}
	}

	if ( ! homedir )
	{
		fprintf ( stderr, "Could not find home directory. Using "
			"current directory.");
		homedir = ".";
	}

	return homedir;
}

int mtkit_file_valid (
	char		* filename
	)
{
	FILE		* fp;


	fp = fopen ( filename, "r" );
	if ( ! fp )
	{
		return ( errno == ENOENT ? -1 : 1 );
	}

	fclose ( fp );

	return 0;
}



#define BUFSIZE 100

char * mtkit_file_readline (
	FILE		* fp,
	int		* len,
	int		* len_nl
	)
{
	char		* line,
			* cp;
	int		bufsize,
			p,
			i,
			nl;


	if ( ! fp )
	{
		return NULL;
	}

	line = malloc ( BUFSIZE + 1 );
	if ( ! line )
	{
		return NULL;
	}

	bufsize = BUFSIZE;
	p = 0;
	nl = 0;

	// Put data from the file into the memory buffer
	do
	{
		i = fgetc ( fp );

		if ( i == EOF )
		{
			break;
		}

		if ( i < 0 )
		{
			goto error;
		}

		line[p] = i;
		p++;

		if ( p >= bufsize )
		{
			bufsize += BUFSIZE;
			cp = realloc ( line, bufsize );

			if ( ! cp )
			{
				goto error;
			}

			line = cp;
		}

		if ( i == '\n' )
		{
			p--;
			nl = 1;

			break;
		}

		if ( i == '\r' )
		{
			p--;
			nl = 1;

			// Check for DOS newline \r\n
			i = fgetc ( fp );

			if ( i == EOF )
			{
				break;
			}

			if ( i < 0 )
			{
				goto error;
			}

			// This isn't a DOS newline so give this byte back
			// (Mac newline)

			if ( i != '\n' )
			{
				ungetc ( i, fp );
			}
			else
			{
				nl ++;
			}

			break;
		}
	}
	while ( 1 );

	if ( (p + nl) < 1 )
	{
		goto error;		// No input
	}

	line[p] = 0;

	if ( len )
	{
		len[0] = p;
	}

	if ( len_nl )
	{
		len_nl[0] = nl;
	}

	if ( p < bufsize )		// Shrink wrap final string
	{
		cp = realloc ( line, p + 1 );
		if ( cp )
		{
			line = cp;
		}
	}

	return line;

error:
	free ( line );

	return NULL;
}



/*
	------- mtFile -------
*/

mtFile * mtkit_file_open_disk (
	char		* filename
	)
{
	mtFile		* mtfp;


	mtfp = calloc ( sizeof ( mtFile ), 1 );

	if ( ! mtfp )
	{
		return NULL;
	}

	mtfp->fp = fopen ( filename, "wb" );
	if ( ! mtfp->fp )
	{
		free ( mtfp );

		return NULL;
	}

	return mtfp;
}

mtFile * mtkit_file_open_mem ( void )
{
	mtFile		* mtfp;


	mtfp = calloc ( sizeof ( mtFile ), 1 );

	if ( ! mtfp )
	{
		return NULL;
	}

	mtfp->buf_start = malloc ( MTKIT_FILE_MEM_BUFLEN );
	if ( ! mtfp->buf_start )
	{
		free ( mtfp );

		return NULL;
	}

	mtfp->buf_next = mtfp->buf_start;
	mtfp->buf_left = MTKIT_FILE_MEM_BUFLEN;

	return mtfp;
}

int mtkit_file_close (
	mtFile		* mtfp
	)
{
	if ( ! mtfp )
	{
		return 1;
	}

	if ( mtfp->fp )
	{
		fclose ( mtfp->fp );
	}
	else
	{
		free ( mtfp->buf_start );
	}

	// Dangling pointers left as we destroy structure
	free ( mtfp );

	return 0;			// Success
}

int mtkit_file_write (
	mtFile		* mtfp,
	void		* mem,
	int64_t		mem_len
	)
{
	if (	! mtfp		||
		! mem		||
		mem_len < 0
		)
	{
		return 1;
	}

	if ( mem_len == 0 )
	{
		return 0;
	}

	// NOTE: We do not check for size of memory chunk getting too big

	if ( mtfp->fp )
	{
		if ( fwrite ( mem, 1, mem_len, mtfp->fp ) != mem_len )
		{
			return 1;
		}
	}
	else
	{
		if ( mem_len > mtfp->buf_left )
		{
			void		* new;
			int64_t		len,
					todo;

			len = mtfp->buf_next - mtfp->buf_start,
			todo = mem_len + MTKIT_FILE_MEM_BUFLEN;

			new = realloc ( mtfp->buf_start, len + todo );
			if ( ! new )
			{
				return 1;
			}

			mtfp->buf_start = new;
			mtfp->buf_next = new + len;
			mtfp->buf_left = todo;
		}

		memcpy ( mtfp->buf_next, mem, mem_len );
		mtfp->buf_next += mem_len;
		mtfp->buf_left -= mem_len;
	}

	return 0;			// Success
}

int mtkit_file_write_string (
	mtFile		* mtfp,
	char		* str
	)
{
	if ( ! str )
	{
		return 1;
	}

	return mtkit_file_write ( mtfp, str, strlen ( str ) );
}

int mtkit_file_get_mem (
	mtFile		* mtfp,
	void		** buf,
	int64_t		* buf_len
	)
{
	if ( ! mtfp )
	{
		return 1;
	}

	if ( buf )
	{
		buf[0] = mtfp->buf_start;
	}

	if ( buf_len )
	{
		buf_len[0] = mtfp->buf_next - mtfp->buf_start;
	}

	return 0;
}

int mtkit_file_header_gz (
	unsigned char	* mem,
	int		mem_size
	)
{
	if (	mem		&&
		mem_size > 20	&&
		mem[0] == 31	&&
		mem[1] == 139
		)
	{
		return 1;
	}

	return 0;
}

int mtkit_file_header_zip (
	unsigned char	* mem,
	int		mem_size
	)
{
	if (	mem		&&
		mem_size > 30	&&
		mem[0] == 0x50	&&
		mem[1] == 0x4b	&&
		mem[2] == 0x03	&&
		mem[3] == 0x04
		)
	{
		return 1;
	}

	return 0;
}

int64_t mtkit_file_size (
	char		* filename
	)
{
	struct stat	buf;


	if ( ! filename )
	{
		return -1;
	}

	if ( stat ( filename, &buf ) )
	{
		return -2;
	}

	return buf.st_size;
}

void mtkit_print_errno (
	int		e
	)
{
	char		* se;


	if ( e == 0 )
	{
		return;
	}

	se = strerror ( e );
	if ( ! se )
	{
		se = "?";
	}

	fprintf ( stderr, "errno = %i - %s\n\n", e, se );
}
