/*
	Copyright (C) 2007-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



static char	mtkit_locale[256];



char * mtkit_utf8_from_locale (
	char		* text
	)
{
	char		* new;


	if (	! text ||
		mtkit_string_encoding_conversion ( text, mtkit_locale, &new,
			"UTF-8" )
		)
	{
		return NULL;
	}

	return new;
}

int mtkit_utf8_remove (
	char		** utf8,
	int		start,
	int		length
	)
{
	char		* a,
			* c,
			* s,
			* new_utf8;
	int		as = 0,
			bs = 0,
			cs = 0,
			i;


	if ( ! utf8 || ! utf8[0] )
	{
		return 1;
	}

	a = s = utf8[0];

	// Calculate size of <a> chunk and beginning of <b>
	for ( as = 0; as < start; as ++ )
	{
		if ( s[0] == 0 )
		{
			break;		// End of string reached
		}

		if ( s[0] >= 0 )
		{
			s++;
		}
		else
		{
			i = mtkit_utf8_to_utf32 ( (unsigned char *)s, NULL );

			if ( i < 1 )
			{
				return 1;	// Error
			}

			s += i;
			as += i - 1;
		}
	}

	// Calculate size of <b> chunk and beginning of <c>
	for ( bs = 0; bs < length; bs ++ )
	{
		if ( s[0] == 0 )
		{
			break;		// End of string reached
		}

		if ( s[0] >= 0 )
		{
			s++;
		}
		else
		{
			i = mtkit_utf8_to_utf32 ( (unsigned char *)s, NULL );

			if ( i < 1 )
			{
				return 1;	// Error
			}

			s += i;
			bs += i - 1;
		}
	}

	c = s;
	cs = strlen ( c );

	i = as + cs + 1;
	new_utf8 = malloc ( i );

	if ( ! new_utf8 )
	{
		return 1;
	}

	new_utf8[i - 1] = 0;		// Terminator

	if ( as )
	{
		memcpy ( new_utf8, a, as );
	}

	if ( cs )
	{
		memcpy ( new_utf8 + as, c, cs );
	}

	free ( utf8[0] );		// Remove old string text
	utf8[0] = new_utf8;		// Put new string text in place

	return 0;			// Success
}

int mtkit_utf8_insert (
	char		** utf8,
	int		n,
	char		* utf8_insert
	)
{
	int		new_len;
	char		* new_utf8;


	if ( ! utf8 )
	{
		return 1;
	}

	if ( ! utf8_insert )
	{
		return 0;		// Nothing to do
	}

	new_len = strlen ( utf8_insert ) + 1;

	if ( new_len == 1 )
	{
		return 0;		// Nothing to do
	}

	if ( utf8[0] )
	{
		new_len += strlen ( utf8[0] );
	}

	new_utf8 = malloc ( new_len );

	if ( ! new_utf8 )
	{
		return 1;
	}

	new_utf8[0] = 0;

	if ( n > 0 && utf8[0] )
	{
		char		* txt = utf8[0];
		int		pos = 0,
				i;


		// Copy the correct number of bytes into the new utf8

		while ( txt[0] && pos < n )
		{
			i = mtkit_utf8_to_utf32 ( (unsigned char *)txt, NULL );

			if ( i < 1 )
			{
				// Error in UTF-8 string so bail out

				free ( new_utf8 );

				return 1;
			}

			txt += i;
			pos ++;
		}

		memcpy ( new_utf8, utf8[0], txt - utf8[0] );
		new_utf8[ txt - utf8[0] ] = 0;

		strcat ( new_utf8, utf8_insert );

		if ( txt[0] )
		{
			strcat ( new_utf8, txt );
		}
	}
	else	// Place 'new' at the beginning
	{
		strcat ( new_utf8, utf8_insert );

		if ( utf8[0] )
		{
			strcat ( new_utf8, utf8[0] );
		}
	}

	free ( utf8[0] );		// Remove old string text
	utf8[0] = new_utf8;		// Put new string text in place

	return 0;			// Success
}

int mtkit_string_encoding_conversion (
	char		* text_in,
	char		* text_in_encoding,
	char		** text_out,
	char		* text_out_encoding
	)
{
	char		* buf_in[1];
	char		* buf_out[2];
	size_t		size_in = strlen ( text_in ),
			size_out = (size_in + 1) * 4 + 5,
			s;
	iconv_t		cd;


	if (	! text_in		||
		! text_in_encoding	||
		! text_out		||
		! text_out_encoding
		)
	{
		return -666;
	}

	buf_in[0] = text_in;
	buf_out[0] = buf_out[1] = calloc ( size_out, 1 );

	if ( ! buf_out[0] )
	{
		return -1;		// Not enough memory
	}

	cd = iconv_open ( text_out_encoding, text_in_encoding );
	if ( cd == (iconv_t)(-1) )
	{
		free ( buf_out[0] );
		return -2;		// iconv_open failure
	}

	s = iconv ( cd, buf_in, &size_in, buf_out, &size_out );
	iconv_close ( cd );
	if ( s )
	{
		free ( buf_out[1] );
		return -3;		// iconv failure
	}

	s = strlen ( buf_out[1] ) + 1;	// Get real size of output buffer
	buf_out[0] = realloc ( buf_out[1], s );	// Remove any slack memory

	if ( ! buf_out[0] )
	{
		return -4;		// realloc failure
	}

	text_out[0] = buf_out[0];

	return 0;			// Success
}

int mtkit_string_argv_free (
	char		** args
	)
{
	int		i;


	if ( ! args )
	{
		return 0;
	}

	for ( i = 0; args[i]; i++ )
	{
		free ( args[i] );
	}

	free ( args );

	return 0;
}

char ** mtkit_string_argv (
	char		* input
	)
{
	char		** args,
			* src,
			* input_end;
	int		arg,
			arg_tot,
			q,
			cpos;


	if ( ! input )
	{
		return NULL;
	}

	// Ignore leading whitespaces
	while ( isspace ( input[0] ) )
	{
		input ++;
	}

	// Count the number of arguments in the input string
	arg = 0;
	q = 0;
	for ( src = input; src[0]; src ++ )
	{
		if ( src[0] == '"' )
		{
			q = ! q;

			continue;
		}
		else if ( src[0] == '\\' )
		{
			src ++;

			if ( src[0] == 0 )
			{
				break;
			}
		}
		else if ( isspace ( src[0] ) )
		{
			if ( ! q )
			{
				while ( isspace ( src[0] ) )
				{
					src ++;
				}

				if ( src[0] == 0 )
				{
					break;
				}

				src --;
				arg ++;

				continue;
			}
		}
	}

	input_end = src;
	arg_tot = arg + 1;

	args = calloc ( arg_tot + 1, sizeof ( char * ) );

	if ( ! args )
	{
		return NULL;
	}

	// Create each of the args
	arg = 0;
	q = 0;
	cpos = 0;
	args[0] = malloc ( input_end - input + 1 );

	if ( ! args[0] )
	{
		goto fail;
	}

	for ( src = input; src[0]; src ++ )
	{
		if ( src[0] == '"' )
		{
			q = ! q;

			continue;
		}
		else if ( src[0] == '\\' )
		{
			src ++;
			if ( src[0] == 0 )
			{
				break;
			}
		}
		else if ( isspace ( src[0] ) )
		{
			if ( ! q )
			{
				while ( isspace ( src[0] ) )
				{
					src ++;
				}

				if ( src[0] == 0 )
				{
					break;
				}

				// Finish off this arg string
				args[arg][cpos] = 0;
				args[arg] = realloc ( args[arg], cpos + 1 );

				src --;
				arg ++;

				// Prepare the next arg string
				if ( src > input_end )
				{
					goto fail;	// Shouldn't happen
				}

				args[arg] = malloc ( input_end - src + 1 );
				cpos = 0;

				if ( ! args[arg] )
				{
					goto fail;
				}

				continue;
			}
		}

		args[arg][cpos] = src[0];
		cpos ++;
	}

	// Finish off the last arg string
	args[arg][cpos] = 0;
	args[arg] = realloc ( args[arg], cpos + 1 );

	return args;

fail:
	mtkit_string_argv_free ( args );

	return NULL;
}


int mtkit_set_locale (
	char		* st
	)
{
	if ( ! st )
	{
		st = "ASCII";
	}
	else if ( st[0] == 0 )
	{
		setlocale ( LC_ALL, "" );
		st = nl_langinfo ( CODESET );
	}

	mtkit_strncpy0 ( mtkit_locale, st, 256 );

	return 0;
}

char * mtkit_string_join (
	char		** str_array,
	char		* separator
	)
{
	int		len = 0,
			sep_len = 0,
			i;
	char		* str = NULL;


	if (	! str_array	||
		! str_array[0]	||
		! separator
		)
	{
		return NULL;
	}

	if ( separator )
	{
		sep_len = strlen ( separator );
	}

	// Count up the total size required
	len += strlen ( str_array[0] ) + 1;

	for ( i = 1; str_array[i]; i++ )
	{
		len += sep_len + strlen ( str_array[i] );
	}

	str = malloc ( len );
	if ( ! str )
	{
		return NULL;
	}

	str[0] = 0;
	strcat ( str, str_array[0] );
	for ( i = 1; str_array[i]; i++ )
	{
		strcat ( str, separator );
		strcat ( str, str_array[i] );
	}

	return str;
}



/*
Portions of this UTF-8 conversion code comes from ConvertUTF.c with
adjustments by Mark Tyler in March 2008 to make it work inside the mtKit
context.
*/

/*
 * Copyright 2001-2004 Unicode, Inc.
 *
 * Disclaimer
 *
 * This source code is provided as is by Unicode, Inc. No claims are
 * made as to fitness for any particular purpose. No warranties of any
 * kind are expressed or implied. The recipient agrees to determine
 * applicability of information provided. If this file has been
 * purchased on magnetic or optical media from Unicode, Inc., the
 * sole remedy for any claim will be exchange of defective media
 * within 90 days of receipt.
 *
 * Limitations on Rights to Redistribute This Code
 *
 * Unicode, Inc. hereby grants the right to freely use the information
 * supplied in this file in the creation of products supporting the
 * Unicode Standard, and to make copies of this file in any form
 * for internal or external distribution as long as this notice
 * remains attached.
 */

/* ---------------------------------------------------------------------

    Conversions between UTF32, UTF-16, and UTF-8. Source code file.
    Author: Mark E. Davis, 1994.
    Rev History: Rick McGowan, fixes & updates May 2001.
    Sept 2001: fixed const & error conditions per
	mods suggested by S. Parent & A. Lillich.
    June 2002: Tim Dodd added detection and handling of incomplete
	source sequences, enhanced error detection, added casts
	to eliminate compiler warnings.
    July 2003: slight mods to back out aggressive FFFE detection.
    Jan 2004: updated switches in from-UTF8 conversions.
    Oct 2004: updated to use UNI_MAX_LEGAL_UTF32 in UTF-32 conversions.

    See the header file "ConvertUTF.h" for complete documentation.

------------------------------------------------------------------------ */

#define UNI_MAX_LEGAL_UTF32	0x0010FFFF
#define UNI_SUR_HIGH_START	0xD800
#define UNI_SUR_HIGH_END	0xDBFF
#define UNI_SUR_LOW_START	0xDC00
#define UNI_SUR_LOW_END		0xDFFF

/*
 * Index into the table below with the first byte of a UTF-8 sequence to
 * get the number of trailing bytes that are supposed to follow it.
 * Note that *legal* UTF-8 values can't have 4 or 5-bytes. The table is
 * left as-is for anyone who may want to do such conversion, which was
 * allowed in earlier algorithms.
 */
static const char	trailingBytesForUTF8[256] =
{
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

/*
 * Magic values subtracted from a buffer value during UTF8 conversion.
 * This table contains as many values as there might be trailing bytes
 * in a UTF-8 sequence.
 */
static const unsigned long offsetsFromUTF8[5] =
{
0, 0x00000000UL, 0x00003080UL, 0x000E2080UL, 0x03C82080UL
};

/*
 * Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
 * into the first byte, depending on how many bytes follow.  There are
 * as many entries in this table as there are UTF-8 sequence types.
 * (I.e., one byte sequence, two byte... etc.). Remember that sequencs
 * for *legal* UTF-8 will be 4 or fewer bytes total.
 */
static const unsigned char firstByteMark[7] =
	{ 0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC };



int mtkit_utf8_is_legal (
	unsigned char	* src,
	int		bytes
	)
{
	unsigned char	a;
	const unsigned char * srcptr = src + bytes;


	if ( ! src )
	{
		return 0;
	}

	switch ( bytes )
	{
		default: return 0;
		/* Everything else falls through when "true"... */
		case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return 0;
		case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return 0;
		case 2: if ((a = (*--srcptr)) > 0xBF) return 0;

		switch ( src[0] )
		{
		/* no fall-through in this inner switch */
			case 0xE0: if (a < 0xA0) return 0; break;
			case 0xED: if (a > 0x9F) return 0; break;
			case 0xF0: if (a < 0x90) return 0; break;
			case 0xF4: if (a > 0x8F) return 0; break;
			default:   if (a < 0x80) return 0;
		}

		case 1:
			if ( src[0] >= 0x80 && src[0] < 0xC2 )
			{
				return 0;
			}
	}

	if ( src[0] > 0xF4 )
	{
		return 0;
	}

	return 1;
}

int mtkit_utf8_string_legal (
	unsigned char	* src,
	int		bytes
	)
{
	int		i;


	if ( bytes < 1 )
	{
		while ( src[0] )
		{
			if ( src[0] < 128 )
			{
				src ++;
			}
			else
			{
				i = mtkit_utf8_to_utf32 ( src, NULL );

				if ( i < 1 )
				{
					return 0;
				}

				src += i;
			}
		}
	}
	else
	{
		unsigned char	* end;


		end = src + bytes;

		while ( src < end )
		{
			if ( src[0] < 128 )
			{
				src ++;

				continue;
			}

			i = mtkit_utf8_to_utf32 ( src, NULL );

			if ( i < 1 )
			{
				return 0;
			}

			src += i;
		}
	}

	return 1;			// Legal
}

int mtkit_utf8_to_utf32 (
	unsigned char	* src,
	unsigned int	* unicode
	)
{
	int		i,
			length;


	if ( ! src )
	{
		return -1;
	}

	length = 1 + trailingBytesForUTF8[ src[0] ];

	if ( length > 4 || length < 1 )
	{
		// Only 1-4 bytes per valid UTF-8 char

		return -1;
	}

	for ( i = 1; i < length; i++ )
	{
		if ( src[i] == 0 )
		{
			// Detect premature end to UTF-8 char

			return -1;
		}
	}

	if ( ! mtkit_utf8_is_legal ( src, length ) )
	{
		return -1;		// Illegal UTF-8 character
	}

	if ( unicode )
	{
		unsigned int	ch = 0;


		switch ( length )
		{
			case 4: ch += *src++; ch <<= 6;
			case 3: ch += *src++; ch <<= 6;
			case 2: ch += *src++; ch <<= 6;
			case 1: ch += *src++;
		}

		ch -= offsetsFromUTF8[ length ];

/*
 * UTF-16 surrogate values are illegal in UTF-32, and anything
 * over Plane 17 (> 0x10FFFF) is illegal.
 */
		if ( ch > UNI_MAX_LEGAL_UTF32 )
		{
			return -1;
		}

		if (	ch >= UNI_SUR_HIGH_START &&
			ch <= UNI_SUR_LOW_END
			)
		{
			return -1;
		}

		unicode[0] = ch;
	}

	return length;
}

int mtkit_utf8_from_utf32 (
	unsigned char	* src,
	unsigned int	unicode
	)
{
	unsigned int	ch = unicode,
			bytesToWrite = 0,
			byteMask = 0xBF,
			byteMark = 0x80;


	if ( ! src )
	{
		return 0;
	}

	/* UTF-16 surrogate values are illegal in UTF-32 */
	if (	ch >= UNI_SUR_HIGH_START &&
		ch <= UNI_SUR_LOW_END
		)
	{
		return 0;
	}

	// Figure out how many bytes the result will require.

	if ( ch < 0x80 ) bytesToWrite = 1;
	else if ( ch < 0x800) bytesToWrite = 2;
	else if ( ch < 0x10000) bytesToWrite = 3;
	else if ( ch <= UNI_MAX_LEGAL_UTF32) bytesToWrite = 4;
	else return 0;			// Illegal char

	src += bytesToWrite;
	*src-- = 0;			// Terminator

	switch ( bytesToWrite )
	{ /* note: everything falls through. */
		case 4:	*src-- = ( (ch | byteMark) & byteMask ); ch >>= 6;
		case 3:	*src-- = ( (ch | byteMark) & byteMask ); ch >>= 6;
		case 2:	*src-- = ( (ch | byteMark) & byteMask ); ch >>= 6;
		case 1:	*src   = ( ch | firstByteMark[bytesToWrite] );
	}

	return bytesToWrite;
}

int mtkit_utf8_len (
	unsigned char	* src,
	int		bytes
	)
{
	unsigned char	* end;
	int		tot = 0,
			i;


	if ( ! src )
	{
		return 0;
	}

	end = src + bytes;
	if ( bytes < 1 )
	{
		while ( src[0] != 0 )
		{
			if ( src[0] < 128 )
			{
				tot ++;
				src ++;

				continue;
			}

			i = mtkit_utf8_to_utf32 ( src, NULL );

			if ( i == -1 )
			{
				return 0;
			}

			src += i;
			tot ++;
		}
	}
	else
	{
		while ( src < end )
		{
			if ( src[0] < 128 )
			{
				tot ++;
				src ++;

				continue;
			}

			i = mtkit_utf8_to_utf32 ( src, NULL );

			if ( i == -1 )
			{
				return 0;
			}

			src += i;
			tot ++;
		}
	}

	return tot;
}

int mtkit_utf8_char_len (
	unsigned char	byte
	)
{
	return ( 1 + trailingBytesForUTF8[byte] );
}

int mtkit_utf8_offset (
	unsigned char	* src,
	int		bytes,
	int		num
	)
{
	int		offset = 0,
			i;


	for ( i = 0; i < num; i++ )
	{
		if ( offset >= bytes )
		{
			return -1;	// UTF-8 string too short
		}

		offset	+= 1 + trailingBytesForUTF8[ src[offset] ];
	}

	return offset;
}

char * mtkit_strnncat (
	char		* dest,
	const char	* src,
	int		max
	)
{
	int		l = strlen ( dest );


	if ( max > l )
	{
		strncpy ( dest + l, src, max - l - 1 );
	}

	dest[max - 1] = 0;

	return dest;
}

int mtkit_strtod (
	char		* input,
	double		* result,
	char		** next,
	int		strict
	)
{
	double		d;
	char		* s;


	errno = 0;
	d = strtod ( input, &s );

	if ( errno || ! s || s == input )
	{
		return 1;	/* Error or nothing parsed */
	}

	if ( strict && mtkit_strnonspaces ( s ) )
	{
		return 1;	/* String not a pure number */
	}

	if ( next )
	{
		next[0] = s;
	}

	if ( result )
	{
		result[0] = d;
	}

	return 0;
}

int mtkit_strtoi (
	char		* input,
	int		* result,
	char		** next,
	int		strict
	)
{
	int		i,
			base = 10;
	char		* s;


	if ( input[0] == '0' && input[1] == 'x' )
	{
		base = 16;
	}

	errno = 0;
	i = (int)strtol ( input, &s, base );

	if ( errno || ! s || s == input )
	{
		return 1;	/* Error or nothing parsed */
	}

	if ( strict && mtkit_strnonspaces ( s ) )
	{
		return 1;	/* String not a pure number */
	}

	if ( next )
	{
		next[0] = s;
	}

	if ( result )
	{
		result[0] = i;
	}

	return 0;
}

char * mtkit_strtok (
	char		* input,
	char		* delim,
	int		n
	)
{
	int		newlen,
			del_found;
	char		* sp,
			* sp2,
			* new,
			* sd;


	if ( ! input || ! delim || n < 0 )
	{
		return NULL;
	}

	// Find the n'th delimeter
	for ( sp = input; n > 0; sp ++ )
	{
		if ( sp[0] == 0 )
		{
			return NULL;	// End of string reached
		}

		for ( sd = delim; sd[0] != 0; sd ++ )
		{
			if ( sp[0] == sd[0] )	// Delimeter found
			{
				n --;

				break;
			}
		}
	}

	// Get length of this field
	del_found = 0;
	for ( sp2 = sp; sp2[0] != 0 && ! del_found; sp2 ++ )
	{
		for ( sd = delim; sd[0] != 0; sd ++ )
		{
			if ( sp2[0] == sd[0] )
			{
				del_found = 1;
				sp2 --;

				break;
			}
		}
	}

	newlen = sp2 - sp;

	new = calloc ( newlen + 1, 1 );

	if ( ! new )
	{
		return NULL;
	}

	if ( newlen > 0 )
	{
		memcpy ( new, sp, newlen );
	}

	return new;
}

int mtkit_strtok_count (
	char		* input,
	char		* delim
	)
{
	int		count = 1;
	char		* sp,
			* sd;


	if ( ! input || ! delim )
	{
		return -1;
	}

	for ( sp = input; sp[0]; sp ++ )
	{
		for ( sd = delim; sd[0]; sd ++ )
		{
			if ( sp[0] == sd[0] )	// Delimeter found
			{
				count ++;

				break;
			}
		}
	}

	return count;
}

int mtkit_strtok_num (
	char		* input,
	char		* delim,
	int		n,
	double		* result
	)
{
	int		res;
	char		* tok;


	tok = mtkit_strtok ( input, delim, n );
	if ( ! tok )
	{
		return 1;
	}

	res = mtkit_strtod ( tok, result, NULL, 0 );
	free ( tok );

	return res;
}

char * mtkit_strcasestr (
	char		* haystack,
	char		* needle
	)
{
	int		i;
	char		* hp;


	if ( ! haystack || ! needle )
	{
		return NULL;
	}

	if ( needle[0] == 0 )
	{
		return haystack;
	}

	for ( hp = haystack; hp[0] != 0; hp ++ )
	{
		for ( i = 0; needle[i] != 0; i++ )
		{
			if ( tolower ( hp[i] ) != tolower ( needle[i] ) )
			{
				break;
			}
		}

		if ( needle[i] == 0 )
		{
			// Whole of needle exists here in the haystack

			return hp;
		}
	}

	return NULL;			// needle not found
}

char * mtkit_strjoinv (
	char		* sep,
	char		** list
	)
{
	char		* str;
	int		i,
			seplen = 0,
			totlen = 0;


	if ( ! list || ! list[0] )
	{
		return NULL;
	}

	if ( sep )
	{
		seplen = strlen ( sep );
	}

	// Calculate size of final string
	totlen = strlen ( list[0] );
	for ( i = 1; list[i]; i++ )
	{
		totlen += seplen;
		totlen += strlen ( list[i] );
	}

	str = calloc ( totlen + 1, 1 );
	if ( ! str )
	{
		return NULL;
	}

	// Form new string
	strcpy ( str, list[0] );
	for ( i = 1; list[i]; i++ )
	{
		if ( sep )
		{
			strcat ( str, sep );
		}

		strcat ( str, list[i] );
	}

	return str;
}

int mtkit_strfreeswap (
	char		** spp,
	char		* str
	)
{
	if ( ! spp )
	{
		return 1;
	}

	if ( str )
	{
		str = strdup ( str );

		if ( ! str )
		{
			return 1;
		}
	}

	free ( spp[0] );
	spp[0] = str;

	return 0;
}



enum
{
	HTML_AMP_LEN	= 5,		// &amp;
	HTML_QUOT_LEN	= 6,		// &quot;
	HTML_LT_LEN	= 4,		// &lt;
	HTML_GT_LEN	= 4		// &gt;
};



char * mtkit_strtohtml (
	char		* input
	)
{
	int		len;
	char		* new = NULL,
			* src,
			* dest;


	if ( ! input )
	{
		goto finish;
	}

	// Calculate the length of the new string
	for ( len = 0, src = input; src[0] != 0; src ++ )
	{
		switch ( src[0] )
		{
		case '&':	len += HTML_AMP_LEN;	break;
		case '"':	len += HTML_QUOT_LEN;	break;
		case '<':	len += HTML_LT_LEN;	break;
		case '>':	len += HTML_GT_LEN;	break;
		default:	len ++;
		}
	}

	new = malloc ( len + 1 );
	if ( ! new )
	{
		goto finish;
	}

	// Populate the new string
	for ( dest = new, src = input; src[0] != 0; src ++ )
	{
		if ( src[0] < 32 && src[0] >= 0 )
		{
			*dest++ = 32;

			continue;
		}

		switch ( src[0] )
		{
		case '&':
			strcpy ( dest, "&amp;" );
			dest += HTML_AMP_LEN;
			break;

		case '"':
			strcpy ( dest, "&quot;" );
			dest += HTML_QUOT_LEN;
			break;

		case '<':
			strcpy ( dest, "&lt;" );
			dest += HTML_LT_LEN;
			break;

		case '>':
			strcpy ( dest, "&gt;" );
			dest += HTML_GT_LEN;
			break;

		default:
			*dest++ = src[0];
		}
	}

	dest[0] = 0;

finish:
	return new;
}



/*
Wilcard expansion for * and ? characters.  A classic problem, and here is how
I solved it from scratch!  Built in support for case (in)sensitivity and a few
of my favourite \ substitutions.
M.Tyler 23-8-2009
*/

// The character after a \ needs to be substituted in some cases
static char char_subst (
	char		ch
	)
{
	switch ( ch )
	{
	case 'n': return '\n';
	case 'r': return '\r';
	case 't': return '\t';
	}

	return ch;
}

// Create new dynamic string with substitutions for \, i.e. "\*" -> "*",
// "\?" -> "?" etc.
// Terminators are 0 * ? characters

static char * str_form (
	char		* st,
	char		** pnext
	)
{
	char		* cp,
			* new,
			* np,
			ch;
	int		tot;


	for ( cp = st, tot = 0; ; cp ++, tot ++ )
	{
		if (	cp[0] == 0	||
			cp[0] == '*'	||
			cp[0] == '?'
			)
		{
			break;
		}

		if ( cp[0] == '\\' )
		{
			cp ++;

			if ( cp[0] == 0 )
			{
				break;
			}
		}
	}

	pnext[0] = cp;

	if ( tot < 1 )
	{
		// Shouldn't happen if str_form called properly

		return NULL;
	}

	new = malloc ( 1 + tot );

	if ( ! new )
	{
		return NULL;
	}

	new[tot] = 0;

	for ( cp = st, np = new; ; cp ++, np ++ )
	{
		ch = cp[0];

		if (	ch == 0		||
			ch == '*'	||
			ch == '?'
			)
		{
			break;
		}

		if ( ch == '\\' )
		{
			cp ++;

			if ( cp[0] == 0 )
			{
				break;
			}

			ch = char_subst ( cp[0] );
		}

		np[0] = ch;
	}

	return new;
}

int mtkit_strmatch (
	char		* string,
	char		* pattern,
	int		mode
	)
{
	int		match = -1,
			sp = 0,
			pp = 0,
			substr_len,
			res;
	char		s,
			p,
			* substr,
			* st,
			* pnext;


	if ( ! string || ! pattern )
	{
		return -2;
	}

	while ( 1 )
	{
		s = string[ sp ++ ];
		p = pattern[ pp ++ ];

		if ( ! ( mode & 1 ) )
		{
			s = tolower ( s );
			p = tolower ( p );
		}

		switch ( p )
		{
		case '?':
			if ( s == 0 )
			{
				// No more characters, but pattern needs one

				return -1;
			}
			break;

		case '*':
			// Once in here, we return and never break
			while ( 1 )
			{
				p = pattern[ pp ++ ];

				switch ( p )
				{
				case 0:
					// Pattern ending in * or ? always
					// matches here

					if ( match < 0 )
					{
						match = 0;
					}

					return match;

				case '*':
					// Repeat *'s can be ignored here
					// because:
					// "*" = "**" = "****", "*?*" = "*?",
					// "*??*" = "*??" = "*?*?*"

					break;

				case '?':
					if ( s == 0 )
					{
						// No more characters, but
						// pattern needs one

						return -1;
					}

					s = string[ sp ++ ];

					// No need to bother with case check
					// for ?

					break;

				default:
					if ( s == 0 )
					{
						// No more characters, but
						// pattern needs one

						return -1;
					}

					substr = str_form ( pattern + pp - 1,
						&pnext );

					if ( ! substr )
					{
						return -3;
					}

					substr_len = strlen ( substr );
					st = string + sp - 2;

					do
					{
						if ( mode & 1 )
						{
							st = strstr ( st + 1,
								substr );
						}
						else
						{
							st = mtkit_strcasestr (
								st + 1,
								substr );
						}

						if ( ! st )
						{
							// No match
							free ( substr );

							return -1;
						}

						res = mtkit_strmatch ( st +
							substr_len, pnext,
							mode );

						if ( res < -1 )
						{
							// Unexpected error
							free ( substr );

							return res;
						}
					}
					while ( res < 0 );

					free ( substr );

					if ( match < 0 )
					{
						match = st - string;
					}

					// Recursion also matches so this was
					// first matching char

					return match;
				}
			}
			break;

		default:
			if ( match < 0 )
			{
				// Set match point if not already done so
				match = sp - 1;
			}

			if ( s == 0 && p == 0 )
			{
				// End of both so they must match

				return match;
			}

			if ( p == '\\' )
			{
				p = char_subst ( pattern[ pp ++ ] );

				if ( s == 0 && p == 0 )
				{
					// End of both so they must match
					// "abcdef" == "abcdef\"

					return match;
				}
			}

			if ( s == 0 || p == 0 )
			{
				// "abc" != "abcdef" or "abcdef" != "abc\"
				// "abcdef" != "abc"

				return -1;
			}

			if ( s != p )
			{
				// Chars don't match so pattern doesn't

				return -1;
			}

			break;		// These chars match so move on in loop
		}
	}

	return -1;			// No match
}

int mtkit_strnonspaces (
	char		* input
	)
{
	while ( input[0] )
	{
		if ( ! isspace ( input[0] ) )
		{
			return 1;
		}

		input ++;
	}

	return 0;
}

int mtkit_strtothou (
	char		* dest,
	char		* src,
	int		dest_size,
	char		separator,
	char		minus,
	char		dpoint,
	int		sep_num,
	int		right_justify
	)
{
	int		i,
			j,
			k,
			start,
			end,
			oldlen,
			newlen;
	char		* ch;


	if ( sep_num < 1 )
	{
		return -1;		// Sanity check
	}

	oldlen = strlen ( src );

	// First character to be subjected to separating
	start = 0;

	// Last character to be subjected to separating
	end = oldlen - 1;

	if ( ( ch = strrchr ( src, dpoint ) ) )
	{
		// Decimal point detected, adjust end accordingly
		end = ch - src - 1;
	}

	if ( ( ch = strchr ( src, minus ) ) )
	{
		// Minus sign detected, adjust start accordingly
		start = ch - src + 1;
	}

	if ( ( end + 1 ) < start )
	{
		// Decimal point is before minus sign so bail out

		return -1;
	}

	newlen = oldlen + ( end - start ) / sep_num;
	if ( newlen + 1 > dest_size )
	{
		// Output buffer is not large enough to hold the result, so
		// bail out

		return -1;
	}

	i = oldlen - 1;			// Input buffer pointer
	k = dest_size - 1;		// Output buffer pointer
	dest[k--] = 0;			// Output string terminator

	while ( i > end )
	{
		dest[ k-- ] = src[ i-- ];

		// Copy characters from the end to the '.'
	}

	for ( j = 0; i >= start; j++ )
	{
		if ( j && ( (j % sep_num) == 0 ) )
		{
			dest[ k-- ] = separator;	// Add separator
		}

		dest[k--] = src[i--];			// Copy char
	}

	while ( i >= 0 )
	{
		// Copy characters from the '-' to the beginning
		dest[ k-- ] = src[ i-- ];
	}

	if ( right_justify )
	{
		// Right align so pad beginning of output with spaces
		while ( k >= 0 )
		{
			dest[ k-- ] = ' ';
		}
	}
	else
	{
		// Left align so shift string flush to beginning
		k++;
		j = 0;

		do
		{
			dest[ j++ ] = dest[ k++ ];
		}
		while ( dest[ j - 1 ] != 0 );
	}

	return 0;
}

char * mtkit_strdupstrip (
	char		* input,
	char		* filter,
	int		mode
	)
{
	unsigned char	* output,
			* src,
			* dest,
			index[256];


	if (	! input		||
		! filter	||
		mode < 0	||
		mode > 1
		)
	{
		return NULL;
	}

	output = (unsigned char *)strdup ( input );
	if ( ! output )
	{
		return NULL;
	}

	memset ( index, ! mode, 256 );
	for ( src = (unsigned char *)filter; src[0]; src ++ )
	{
		index[ src[0] ] = mode;
	}

	for ( src = dest = output; src[0]; src ++ )
	{
		if ( index[ src[0] ] == 1 )
		{
			*dest++ = src[0];
		}
	}

	dest[0] = 0;

	if ( src != dest )		// Shrink if required
	{
		src = (unsigned char *)strdup ( (char *)output );
		free ( output );

		if ( ! src )
		{
			return NULL;
		}

		output = src;
	}

	return (char *)output;
}

int mtkit_strtoddt (
	char		* input,
	double		* result
	)
{
	int		day = 1,
			month = 1,
			year = 0,
			hour = 0,
			minute = 0,
			second = 0;
	char		* s = input,
			ch;


	if ( mtkit_strtoi ( s, &day, &s, 0 ) )	// DAY (or possibly HOUR)
	{
		return -1;
	}

	while ( isspace ( s[0] ) )
	{
		s++;
	}

	ch = *s++;
	switch ( ch )
	{
		case ':':		// String starts witha time
			hour = day;
			day = 1;	// Set date to 1/1/0

			goto get_time;

		case '-':
		case '/':		// String starts with a date
			break;

		default:
			return 2;	// Any other char = error
	}

	if ( mtkit_strtoi ( s, &month, &s, 0 ) )	// MONTH
	{
		return 3;
	}

	while ( isspace ( s[0] ) )
	{
		s++;
	}

	ch = *s++;

	if ( ch != '/' && ch != '-' )
	{
		// Not valid date separator

		return 4;
	}

	if ( mtkit_strtoi ( s, &year, &s, 0 ) )
	{
		// YEAR

		return 5;
	}

	if ( day > 31 )
	{
		// If day is > 31 interpret date text as y-m-d

		hour = year;
		year = day;
		day = hour;
		hour = 0;
	}

	while ( isspace ( s[0] ) )
	{
		s++;
	}

	if ( s[0] == 0 )
	{
		// String doesn't contain a time

		goto finish;
	}

	if ( mtkit_strtoi ( s, &hour, &s, 0 ) )
	{
		// HOUR

		return 6;
	}

	while ( isspace ( s[0] ) )
	{
		s++;
	}

	if ( *s++ != ':' )
	{
		// ':' should be present after hour

		return 7;
	}

get_time:

	if ( mtkit_strtoi ( s, &minute, &s, 0 ) )
	{
		// MINUTE

		return 8;
	}

	while ( isspace ( s[0] ) )
	{
		s++;
	}

	if ( s[0] == ':' )		// Second ':' so seconds given
	{
		s++;

		if ( mtkit_strtoi ( s, &second, &s, 0 ) )
		{
			// SECOND

			return 9;
		}
	}

	if ( mtkit_strnonspaces ( s ) )
	{
		// Trailing non-whitespaces are illegal

		return 10;
	}

finish:

	return mtkit_itoddt ( day, month, year, hour, minute, second, result );
}



// WARNING - February must be set by calling func!
static int mlen[12] = { 31, 0, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };



int mtkit_is_leap_year (
	int		year
	)
{
	if ( (year % 400 ) == 0 ) return 1;
	if ( (year % 100 ) == 0 ) return 0;
	if ( (year % 4) == 0 ) return 1;	// Leap year

	return 0;				// Not leap year
}

int mtkit_itoddt (
	int		day,
	int		month,
	int		year,
	int		hour,
	int		minute,
	int		second,
	double		* result
	)
{
	int		i,
			j;


	// General bounds
	if ( day < 1 || day > 31 )		return 1;
	if ( month < 1 || month > 12 )		return 1;
	if ( year < 0 )				return 1;
	if ( hour < 0 || hour > 23 )		return 1;
	if ( minute < 0 || minute > 59 )	return 1;
	if ( second < 0 || second > 59 )	return 1;

	// Leap year
	if ( mtkit_is_leap_year ( year ) )
	{
		mlen[1] = 29;
	}
	else
	{
		mlen[1] = 28;
	}

	// Check day is valid for given month
	if ( day > mlen[month-1] )
	{
		return 1;
	}

	if ( result )
	{
		result[0] = ( (double)(hour * 3600 + minute * 60 + second) ) /
			86400;

		// Whole years
		result[0] += 365 * year;

		// Leap days (whole 400 year chunks before this year)
		i = year / 400;

		// 97 leap days per 400 years = 25 + 24 + 24 + 24
		i = i * 97;

		result[0] += i;

		// Leap days (whole century chunk before this year)
		i = year % 400;

		// We have passed the first year every 400 years, e.g. 2000
		if ( i > 0 )
		{
			result[0] += 1;
			// 0, 400, ..., 2000 have leap days
		}

		j = (i - 1) / 100;
		if ( i > 0 )
		{
			// *04, *08, *12, etc have leap days
			// but not 100, 200, 300
			result[0] += (i-1) / 4 - j;
		}

		// Add full months (before this month)
		for ( i = 0; i < ( month - 1 ); i++ )
		{
			result[0] += mlen[i];
		}

		result[0] += day - 1;
	}

	return 0;
}

static void tweak_final_century (
	int		* i,
	int		* tot,
	int		* yr
	)
{
	int		j = 0;


	// Remaining years follow 365, 365, 365, 366 pattern
	if ( i[0] >= 1461 )	// 4 year chunk: 365 + 365 + 365 + 366
	{
		j = i[0] / 1461;
		i[0] -= 1461 * j;
		tot[0] += 1461 * j;
		yr[0] += j * 4;
	}

	if ( i[0] == 1460 )
	{
		// Catch when on last day of fourth year,
		// i.e. 365 + 365 + 365 + 365

		j = 3;
	}
	else
	{
		// 0..3 years, each 365 days

		j = i[0] / 365;
	}

	i[0] -= 365 * j;
	tot[0] += 365 * j;
	yr[0] += j;
}

int mtkit_ddttoi (
	double		datetime,
	int		* day,
	int		* month,
	int		* year,
	int		* hour,
	int		* minute,
	int		* second
	)
{
	double		time;
	int		days,
			secs,
			i,
			j,
			tot,
			yr;


	days = (int)datetime;
	if ( datetime < 0 || days < 0 )
	{
		return 1;
	}

	time = datetime - days;
	secs = 0.5 + time * 86400;
		// 0.5 rounds as time is always positive (no need for rint)

	if ( second )
	{
		second[0] = secs % 60;
	}

	secs /= 60;

	if ( minute )
	{
		minute[0] = secs % 60;
	}

	secs /= 60;

	if ( hour )
	{
		hour[0] = secs % 24;
	}

	tot = 0;

	i = days / 146097;		// How many whole 4 centuries are here?
	tot += i * 146097;
	yr = 400 * i;

	i = days - tot;			// i = remaining days (0..399 years)

	if ( i >= 36525 )		// 100-399 years left
	{
		i -= 36525;		// Lose the whole of the first century
		tot += 36525;
		yr += 100;

		j = i / 36524;		// Remaining whole centuries
		i -= j * 36524;
		tot += j * 36524;
		yr += j * 100;

		if ( i >= 365 )
		{
			// First year of final century has 365 days
			i -= 365;
			tot += 365;
			yr += 1;

			tweak_final_century ( &i, &tot, &yr );
		}
	}
	else				// 0-99 years left
	{
		if ( i >= 366 )
		{
			// First year of final century has 366 days

			// 1-99 years left

			i -= 366;
			tot += 366;
			yr += 1;

			tweak_final_century ( &i, &tot, &yr );
		}
	}

	// Year found
	if ( year )
	{
		year[0] = yr;
	}

	if ( mtkit_is_leap_year ( yr ) )
	{
		mlen[1] = 29;
	}
	else
	{
		mlen[1] = 28;
	}

	for ( i = 0; i < 12; i++ )
	{
		if ( ( tot + mlen[i] ) > days )
		{
			break;
		}

		tot += mlen[i];
	}

	// Month found
	if ( month )
	{
		month[0] = i + 1;
	}

	// Day is the remainder
	if ( day )
	{
		day[0] = days - tot + 1;
	}

	return 0;
}

int mtkit_ddt_weekday (
	double		datetime
	)
{
	int		i = datetime;


	if ( i < 0 )
	{
		return -1;
	}

	return ( ( i + 6 ) % 7 );
}



typedef struct
{
	char len, c1, c2;
} iso2utf;



static const iso2utf iso2utf_table[256] = {
{ 1, 0, 0 }, { 1, 1, 0 }, { 1, 2, 0 }, { 1, 3, 0 },
{ 1, 4, 0 }, { 1, 5, 0 }, { 1, 6, 0 }, { 1, 7, 0 },
{ 1, 8, 0 }, { 1, 9, 0 }, { 1, 10, 0 }, { 1, 11, 0 },
{ 1, 12, 0 }, { 1, 13, 0 }, { 1, 14, 0 }, { 1, 15, 0 },
{ 1, 16, 0 }, { 1, 17, 0 }, { 1, 18, 0 }, { 1, 19, 0 },
{ 1, 20, 0 }, { 1, 21, 0 }, { 1, 22, 0 }, { 1, 23, 0 },
{ 1, 24, 0 }, { 1, 25, 0 }, { 1, 26, 0 }, { 1, 27, 0 },
{ 1, 28, 0 }, { 1, 29, 0 }, { 1, 30, 0 }, { 1, 31, 0 },
{ 1, 32, 0 }, { 1, 33, 0 }, { 1, 34, 0 }, { 1, 35, 0 },
{ 1, 36, 0 }, { 1, 37, 0 }, { 1, 38, 0 }, { 1, 39, 0 },
{ 1, 40, 0 }, { 1, 41, 0 }, { 1, 42, 0 }, { 1, 43, 0 },
{ 1, 44, 0 }, { 1, 45, 0 }, { 1, 46, 0 }, { 1, 47, 0 },
{ 1, 48, 0 }, { 1, 49, 0 }, { 1, 50, 0 }, { 1, 51, 0 },
{ 1, 52, 0 }, { 1, 53, 0 }, { 1, 54, 0 }, { 1, 55, 0 },
{ 1, 56, 0 }, { 1, 57, 0 }, { 1, 58, 0 }, { 1, 59, 0 },
{ 1, 60, 0 }, { 1, 61, 0 }, { 1, 62, 0 }, { 1, 63, 0 },
{ 1, 64, 0 }, { 1, 65, 0 }, { 1, 66, 0 }, { 1, 67, 0 },
{ 1, 68, 0 }, { 1, 69, 0 }, { 1, 70, 0 }, { 1, 71, 0 },
{ 1, 72, 0 }, { 1, 73, 0 }, { 1, 74, 0 }, { 1, 75, 0 },
{ 1, 76, 0 }, { 1, 77, 0 }, { 1, 78, 0 }, { 1, 79, 0 },
{ 1, 80, 0 }, { 1, 81, 0 }, { 1, 82, 0 }, { 1, 83, 0 },
{ 1, 84, 0 }, { 1, 85, 0 }, { 1, 86, 0 }, { 1, 87, 0 },
{ 1, 88, 0 }, { 1, 89, 0 }, { 1, 90, 0 }, { 1, 91, 0 },
{ 1, 92, 0 }, { 1, 93, 0 }, { 1, 94, 0 }, { 1, 95, 0 },
{ 1, 96, 0 }, { 1, 97, 0 }, { 1, 98, 0 }, { 1, 99, 0 },
{ 1, 100, 0 }, { 1, 101, 0 }, { 1, 102, 0 }, { 1, 103, 0 },
{ 1, 104, 0 }, { 1, 105, 0 }, { 1, 106, 0 }, { 1, 107, 0 },
{ 1, 108, 0 }, { 1, 109, 0 }, { 1, 110, 0 }, { 1, 111, 0 },
{ 1, 112, 0 }, { 1, 113, 0 }, { 1, 114, 0 }, { 1, 115, 0 },
{ 1, 116, 0 }, { 1, 117, 0 }, { 1, 118, 0 }, { 1, 119, 0 },
{ 1, 120, 0 }, { 1, 121, 0 }, { 1, 122, 0 }, { 1, 123, 0 },
{ 1, 124, 0 }, { 1, 125, 0 }, { 1, 126, 0 }, { 1, 127, 0 },
{ 2, -62, -128 }, { 2, -62, -127 }, { 2, -62, -126 }, { 2, -62, -125 },
{ 2, -62, -124 }, { 2, -62, -123 }, { 2, -62, -122 }, { 2, -62, -121 },
{ 2, -62, -120 }, { 2, -62, -119 }, { 2, -62, -118 }, { 2, -62, -117 },
{ 2, -62, -116 }, { 2, -62, -115 }, { 2, -62, -114 }, { 2, -62, -113 },
{ 2, -62, -112 }, { 2, -62, -111 }, { 2, -62, -110 }, { 2, -62, -109 },
{ 2, -62, -108 }, { 2, -62, -107 }, { 2, -62, -106 }, { 2, -62, -105 },
{ 2, -62, -104 }, { 2, -62, -103 }, { 2, -62, -102 }, { 2, -62, -101 },
{ 2, -62, -100 }, { 2, -62, -99 }, { 2, -62, -98 }, { 2, -62, -97 },
{ 2, -62, -96 }, { 2, -62, -95 }, { 2, -62, -94 }, { 2, -62, -93 },
{ 2, -62, -92 }, { 2, -62, -91 }, { 2, -62, -90 }, { 2, -62, -89 },
{ 2, -62, -88 }, { 2, -62, -87 }, { 2, -62, -86 }, { 2, -62, -85 },
{ 2, -62, -84 }, { 2, -62, -83 }, { 2, -62, -82 }, { 2, -62, -81 },
{ 2, -62, -80 }, { 2, -62, -79 }, { 2, -62, -78 }, { 2, -62, -77 },
{ 2, -62, -76 }, { 2, -62, -75 }, { 2, -62, -74 }, { 2, -62, -73 },
{ 2, -62, -72 }, { 2, -62, -71 }, { 2, -62, -70 }, { 2, -62, -69 },
{ 2, -62, -68 }, { 2, -62, -67 }, { 2, -62, -66 }, { 2, -62, -65 },
{ 2, -61, -128 }, { 2, -61, -127 }, { 2, -61, -126 }, { 2, -61, -125 },
{ 2, -61, -124 }, { 2, -61, -123 }, { 2, -61, -122 }, { 2, -61, -121 },
{ 2, -61, -120 }, { 2, -61, -119 }, { 2, -61, -118 }, { 2, -61, -117 },
{ 2, -61, -116 }, { 2, -61, -115 }, { 2, -61, -114 }, { 2, -61, -113 },
{ 2, -61, -112 }, { 2, -61, -111 }, { 2, -61, -110 }, { 2, -61, -109 },
{ 2, -61, -108 }, { 2, -61, -107 }, { 2, -61, -106 }, { 2, -61, -105 },
{ 2, -61, -104 }, { 2, -61, -103 }, { 2, -61, -102 }, { 2, -61, -101 },
{ 2, -61, -100 }, { 2, -61, -99 }, { 2, -61, -98 }, { 2, -61, -97 },
{ 2, -61, -96 }, { 2, -61, -95 }, { 2, -61, -94 }, { 2, -61, -93 },
{ 2, -61, -92 }, { 2, -61, -91 }, { 2, -61, -90 }, { 2, -61, -89 },
{ 2, -61, -88 }, { 2, -61, -87 }, { 2, -61, -86 }, { 2, -61, -85 },
{ 2, -61, -84 }, { 2, -61, -83 }, { 2, -61, -82 }, { 2, -61, -81 },
{ 2, -61, -80 }, { 2, -61, -79 }, { 2, -61, -78 }, { 2, -61, -77 },
{ 2, -61, -76 }, { 2, -61, -75 }, { 2, -61, -74 }, { 2, -61, -73 },
{ 2, -61, -72 }, { 2, -61, -71 }, { 2, -61, -70 }, { 2, -61, -69 },
{ 2, -61, -68 }, { 2, -61, -67 }, { 2, -61, -66 }, { 2, -61, -65 }
};

char * mtkit_iso8859_to_utf8 (
	char		* input,
	int		bytes,
	int		* new_size
	)
{
	unsigned char	* src,
			* end,
			uc;
	char		* newstr,
			* dest;
	int		totlen;


	if ( bytes < 0 || ! input )
	{
		return NULL;
	}

	if ( bytes == 0 )
	{
		bytes = strlen ( input );
	}

	// Pass 1 - Calculate length of new string in bytes

	totlen = 1;		// NUL terminator
	src = (unsigned char *) input;
	end = src + bytes;

	while ( src < end )
	{
		uc = *src++;
		totlen += iso2utf_table[uc].len;
	}

	newstr = calloc ( totlen, 1 );

	if ( ! newstr )
	{
		return NULL;
	}

	if ( new_size )
	{
		new_size[0] = totlen;
	}

	// Pass 2 - Do conversion

	src = (unsigned char *) input;
	dest = newstr;
	end = src + bytes;

	while ( src < end )
	{
		uc = *src++;
		if ( iso2utf_table[uc].len == 1 )
		{
			*dest++ = iso2utf_table[uc].c1;
		}
		else
		{
			*dest++ = iso2utf_table[uc].c1;
			*dest++ = iso2utf_table[uc].c2;
		}
	}

	return newstr;
}

