/*
	Copyright (C) 2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mtkit.h>
#include <mtcelledit.h>

#include "mtotp.h"



static mtOTP		* mtotp
			;
static int		mash_filetot,	// Next free slot in mash_list[]
			i_error,
			i_filetot,
			i_mode,
			i_tmp,
			i_verbose	// 1 = Verbose output
			;
static char		* mash_list [ MTOTP_MASH_MAX ],
			* s_output,	// Output filename
			* s_comment,	// Pad comment
			* s_root	// Pad root directory manual setting
			;



enum
{
	MODE_AUTO		= 0,
	MODE_GENERATE		= 1,
	MODE_ENCRYPT		= 2,
	MODE_DECRYPT		= 3,
	MODE_ANALYSE		= 4,
	MODE_MASH		= 5,

	MODE_TOTAL
};



static int mash_file_add (
	char		* filename
	)
{
	if ( mash_filetot >= MTOTP_MASH_MAX )
	{
		fprintf ( stderr, "Too many files.  "
			"Maximum mash files = %i\n\n", MTOTP_MASH_MAX );

		return 1;
	}

	mash_list[ mash_filetot ] = filename;
	mash_filetot ++;

	return 0;
}

static int arg_file_func (
	char		* filename
	)
{
	int		mode = i_mode;


	i_filetot ++;

	if ( mode == MODE_MASH )
	{
		if ( mash_file_add ( filename ) )
		{
			i_error = 1;

			return 1;
		}

		return 0;
	}

	if ( mode == MODE_AUTO )
	{
		if ( mtotp_query_filename ( filename, NULL, NULL ) )
		{
			mode = MODE_ENCRYPT;
		}
		else
		{
			mode = MODE_DECRYPT;
		}
	}

	switch ( mode )
	{
	case MODE_GENERATE:
		i_error = mtotp_generate ( mtotp, filename, s_comment,
			i_verbose );
		break;

	case MODE_ENCRYPT:
		i_error = mtotp_encrypt ( mtotp, filename, i_verbose );
		break;

	case MODE_DECRYPT:
		i_error = mtotp_decrypt ( mtotp, filename, i_verbose );
		break;

	case MODE_ANALYSE:
		i_error = mtotp_analyse ( mtotp, filename, i_verbose );
		break;

	default:
		// Should never happen

		fprintf ( stderr, "Error: file_func: invalid mode %i\n",
			i_mode );
		i_error = 1;

		return 1;		// Stop
	}

	return i_error;
}

static int arg_error_func (
	int		error,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	fprintf ( stderr, "error_func: Argument ERROR! - num = %i arg = %i/%i",
		error, arg, argc );

	if ( arg < argc )
	{
		fprintf ( stderr, " '%s'", argv[arg] );
	}

	fprintf ( stderr, "\n" );

	i_error = 1;

	return 1;			// Stop parsing
}

static int arg_output_file (
	mtArg		* mtarg,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	if ( i_mode == MODE_MASH )
	{
		i_error = mtotp_mash ( mtotp, mash_list, mash_filetot, s_output,
			i_verbose );
	}

	return i_error;
}

static int arg_root (
	mtArg		* mtarg,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	if ( mtotp_is_pad_loaded ( mtotp ) == MTOTP_PAD_LOADED )
	{
		fprintf ( stderr, "Root directory already set.\n\n" );
		i_error = 1;

		return 1;
	}

	if ( mtotp_set_root ( mtotp, s_root ) )
	{
		i_error = 1;

		return 1;
	}

	return 0;
}

static int arg_mode (
	mtArg		* mtarg,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	if ( i_mode == 0 )
	{
		i_mode = i_tmp;
	}
	else
	{
		fprintf ( stderr, "Error - You cannot set the mode twice."
			"\n\n" );
		i_error = 1;
	}

	return i_error;
}

static int arg_print_version (
	mtArg		* mtarg,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	printf ( "%s\n\n", VERSION );

	i_error = -1;			// Stop actions, but terminate with 0

	return 1;			// Stop parsing
}

static int arg_print_help (
	mtArg		* mtarg,
	int		arg,
	int		argc,
	char		* argv[]
	)
{
	arg_print_version ( mtarg, arg, argc, argv );

	printf ( "Usage: %s [OPTION]... [FILE]...\n"
"\n"
"Options:\n"
"  --help               Output this help.\n"
"  --version            Output version information.\n"
"  -a                   Analysis mode.\n"
"  -c STRING            Add this comment to pad database when generating.\n"
"  -d                   Decryption mode.\n"
"  -e                   Encryption mode.\n"
"  -g                   Generation mode.\n"
"  -m                   Mash mode.\n"
"  -o STRING            Output file for mash mode.\n"
"  -root STRING         Use this directory for the pad and index files.\n"
"  -v                   Verbose.\n"
"\n"
"When no files are passed the program returns pad state information.\n"
"If no mode is set and a file is passed, encryption or\n"
"decryption is set by checking the input filename suffix (i.e. if it fits\n"
"the form FILE.mtotp.ROW.OFFSET then the mode is decryption).\n"
"Once the mode has been set it cannot be changed.\n"
"Once the root directory has been set it cannot be changed.\n"
"\n"
		, BIN_NAME );

	i_error = -1;			// Stop actions, but terminate with 0

	return 1;			// Stop parsing
}

int main (
	int		argc,
	char		* argv[]
	)
{
	static mtArg	arg_list[] = {
	{ "-help",	MTKIT_ARG_SWITCH, &i_tmp, 1, arg_print_help },
	{ "-version",	MTKIT_ARG_SWITCH, &i_tmp, 2, arg_print_version },

	{ "a",		MTKIT_ARG_SWITCH, &i_tmp, MODE_ANALYSE, arg_mode },
	{ "c",		MTKIT_ARG_STRING, &s_comment },
	{ "d",		MTKIT_ARG_SWITCH, &i_tmp, MODE_DECRYPT, arg_mode },
	{ "e",		MTKIT_ARG_SWITCH, &i_tmp, MODE_ENCRYPT, arg_mode },
	{ "g",		MTKIT_ARG_SWITCH, &i_tmp, MODE_GENERATE, arg_mode },
	{ "m",		MTKIT_ARG_SWITCH, &i_tmp, MODE_MASH, arg_mode },
	{ "o",		MTKIT_ARG_STRING, &s_output, 0, arg_output_file },
	{ "root",	MTKIT_ARG_STRING, &s_root, 0, arg_root },
	{ "v",		MTKIT_ARG_SWITCH, &i_verbose, 1 },

	{ NULL }
	};


	ced_init ();			// Set up random seed
	mtotp = mtotp_new ();

	if ( ! mtotp )
	{
		return 1;
	}

	// Parse & action the command line arguments
	mtkit_arg_parse ( argc, argv, arg_list, arg_file_func, arg_error_func );

	if ( i_error == 0 )
	{
		if ( i_filetot == 0 )
		{
			i_error = mtotp_query ( mtotp, NULL, 1, i_verbose );
		}
		else
		{
			if ( i_verbose )
			{
				printf ( "%i file(s) processed.\n\n",
					i_filetot );
			}
		}
	}

	mtotp_free ( mtotp );

	if ( i_error == -1 )
	{
		return 0;
	}

	return i_error;
}
