/*
	Copyright (C) 2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>

#include <mtkit.h>
#include <mtcelledit.h>

#include "mtotp.h"



/*
NOTE:
BUFLEN == PADSIZE_MIN as this makes the code easier to manage as there is never
more than 2 pad files required to fill a buffer.
*/

#define BUFLEN			1048576
#define RANDOM_NUM_8		( ( rand () >> 5 ) & 255 )

#define COUNTER_MAX		9999
#define COUNTER_PRINTF_DOUBLE	"%04.0f"
#define COUNTER_PRINTF_INT	"%04i"

#define PADSIZE_MIN		1048576.0
#define PADSIZE_MAX		35184372088832.0

#define MTOTP_DIR_HOME		".config/mtotp"
#define MTOTP_DIR_LOCK		"conf/lock"
#define MTOTP_DIR_INDEX		"conf/index.tsv"
#define MTOTP_DIR_PAD		"pad_"

#define SHEET_COL_PADNUM	1
#define SHEET_COL_PADSIZE	2
#define SHEET_COL_PADUSED	3
#define SHEET_COL_PADFILENAME	4
#define SHEET_COL_PADCOMMENT	5



struct _mtOTP
{
	int		error
			;
	char		filename_root [ MTKIT_PATHBUF_MAX ],
			filename_lock [ MTKIT_PATHBUF_MAX ],
			filename_index [ MTKIT_PATHBUF_MAX ]
			;
	CedSheet	* sheet_index;	// Current OTP file index
	unsigned char	buf_in [ BUFLEN ],
			buf_xor [ BUFLEN ]
			;
	unsigned int	rowtot,		// Total rows in index sheet
			rowenc		// Next row to be used for encryption
			;
};



mtOTP * mtotp_new ( void )
{
	mtOTP		* mtotp;


	mtotp = calloc ( 1, sizeof ( mtOTP ) );

	if ( ! mtotp )
	{
		fprintf ( stderr, "Unable to create new mtOTP.\n\n" );
	}

	return mtotp;
}

static int save_index (
	mtOTP		* mtotp
	)
{
	if ( mtotp->filename_index [0] == 0 )
	{
		return 0;		// Nothing to do
	}

	if ( ced_sheet_save ( mtotp->sheet_index, mtotp->filename_index,
		CED_FILE_TYPE_TSV_CONTENT ) )
	{
		fprintf ( stderr, "Unable to save index file: %s\n\n",
			mtotp->filename_index );

		return 1;
	}

	return 0;			// Success
}

static int remove_lock_file (
	mtOTP		* mtotp
	)
{
	if ( mtotp->filename_lock [0] == 0 )
	{
		return 0;		// Nothing to do
	}

	if ( remove ( mtotp->filename_lock ) )
	{
		fprintf ( stderr, "Unable to remove lock file: %s\n\n",
			mtotp->filename_lock );

		return 1;
	}

	return 0;
}

static void rename_current_index (
	mtOTP		* mtotp
	)
{
	time_t		now;
	struct tm	* now_tm;
	char		tbuf[ MTKIT_PATHBUF_MAX ] = { 0 };
	int		i;


	if (	mtotp->filename_index [0] == 0 ||
		mtkit_file_valid ( mtotp->filename_index )
		)
	{
		return;			// No file to rename
	}

	now = time ( NULL );
	now_tm = localtime ( &now );

	for ( i = 0; i <= COUNTER_MAX; i++ )
	{
		snprintf ( tbuf, MTKIT_PATHBUF_MAX, "%s_%04i%02i%02i_%04i",
			mtotp->filename_index, now_tm->tm_year + 1900,
			now_tm->tm_mon + 1, now_tm->tm_mday, i );

		if ( mtkit_file_valid ( tbuf ) )
		{
			// No valid file so we can use this one
			break;
		}
	}

	if ( i > COUNTER_MAX )
	{
		fprintf ( stderr, "Unable to find available backup index "
			"number.\n\n" );

		return;
	}

	if ( rename ( mtotp->filename_index, tbuf ) )
	{
		fprintf ( stderr, "Unable to rename index file to: %s\n\n",
			tbuf );

		return;
	}

	// Success
}

int mtotp_free (
	mtOTP		* mtotp
	)
{
	if ( ! mtotp )
	{
		return 1;
	}

	rename_current_index ( mtotp );
	save_index ( mtotp );
	remove_lock_file ( mtotp );

	ced_sheet_destroy ( mtotp->sheet_index );
	free ( mtotp );

	return 0;
}

static double get_cell_value (
	CedSheet	* sheet,
	int		row,
	int		col
	)
{
	CedCell		* cell;


	cell = ced_sheet_get_cell ( sheet, row, col );
	if ( cell )
	{
		return cell->value;
	}

	return 0;
}

static int cb_scan_rowenc (
	CedSheet	* sheet,
	CedCell		* cell,
	unsigned int	row,
	unsigned int	col,
	void		* user_data
	)
{
	mtOTP		* mtotp = user_data;
	double		used = 0;


	used = get_cell_value ( mtotp->sheet_index, row, SHEET_COL_PADUSED );

	if ( used < 0 || used > cell->value )
	{
		fprintf ( stderr, "Index corruption at row %ui.\n\n",
			row );

		mtotp->error = 1;

		return 1;
	}

	if ( used < cell->value )
	{
		mtotp->rowenc = row;

		return 1;		// Stop scanning
	}

	return 0;			// Continue scanning
}

static int validate_filename (
	char		* filename
	)
{
	if ( ! filename )
	{
		fprintf ( stderr, "NULL Filename.\n\n" );

		return 1;
	}

	if ( strlen ( filename ) > (MTKIT_PATHBUF_MAX - 32) )
	{
		fprintf ( stderr, "Filename too long '%s'.\n\n", filename );

		return 1;
	}

	return 0;
}

static int set_pad_root (
	mtOTP		* mtotp,	// Caller must check this exists
	char		* root,
	int		home
	)
{
	if ( ! root )
	{
		fprintf ( stderr, "Bad root directory name (NULL).\n\n" );

		return 1;
	}

	if ( validate_filename ( root ) )
	{
		return 1;
	}

	if ( mtotp->sheet_index )
	{
		rename_current_index ( mtotp );
		save_index ( mtotp );
		remove_lock_file ( mtotp );

		memset ( mtotp, 0, sizeof ( mtotp[0] ) );
	}

	if ( home )
	{
		snprintf ( mtotp->filename_root, MTKIT_PATHBUF_MAX, "%s/%s",
			root, MTOTP_DIR_HOME );
	}
	else
	{
		snprintf ( mtotp->filename_root, MTKIT_PATHBUF_MAX, "%s",
			root );
	}

	snprintf ( mtotp->filename_lock, MTKIT_PATHBUF_MAX, "%s/%s",
		mtotp->filename_root, MTOTP_DIR_LOCK );
	snprintf ( mtotp->filename_index, MTKIT_PATHBUF_MAX, "%s/%s",
		mtotp->filename_root, MTOTP_DIR_INDEX );

	if ( 0 == mtkit_file_valid ( mtotp->filename_lock ) )
	{
		fprintf ( stderr, "Lock file already exists: '%s'\n\n",
			mtotp->filename_lock );

		return 1;
	}

	if ( mtkit_file_save ( mtotp->filename_lock, mtotp->filename_lock,
		0, 0 ) )
	{
		fprintf ( stderr, "Unable to create lock file: '%s'\n\n",
			mtotp->filename_lock );

		return 1;
	}

	mtotp->sheet_index = ced_sheet_load ( mtotp->filename_index, NULL,
		NULL );

	if ( ! mtotp->sheet_index )
	{
		mtotp->sheet_index = ced_sheet_new ();
		if ( ! mtotp->sheet_index )
		{
			fprintf ( stderr, "Unable to create new sheet "
				"index in memory.\n\n" );

			return 1;
		}
	}

	ced_sheet_get_geometry ( mtotp->sheet_index, &mtotp->rowtot, NULL );

	ced_sheet_scan_area ( mtotp->sheet_index, 1, SHEET_COL_PADSIZE, 0, 1,
		cb_scan_rowenc, mtotp );

	if ( mtotp->error )
	{
		mtotp->error = 0;

		return 1;
	}

	return 0;
}

int mtotp_set_root (
	mtOTP		* mtotp,
	char		* root
	)
{
	if ( ! mtotp )
	{
		return 1;
	}

	return set_pad_root ( mtotp, root, 0 );
}

static int mtotp_check_setup (		// Set root to HOME if not already set
	mtOTP		* mtotp
	)
{
	if ( ! mtotp )
	{
		return 1;
	}

	if ( ! mtotp->sheet_index )
	{
		return set_pad_root ( mtotp, mtkit_file_home (), 1 );
	}

	return 0;
}

int mtotp_query (
	mtOTP		* mtotp,
	double		* total,
	int		output,
	int		verbose
	)
{
	int		i,
			empty = 0;
	double		grand = 0.0,
			available = 0.0,
			avtot = 0.0,
			used,
			size;


	if ( mtotp_check_setup ( mtotp ) )
	{
		return 1;
	}

	for ( i = 1; i <= mtotp->rowtot; i++ )
	{
		size = get_cell_value ( mtotp->sheet_index, i,
			SHEET_COL_PADSIZE );

		used = get_cell_value ( mtotp->sheet_index, i,
			SHEET_COL_PADUSED );

		available = size - used;

		if (	size < 0	||
			used < 0	||
			used > size	||
			( empty && used != 0 )
			)
		{
			fprintf ( stderr, "Error in index row %i\n", i );

			return 1;
		}

		grand += size;
		avtot += available;

		if ( output && verbose )
		{
			printf ( "Pad %i = %.3f GB Available\n", i,
				available / 1024 / 1024 / 1024 );
		}

		if ( available > 0 )
		{
			empty = 1;
			// All of the following pads should be unused.
			// If not that the index is corrupt.
		}
	}

	if ( output )
	{
		if ( verbose )
		{
			puts ( "" );
		}

		printf ( "Next available pad    %i\n\n", mtotp->rowenc );
		printf ( "Total Pad Size        %.3f GB\n",
			grand / 1024 / 1024 / 1024 );
		printf ( "Total Pad Available   %.3f GB\n",
			avtot / 1024 / 1024 / 1024 );
	}

	if ( total )
	{
		total[0] = avtot;
	}

	return 0;			// Success
}

int mtotp_is_pad_loaded (
	mtOTP		* mtotp
	)
{
	if ( ! mtotp || ! mtotp->sheet_index )
	{
		return 1;
	}

	return 0;
}

char * mtotp_next_char_left (
	char		* a,
	char		* b,
	char		ch
	)
{
	for ( ; b >= a; b -- )
	{
		if ( b[0] == ch )
		{
			return b;
		}
	}

	return NULL;
}

int mtotp_query_filename (
	char		* filename,
	double		* row,
	double		* offset
	)
{
	double		num_r,
			num_o
			;
	char		* s1 = NULL,
			* s2 = NULL,
			* s3 = NULL;


	if ( ! filename )
	{
		return 1;
	}

	// Is filename of this form?
	//	S.mtotp.A.B
	// A = Double > 0 && < CED_MAX_ROW
	// B = Double > 0
	// S = At least one character

	s1 = strrchr ( filename, '.' );
	if ( ! s1 )
	{
		// .B is missing
		goto error;
	}

	s1[0] = 0;

	s2 = mtotp_next_char_left ( filename, s1 - 1, '.' );
	if ( ! s2 )
	{
		// .A is missing
		goto error;
	}

	s2[0] = 0;

	s3 = mtotp_next_char_left ( filename, s2 - 1, '.' );
	if ( ! s3 || ! s3[1] )
	{
		// .mtotp is missing
		goto error;
	}

	if ( s3 == filename )
	{
		// S is empty
		goto error;
	}

	if ( 0 != strcmp ( s3 + 1, "mtotp" ) )
	{
		goto error;
	}

	if (	mtkit_strtod ( s2 + 1, &num_r, NULL, 1 )	||
		num_r <= 0					||
		num_r > CED_MAX_ROW
		)
	{
		goto error;
	}

	if (	mtkit_strtod ( s1 + 1, &num_o, NULL, 1 ) ||
		num_o < 0
		)
	{
		goto error;
	}

	s1[0] = '.';
	s2[0] = '.';

	if ( row )
	{
		row[0] = num_r;
	}

	if ( offset )
	{
		offset[0] = num_o;
	}

	return 0;			// Success

error:
	if ( s1 )
	{
		s1[0] = '.';
	}

	if ( s2 )
	{
		s2[0] = '.';
	}

	return 1;
}

static void print_errno (
	int		e
	)
{
	char		* se;


	se = strerror ( e );
	if ( ! se )
	{
		se = "?";
	}

	fprintf ( stderr, "errno = %i - %s\n\n", e, se );
}

static void print_title (
	char		* txt
	)
{
	char		* ds,
			* src;


	ds = strdup ( txt );

	if ( ! ds )
	{
		puts ( txt );

		return;
	}

	for ( src = ds; src[0] != 0; src ++ )
	{
		src[0] = '-';
	}

	puts ( ds );
	puts ( txt );
	puts ( ds );

	free ( ds );
}

static int fio_check_no_output_file (
	char		* filename
	)
{
	if ( 0 == mtkit_file_valid ( filename ) )
	{
		fprintf ( stderr, "Output filename '%s' already exists.\n\n",
			filename );

		return 1;
	}

	return 0;			// No file
}

static FILE * fio_open (
	char		* filename,
	char		* rw
	)
{
	FILE		* fp;


	fp = fopen ( filename, rw );
	if ( ! fp )
	{
		fprintf ( stderr, "Unable to open '%s'.\n\n", filename );

		return NULL;
	}

	return fp;			// Success
}

static void fio_close (
	FILE		* fp
	)
{
	if ( fp )
	{
		fclose ( fp );
	}
}

static int fio_read (
	FILE		* fp,
	void		* buf,
	int		tot,
	int		* buflen,
	char		* filename
	)
{
	buflen[0] = fread ( buf, 1, tot, fp );
	if ( buflen[0] < 1 )
	{
		if ( ! feof ( fp ) )
		{
			fprintf ( stderr, "Error reading file '%s'.\n\n",
				filename );

			return 1;
		}

		// EOF
	}

	return 0;		// Success
}

static int fio_write (
	FILE		* fp,
	void		* buf,
	int		tot,
	char		* filename
	)
{
	if ( fwrite ( buf, 1, tot, fp ) != tot )
	{
		fprintf ( stderr, "Error writing file '%s'.\n\n", filename );

		return 1;
	}

	return 0;		// Success
}

static void print_update_progress (
	int		verbose,
	double		btot,
	double		filesize
	)
{
	if ( verbose )
	{
		fprintf ( stderr, "\rProgress = %.1f%%",
			100 * btot / filesize );
		fflush ( stderr );
	}
}

static void print_update_progress_nl (
	int		verbose
	)
{
	if ( verbose )
	{
		fprintf ( stderr, "\n" );
	}
}

static double fio_get_filesize (
	char		* filename
	)
{
	double		size;


	size = mtkit_file_size ( filename );

	if ( size < 0 )
	{
		// Store current errno value just in case it gets changed
		int		e = errno;


		fprintf ( stderr, "Unable to access file '%s'\n\n", filename );
		mtkit_print_errno ( e );

		return -1;
	}

	return size;
}

static void xor_buffer (
	mtOTP		* mtotp,
	int		len
	)
{
	int		i;


	if ( len < 1 )
	{
		len = BUFLEN;
	}

	for ( i = 0; i < len; i++ )
	{
		mtotp->buf_in[ i ] ^= mtotp->buf_xor[ i ];
	}
}

int mtotp_mash (
	mtOTP		* mtotp,
	char		* mash_list[],
	int		mash_filetot,
	char		* output,
	int		verbose
	)
{
	FILE		* fp,
			* fp_mash[ MTOTP_MASH_MAX + 1 ] = { NULL }
			;
	double		max_filesize = 0,
			btot,
			btodo,
			bdone,
			fail;
	int		i,
			res = 0;
	unsigned char	* dest;


	if (	! mash_list			||
		mash_filetot < 1		||
		mash_filetot > MTOTP_MASH_MAX
		)
	{
		return 1;
	}

	if ( fio_check_no_output_file ( output ) )
	{
		return 1;
	}

	// Get largest filesize of all input files and use as target later
	for ( i = 0; i < mash_filetot; i++ )
	{
		btot = fio_get_filesize ( mash_list[ i ] );

		if ( btot < 0 )
		{
			return 1;
		}

		max_filesize = MAX ( max_filesize, btot );
	}

	// Prepare output file
	fp = fio_open ( output, "w" );
	if ( ! fp )
	{
		return 1;
	}

	// Prepare input file(s)
	for ( i = 0; i < mash_filetot; i++ )
	{
		fp_mash[ i ] = fio_open ( mash_list[ i ], "r" );

		if ( ! fp_mash[ i ] )
		{
			res = 1;

			goto finish;
		}
	}

	// Mash all files, one buffer at a time
	for ( btot = 0; btot < max_filesize; btot += BUFLEN )
	{
		dest = mtotp->buf_in;

		for ( i = 0; i < mash_filetot; i++ )
		{
			// Collect the next chunk of input file data
			fail = 0;

			for ( btodo = BUFLEN; btodo > 0; btodo -= bdone )
			{
				bdone = fread ( dest, 1, btodo, fp_mash[ i ] );

				if ( bdone < 1 )
				{
					if (	feof ( fp_mash[ i ] ) &&
						fail < 2
						)
					{
						// EOF reached, start again
						rewind ( fp_mash[ i ] );
						fail ++;
/*
NOTE
Using the fail variable we avoid 2 issues:
	* Zero length files causing infinite loop.
	* Input files with a size a multiple of BUFLEN.
*/

						continue;
					}

					fprintf ( stderr,
						"Problem loading %s\n\n",
						mash_list [ i ] );

					res = 1;

					goto finish;
				}

				dest += (int)bdone;
				fail = 0;
			}

			if ( i > 0 )
			{
				xor_buffer ( mtotp, 0 );
			}

			dest = mtotp->buf_xor;
		}

		btodo = MIN ( (max_filesize - btot), BUFLEN );

		if ( fio_write ( fp, mtotp->buf_in, btodo, output ) )
		{
			res = 1;

			goto finish;
		}

		print_update_progress ( verbose,
			MIN ( btot + BUFLEN, max_filesize ), max_filesize );
	}

finish:
	print_update_progress_nl ( verbose );

	for ( i = 0; fp_mash[ i ]; i++ )
	{
		fclose ( fp_mash[ i ] );
	}

	fclose ( fp );

	return res;
}



#define SEQ_MAX		5



int mtotp_analyse (
	mtOTP		* mtotp,
	char		* filename,
	int		verbose
	)
{
	FILE		* fp;
	int		buflen,
			i,
			seq = 0,
			last = -1,
			byte,
			delta
			;
	long long int	freq[256] = { 0 },
			ldelt[256] = { 0 },
			numrep[ SEQ_MAX ] = { 0 }
			;
	double		numrep_exp[ SEQ_MAX ],
			div,
			diff,
			norm,
			max,
			cum,
			filesize = 0,
			btot = 0
			;
	static char	* numrep_str[ SEQ_MAX ] = {
			"2",
			"3",
			"4",
			"5",
			"6+"
			}
			;
	unsigned char	* src,
			* end
			;


	if ( ! filename )
	{
		return 1;
	}

	fp = fio_open ( filename, "r" );
	if ( ! fp )
	{
		return 1;
	}

	filesize = fio_get_filesize ( filename );
	if ( filesize < 0 )
	{
		return 1;
	}

	printf ( "Analysing '%s'\n", filename );

	if ( fio_read ( fp, mtotp->buf_in, BUFLEN, &buflen, filename ) )
	{
		goto finish;		// I/O error
	}

	if ( buflen < 1 )
	{
		goto finish;		// EOF
	}

	src = mtotp->buf_in;
	end = src + buflen - 1;
	btot += buflen;

	byte = *src++;
	freq[ byte ] ++;
	last = byte;

	while ( 1 )
	{
		if ( src > end )
		{
			if ( fio_read ( fp, mtotp->buf_in, BUFLEN, &buflen,
				filename ) )
			{
				goto finish;	// I/O Error
			}

			if ( buflen < 1 )
			{
				break;		// EOF
			}

			src = mtotp->buf_in;
			end = src + buflen - 1;

			btot += buflen;
			print_update_progress ( verbose, btot, filesize );
		}

		byte = *src++;

		freq[ byte ] ++;

		delta = byte - last;

		if ( delta < 0 )
		{
			delta += 256;
		}

		ldelt[ delta ] ++;

		if ( byte == last )
		{
			seq ++;
		}
		else
		{
			if ( seq )
			{
				// Record the sequence number in table:
				// 2,3,4,5,6+

				seq --;

				if ( seq >= SEQ_MAX )
				{
					seq = SEQ_MAX - 1;
				}

				numrep[ seq ] ++;
			}

			seq = 0;
		}

		last = byte;
	}

	if ( seq )
	{
		// Record the sequence number in table:
		// 2,3,4,5,6+

		seq --;

		if ( seq >= SEQ_MAX )
		{
			seq = SEQ_MAX - 1;
		}

		numrep[ seq ] ++;

		seq = 0;
	}

finish:
	print_update_progress_nl ( verbose );

	fclose ( fp );

	printf ( "Bytes read = %.1f\n\n", btot );
	print_title ( "Byte Frequencies" );
	printf ( "\n" );

	if ( verbose )
	{
		printf ( "Byte\tTot\tNorm\tDiff%%\n" );
	}

	norm = btot / 256.0;
	max = 0;
	cum = 0;

	for ( i = 0; i < 256; i++ )
	{
		diff = -100 * ( norm - (double)freq[i] ) / norm;

		cum += fabs ( diff );
		max = MAX ( max, fabs ( diff ) );

		if ( verbose )
		{
			printf ( "%i\t%lli\t%.1f\t%.3f\n",
				i, freq[i], norm, diff );
		}
	}

	if ( verbose )
	{
		puts ( "" );
	}

	printf ( "Maximum Variation = %.3f\n", max );
	printf ( "Average Variation = %.3f\n", cum / 256.0 );
	printf ( "\n" );
	print_title ( "Delta Byte Frequencies" );
	printf ( "\n" );

	if ( verbose )
	{
		printf ( "Delta\tTot\tNorm\tDiff%%\n" );
	}

	max = 0;
	cum = 0;

	for ( i = 0; i < 256; i++ )
	{
		diff = -100 * ( norm - (double)ldelt[i] ) / norm;

		cum += fabs ( diff );
		max = MAX ( max, fabs ( diff ) );

		if ( verbose )
		{
			printf ( "%i\t%lli\t%.1f\t%.3f\n",
				i, ldelt[i], norm, diff );
		}
	}

	if ( verbose )
	{
		puts ( "" );
	}

	printf ( "Maximum Variation = %.3f\n", max );
	printf ( "Average Variation = %.3f\n", cum / 511.0 );
	printf ( "\n" );
	print_title ( "Identical Consecutive Bytes" );
	printf ( "\n" );
	printf ( "ICB's\tTot\tNorm\tDiff%%\n" );

	// Calculate expected number based on file size
	div = 256.0;

	for ( i = 0; i < SEQ_MAX; i++, div *= 256.0 )
	{
		numrep_exp[i] = btot / div;

		if ( numrep_exp[i] == 0 )
		{
			diff = 0.0;
		}
		else
		{
			diff = -100 * ( numrep_exp[i] - (double)numrep[i] ) /
				numrep_exp[i];
		}

		printf ( "%s\t%lli\t%.1f\t%.3f\n", numrep_str[i], numrep[i],
			numrep_exp[i], diff );
	}

	printf ( "\n\n" );

	return 0;			// Success
}

static FILE * fio_prep_pad (
	mtOTP		* mtotp,
	double		row,
	double		offset
	)
{
	double		num;
	char		buf [ MTKIT_PATHBUF_MAX ];
	FILE		* fp;


	num = get_cell_value ( mtotp->sheet_index, row, SHEET_COL_PADNUM );

	if ( num < 0 || num > COUNTER_MAX )
	{
		fprintf ( stderr, "Bad pad number ID at row %.0f\n\n", row );

		return NULL;
	}

	snprintf ( buf, MTKIT_PATHBUF_MAX, "%s/%s"COUNTER_PRINTF_DOUBLE".mtotp",
		mtotp->filename_root, MTOTP_DIR_PAD, num );

	fp = fio_open ( buf, "r" );
	if ( fp )
	{
		if ( fseeko ( fp, offset, SEEK_SET ) )
		{
			int		e = errno;


			fclose ( fp );
			fp = NULL;

			fprintf ( stderr, "Seek failure on file '%s' to "
				"offset %.0f.\n\n", buf, offset );
			print_errno ( e );
		}
	}
	else
	{
		fprintf ( stderr, "Unable to open pad file '%s'\n\n", buf );
	}

	return fp;
}

static int fio_xor_loop (
	mtOTP		* mtotp,
	int		verbose,
	char		* filename_in,
	char		* filename_out,
	double		row,
	double		offset,
	int		update_sheet
	)
{
	FILE		* fp_in = NULL,
			* fp_out = NULL,
			* fp_pad = NULL
			;
	CedSheet	* sh_temp = NULL;
	int		i_error = 0,
			buflen,
			bdone,
			btodo;
	double		filesize,
			btot = 0;


	filesize = fio_get_filesize ( filename_in );
	if ( filesize < 0 )
	{
		return 1;
	}

	if ( fio_check_no_output_file ( filename_out ) )
	{
		return 1;
	}

	fp_in = fio_open ( filename_in, "r" );
	if ( ! fp_in )
	{
		return 1;
	}

	fp_out = fio_open ( filename_out, "w" );
	if ( ! fp_out )
	{
		i_error = 1;

		goto finish;
	}

	fp_pad = fio_prep_pad ( mtotp, row, offset );
	if ( ! fp_pad )
	{
		i_error = 1;

		goto finish;
	}

	if ( update_sheet )
	{
		sh_temp = ced_sheet_copy_area ( mtotp->sheet_index, 1,
			SHEET_COL_PADUSED, 0, 1 );

		if ( ! sh_temp )
		{
			i_error = 1;

			goto finish;
		}
	}

	while ( 1 )
	{
		// Read next chunk from INPUT
		if ( fio_read ( fp_in, mtotp->buf_in, BUFLEN, &buflen,
			filename_in ) )
		{
			i_error = 1;

			goto finish;
		}

		if ( buflen == 0 )
		{
			break;		// EOF
		}

		btot += buflen;

		// Read next chunk from PAD
		if ( fio_read ( fp_pad, mtotp->buf_xor, buflen, &bdone, "PAD" )
			)
		{
			i_error = 1;

			goto finish;
		}

		btodo = buflen - bdone;
		offset += bdone;

		if ( feof ( fp_pad ) )
		{
			if ( offset != get_cell_value ( mtotp->sheet_index,
				row, SHEET_COL_PADSIZE ) )
			{
				fprintf ( stderr, "Error - End of pad reached "
					"but recorded size doesn't match.\n\n"
					);

				i_error = 1;

				goto finish;
			}

			ced_sheet_set_cell_value ( sh_temp, row, 1, offset );

			row ++;
			fclose ( fp_pad );
			fp_pad = fio_prep_pad ( mtotp, row, 0 );

			if ( ! fp_pad )
			{
				i_error = 1;

				goto finish;
			}

			offset = btodo;
			btodo -= fread ( mtotp->buf_xor + bdone, 1, btodo,
				fp_pad );
		}

		if ( btodo != 0 )
		{
			fprintf ( stderr, "Unexpected error reading pad "
				"%.0f\n\n", row );

			i_error = 1;

			goto finish;
		}

		xor_buffer ( mtotp, buflen );

		if ( fio_write ( fp_out, mtotp->buf_in, buflen, filename_out ) )
		{
			goto finish;
		}

		print_update_progress ( verbose, btot, filesize );
	}

	if ( i_error == 0 )
	{
		print_update_progress_nl ( verbose );

		if ( filesize != btot )
		{
			fprintf ( stderr,
				"Output file not same length as input.\n" );
			i_error = 1;
		}
	}

finish:
	if ( sh_temp )
	{
		if ( i_error == 0 )
		{
			// Everything succeeded so fix changes

			ced_sheet_set_cell_value ( sh_temp, row, 1, offset );

			i_error = ced_sheet_paste_area ( mtotp->sheet_index,
				sh_temp, 1, SHEET_COL_PADUSED, 0, 1, 0, 0, 0 );

			if ( i_error )
			{
				fprintf ( stderr, "Error writing changes to "
					"sheet after encryption.\n\n" );
			}
			else
			{
				mtotp->rowenc = row;
			}
		}

		ced_sheet_destroy ( sh_temp );
		sh_temp = NULL;
	}

	fio_close ( fp_in );
	fio_close ( fp_out );
	fio_close ( fp_pad );

	return i_error;
}

int mtotp_encrypt (
	mtOTP		* mtotp,
	char		* filename,
	int		verbose
	)
{
	char		buf [ MTKIT_PATHBUF_MAX ];
	double		filesize,
			pad_left,
			row,
			offset;


	if (	mtotp_check_setup ( mtotp )	||
		validate_filename ( filename )
		)
	{
		return 1;
	}

	filesize = fio_get_filesize ( filename );
	if ( filesize < 0 )
	{
		return 1;
	}

	if ( mtotp_query ( mtotp, &pad_left, 0, verbose ) )
	{
		return 1;
	}

	if ( filesize > pad_left )
	{
		fprintf ( stderr, "The pad isn't large enough to encrypt "
			"'%s'.\n\n", filename );

		return 1;
	}

	// query_pad_info checks index integrity for us.
	// rowenc is populated when loading the pad.
	row = mtotp->rowenc;
	offset = get_cell_value ( mtotp->sheet_index, row, SHEET_COL_PADUSED );

	snprintf ( buf, MTKIT_PATHBUF_MAX, "%s.mtotp.%.0f.%.0f",
		filename, row, offset );

	if ( fio_xor_loop ( mtotp, verbose, filename, buf, row, offset, 1 ) )
	{
		return 1;
	}

	return 0;			// Success
}

int mtotp_decrypt (
	mtOTP		* mtotp,
	char		* filename,
	int		verbose
	)
{
	char		filename_out [ MTKIT_PATHBUF_MAX ],
			* sp;
	double		row,
			offset;


	if (	mtotp_check_setup ( mtotp )	||
		validate_filename ( filename )
		)
	{
		return 1;
	}

	if ( mtotp_query_filename ( filename, &row, &offset ) )
	{
		fprintf ( stderr, "Filename doesn't contain required row and \n"
			"offset.\n\n" );

		return 1;
	}

	mtkit_strncpy0 ( filename_out, filename, MTKIT_PATHBUF_MAX );

	// Strip off .mtotp.%i.%i - query_filename () has already been
	// called successfully so we know the filename has this format
	sp = strrchr ( filename_out, '.' );
	sp[0] = 0;

	sp = mtotp_next_char_left ( filename_out, sp - 1, '.' );
	sp[0] = 0;

	sp = mtotp_next_char_left ( filename_out, sp - 1, '.' );
	sp[0] = 0;

	if ( fio_xor_loop ( mtotp, verbose, filename, filename_out, row,
		offset, 0 ) )
	{
		return 1;
	}

	return 0;
}

int mtotp_generate (
	mtOTP		* mtotp,
	char		* filename,
	char		* comment,
	int		verbose
	)
{
	double		filesize = 0,
			btot = 0
			;
	FILE		* fp_in,
			* fp_out
			;
	int		buflen,
			i,
			tmp,
			pos,
			res = 0,
			padfile_num
			;
	char		filename_out [ MTKIT_PATHBUF_MAX ] = { 0 };


	if (	mtotp_check_setup ( mtotp )	||
		validate_filename ( filename )
		)
	{
		return 1;
	}

	filesize = fio_get_filesize ( filename );
	if ( filesize < 0 )
	{
		return 1;
	}

	if ( filesize < PADSIZE_MIN || filesize > PADSIZE_MAX )
	{
		fprintf ( stderr, "File size is bad for '%s'.  "
			"%.0f <= byte size <= %.0f\n\n",
			filename, PADSIZE_MIN, PADSIZE_MAX );

		return 1;
	}

	// Get new pad filename
	for (	padfile_num = mtotp->rowtot;
		padfile_num <= COUNTER_MAX;
		padfile_num ++
		)
	{
		snprintf ( filename_out, MTKIT_PATHBUF_MAX,
			"%s/%s"COUNTER_PRINTF_INT".mtotp",
			mtotp->filename_root, MTOTP_DIR_PAD, padfile_num );

		if ( mtkit_file_valid ( filename_out ) )
		{
			break;
		}
	}

	if ( padfile_num > COUNTER_MAX )
	{
		fprintf ( stderr, "No spare pad number found.\n\n" );

		return 1;
	}

	fp_in = fio_open ( filename, "r" );
	if ( ! fp_in )
	{
		return 1;
	}

	fp_out = fio_open ( filename_out, "w" );
	if ( ! fp_out )
	{
		fclose ( fp_in );

		return 1;
	}

	if ( verbose )
	{
		printf ( "%.0f bytes in '%s'.\n\n", filesize, filename );
	}

	while ( 1 )
	{
		if ( fio_read ( fp_in, mtotp->buf_in, BUFLEN, &buflen,
			filename ) ||
			buflen < 1
			)
		{
			break;
		}

		// PRNG XOR mashup
		for ( i = 0; i < buflen; i++ )
		{
			mtotp->buf_in[ i ] ^= RANDOM_NUM_8;
		}

		// Shuffle all of the bytes around
		for ( i = 0; i < buflen; i++ )
		{
			pos = rand () % buflen;

			tmp = mtotp->buf_in[ i ];
			mtotp->buf_in[ i ] = mtotp->buf_in[ pos ];
			mtotp->buf_in[ pos ] = tmp;
		}

		if ( fio_write ( fp_out, mtotp->buf_in, buflen, filename_out ) )
		{
			res = 1;

			break;
		}

		btot += buflen;
		print_update_progress ( verbose, btot, filesize );
	}

	print_update_progress_nl ( verbose );

	if ( verbose )
	{
		printf ( "\n%.0f bytes written to '%s'.\n\n", btot,
			filename_out );
	}

	fclose ( fp_in );
	fclose ( fp_out );

	if ( res == 0 )
	{
		mtotp->rowtot ++;

		ced_sheet_set_cell_value ( mtotp->sheet_index, mtotp->rowtot,
			SHEET_COL_PADNUM, padfile_num );

		ced_sheet_set_cell_value ( mtotp->sheet_index, mtotp->rowtot,
			SHEET_COL_PADSIZE, filesize );

		ced_sheet_set_cell_text ( mtotp->sheet_index, mtotp->rowtot,
			SHEET_COL_PADFILENAME, filename );

		if ( comment )
		{
			ced_sheet_set_cell_text ( mtotp->sheet_index,
				mtotp->rowtot, SHEET_COL_PADCOMMENT, comment );
		}
	}

	return res;
}
