/*
	Copyright (C) 2004-2013 Mark Tyler
	Copyright (C) 2006-2008 Mark Tyler and Dmitry Groshev

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



enum	// File format flags
{
	MTPIXEL_FF_RGB		= 1 << 0,
	MTPIXEL_FF_INDEXED	= 1 << 1,
};



static mtImage * load_bmp (
	char		* filename
	)
{
	mtImage		* image = NULL;
	unsigned char	head[1024],
			* buf = NULL,
			pix,
			* src,
			* dest;
	int		width,
			height,
			readin,
			bitcount,
			compression,
			cols,
			i,
			j,
			k,
			bpl = 0,
			bpp;
	FILE		* fp;


	fp = fopen ( filename, "rb" );
	if ( fp == NULL )
	{
		return NULL;
	}

	readin = fread ( head, 1, 54, fp );
	if ( readin < 54 )
	{
		goto fail;		// No proper header
	}

	if ( head[0] != 'B' || head[1] != 'M' )
	{
		goto fail;		// Signature
	}

	width = head[18] + (head[19] << 8) +
		(head[20] << 16) + (head[21] << 24);

	height = head[22] + (head[23] << 8) + (head[24] << 16) +
		(head[25] << 24);

	bitcount = head[28] + (head[29] << 8);

	compression = head[30] + (head[31] << 8) + (head[32] << 16) +
		(head[33] << 24);

	cols = head[46] + (head[47] << 8) + (head[48] << 16) + (head[49] << 24);

	if (	bitcount != 1 &&
		bitcount != 4 &&
		bitcount != 8 &&
		bitcount != 24 )
	{
		goto fail;
	}

	if ( compression != 0 )
	{
		goto fail;
	}

	if ( bitcount == 24 )
	{
		bpp = MTPIXEL_IMAGE_RGB;
	}
	else
	{
		bpp = MTPIXEL_IMAGE_INDEXED;
	}

	image = mtpixel_image_new ( width, height, bpp,
		MTPIXEL_CHANNEL_MASK_IMAGE );

	if ( ! image )
	{
		goto fail;
	}

	bpl = width;			// Calculate bytes per line

	if ( bitcount == 24 )
	{
		bpl = width * 3;
	}

	if ( bitcount == 4 )
	{
		bpl = (width + 1) / 2;
	}

	if ( bitcount == 1 )
	{
		bpl = (width + 7) / 8;
	}

	if ( bpl % 4 != 0 )
	{
		// 4 byte boundary for pixels
		bpl = bpl + 4 - ( bpl % 4 );
	}

	buf = malloc ( bpl );
	if ( ! buf )
	{
		goto fail;
	}

	if ( bitcount == 24 )		// RGB image
	{
		for ( j = 0; j < height; j++ )
		{
			// Read in line of pixels
			readin = fread ( buf, 1, bpl, fp );

			if ( readin != bpl )
			{
				goto fail;
			}

			src = buf;
			dest = image->channel[MTPIXEL_CHANNEL_IMAGE] +
				3 * width * ( height - 1 - j );

			for ( i = 0; i < width; i++ )
			{
				dest[2] = *src++;
				dest[1] = *src++;
				dest[0] = *src++;
				dest += 3;
			}
		}
	}
	else				// Indexed palette image
	{
		if ( cols == 0 )
		{
			cols = 1 << bitcount;
		}

		if ( cols < 2 || cols > 256 )
		{
			goto fail;
		}

		image->palette.size = cols;

		// Read in colour table
		readin = fread ( head, 1, cols * 4, fp );

		if ( readin != cols * 4 )
		{
			goto fail;
		}

		for ( i = 0; i < cols; i++ )
		{
			image->palette.col[i][0] = head[2 + 4 * i];
			image->palette.col[i][1] = head[1 + 4 * i];
			image->palette.col[i][2] = head[0 + 4 * i];
		}

		for ( j = 0; j < height; j++ )
		{
			// Read in line of pixels
			readin = fread ( buf, 1, bpl, fp );

			if ( readin != bpl )
			{
				goto fail;
			}

			src = buf;
			dest = image->channel[MTPIXEL_CHANNEL_IMAGE] +
				width * ( height - 1 - j );

			if ( bitcount == 8 )
			{
				memcpy ( dest, src, width );
			}
			else if ( bitcount == 4 )
			{
				for ( i = 0; i < width; i += 2 )
				{
					pix = *src++;
					*dest++ = pix >> 4;

					if ( (i + 1) < width )
					{
						*dest++ = pix % 16;
					}
				}
			}
			else if ( bitcount == 1 )
			{
				for ( i = 0; i < width; i += 8 )
				{
					pix = *src++;
					k = 0;

					while ( k < 8 && (i + k) < width )
					{
						*dest++ = (pix >> (7 - k) ) % 2;
						k++;
					}
				}
			}
		}
	}

	free ( buf );
	fclose ( fp );

	return image;			// Success

fail:
	mtpixel_image_destroy ( image );
	free ( buf );
	fclose ( fp );

	return NULL;
}

static int save_bmp (
	mtImage		* image,
	char		* filename,
	int		fileinfo
	)
{
	unsigned char	head[1024] = {0},
			* buf,
			* src,
			* dest;
	int		written,
			i,
			j,
			bpl,
			bpp,
			filesize,
			headsize,
			w,
			h,
			cols;
	FILE		* fp;


	switch ( image->type )
	{
		case MTPIXEL_IMAGE_INDEXED:
			bpp = 1;
			break;

		case MTPIXEL_IMAGE_RGB:
			bpp = 3;
			break;

		default:
			return 1;
	}

	w = image->width;
	h = image->height;
	cols = image->palette.size;

	bpl = w * bpp;

	if ( bpl % 4 != 0 )
	{
		bpl = bpl + 4 - (bpl % 4);	// Adhere to 4 byte boundaries
	}

	filesize = 54 + bpl * h;

	if ( bpp == 1 )
	{
		filesize = filesize + cols * 4;
	}

	headsize = filesize - bpl * h;

	fp = fopen ( filename, "wb" );
	if ( fp == NULL )
	{
		return 1;
	}

	head[0] = 'B'; head[1] = 'M';	// Signature

	head[2] = filesize % 256;
	head[3] = (filesize >> 8) % 256;
	head[4] = (filesize >> 16) % 256;
	head[5] = (filesize >> 24) % 256;

	head[10] = headsize % 256;
	head[11] = headsize / 256;

	head[14] = 40; head[26] = 1;

	head[18] = w % 256;
	head[19] = w / 256;
	head[20] = 0;
	head[21] = 0;
	head[22] = h % 256;
	head[23] = h / 256;
	head[24] = 0;
	head[25] = 0;

	head[28] = bpp * 8;
	head[29] = 0;			// Bits per pixel
	head[30] = 0;
	head[31] = 0;
	head[32] = 0;
	head[33] = 0;			// No compression

	head[34] = bpl * h % 256;
	head[35] = (bpl * h >> 8) % 256;
	head[36] = (bpl * h >> 16) % 256;
	head[37] = (bpl * h >> 24) % 256;

	head[38] = 18;
	head[39] = 11;
	head[42] = 18;
	head[43] = 11;

	if ( bpp == 1 )
	{
		head[46] = cols % 256;
		head[47] = cols / 256;
		head[50] = head[46];
		head[51] = head[47];
	}

	written = fwrite ( head, 1, 54, fp );
	if ( written < 54 )
	{
		goto fail;
	}

	buf = calloc ( bpl, 1 );
	if ( ! buf )
	{
		goto fail;
	}

	if ( bpp == 3 )			// RGB image
	{
		for ( j = h - 1; j >= 0; j-- )
		{
			src = image->channel[ MTPIXEL_CHANNEL_IMAGE ] +
				3 * w * j;
			dest = buf;

			for ( i = 0; i < w; i++ )
			{
				*dest++ = src[2];
				*dest++ = src[1];
				*dest++ = src[0];
				src += 3;
			}

			if ( fwrite ( buf, 1, bpl, fp ) != bpl )
			{
				free ( buf );
				goto fail;
			}
		}
	}

	if ( bpp == 1 )		// Indexed palette image
	{
		for ( i = 0; i < cols; i++ )
		{
			head[3 + 4 * i] = 0;
			head[2 + 4 * i] = image->palette.col[i][0];
			head[1 + 4 * i] = image->palette.col[i][1];
			head[0 + 4 * i] = image->palette.col[i][2];
		}

		if ( fwrite ( head, 1, cols * 4, fp ) != cols * 4 )
		{
			free ( buf );
			goto fail;
		}

		for ( j = h - 1; j >= 0; j-- )
		{
			src = image->channel[MTPIXEL_CHANNEL_IMAGE] + w * j;

			// We can't save directly from the image as bpl might
			// be > w

			memcpy ( buf, src, w );

			if ( fwrite ( buf, 1, bpl, fp ) != bpl )
			{
				free ( buf );
				goto fail;
			}
		}
	}

	free ( buf );
	fclose ( fp );

	return 0;			// Success

fail:
	fclose ( fp );

	return 1;
}



#ifdef U_PNG

#define PNG_READ_PACK_SUPPORTED
#include <png.h>
#include <zlib.h>



#define PNG_BYTES_TO_CHECK 8
#define PNG_HANDLE_CHUNK_ALWAYS 3



static const char * chunk_names[MTPIXEL_CHANNEL_TOTAL] = {
			"",
			"alPh",
			"seLc",
			"maSk"
			};



static mtImage * load_png (
	char		* filename
	)
{
	mtImage		* image = NULL;
	/* Description of PNG interlacing passes as X0, DX, Y0, DY */
	static const unsigned char png_interlace[8][4] = {
		{ 0, 1, 0, 1 }, /* One pass for non-interlaced */
		{ 0, 8, 0, 8 }, /* Seven passes for Adam7 interlaced */
		{ 4, 8, 0, 8 },
		{ 0, 4, 4, 8 },
		{ 2, 4, 0, 4 },
		{ 0, 2, 2, 4 },
		{ 1, 2, 0, 2 },
		{ 0, 1, 1, 2 }
			};
	static png_bytep * row_pointers;
	png_structp	png_ptr;
	png_infop	info_ptr;
	png_color_16p	trans_rgb;
	png_unknown_chunkp uk_p;
	png_bytep	trans;
	png_colorp	png_palette;
	png_uint_32	pwidth,
			pheight;
	char		buf[PNG_BYTES_TO_CHECK + 1];
	unsigned char	* src,
			* dest,
			* dsta;
	unsigned long	dest_len;
	FILE		* fp;
	int		i,
			j,
			k,
			bit_depth,
			color_type,
			interlace_type,
			num_uk,
			maxpass,
			x0,
			dx,
			y0,
			dy,
			n,
			height,
			width,
			bpp;


	fp = fopen ( filename, "rb" );
	if ( fp == NULL )
	{
		return NULL;
	}

	i = fread ( buf, 1, PNG_BYTES_TO_CHECK, fp );

	if ( i != PNG_BYTES_TO_CHECK )
	{
		goto fail;
	}

	if ( png_sig_cmp ( (unsigned char *)buf, 0, PNG_BYTES_TO_CHECK ) )
	{
		goto fail;
	}

	png_ptr = png_create_read_struct ( PNG_LIBPNG_VER_STRING, NULL, NULL,
		NULL);

	if ( ! png_ptr )
	{
		goto fail;
	}

	row_pointers = NULL;
	info_ptr = png_create_info_struct ( png_ptr );

	if ( ! info_ptr )
	{
		goto fail2;
	}

	if ( setjmp ( png_jmpbuf ( png_ptr ) ) )
	{
		goto fail2;
	}

	png_init_io ( png_ptr, fp );
	png_set_sig_bytes ( png_ptr, PNG_BYTES_TO_CHECK );

	/* Stupid libpng handles private chunks on all-or-nothing basis */
	png_set_keep_unknown_chunks ( png_ptr, PNG_HANDLE_CHUNK_ALWAYS, NULL,
		0 );

	png_read_info ( png_ptr, info_ptr );
	png_get_IHDR ( png_ptr, info_ptr, &pwidth, &pheight, &bit_depth,
		&color_type, &interlace_type, NULL, NULL );

	width = (int)pwidth;
	height = (int)pheight;

	if ( width < 1 || height < 1 )
	{
		goto fail2;
	}

	if (	color_type != PNG_COLOR_TYPE_PALETTE ||
		bit_depth > 8
		)
	{
		bpp = MTPIXEL_IMAGE_RGB;
	}
	else
	{
		bpp = MTPIXEL_IMAGE_INDEXED;
	}

	i = MTPIXEL_CHANNEL_MASK_IMAGE;

	if (	color_type == PNG_COLOR_TYPE_RGB_ALPHA ||
		color_type == PNG_COLOR_TYPE_GRAY_ALPHA
		)
	{
		i |= MTPIXEL_CHANNEL_MASK_ALPHA;
	}

	image = mtpixel_image_new ( width, height, bpp, i );
	if ( ! image )
	{
		goto fail2;
	}

	i = sizeof ( png_bytep ) * height;
	row_pointers = malloc ( i + width * 4 );

	if ( ! row_pointers )
	{
		goto fail2;
	}

	row_pointers[0] = (png_bytep) ( (char *)row_pointers + i );

	if ( png_get_valid ( png_ptr, info_ptr, PNG_INFO_PLTE ) )
	{
		png_get_PLTE ( png_ptr, info_ptr, &png_palette,
			&image->palette.size );

		for ( i = 0; i < image->palette.size; i++ )
		{
			image->palette.col[i][0] = png_palette[i].red;
			image->palette.col[i][1] = png_palette[i].green;
			image->palette.col[i][2] = png_palette[i].blue;
		}
	}

	/* RGB PNG file */
	if ( bpp == MTPIXEL_IMAGE_RGB )
	{
		png_set_strip_16 ( png_ptr );
		png_set_expand_gray_1_2_4_to_8 ( png_ptr );
		png_set_palette_to_rgb ( png_ptr );
		png_set_gray_to_rgb ( png_ptr );

		/* Is there a transparent color? */
		if ( png_get_valid ( png_ptr, info_ptr, PNG_INFO_tRNS ) )
		{
			png_get_tRNS ( png_ptr, info_ptr, 0, 0, &trans_rgb );

			if ( color_type == PNG_COLOR_TYPE_GRAY )
			{
				if ( bit_depth == 4 )
				{
					i = trans_rgb->gray * 17;
				}
				else if ( bit_depth == 8 )
				{
					i = trans_rgb->gray;
					// Hope libpng compiled w/o accurate
					// transform
				}
				else if ( bit_depth == 16 )
				{
					i = trans_rgb->gray >> 8;
				}

				image->palette.trans = MTPIXEL_RGB_2_INT ( i, i,
					i );
			}
			else
			{
				image->palette.trans = MTPIXEL_RGB_2_INT (
					trans_rgb->red, trans_rgb->green,
					trans_rgb->blue );
			}
		}
		else
		{
			image->palette.trans = -1;
		}

		if ( image->channel[ MTPIXEL_CHANNEL_ALPHA ] ) /* RGBA */
		{
			/* Have to do deinterlacing myself */
			if ( interlace_type == PNG_INTERLACE_NONE )
			{
				k = 0;
				maxpass = 1;
			}
			else if ( interlace_type == PNG_INTERLACE_ADAM7 )
			{
				k = 1;
				maxpass = 8;
			}
			else
			{
				/* Unknown type */
				goto fail2;
			}

			for ( n = 0; k < maxpass; k++ )
			{
				x0 = png_interlace[k][0];
				dx = png_interlace[k][1];
				y0 = png_interlace[k][2];
				dy = png_interlace[k][3];

				for ( i = y0; i < height; i += dy , n++ )
				{
					png_read_rows ( png_ptr,
						&row_pointers[0], NULL, 1 );
					src = row_pointers[0];
					dest = image->channel[
						MTPIXEL_CHANNEL_IMAGE ] +
						(i * width + x0) * 3;
					dsta = image->channel[
						MTPIXEL_CHANNEL_ALPHA ] +
						i * width;

					for ( j = x0; j < width; j += dx )
					{
						dest[0] = src[0];
						dest[1] = src[1];
						dest[2] = src[2];
						dsta[j] = src[3];
						src += 4; dest += 3 * dx;
					}
				}
			}
		}
		else	/* RGB */
		{
			png_set_strip_alpha ( png_ptr );

			for ( i = 0; i < height; i++ )
			{
				row_pointers[i] = image->channel[
					MTPIXEL_CHANNEL_IMAGE ] +
					i * width * 3;
			}

			png_read_image ( png_ptr, row_pointers );
		}
	}
	else	/* Paletted PNG file */
	{
		/* Is there a transparent index? */
		if ( png_get_valid ( png_ptr, info_ptr, PNG_INFO_tRNS ) )
		{
			/* !!! Currently no support for partial transparency */
			png_get_tRNS ( png_ptr, info_ptr, &trans, &i, 0 );
			image->palette.trans = -1;

			for ( j = 0; j < i; j++ )
			{
				if ( trans[j] )
				{
					continue;
				}

				image->palette.trans = j;

				break;
			}
		}

		png_set_strip_16 ( png_ptr );
		png_set_strip_alpha ( png_ptr );
		png_set_packing ( png_ptr );

		if (	color_type == PNG_COLOR_TYPE_GRAY &&
			bit_depth < 8
			)
		{
			png_set_expand_gray_1_2_4_to_8 ( png_ptr );
		}

		for ( i = 0; i < height; i++ )
		{
			row_pointers[i] = image->channel[
				MTPIXEL_CHANNEL_IMAGE ] + i * width;
		}

		png_read_image ( png_ptr, row_pointers );
	}

	png_read_end ( png_ptr, info_ptr );

	num_uk = png_get_unknown_chunks ( png_ptr, info_ptr, &uk_p );
	if ( num_uk )	/* File contains mtPaint's private chunks */
	{
		for ( i = 0; i < num_uk; i++ )	/* Examine each chunk */
		{
			for (	j = MTPIXEL_CHANNEL_ALPHA;
				j < MTPIXEL_CHANNEL_TOTAL;
				j++
				)
			{
				if ( ! strcmp ( (char *)uk_p[i].name,
					chunk_names[j] )
					)
				{
					break;
				}
			}

			if ( j >= MTPIXEL_CHANNEL_TOTAL )
			{
				continue;
			}

			/* Try to allocate a channel */
			if ( mtpixel_channel_new ( image, j ) )
			{
				break;
			}

			/* Skip if not allocated */
			if ( ! image->channel[j] )
			{
				continue;
			}

			dest_len = width * height;
			uncompress ( image->channel[j], &dest_len, uk_p[i].data,
				uk_p[i].size );
		}

		/* !!! Is this call really needed? */
		png_free_data ( png_ptr, info_ptr, PNG_FREE_UNKN, -1 );
	}

	free ( row_pointers );
	png_destroy_read_struct ( &png_ptr, &info_ptr, NULL );
	fclose ( fp );

	return image;

fail2:
	free ( row_pointers );
	png_destroy_read_struct ( &png_ptr, &info_ptr, NULL );
	mtpixel_image_destroy ( image );

fail:
	fclose ( fp );

	return NULL;
}



#ifndef PNG_AFTER_IDAT
#define PNG_AFTER_IDAT 8
#endif



static int save_png (
	mtImage		* image,
	char		* filename,
	int		fileinfo
	)
{
	png_unknown_chunk unknown0;
	png_structp	png_ptr;
	png_infop	info_ptr;
	FILE		* fp;
	int		i,
			j,
			h = image->height,
			w = image->width,
			chunks = 0;
	unsigned long	dest_len,
			tmp_len;
	unsigned char	* rgba_row = NULL,
			* tmp,
			* tmi,
			* tma;


	if (	image->type == MTPIXEL_IMAGE_RGB &&
		image->channel[MTPIXEL_CHANNEL_ALPHA]
		)
	{
		rgba_row = malloc ( w * 4 );
		if ( ! rgba_row )
		{
			return -1;
		}
	}

	fp = fopen ( filename, "wb" );
	if ( fp == NULL )
	{
		goto exit0;
	}

	png_ptr = png_create_write_struct ( PNG_LIBPNG_VER_STRING,
		(png_voidp)NULL, NULL, NULL );

	if ( ! png_ptr )
	{
		goto exit1;
	}

	info_ptr = png_create_info_struct ( png_ptr );
	if ( ! info_ptr )
	{
		goto exit2;
	}

	png_init_io ( png_ptr, fp );
	png_set_compression_level ( png_ptr, fileinfo );

	if ( image->type == MTPIXEL_IMAGE_INDEXED )
	{
		png_set_IHDR ( png_ptr, info_ptr, w, h,
			8, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );
	}

	if ( image->type == MTPIXEL_IMAGE_RGB )
	{
		png_set_IHDR ( png_ptr, info_ptr, w, h,
			8, image->channel[MTPIXEL_CHANNEL_ALPHA] ?
				PNG_COLOR_TYPE_RGB_ALPHA : PNG_COLOR_TYPE_RGB,
			PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT );
	}

	if ( image->palette.size > 0 )
	{
		png_color_16	trans_rgb;
		png_color	pal[256];


		for ( i = 0; i < image->palette.size; i++ )
		{
			pal[i].red = image->palette.col[i][0];
			pal[i].green = image->palette.col[i][1];
			pal[i].blue = image->palette.col[i][2];
		}

		png_set_PLTE ( png_ptr, info_ptr, pal, image->palette.size );

		/* Transparent index in use */
		if ( image->palette.trans != -1 )
		{
			if ( image->type == MTPIXEL_IMAGE_INDEXED &&
				image->palette.trans < 256
				)
			{
				unsigned char	trans[256];


				memset ( trans, 255, 256 );
				trans[image->palette.trans] = 0;
				png_set_tRNS ( png_ptr, info_ptr, trans,
					image->palette.size, 0 );
			}

			if (	image->type == MTPIXEL_IMAGE_RGB &&
				! image->channel[MTPIXEL_CHANNEL_ALPHA]
				)
			{
				trans_rgb.red = MTPIXEL_INT_2_R (
					image->palette.trans );
				trans_rgb.green = MTPIXEL_INT_2_G (
					image->palette.trans );
				trans_rgb.blue = MTPIXEL_INT_2_B (
					image->palette.trans );
				png_set_tRNS ( png_ptr, info_ptr, 0, 1,
					&trans_rgb );
			}
		}
	}

	png_write_info ( png_ptr, info_ptr );

	if (	image->type == MTPIXEL_IMAGE_INDEXED ||
		! image->channel[MTPIXEL_CHANNEL_ALPHA]
		)
	{
		/* Flat RGB/Indexed image */
		if ( image->type == MTPIXEL_IMAGE_RGB )
		{
			w *= 3;
		}

		for ( i = 0; i < h; i++ )
		{
			png_write_row ( png_ptr,
				(png_bytep)(image->channel[
				MTPIXEL_CHANNEL_IMAGE ] + i * w) );
		}
	}
	else if ( rgba_row ) /* RGBA image */
	{
		tmi = image->channel[MTPIXEL_CHANNEL_IMAGE];
		tma = image->channel[MTPIXEL_CHANNEL_ALPHA];

		for ( i = 0; i < h; i++ )
		{
			tmp = rgba_row;

			for ( j = 0; j < w; j++ ) /* Combine RGB and alpha */
			{
				tmp[0] = tmi[0];
				tmp[1] = tmi[1];
				tmp[2] = tmi[2];
				tmp[3] = tma[0];
				tmp += 4;
				tmi += 3;
				tma ++;
			}

			png_write_row ( png_ptr, (png_bytep)rgba_row );
		}
	}

	/* Save private chunks into PNG file if we need to */
	if ( image->type == MTPIXEL_IMAGE_INDEXED )
	{
		j = MTPIXEL_CHANNEL_ALPHA;
	}
	else
	{
		j = MTPIXEL_CHANNEL_ALPHA + 1;
	}

	for ( i = j; ! image->channel[i] && (i < MTPIXEL_CHANNEL_TOTAL); i++ )
	{
	}

	if ( i < MTPIXEL_CHANNEL_TOTAL )
	{
		/* Get size required for each zlib compress */
		w = image->width * image->height;

#if ZLIB_VERNUM >= 0x1200
		tmp_len = compressBound ( w );
#else
		tmp_len = w + (w >> 8) + 32;
#endif

		// Temporary space for compression
		tmp = malloc ( tmp_len );

		if ( ! tmp )
		{
			goto exit3;
		}

		for ( ; i < MTPIXEL_CHANNEL_TOTAL; i++ )
		{
			if ( ! image->channel[i] )
			{
				continue;
			}

			dest_len = tmp_len;
			if ( compress2 ( tmp, &dest_len, image->channel[i], w,
				fileinfo) != Z_OK
				)
			{
				continue;
			}

			strncpy ( (char *)unknown0.name, chunk_names[i], 5 );

			unknown0.data = tmp;
			unknown0.size = dest_len;
			png_set_unknown_chunks ( png_ptr, info_ptr, &unknown0,
				1 );
			png_set_unknown_chunk_location ( png_ptr, info_ptr,
				chunks ++, PNG_AFTER_IDAT );
		}

		free ( tmp );
	}

	png_write_end ( png_ptr, info_ptr );
	png_destroy_write_struct ( &png_ptr, &info_ptr );
	fclose ( fp );
	free ( rgba_row );

	return 0;

exit3:
	png_write_end ( png_ptr, info_ptr );
exit2:
	png_destroy_write_struct ( &png_ptr, &info_ptr );
exit1:
	fclose ( fp );
exit0:
	free ( rgba_row );

	return 1;
}

#endif	// U_PNG



#ifdef U_GIF
#include <gif_lib.h>

/*
FIXME - This is to get mtPixel working with Giflib 4 & 5.
When all systems are using Giflib 5, remove all U_GIFLIB_OLD and migrate to the
new code (circa 2020 I would guess).
*/

#ifdef GIF_LIB_VERSION
	#define U_GIFLIB_OLD
#endif



static mtImage * load_gif (
	char		* filename
	)
{
	/* GIF interlace pattern: Y0, DY, ... */
	unsigned char	interlace[10] = { 0, 1, 0, 8, 4, 8, 2, 4, 1, 2 };
	GifFileType	* giffy;
	GifRecordType	gif_rec;
	GifByteType	* byte_ext;
	ColorMapObject	* cmap = NULL;
	int		i,
			j,
			k,
			kx,
			n,
			w,
			h,
			dy,
			frame = 0,
			val,
			trans = -1;
	mtImage		* image = NULL;


#ifdef U_GIFLIB_OLD
	giffy = DGifOpenFileName ( filename );
#else
	giffy = DGifOpenFileName ( filename, NULL );
#endif

	if ( ! giffy )
	{
		return NULL;
	}

	while ( 1 )
	{
		if ( DGifGetRecordType ( giffy, &gif_rec) == GIF_ERROR )
		{
			goto fail;
		}

		if ( gif_rec == TERMINATE_RECORD_TYPE )
		{
			break;
		}
		else if ( gif_rec == EXTENSION_RECORD_TYPE )
		{
			if ( DGifGetExtension ( giffy, &val, &byte_ext ) ==
				GIF_ERROR )
			{
				goto fail;
			}

			while ( byte_ext )
			{
				if ( val == GRAPHICS_EXT_FUNC_CODE )
				{
					trans = byte_ext[1] & 1 ?
						byte_ext[4] : -1;
				}

				if ( DGifGetExtensionNext ( giffy, &byte_ext )
					== GIF_ERROR )
				{
					goto fail;
				}
			}
		}
		else if ( gif_rec == IMAGE_DESC_RECORD_TYPE )
		{
			if ( frame ++ ) /* Multipage GIF - notify user */
			{
				goto fail;
			}

			if ( DGifGetImageDesc ( giffy ) == GIF_ERROR )
			{
				goto fail;
			}

			/* Get palette */
			cmap = giffy->SColorMap ? giffy->SColorMap :
				giffy->Image.ColorMap;
			if ( ! cmap )
			{
				goto fail;
			}

			j = cmap->ColorCount;
			if ( j > 256 || j < 1 )
			{
				goto fail;
			}

			w = giffy->Image.Width;
			h = giffy->Image.Height;

			image = mtpixel_image_new ( w, h, MTPIXEL_IMAGE_INDEXED,
					MTPIXEL_CHANNEL_MASK_IMAGE );
			if ( ! image )
			{
				goto fail;
			}

			image->palette.trans = trans;
			image->palette.size = j;

			for ( i = 0; i < j; i++ )
			{
				image->palette.col[i][0] = cmap->Colors[i].Red;
				image->palette.col[i][1] =
					cmap->Colors[i].Green;
				image->palette.col[i][2] = cmap->Colors[i].Blue;
			}

			if ( giffy->Image.Interlace )
			{
				k = 2;
				kx = 10;
			}
			else
			{
				k = 0;
				kx = 2;
			}

			for ( n = 0; k < kx; k += 2 )
			{
				dy = interlace[k + 1];

				for ( i = interlace[k]; i < h; n++ , i += dy )
				{
					if ( DGifGetLine ( giffy,
						image->channel[
						MTPIXEL_CHANNEL_IMAGE ] +
						i * w, w) == GIF_ERROR )
					{
						goto fail;
					}
				}
			}
		}
	}

	DGifCloseFile ( giffy );

	return image;

fail:
	mtpixel_image_destroy ( image );
	DGifCloseFile ( giffy );

	return NULL;
}

static int save_gif (
	mtImage		* image,
	char		* filename,
	int		fileinfo
	)
{
	mode_t		mode;
	ColorMapObject	* gif_map;
	GifFileType	* giffy;
	unsigned char	gif_ext_data[8];
	int		i,
			w,
			h;


	w = image->width;
	h = image->height;

#ifdef U_GIFLIB_OLD
	gif_map = MakeMapObject ( 256, NULL );
#else
	gif_map = GifMakeMapObject ( 256, NULL );
#endif

	if ( ! gif_map )
	{
		return 1;
	}

#ifdef U_GIFLIB_OLD
	giffy = EGifOpenFileName ( filename, FALSE );
#else
	giffy = EGifOpenFileName ( filename, 0, NULL );
#endif

	if ( ! giffy )
	{
		goto fail0;
	}

	for ( i = 0; i < image->palette.size; i++ )
	{
		gif_map->Colors[i].Red = image->palette.col[i][0];
		gif_map->Colors[i].Green = image->palette.col[i][1];
		gif_map->Colors[i].Blue = image->palette.col[i][2];
	}

	for ( ; i < 256; i++ )
	{
		gif_map->Colors[i].Red =
			gif_map->Colors[i].Green =
			gif_map->Colors[i].Blue	= 0;
	}

	if ( EGifPutScreenDesc ( giffy, w, h, 256, 0, gif_map ) == GIF_ERROR )
	{
		goto fail;
	}

	if ( image->palette.trans >= 0 )
	{
		gif_ext_data[0] = 1;
		gif_ext_data[1] = 0;
		gif_ext_data[2] = 0;
		gif_ext_data[3] = image->palette.trans;
		EGifPutExtension ( giffy, GRAPHICS_EXT_FUNC_CODE, 4,
			gif_ext_data );
	}

	if ( EGifPutImageDesc ( giffy, 0, 0, w, h, 0, NULL ) == GIF_ERROR )
	{
		goto fail;
	}

	for ( i = 0; i < h; i++ )
	{
		EGifPutLine ( giffy, image->channel[MTPIXEL_CHANNEL_IMAGE] +
			i * w, w );
	}

	EGifCloseFile ( giffy );

#ifdef U_GIFLIB_OLD
	FreeMapObject ( gif_map );
#else
	GifFreeMapObject ( gif_map );
#endif

	return 0;

fail:
	EGifCloseFile ( giffy );
	mode = umask ( 0022 );
	umask ( mode );
	chmod ( filename, 0666 & ~mode );

fail0:
#ifdef U_GIFLIB_OLD
	FreeMapObject ( gif_map );
#else
	GifFreeMapObject ( gif_map );
#endif

	return 1;
}

#endif	// U_GIF


static void create_grey_palette (
	mtPalette	* pal
	)
{
	int		i;


	pal->size = 256;
	for ( i = 0; i < 256; i++ )
	{
		pal->col[i][0] = i;
		pal->col[i][1] = i;
		pal->col[i][2] = i;
	}
}


#ifdef U_JPEG
#include <jpeglib.h>


struct my_error_mgr
{
	struct jpeg_error_mgr	pub;		// "public" fields
	jmp_buf			setjmp_buffer;	// for return to caller
};



typedef struct my_error_mgr * my_error_ptr;



METHODDEF ( void ) my_error_exit (
	j_common_ptr cinfo
	)
{
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	longjmp ( myerr->setjmp_buffer, 1 );
}



struct my_error_mgr	jerr;



static mtImage * load_jpeg (
	char		* filename
	)
{
	struct jpeg_decompress_struct cinfo;
	unsigned char	xb = 0,
			* memp,
			* memx = NULL;
	FILE		* fp;
	int		i,
			width,
			height,
			bpp;
	mtImage		* image = NULL;


	fp = fopen ( filename, "rb" );
	if ( fp == NULL )
	{
		return NULL;
	}

	jpeg_create_decompress ( &cinfo );
	cinfo.err = jpeg_std_error ( &jerr.pub );
	jerr.pub.error_exit = my_error_exit;

	if ( setjmp ( jerr.setjmp_buffer ) )
	{
		goto fail;
	}

	jpeg_stdio_src ( &cinfo, fp );
	jpeg_read_header ( &cinfo, TRUE );
	jpeg_start_decompress ( &cinfo );

	width = cinfo.output_width;
	height = cinfo.output_height;
	bpp = MTPIXEL_IMAGE_RGB;

	switch ( cinfo.out_color_space )
	{
	case JCS_RGB:
		break;

	case JCS_GRAYSCALE:
		bpp = MTPIXEL_IMAGE_INDEXED;
		break;

	case JCS_CMYK:
		/* Photoshop writes CMYK data inverted */
		xb = cinfo.saw_Adobe_marker ? 0 : 255;
		memx = malloc ( cinfo.output_width * 4 );

		if ( memx )
		{
			break;
		}
		// Fallthrough

	default:
		/* Unsupported colorspace */
		goto fail;
	}

	image = mtpixel_image_new ( width, height, bpp,
		MTPIXEL_CHANNEL_MASK_IMAGE );
	if ( ! image )
	{
		goto fail;
	}

	if ( bpp == MTPIXEL_IMAGE_INDEXED )
	{
		create_grey_palette ( &image->palette );
	}

	for ( i = 0; i < height; i++ )
	{
		memp = image->channel[ MTPIXEL_CHANNEL_IMAGE ] +
			width * i * bpp;
		jpeg_read_scanlines ( &cinfo, memx ? &memx : &memp, 1 );

		if ( memx ) /* Simple CMYK->RGB conversion */
		{
			unsigned char	* src = memx,
					* dest = memp;
			int		j,
					k,
					r,
					g,
					b;


			for ( j = 0; j < width; j++ , src += 4 , dest += 3 )
			{
				k = src[3] ^ xb;
				r = (src[0] ^ xb) * k;
				dest[0] = (r + (r >> 8) + 1) >> 8;
				g = (src[1] ^ xb) * k;
				dest[1] = (g + (g >> 8) + 1) >> 8;
				b = (src[2] ^ xb) * k;
				dest[2] = (b + (b >> 8) + 1) >> 8;
			}
		}
	}

	jpeg_finish_decompress ( &cinfo );
	jpeg_destroy_decompress ( &cinfo );
	fclose ( fp );
	free ( memx );

	return image;

fail:
	mtpixel_image_destroy ( image );

	jpeg_finish_decompress ( &cinfo );
	jpeg_destroy_decompress ( &cinfo );
	fclose ( fp );
	free ( memx );

	return NULL;
}

static int save_jpeg (
	mtImage		* image,
	char		* filename,
	int		fileinfo
	)
{
	struct jpeg_compress_struct cinfo;
	JSAMPROW	row_pointer;
	FILE		* fp;
	int		i;


	fp = fopen ( filename, "wb" );
	if ( fp == NULL )
	{
		return 1;
	}

	cinfo.err = jpeg_std_error ( &jerr.pub );
	jerr.pub.error_exit = my_error_exit;

	if ( setjmp ( jerr.setjmp_buffer ) )
	{
		jpeg_destroy_compress ( &cinfo );
		fclose ( fp );

		return 1;
	}

	jpeg_create_compress ( &cinfo );
	jpeg_stdio_dest ( &cinfo, fp );

	cinfo.image_width = image->width;
	cinfo.image_height = image->height;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;

	jpeg_set_defaults ( &cinfo );
	jpeg_set_quality ( &cinfo, fileinfo, TRUE );
	jpeg_start_compress ( &cinfo, TRUE );

	row_pointer = image->channel[ MTPIXEL_CHANNEL_IMAGE ];

	for ( i = 0; i < image->height; i++ )
	{
		jpeg_write_scanlines ( &cinfo, &row_pointer, 1 );
		row_pointer += 3 * image->width;
	}

	jpeg_finish_compress ( &cinfo );
	jpeg_destroy_compress ( &cinfo );
	fclose ( fp );

	return 0;
}
#endif	// U_JPEG




typedef mtImage * (* func_load) (
	char		* filename
	);

typedef int (* func_save) (
	mtImage		* image,
	char		* filename,
	int		fileinfo
	);



typedef struct
{
	mtFileType	type;
	char		* desc,
			* ext_short,
			* ext_long;
	int		flags;
	func_load	load;
	func_save	save;
} _mtpixel_ff;



// The ordering must be as in the enum

static _mtpixel_ff	ff_table[ MTPIXEL_FILE_TYPE_TOTAL - 1 ] =
{
{ MTPIXEL_FILE_TYPE_BMP, "BMP", "bmp", "", MTPIXEL_FF_RGB | MTPIXEL_FF_INDEXED, load_bmp, save_bmp },

#ifdef U_PNG
{ MTPIXEL_FILE_TYPE_PNG, "PNG", "png", "", MTPIXEL_FF_RGB | MTPIXEL_FF_INDEXED, load_png, save_png },
#else
{ MTPIXEL_FILE_TYPE_NONE, "", "", "", 0 },
#endif

#ifdef U_GIF
{ MTPIXEL_FILE_TYPE_GIF, "GIF", "gif", "", MTPIXEL_FF_INDEXED, load_gif, save_gif },
#else
{ MTPIXEL_FILE_TYPE_NONE, "", "", "", 0 },
#endif

#ifdef U_JPEG
{ MTPIXEL_FILE_TYPE_JPEG, "JPEG", "jpg", "jpeg", MTPIXEL_FF_RGB, load_jpeg, save_jpeg },
#else
{ MTPIXEL_FILE_TYPE_NONE, "", "", "", 0 },
#endif
};



int mtpixel_filetype (
	int		type,
	int		* flags,
	char		** desc,
	char		** ext_short,
	char		** ext_long
	)
{
	if (	type > MTPIXEL_FILE_TYPE_NONE	&&
		type < MTPIXEL_FILE_TYPE_TOTAL	&&
		ff_table[ type - 1 ].type == type
		)
	{
		if ( flags )
		{
			flags[0] = ff_table[ type - 1 ].flags;
		}

		if ( desc )
		{
			desc[0] = ff_table[ type - 1].desc;
		}

		if ( ext_short )
		{
			ext_short[0] = ff_table[ type - 1 ].ext_short;
		}

		if ( ext_long )
		{
			ext_long[0] = ff_table[ type - 1 ].ext_long;
		}

		return 1;		// Supported
	}

	return 0;			// Not supported
}

static int detect_image_format (
	char		* filename
	)
{
	unsigned char	buf[66] = {0};
	FILE		* fp;
	int		tot;


	fp = fopen ( filename, "rb" );
	if ( ! fp )
	{
		return -1;
	}

	tot = fread ( buf, 1, 64, fp );
	fclose ( fp );

	if ( tot < 4 )
	{
		return MTPIXEL_FILE_TYPE_NONE;
	}

	/* Check all unambiguous signatures */
#ifdef U_PNG
	if ( 0 == memcmp ( buf, "\x89PNG", 4 ) )
	{
		return MTPIXEL_FILE_TYPE_PNG;
	}
#endif

#ifdef U_JPEG
	if ( 0 == memcmp ( buf, "\xFF\xD8", 2 ) )
	{
		return MTPIXEL_FILE_TYPE_JPEG;
	}
#endif

#ifdef U_GIF
	if ( 0 == memcmp ( buf, "GIF8", 4 ) )
	{
		return MTPIXEL_FILE_TYPE_GIF;
	}
#endif

	if ( 0 == memcmp ( buf, "BM", 2 ) )
	{
		return MTPIXEL_FILE_TYPE_BMP;
	}

	return MTPIXEL_FILE_TYPE_NONE;
}

static int file_type_ref (
	int		filetype
	)
{
	int		i;


	for ( i = 0; i < ( MTPIXEL_FILE_TYPE_TOTAL - 1 ); i++ )
	{
		if ( ff_table[i].type == filetype )
		{
			break;
		}
	}

	if ( i >= ( MTPIXEL_FILE_TYPE_TOTAL - 1 ) )
	{
		return -1;
	}

	return i;
}

mtImage * mtpixel_image_load (
	char		* filename
	)
{
	mtImage		* image = NULL;
	int		ftype,
			i;


	ftype = detect_image_format ( filename );
	if ( ftype == MTPIXEL_FILE_TYPE_NONE )
	{
		return NULL;
	}

	i = file_type_ref ( ftype );
	if ( i < 0 )
	{
		return NULL;
	}

	image = ff_table[i].load ( filename );
	if ( ! image )
	{
		return NULL;
	}

	return image;
}

int mtpixel_image_save (
	mtImage		* image,
	char		* filename,
	mtFileType	filetype,
	int		fileinfo
	)
{
	int		i,
			res = 1;


	if (	! image					||
		! filename				||
		filetype <= MTPIXEL_FILE_TYPE_NONE	||
		filetype >= MTPIXEL_FILE_TYPE_TOTAL	||
		! image->channel[MTPIXEL_CHANNEL_IMAGE]
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	i = file_type_ref ( filetype );
	if ( i < 0 )
	{
		return MTPIXEL_ERRNO_FILETYPE_NOT_SUPPORTED;
	}

	// Ensure the image type is supported by this format
	if (	image->type == MTPIXEL_IMAGE_RGB &&
		! ( ff_table[i].flags & MTPIXEL_FF_RGB )
		)
	{
		return MTPIXEL_ERRNO_FILETYPE_NOT_SUPPORTED;
	}

	if (	image->type == MTPIXEL_IMAGE_INDEXED &&
		! ( ff_table[i].flags & MTPIXEL_FF_INDEXED )
		)
	{
		return MTPIXEL_ERRNO_FILETYPE_NOT_SUPPORTED;
	}

	res = ff_table[i].save ( image, filename, fileinfo );
	if ( res )
	{
		return MTPIXEL_ERRNO_FILE_SAVE;
	}

	return 0;
}

int mtpixel_channel_load (
	mtImage		* image,
	char		* filename,
	mtChannelType	channel
	)
{
	mtImage		* new_image;


	if (	! image			||
		! filename		||
		channel < 0		||
		channel >= MTPIXEL_CHANNEL_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	new_image = mtpixel_image_load ( filename );
	if ( ! new_image )
	{
		return MTPIXEL_ERRNO_FILE_LOAD;
	}

	if (	new_image->width != image->width		||
		new_image->height != image->height		||
		! new_image->channel[MTPIXEL_CHANNEL_IMAGE]	||
		(	channel != MTPIXEL_CHANNEL_IMAGE &&
			new_image->type != MTPIXEL_IMAGE_INDEXED
			)
		)
	{
		mtpixel_image_destroy ( new_image );

		return MTPIXEL_ERRNO_CHANNEL_GEOMETRY;
	}

	// Remove old channel
	mtpixel_channel_destroy ( image, channel );

	if ( channel == MTPIXEL_CHANNEL_IMAGE )
	{
		// Import the palette if we are loading into the image channel
		memcpy ( &image->palette, &new_image->palette,
			sizeof ( image->palette ) );
		image->type = new_image->type;
	}

	// Transfer the data from one structure to another
	image->channel[ channel ] = new_image->channel[ MTPIXEL_CHANNEL_IMAGE ];
	new_image->channel[ MTPIXEL_CHANNEL_IMAGE ] = NULL;

	mtpixel_image_destroy ( new_image );

	return 0;
}

int mtpixel_channel_save (
	mtImage		* image,
	char		* filename,
	mtChannelType	channel,
	mtFileType	filetype,
	int		fileinfo
	)
{
	mtImage		tmp_image;
	int		i;


	if (	! image					||
		! filename				||
		filetype <= MTPIXEL_FILE_TYPE_NONE	||
		filetype >= MTPIXEL_FILE_TYPE_TOTAL	||
		channel < 0				||
		channel >= MTPIXEL_CHANNEL_TOTAL	||
		! image->channel[ channel ]
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	i = file_type_ref ( filetype );
	if ( i < 0 )
	{
		return MTPIXEL_ERRNO_FILETYPE_NOT_SUPPORTED;
	}

	// Created mock image by borrowing the image channel ref
	memcpy ( &tmp_image, image, sizeof ( tmp_image ) );
	memset ( &tmp_image.channel, 0, sizeof ( tmp_image.channel ) );

	tmp_image.channel[ MTPIXEL_CHANNEL_IMAGE ] = image->channel[ channel ];
	if ( channel != MTPIXEL_CHANNEL_IMAGE )
	{
		// Create grey palette for indexed channel
		tmp_image.type = MTPIXEL_IMAGE_INDEXED;
		create_grey_palette ( &tmp_image.palette );
	}

	i = mtpixel_image_save ( &tmp_image, filename, filetype, fileinfo );
	if ( i )
	{
		return i;
	}

	return 0;
}

