/*
	Copyright (C) 2006-2009 Dmitry Groshev
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



/*
These algorithms were first created for mtPaint by Dmitry Groshev. I made a few
very minor tweaks to the interface to get them into an mtPixel context but
otherwise the work and all credit should go to Dmitry.
M.Tyler.
*/

/*
Max-Min quantization algorithm - good for preserving saturated colors,
and because of that, bad when used without dithering - WJ
*/

#define HISTSIZE (64 * 64 * 64)


static int maxminquan (
	unsigned char	* inbuf,
	int		width,
	int		height,
	int		quant_to,
	mtPalette	* pal
	)
{
	int		i,
			j,
			k,
			ii,
			r,
			g,
			b,
			dr,
			dg,
			db,
			l = width * height,
			* hist;

	/* Allocate histogram */
	hist = calloc ( 1, HISTSIZE * sizeof ( int ) );
	if ( ! hist )
	{
		return 1;
	}

	/* Fill histogram */
	for ( i = 0; i < l; i++ )
	{
		++hist[ ( ( inbuf[0] & 0xFC ) << 10 ) +
			( ( inbuf[1] & 0xFC ) << 4 ) +
			( inbuf[2] >> 2) ];

		inbuf += 3;
	}

	/* Find the most frequent color */
	j = k = -1;

	for ( i = 0; i < HISTSIZE; i++ )
	{
		if ( hist[i] <= k )
		{
			continue;
		}

		j = i;
		k = hist[i];
	}

	/* Make it first */
	pal->col[0][0] = r = j >> 12;
	pal->col[0][1] = g = (j >> 6) & 0x3F;
	pal->col[0][2] = b = j & 0x3F;

	/* Find distances from all others to it */
	if ( quant_to > 1 )
	{
		for ( i = 0; i < HISTSIZE; i++ )
		{
			if ( ! hist[i] )
			{
				continue;
			}

			dr = ( i >> 12 ) - r;
			dg = ( ( i >> 6 ) & 0x3F ) - g;
			db = ( i & 0x3F ) - b;
			hist[i] = dr * dr + dg * dg + db * db;
		}
	}

	/* Add more colors */
	for ( ii = 1; ii < quant_to; ii++ )
	{
		/* Find farthest color */
		j = -1;

		for ( i = k = 0; i < HISTSIZE; i++ )
		{
			if ( hist[i] <= k )
			{
				continue;
			}

			j = i;
			k = hist[i];
		}

		/* No more colors? */
		if ( j < 0 )
		{
			break;
		}

		/* Store into palette */
		pal->col[ii][0] = r = j >> 12;
		pal->col[ii][1] = g = (j >> 6) & 0x3F;
		pal->col[ii][2] = b = j & 0x3F;

		/* Update distances */
		for ( i = 0; i < HISTSIZE; i++ )
		{
			if ( ! hist[i] )
			{
				continue;
			}

			dr = ( i >> 12 ) - r;
			dg = ( ( i >> 6 ) & 0x3F ) - g;
			db = ( i & 0x3F ) - b;
			k = dr * dr + dg * dg + db * db;

			if ( k < hist[i] )
			{
				hist[i] = k;
			}
		}
	}

	/* Upconvert colors */
	for ( i = 0; i < ii; i++ )
	{
		pal->col[i][0] = (pal->col[i][0] << 2) + (pal->col[i][0] >> 4);
		pal->col[i][1] = (pal->col[i][1] << 2) + (pal->col[i][1] >> 4);
		pal->col[i][2] = (pal->col[i][2] << 2) + (pal->col[i][2] >> 4);
	}

	free ( hist );

	return 0;
}

/*
Pairwise Nearest Neighbor quantization algorithm - minimizes mean square
error measure; time used is proportional to number of bins squared - WJ
*/

typedef struct
{
	double		rc,
			gc,
			bc,
			err;
	int		cnt;
	unsigned short	nn,
			fw,
			bk,
			tm,
			mtm;
} pnnbin;



static void find_nn (
	pnnbin		* bins,
	int		idx
	)
{
	pnnbin		* bin1,
			* bin2;
	int		i,
			nn = 0;
	double		n1,
			n2,
			dr,
			dg,
			db,
			nerr,
			err = 1e100;


	bin1 = bins + idx;
	n1 = bin1->cnt;

	for ( i = bin1->fw; i; i = bin2->fw )
	{
		bin2 = bins + i;
		n2 = bin2->cnt;
		dr = bin2->rc - bin1->rc;
		dg = bin2->gc - bin1->gc;
		db = bin2->bc - bin1->bc;
		nerr = ( ( ( dr * dr + dg * dg + db * db ) / ( n1 + n2 ) ) *
			n1 * n2 );

		if ( nerr >= err )
		{
			continue;
		}

		err = nerr;
		nn = i;
	}

	bin1->err = err;
	bin1->nn = nn;
}

static int pnnquan (
	unsigned char	* inbuf,
	int		width,
	int		height,
	int		quant_to,
	mtPalette	* pal
	)
{
	unsigned short	heap[ 32769 ];
	pnnbin		* bins,
			* tb,
			* nb;
	double		d,
			err,
			n1,
			n2;
	int		i,
			j,
			k,
			l,
			l2,
			h,
			b1,
			maxbins,
			extbins;


	heap[0] = 0;			// Empty

	bins = calloc ( 32768, sizeof ( pnnbin ) );

	if ( ! bins )
	{
		return 1;
	}

	/* Build histogram */
	k = width * height;
	for ( i = 0; i < k; i++ , inbuf += 3 )
	{
// !!! Can throw gamma correction in here, but what to do about perceptual
// !!! nonuniformity then?
		j = ( ( inbuf[0] & 0xF8 ) << 7 ) +
			( ( inbuf[1] & 0xF8 ) << 2) +
			( inbuf[2] >> 3 );

		tb = bins + j;
		tb->rc += inbuf[0];
		tb->gc += inbuf[1];
		tb->bc += inbuf[2];
		tb->cnt ++;
	}

	/* Cluster nonempty bins at one end of array */
	tb = bins;
	for ( i = 0; i < 32768; i++ )
	{
		if ( ! bins[i].cnt )
		{
			continue;
		}

		*tb = bins[i];
		d = 1.0 / (double)tb->cnt;
		tb->rc *= d; tb->gc *= d; tb->bc *= d;
		tb ++;
	}

	maxbins = tb - bins;

	for ( i = 0; i < maxbins - 1; i++ )
	{
		bins[i].fw = i + 1;
		bins[i + 1].bk = i;
	}

// !!! Already zeroed out by calloc ()
//	bins[0].bk = bins[i].fw = 0;

	/* Initialize nearest neighbors and build heap of them */
	for ( i = 0; i < maxbins; i++ )
	{
		find_nn ( bins, i );
		/* Push slot on heap */
		err = bins[i].err;

		for ( l = ++heap[0]; l > 1; l = l2 )
		{
			l2 = l >> 1;

			if ( bins[h = heap[l2]].err <= err )
			{
				break;
			}

			heap[l] = h;
		}

		heap[l] = i;
	}

	/* Merge bins which increase error the least */
	extbins = maxbins - quant_to;
	for ( i = 0; i < extbins; )
	{
		/* Use heap to find which bins to merge */
		while ( 1 )
		{
			/* One with least error */
			tb = bins + ( b1 = heap[1] );

			/* Is stored error up to date? */

			if (	( tb->tm >= tb->mtm ) &&
				( bins[tb->nn].mtm <= tb->tm )
				)
			{
				break;
			}

			if ( tb->mtm == 0xFFFF )
			{
				/* Deleted node */
				b1 = heap[1] = heap[ heap[0] -- ];
			}
			else
			{
				/* Too old error value */
				find_nn ( bins, b1 );
				tb->tm = i;
			}

			/* Push slot down */
			err = bins[b1].err;

			for ( l = 1; (l2 = l + l) <= heap[0]; l = l2 )
			{
				if (	l2 < heap[0] &&
					bins[heap[l2]].err >
						bins[ heap[ l2 + 1 ] ].err
					)
				{
					l2++;
				}

				if ( err <= bins[ h = heap[ l2 ] ].err )
				{
					break;
				}

				heap[l] = h;
			}

			heap[l] = b1;
		}

		/* Do a merge */
		nb = bins + tb->nn;
		n1 = tb->cnt;
		n2 = nb->cnt;
		d = 1.0 / ( n1 + n2 );
		tb->rc = d * rint ( n1 * tb->rc + n2 * nb->rc );
		tb->gc = d * rint ( n1 * tb->gc + n2 * nb->gc );
		tb->bc = d * rint ( n1 * tb->bc + n2 * nb->bc );
		tb->cnt += nb->cnt;
		tb->mtm = ++i;

		/* Unchain deleted bin */
		bins[ nb->bk ].fw = nb->fw;
		bins[ nb->fw ].bk = nb->bk;
		nb->mtm = 0xFFFF;
	}

	/* Fill palette */
	i = j = 0;
	while ( 1 )
	{
		pal->col[j][0] = rint ( bins[i].rc );
		pal->col[j][1] = rint ( bins[i].gc );
		pal->col[ j++ ][2] = rint ( bins[i].bc );

		if ( ! (i = bins[i].fw) )
		{
			break;
		}
	}

	free ( bins );

	return 0;
}


int mtpixel_image_quantize (
	mtImage		* image,
	mtPalette	* palette,
	int		cols,
	mtQuantizeType	type
	)
{
	int		i,
			res = 0;


	if (	! image					||
		! image->channel[MTPIXEL_CHANNEL_IMAGE]	||
		image->type != MTPIXEL_IMAGE_RGB	||
		cols < 2				||
		cols > 256				||
		type <= MTPIXEL_QUANTIZE_NONE		||
		type >= MTPIXEL_QUANTIZE_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( ! palette )
	{
		palette = &image->palette;
	}

	switch ( type )
	{
	case MTPIXEL_QUANTIZE_PNN:
		res = pnnquan ( image->channel[ MTPIXEL_CHANNEL_IMAGE ],
			image->width, image->height, cols, palette );
		break;

	case MTPIXEL_QUANTIZE_MAXMIN:
		res = maxminquan ( image->channel[ MTPIXEL_CHANNEL_IMAGE ],
			image->width, image->height, cols, palette );
		break;

	default:
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( res )
	{
		return MTPIXEL_ERRNO_MALLOC;
	}

	image->palette.size = cols;

	/* Clear empty slots */
	for ( i = cols; i < 256; i++ )
	{
		image->palette.col[i][0] = 0;
		image->palette.col[i][1] = 0;
		image->palette.col[i][2] = 0;
	}

	return 0;
}

