/*
	Copyright (C) 2004-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "private.h"

#include "icon_xpm.xpm"



#define MAX_COMBO_ITEMS		16
#define MAX_TABS		10
#define TOT_COLS		9
				// 9th column = empty to keep 8th column nicely
				// spaced.



typedef struct
{
	char		path[ PATH_LEN ];
	GtkWidget	* clist;
	int		sort_column;
	GtkSortType	sort_direction;
	GtkWidget	* sort_arrows[ TOT_COLS ];
	CedSheet	* sheet;
} Tab_node;



static GtkWidget	* main_window,
			* combo,
			* combo_entry,
			* progressbar,
			* notebook,
			* button_analyse,
			* button_copy;
static GList		* combo_list;
static char		combo_items[ MAX_COMBO_ITEMS ][ PATH_LEN ];
static int		stop_flag;
static double		progress = -1.0;
static Tab_node		tabarr[ MAX_TABS ];



static gint delete_event (
	GtkWidget	* widget,
	GdkEvent	* event,
	gpointer	data
	)
{
	gint		x,
			y,
			width,
			height;
	int		i;
	char		tmptxt[32];


	if ( progress > 0 )
	{
		// In the middle of analysing so signal a bail out
		stop_flag = TRUE;

		return TRUE;
	}

	gdk_window_get_size ( main_window->window, &width, &height );
	gdk_window_get_root_origin ( main_window->window, &x, &y );

	prefs_set_int ( PREFS_WINDOW_X, x );
	prefs_set_int ( PREFS_WINDOW_Y, y );
	prefs_set_int ( PREFS_WINDOW_W, width );
	prefs_set_int ( PREFS_WINDOW_H, height );

	for ( i = 0; i < MAX_COMBO_ITEMS; i++ )
	{
		sprintf ( tmptxt, PREFS_DIR_PREFIX"%02i", i );
		prefs_set_string ( tmptxt, combo_items[i] );
	}

	gtk_main_quit ();

	for ( i = 0; i < MAX_TABS; i++ )
	{
		ced_sheet_destroy ( tabarr[ i ].sheet );
		tabarr[ i ].sheet = NULL;
	}

	return FALSE;
}

static int get_page ()
{
	int		page;


	page = gtk_notebook_get_current_page ( GTK_NOTEBOOK ( notebook ) );
	if ( page < 0 || page >= TOT_COLS )
	{
		// Trap stupidity
		return -1;
	}

	return page;
}

static gint copy_button (
	GtkWidget	* widget,
	GdkEvent	* event,
	gpointer	data
	)
{
	char		ch = 0,
			* tbuf = NULL,
			* txt = NULL;
	int		page;
	mtFile		* memfile;
	void		* vb;


	page = get_page ();
	if ( page < 0 )
	{
		return FALSE;
	}

	memfile = ced_sheet_save_mem ( tabarr[ page ].sheet,
		CED_FILE_TYPE_OUTPUT_TSV );
	if ( ! memfile )
	{
		return FALSE;
	}

	// NUL terminate so we can use as string
	if (	mtkit_file_write ( memfile, &ch, 1 )	||
		mtkit_file_get_mem ( memfile, &vb, NULL )
		)
	{
		goto finish;
	}

	txt = vb;

	if ( ! mtkit_utf8_string_legal ( (unsigned char *)txt, 0 ) )
	{
		// Create proper UTF-8 for GTK+
		tbuf = mtkit_iso8859_to_utf8 ( txt, 0, NULL );
		if ( ! tbuf )
		{
			goto finish;
		}

		txt = tbuf;
	}

	gtk_clipboard_set_text ( gtk_clipboard_get ( GDK_SELECTION_CLIPBOARD ),
		txt, -1 );

finish:
	mtkit_file_close ( memfile );
	free ( tbuf );

	return FALSE;
}

static gint analyse_button (
	GtkWidget	* widget,
	GdkEvent	* event,
	gpointer	data
	)
{
	const char	* input;


	input = gtk_entry_get_text ( GTK_ENTRY ( combo_entry ) );
	raft_gui_analyse ( input );

	return FALSE;
}

void raft_gui_init ( void )
{
	int		i;
	char		tmptxt[32];
	GtkWidget	* vbox,
			* hbox,
			* button;
	GtkAccelGroup	* ag;


	gtk_init ( NULL, NULL );

	main_window = gtk_window_new ( GTK_WINDOW_TOPLEVEL );
	ag = gtk_accel_group_new ();

	// Set minimum width/height
	gtk_widget_set_usize ( main_window, 400, 300);

	gtk_window_set_default_size ( GTK_WINDOW ( main_window ),
		prefs_get_int ( PREFS_WINDOW_W ),
		prefs_get_int ( PREFS_WINDOW_H )
		);
	gtk_widget_set_uposition ( main_window,
		prefs_get_int ( PREFS_WINDOW_X ),
		prefs_get_int ( PREFS_WINDOW_Y )
		);

	gtk_window_set_title ( GTK_WINDOW ( main_window ), VERSION );

	vbox = gtk_vbox_new ( FALSE, 0 );
	gtk_widget_show ( vbox );
	gtk_container_add ( GTK_CONTAINER ( main_window ), vbox );

	hbox = gtk_hbox_new ( FALSE, 0 );
	gtk_widget_show ( hbox );
	gtk_box_pack_start ( GTK_BOX ( vbox ), hbox, FALSE, FALSE, 5 );

/////////	TOP LINE BUTTONS

	button = gtk_button_new_with_label ( "Quit" );
	gtk_widget_show ( button );
	gtk_box_pack_start ( GTK_BOX ( hbox ), button, FALSE, FALSE, 0);
	gtk_signal_connect ( GTK_OBJECT ( button ), "clicked",
		GTK_SIGNAL_FUNC ( delete_event ), NULL );
	gtk_widget_add_accelerator ( button, "clicked", ag, GDK_Q,
		GDK_CONTROL_MASK, (GtkAccelFlags) 0 );

	progressbar = gtk_progress_bar_new ();
	gtk_widget_show ( progressbar );
	gtk_box_pack_start ( GTK_BOX ( hbox ), progressbar, TRUE, TRUE, 0 );
	gtk_progress_set_percentage ( GTK_PROGRESS ( progressbar ), 0.0 );
	gtk_progress_set_format_string ( GTK_PROGRESS ( progressbar ),
		"Status: Idle" );
	gtk_progress_set_show_text ( GTK_PROGRESS ( progressbar ), TRUE );

	button = gtk_button_new_with_label ( "Analyse" );
	gtk_widget_show ( button );
	gtk_box_pack_start ( GTK_BOX ( hbox ), button, FALSE, FALSE, 0 );
	gtk_signal_connect ( GTK_OBJECT ( button ), "clicked",
		GTK_SIGNAL_FUNC ( analyse_button ), NULL );
	button_analyse = button;

	button = gtk_button_new_with_label ( "Copy to Clipboard" );
	gtk_widget_show ( button );
	gtk_box_pack_start ( GTK_BOX ( hbox ), button, FALSE, FALSE, 0 );
	gtk_signal_connect ( GTK_OBJECT ( button ), "clicked",
		GTK_SIGNAL_FUNC ( copy_button ), NULL );
	button_copy = button;
	gtk_widget_add_accelerator ( button, "clicked", ag, GDK_C,
		GDK_CONTROL_MASK, (GtkAccelFlags) 0 );

	gtk_widget_set_sensitive ( button_copy, FALSE );

/////////	COMBO

	combo = gtk_combo_new ();
	gtk_widget_show ( combo );
	gtk_box_pack_start ( GTK_BOX ( vbox ), combo, FALSE, FALSE, 0 );
	combo_entry = GTK_COMBO ( combo )->entry;

	for ( i = 0; i < MAX_COMBO_ITEMS; i++ )
	{
		sprintf ( tmptxt, PREFS_DIR_PREFIX"%02i", i );
		mtkit_strncpy0 ( combo_items[i], prefs_get_string ( tmptxt ),
			PATH_LEN );
		combo_list = g_list_append ( combo_list, combo_items[i] );
	}

	gtk_combo_set_popdown_strings ( GTK_COMBO ( combo ), combo_list );

/////////	NOTEBOOK

	notebook = gtk_notebook_new ();
	gtk_box_pack_start ( GTK_BOX ( vbox ), notebook, TRUE, TRUE, 0 );
	gtk_notebook_set_scrollable ( GTK_NOTEBOOK ( notebook ), TRUE );
	gtk_notebook_set_tab_pos ( GTK_NOTEBOOK ( notebook ), GTK_POS_BOTTOM );
	gtk_widget_show ( notebook );

/////////	End of main window widget setup

	gtk_signal_connect_object ( GTK_OBJECT ( main_window ), "delete_event",
		GTK_SIGNAL_FUNC ( delete_event ), NULL );
	gtk_container_set_border_width ( GTK_CONTAINER ( main_window ), 5 );

	gtk_widget_show ( main_window );
	gtk_window_add_accel_group ( GTK_WINDOW ( main_window ), ag );

	gtk_window_set_icon ( (GtkWindow *)main_window,
		gdk_pixbuf_new_from_xpm_data ( (const char **)icon_xpm ) );
}

void raft_gui_start ( void )
{
	gtk_main ();
}

static int scan_cb (
	CedSheet	* sheet,
	int		row,
	void		* user_data
	)
{
	progress = progress + 0.01;

	if ( progress > 1.0 )
	{
		progress = 0;
	}

	gtk_progress_set_percentage ( GTK_PROGRESS ( progressbar ), progress );

	while ( gtk_events_pending () )
	{
		gtk_main_iteration ();	// Keep GUI interactive
	}

	if ( stop_flag )
	{
		return 1;		// User wants to stop
	}

	return 0;			// Keep scanning
}

static void fit_columns (
	GtkWidget	* clist
	)
{
	int		i;


	for ( i = 0; i < TOT_COLS; i++ )
	{
		// Adjust column widths
		gtk_clist_set_column_width ( GTK_CLIST ( clist ), i,
			20 + gtk_clist_optimal_column_width (
			GTK_CLIST ( clist ), i ) );
	}
}

static void sort_and_populate (
	int		tab
	)
{
	gchar		* rtext[ RAFT_COL_TOTAL ] = { NULL };
	unsigned int	cols[] = { 1, 0 };
	int		r,
			c;
	CedSheet	* sheet;
	CedCell		* cell;
	GtkCList	* clist;


	sheet = tabarr[ tab ].sheet;
	clist = (GtkCList *)tabarr[ tab ].clist;

	if ( tabarr[ tab ].sort_direction == GTK_SORT_ASCENDING )
	{
		r = 0;
	}
	else
	{
		r = CED_SORT_MODE_DESCENDING;
	}

	cols[0] = tabarr[ tab ].sort_column + 1;
	ced_sheet_sort_rows ( sheet, 1, 0, cols, r, NULL );

	gtk_clist_freeze ( clist );
	gtk_clist_clear ( clist );

	// Populate clist from sheet
	for ( r = 1; ; r++ )
	{
		for ( c = 1; c < RAFT_COL_TOTAL; c++ )
		{
			cell = ced_sheet_get_cell ( sheet, r, c );
			if ( ! cell )
			{
				goto finish;
			}

			rtext[ c - 1 ] = ced_cell_create_output ( cell, NULL );

			if ( ! mtkit_utf8_string_legal (
				(unsigned char *)rtext[ c - 1 ], 0 )
				)
			{
				char		* tmp,
						* txt = rtext[ c - 1 ];

				// Create proper UTF-8 for GTK+
				tmp = mtkit_iso8859_to_utf8 ( txt, 0, NULL );

				if ( ! tmp )
				{
					tmp = strdup ( "???" );
				}

				free ( txt );
				rtext[ c - 1 ] = tmp;
			}
		}

		gtk_clist_append ( clist, rtext );

		for ( c = 1; c < RAFT_COL_TOTAL; c++ )
		{
			free ( rtext [ c - 1 ] );
		}
	}

finish:

	gtk_clist_thaw ( clist );
}

static void column_button (
	GtkWidget	* widget,
	gint		col,
	gpointer	* pointer
	)
{
	int		page;
	GtkSortType	direction;


	page = get_page ();
	if ( page < 0 )
	{
		return;
	}

	// Reverse the sorting direction if the list is already sorted by this
	// col

	if ( tabarr[page].sort_column == col )
	{
		direction = ( tabarr[page].sort_direction == GTK_SORT_ASCENDING
				? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING );
	}
	else
	{
		direction = GTK_SORT_DESCENDING;
	}


	// Hide old arrow
	gtk_widget_hide ( tabarr[page].sort_arrows[tabarr[page].sort_column] );

	tabarr[page].sort_column = col;
	tabarr[page].sort_direction = direction;

	gtk_arrow_set ( GTK_ARROW (
		tabarr[page].sort_arrows[tabarr[page].sort_column] ),
		( tabarr[page].sort_direction == GTK_SORT_ASCENDING ?
			GTK_ARROW_DOWN : GTK_ARROW_UP ),
		GTK_SHADOW_IN );

	// Show new arrow
	gtk_widget_show ( tabarr[page].sort_arrows[tabarr[page].sort_column] );

	sort_and_populate ( page );
}

static gint clist_event (
	GtkWidget	* widget,
	GdkEvent	* event
	)
{
	if ( progress >= 0 )
	{
		// Program is busy so do nothing
		return FALSE;
	}

	if (	event->type == GDK_BUTTON_PRESS &&
		( (GdkEventButton *) event)->button == 3 // Right click only
		)
	{
		analyse_button ( NULL, NULL, NULL );

		return TRUE;		// Event handled
	}

	if (	event->type == GDK_KEY_PRESS &&
		(
		( (GdkEventKey *) event)->keyval == GDK_Return ||
		( (GdkEventKey *) event)->keyval == GDK_KP_Enter
		)
		)
	{
		analyse_button ( NULL, NULL, NULL );

		return TRUE;		// Event handled
	}

	return FALSE;			// Event not handled
}

static gint clist_select_row (
	GtkWidget	* clist,
	gint		row,
	gint		col,
	GdkEvent	* event,
	gpointer	* pointer
	)
{
	char		tx[PATH_LEN],
			* row_tex[1];
	int		page;


	if ( progress >= 0 )
	{
		// Program is busy so do nothing
		return FALSE;
	}

	page = gtk_notebook_get_current_page ( GTK_NOTEBOOK ( notebook ) );

	gtk_clist_get_text ( GTK_CLIST ( clist ), row, 0, row_tex );
	mtkit_strncpy0 ( tx, tabarr[page].path, PATH_LEN - 2 );

	if ( row_tex[0][0] != '<' )
	{
		// Don't add <.> or <TOTAL>

		strncat ( tx, row_tex[0], PATH_LEN - 2 );
		strncat ( tx, "/", PATH_LEN - 2 );
	}

	gtk_entry_set_text ( GTK_ENTRY ( combo_entry ), tx );

	return TRUE;
}

static int tab_add (
	int		tab
	)
{
	GtkWidget	* temp_hbox,
			* temp_label,
			* scrolledwindow1,
			* label2,
			* clist;
	char		* col_titles[] = {
				"Name",
				"Files",
				"%",
				"Bytes",
				"MB",
				"%",
				"Subdirs",
				"Other",
				" "
				};
	char		tab_text[32],
			* path;
	int		i;


	path = tabarr[tab].path;

	scrolledwindow1 = gtk_scrolled_window_new ( NULL, NULL );
	gtk_scrolled_window_set_policy (
		GTK_SCROLLED_WINDOW ( scrolledwindow1 ),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_show ( scrolledwindow1 );

	gtk_container_add ( GTK_CONTAINER ( notebook ), scrolledwindow1 );
	gtk_container_set_border_width ( GTK_CONTAINER ( scrolledwindow1 ), 5 );

/////////	Analysis results area

	clist = gtk_clist_new ( TOT_COLS );

	tabarr[tab].clist = clist;
	tabarr[tab].sort_direction = GTK_SORT_DESCENDING;
	tabarr[tab].sort_column = 3;

	for ( i = 0; i < TOT_COLS; i++ )
	{
		temp_hbox = gtk_hbox_new ( FALSE, 0 );
		temp_label = gtk_label_new ( col_titles[i] );
		tabarr[tab].sort_arrows[i] = gtk_arrow_new ( GTK_ARROW_DOWN,
			GTK_SHADOW_IN );

		if ( i != 0 )
		{
			gtk_box_pack_end ( GTK_BOX ( temp_hbox ), temp_label,
				FALSE, FALSE, 0 );
			gtk_box_pack_start ( GTK_BOX ( temp_hbox ),
				tabarr[tab].sort_arrows[i], FALSE, FALSE, 0 );
		}
		else
		{
			gtk_box_pack_start ( GTK_BOX ( temp_hbox ), temp_label,
				FALSE, FALSE, 0 );
			gtk_box_pack_end ( GTK_BOX ( temp_hbox ),
				tabarr[tab].sort_arrows[i], FALSE, FALSE, 0 );
		}

		gtk_widget_show ( temp_label );
		gtk_widget_show ( temp_hbox );
		gtk_clist_set_column_widget ( GTK_CLIST ( clist ), i,
			temp_hbox );
		GTK_WIDGET_UNSET_FLAGS ( GTK_CLIST ( clist )->column[i].button,
			GTK_CAN_FOCUS );

		if ( i > 0 )
		{
			gtk_clist_set_column_justification (
				GTK_CLIST ( clist ), i, GTK_JUSTIFY_RIGHT );
		}
	}

	gtk_widget_show ( tabarr[tab].sort_arrows[3] );
	gtk_arrow_set ( GTK_ARROW ( tabarr[tab].sort_arrows[3] ),
		( tabarr[tab].sort_direction == GTK_SORT_ASCENDING ?
			GTK_ARROW_DOWN : GTK_ARROW_UP ), GTK_SHADOW_IN );

	gtk_clist_column_titles_show ( GTK_CLIST ( clist ) );
	gtk_clist_set_selection_mode ( GTK_CLIST ( clist ),
		GTK_SELECTION_BROWSE );

	gtk_container_add ( GTK_CONTAINER ( scrolledwindow1 ), clist );
	gtk_widget_show ( clist );

	gtk_signal_connect ( GTK_OBJECT ( clist ), "click_column",
		GTK_SIGNAL_FUNC ( column_button ), clist );
	gtk_signal_connect ( GTK_OBJECT ( clist ), "select_row",
		GTK_SIGNAL_FUNC ( clist_select_row ), NULL );
	gtk_signal_connect ( GTK_OBJECT ( clist ), "event",
		GTK_SIGNAL_FUNC ( clist_event ), NULL);

	if ( strlen ( path ) < 21 )
	{
		strcpy ( tab_text, path );
	}
	else
	{
		// Ensure tab title is not too long

		tab_text[0] = 0;
		strcat ( tab_text, " ... " );
		strncat ( tab_text, path + ( strlen ( path ) - 20 ), 20 );
	}

	label2 = gtk_label_new ( tab_text );
	gtk_widget_show ( label2 );
	gtk_notebook_set_tab_label ( GTK_NOTEBOOK ( notebook ),
		gtk_notebook_get_nth_page ( GTK_NOTEBOOK ( notebook ), tab ),
		label2 );
	gtk_label_set_justify ( GTK_LABEL ( label2 ), GTK_JUSTIFY_LEFT );
	gtk_misc_set_alignment ( GTK_MISC ( label2 ), 0, 0.5 );

	gtk_notebook_set_page ( GTK_NOTEBOOK ( notebook ), tab );

	sort_and_populate ( tab );
	fit_columns ( tabarr[tab].clist );

	gtk_widget_set_sensitive ( button_copy, TRUE );

	gtk_widget_grab_focus ( clist );

	return 0;			// Success
}

void raft_gui_analyse (
	const char	* directory
	)
{
	int		i,
			pages;
	char		temp_tex[PATH_LEN];


	if ( progress != -1.0 )
	{
		return;			// Program is already busy
	}

	pages = gtk_notebook_get_n_pages ( GTK_NOTEBOOK ( notebook ) );

	if (	! directory		||
		! directory[0]		||
		pages < 0		||
		pages >= MAX_TABS
		)
	{
		return;
	}

	if ( directory[ strlen ( directory ) - 1 ] != '/' )
	{
		// Ensure path has '/' at end

		mtkit_strncpy0 ( temp_tex, directory, PATH_LEN );
		strcat ( temp_tex, "/" );
		gtk_entry_set_text ( GTK_ENTRY ( combo_entry ), temp_tex );
		directory = temp_tex;
	}

	gtk_progress_set_percentage ( GTK_PROGRESS ( progressbar ), 0.0 );
	gtk_progress_set_format_string ( GTK_PROGRESS ( progressbar ),
		"Status: Working" );

	progress = 0.0;
	stop_flag = 0;
	gtk_widget_set_sensitive ( button_analyse, FALSE );

	if ( raft_scan_sheet ( directory, &tabarr[ pages ].sheet, scan_cb,
		NULL ) )
	{
		goto finish;
	}

	mtkit_strncpy0 ( tabarr[pages].path, directory, PATH_LEN );

	if ( tab_add ( pages ) )
	{
		goto finish;
	}

	for ( i = 0 ; i < (MAX_COMBO_ITEMS - 1); i++ )
	{
		// Does this text already exist in the list?
		if ( ! strcmp ( directory, combo_items[i] ) )
		{
			break;
		}
	}

	for ( ; i > 0; i-- )
	{
		// Make space for the text
		mtkit_strncpy0 ( combo_items[i], combo_items[i - 1], PATH_LEN );
	}

	// Add item to list
	mtkit_strncpy0 ( combo_items[0], directory, PATH_LEN );

	gtk_combo_set_popdown_strings ( GTK_COMBO ( combo ), combo_list );

finish:
	gtk_progress_set_percentage ( GTK_PROGRESS ( progressbar ), 0.0 );
	gtk_progress_set_format_string ( GTK_PROGRESS ( progressbar ),
		"Status: Idle" );

	progress = -1.0;		// Signal that we are done
	gtk_widget_set_sensitive ( button_analyse, TRUE );
}
