/*
	Copyright (C) 2004-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



typedef struct
{
	double		files,
			bytes,
			sdirs,
			other;
} scanTot;

typedef struct
{
	CedSheet	* sheet;
	raftScanFunc	callback;
	void		* user_data;
	int		row;
} scanState;



double get_cell_value (
	CedSheet	* sheet,
	int		row,
	int		column
	)
{
	CedCell		* cell;


	cell = ced_sheet_get_cell ( sheet, row, column );
	if ( ! cell )
	{
		return 0;
	}

	return cell->value;
}

static void add_total (			// a += b
	scanTot		* a,
	scanTot		* b
	)
{
	a->files += b->files;
	a->bytes += b->bytes;
	a->sdirs += b->sdirs;
	a->other += b->other;
}

static void set_row_prefs (
	scanState	* state
	)
{
	CedCellPrefs	* pref_num,
			* pref_decimal,
			* pref_percent;


	pref_num = ced_cell_prefs_new ();
	pref_decimal = ced_cell_prefs_new ();
	pref_percent = ced_cell_prefs_new ();

	pref_num->format = 2;
	mtkit_strfreeswap ( &pref_num->num_thousands, "," );

	pref_decimal->format = 2;
	pref_decimal->num_decimal_places = 2;
	mtkit_strfreeswap ( &pref_decimal->num_thousands, "," );

	pref_percent->format = 2;
	pref_percent->num_decimal_places = 1;

	if ( ! pref_num || ! pref_decimal || ! pref_percent )
	{
		goto finish;
	}

	ced_sheet_set_cell_prefs ( state->sheet, state->row, RAFT_COL_FILES,
		pref_num, NULL );
	ced_sheet_set_cell_prefs ( state->sheet, state->row, RAFT_COL_BYTES,
		pref_num, NULL );
	ced_sheet_set_cell_prefs ( state->sheet, state->row, RAFT_COL_SUBDIRS,
		pref_num, NULL );
	ced_sheet_set_cell_prefs ( state->sheet, state->row, RAFT_COL_OTHER,
		pref_num, NULL );

	ced_sheet_set_cell_prefs ( state->sheet, state->row, RAFT_COL_MB,
		pref_decimal, NULL );

	ced_sheet_set_cell_prefs ( state->sheet, state->row,
		RAFT_COL_FILES_PERCENT, pref_percent, NULL );
	ced_sheet_set_cell_prefs ( state->sheet, state->row,
		RAFT_COL_MB_PERCENT, pref_percent, NULL );

finish:
	ced_cell_prefs_destroy ( pref_num );
	ced_cell_prefs_destroy ( pref_decimal );
	ced_cell_prefs_destroy ( pref_percent );
}

static void add2table (
	scanState	* state,
	char		* dirname,
	scanTot		* tot
	)
{
	ced_sheet_set_cell_text ( state->sheet, state->row, RAFT_COL_NAME,
		dirname );

	ced_sheet_set_cell_value ( state->sheet, state->row, RAFT_COL_FILES,
		tot->files );

	ced_sheet_set_cell_value ( state->sheet, state->row, RAFT_COL_BYTES,
		tot->bytes );

	ced_sheet_set_cell_value ( state->sheet, state->row, RAFT_COL_MB,
		tot->bytes / 1024 / 1024 );

	ced_sheet_set_cell_value ( state->sheet, state->row, RAFT_COL_SUBDIRS,
		tot->sdirs );

	ced_sheet_set_cell_value ( state->sheet, state->row, RAFT_COL_OTHER,
		tot->other );

	set_row_prefs ( state );

	state->row ++;
}

static void add_percentages (
	scanState	* state
	)
{
	int		i;
	double		tot_files,
			tot_bytes,
			perc;


	tot_files = get_cell_value ( state->sheet, state->row, RAFT_COL_FILES );
	tot_bytes = get_cell_value ( state->sheet, state->row, RAFT_COL_BYTES );

	for ( i = 1; i < state->row; i++ )
	{
		perc = get_cell_value ( state->sheet, i, RAFT_COL_FILES );
		perc = 100 / tot_files * perc;

		ced_sheet_set_cell_value ( state->sheet, i,
			RAFT_COL_FILES_PERCENT, perc );

		perc = get_cell_value ( state->sheet, i, RAFT_COL_BYTES );
		perc = 100 / tot_bytes * perc;

		ced_sheet_set_cell_value ( state->sheet, i,
			RAFT_COL_MB_PERCENT, perc );
	}
}

static void add_totals (
	scanState	* state
	)
{
	int		i;


	ced_sheet_set_cell_text ( state->sheet, state->row, RAFT_COL_NAME,
		" < TOTAL >" );

	// Set up totals formulas
	for ( i = RAFT_COL_FILES; i < RAFT_COL_TOTAL; i++ )
	{
		ced_sheet_set_cell ( state->sheet, state->row, i,
			"=sum ( r1c:r[-1]c )" );
	}

	// Calculate Totals
	ced_sheet_recalculate ( state->sheet, NULL, 0 );
	ced_sheet_recalculate ( state->sheet, NULL, 1 );

	// Convert formulas into numbers
	for ( i = RAFT_COL_FILES; i < RAFT_COL_TOTAL; i++ )
	{
		ced_sheet_set_cell_value ( state->sheet, state->row, i,
			get_cell_value ( state->sheet, state->row, i )
			);
	}

	ced_sheet_set_cell_value ( state->sheet, state->row,
		RAFT_COL_FILES_PERCENT, 100 );

	ced_sheet_set_cell_value ( state->sheet, state->row,
		RAFT_COL_MB_PERCENT, 100 );

	set_row_prefs ( state );
}

static int scan_recurse (
	const char	* path,
	scanState	* state,
	int		recurse,	// 1 = No Recurse <.>
					// 2 = Top level dirs
					// 3 = Recurse
	scanTot		* sum		// Subdirectory total for (2->3)
	)
{
	int		res;
	scanTot		tot = { 0, 0, 0, 0 };
	DIR		* dp;
	struct dirent	* ep;
	struct stat	buf;
	char		newp[PATH_LEN],
			fullp[PATH_LEN];


	if ( state->callback ( state->sheet, state->row, state->user_data ) )
	{
		return 1;		// User termination
	}

	// Get full name of file/directory
	mtkit_strncpy0 ( fullp, path, PATH_LEN );

	// Open pathname given
	dp = opendir ( fullp );

	if ( ! dp )
	{
		fprintf ( stderr, "Path doesn't exist: %s\n", fullp );

		return -1;
	}

	while ( ( ep = readdir ( dp ) ) )
	{
		if (	! strcmp ( ep->d_name, "." ) ||
			! strcmp ( ep->d_name, ".." )
			)
		{
			continue;
		}

		mtkit_strncpy0 ( newp, fullp, PATH_LEN );
		strncat ( newp, ep->d_name, PATH_LEN - 1 );
		// Get full name of file/directory

		stat ( newp, &buf );	// Get file details

		switch ( ep->d_type )
		{
		case 0:			// Unknown - crops up on cdroms
			if ( S_ISDIR ( buf.st_mode ) != 0 )
			{
				if ( recurse != 1 )
				{
					scanTot		ltot = { 0, 0, 0, 0 };


					mtkit_strncpy0 ( newp, path, PATH_LEN );
					strncat ( newp, ep->d_name, PATH_LEN - 1
						);
					strncat ( newp, "/", PATH_LEN - 1 );

					res = scan_recurse ( newp, state, 3,
						recurse == 2 ? &ltot : sum );

					if ( res )
					{
						return res;
					}

					if ( recurse == 2 )
					{
						// Top level recursion

						add2table ( state, ep->d_name,
							&ltot );
					}
				}

				if ( recurse != 2 )
				{
					tot.sdirs ++;
				}
			}
			else
			{
				if ( recurse != 2 )
				{
					if ( ep->d_type != 10 )
					{
						tot.files ++;
						tot.bytes = tot.bytes +
							buf.st_size;
					}
					else
					{
						tot.other ++;
					}
				}
			}
			break;

		case DT_DIR:
			// Sub-directory found
			// don't count or follow '.' or '..'

			if (	strcmp ( ep->d_name, "." ) != 0 &&
				strcmp ( ep->d_name, ".." ) != 0
				)
			{
				if ( recurse != 1 )
				{
					scanTot		ltot = { 0, 0, 0, 0 };


					mtkit_strncpy0 ( newp, path, PATH_LEN );
					strncat ( newp, ep->d_name, PATH_LEN - 1
						);
					strncat ( newp, "/", PATH_LEN - 1 );

					res = scan_recurse ( newp, state, 3,
						recurse == 2 ? &ltot : sum );

					if ( res )
					{
						return res;
					}

					if ( recurse == 2 )
					{
						// Top level recursion

						add2table ( state, ep->d_name,
							&ltot );
					}
				}

				if ( recurse != 2 )
				{
					tot.sdirs ++;
				}
			}
			break;

		default:
			// File found - normal, link or other

			if ( recurse != 2 )
			{
				if ( ep->d_type != 10 )
				{
					tot.files ++;
					tot.bytes = tot.bytes + buf.st_size;
				}
				else
				{
					tot.other ++;
				}
			}
		}
	}

	if ( recurse == 1 )
	{
		add2table ( state, " <.>", &tot );
	}
	else if ( recurse == 3 )
	{
		add_total ( sum, &tot );
	}

	closedir ( dp );

	return 0;			// Keep scanning
}

int raft_scan_sheet (
	const char	* path,
	CedSheet	** sheet,
	raftScanFunc	callback,
	void		* user_data
	)
{
	scanState	state = { NULL, callback, user_data, 1 };
	int		res = 0;


	if ( ! path || ! sheet )
	{
		return -1;		// Fail
	}

	sheet[0] = ced_sheet_new ();
	if ( ! sheet[0] )
	{
		return -1;		// Fail
	}

	state.sheet = sheet[0];

	res = scan_recurse ( path, &state, 1, NULL );
	if ( res == 0 )
	{
		res = scan_recurse ( path, &state, 2, NULL );
	}

	if ( res )
	{
		ced_sheet_destroy ( sheet[0] );
		sheet[0] = NULL;

		return res;
	}

	add_totals ( &state );
	add_percentages ( &state );

	return 0;
}
