/*
	Copyright (C) 2004-2013 Mark Tyler
	Copyright (C) 2006-2008 Mark Tyler and Dmitry Groshev

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"

#include <ft2build.h>
#include FT_FREETYPE_H

#include <iconv.h>



/* Valid for x = 0..5, which is enough here */
#define MOD3( x )	( ( ( (x) * 5 + 1 ) >> 2 ) & 3 )

/*
Use sRGB gamma if defined, ITU-R 709 gamma otherwise
From my point of view, ITU-R 709 is a better model of a real CRT
*/

// #define SRGB

/*
Use distorted L*X*N* model; the distortion possibly makes it better for
smaller colour differences, but at extrema, error becomes unacceptable
*/

// #define DISTORT_LXN



#define CIENUM		8192
#define EXPNUM		1024
#define EXPLOW		( -0.1875 )



static double	gamma256[256],
		gamma64[64],
		midgamma256[256],
		midgamma64[64],
		wXN[2];
static float	CIE[CIENUM + 2],
		EXP[EXPNUM];



/*
This nightmarish code does conversion from CIE XYZ into my own perceptually
uniform colour space L*X*N*. To produce it, I combined McAdam's colour space
and CIE lightness function, like it was done for L*u*v* space - the result
is a bitch to evaluate and impossible to reverse, but works much better
than both L*a*b and L*u*v for colour comparison - WJ
*/

static void xyz2XN (
	double		* XN,
	double		x,
	double		y,
	double		z
	)
{
	double		xyz = x + y + z,
			k,
			xk,
			yk,
			sxk;


	/* Black is a darker white ;) */
	if ( xyz == 0.0 )
	{
		XN[0] = wXN[0];
		XN[1] = wXN[1];

		return;
	}

	k = 10.0 / ( x * 2.4 + y * 34.0 + xyz );
	xk = x * k;
	yk = y * k;
	sxk = sqrt ( xk );
	XN[0] = (xk * xk * ( 3751.0 - xk * xk * 10.0 ) -
		yk * yk * ( 520.0 - yk * 13295.0 ) +
		xk * yk * ( 32327.0 - xk * 25491.0 - yk * 41672.0 + xk * xk *
			10.0) - sxk * 5227.0 + sqrt ( sxk ) * 2952.0 ) / 900.0;

#ifndef DISTORT_LXN
	/* Do transform properly */
	k = 10.0 / ( y * 4.2 - x + xyz );
#else
	/*
	This equation is incorrect, but I felt that in reality it's better -
	it compresses the colour plane so that green and blue are farther
	from red than from each other, which conforms to human perception.
	Yet for pure colours, distortion tends to become too high. - WJ
	*/

	k = 10.0 / ( y * 5.2 - x + xyz );
#endif

	xk = x * k; yk = y * k;
	XN[1] = (yk * (404.0 - yk * (185.0 - yk * 52.0) ) +
		xk * (69.0 - yk * (yk * (69.0 - yk * 30.0) + xk * 3.0) ) ) /
		900.0;
}

static double CIElum (
	double		x
	)
{
	int		n;


	x *= CIENUM;
	n = x;

	return (CIE[n] + (CIE[n + 1] - CIE[n]) * (x - n) );
}

static double exp10 (
	double		x
	)
{
	int		n;


	x = (x - EXPLOW) * (EXPNUM + EXPNUM);
	n = x;

	return (EXP[n] + (EXP[n + 1] - EXP[n]) * (x - n));
}



#define XX1	(16.0 / 116.0)
#define XX2	( (116.0 * 116.0) / (24.0 * 24.0 * 3.0) )
#define XX3	( (24.0 * 24.0 * 24.0) / (116.0 * 116.0 * 116.0) )

static double CIEpow (
	double		x
	)
{
	return ( ( x > XX3 ) ? cbrt ( x ) : x * XX2 + XX1 );
}


static double	rxyz[3],
		gxyz[3],
		bxyz[3],
		wy;

static void _rgb2LXN (
	double		* tmp,
	double		r,
	double		g,
	double		b
	)
{
	double		x = r * rxyz[0] + g * gxyz[0] + b * bxyz[0];
	double		y = r * rxyz[1] + g * gxyz[1] + b * bxyz[1];
	double		z = r * rxyz[2] + g * gxyz[2] + b * bxyz[2];
	double		L = CIElum ( y );
	double		XN[2];


	xyz2XN ( XN, x, y, z );

	/* Luminance's range must be near to chrominance's _diameter_ */
#ifndef DISTORT_LXN
	/* As recommended in literature (but sqrt ( 3 ) may be better) */
	tmp[0] = L * 2.0;
#else
	/* As felt good in practice */
	tmp[0] = L * M_SQRT2;
#endif
	tmp[1] = (XN[0] - wXN[0]) * L * 13.0;
	tmp[2] = (XN[1] - wXN[1]) * L * 13.0;
}

/* Get subjective brightness measure, by using Ware-Cowan formula */
static double _rgb2B (
	double		r,
	double		g,
	double		b
	)
{
	double		x = r * rxyz[0] + g * gxyz[0] + b * bxyz[0];
	double		y = r * rxyz[1] + g * gxyz[1] + b * bxyz[1];
	double		z = r * rxyz[2] + g * gxyz[2] + b * bxyz[2];

	z += x + y;

	if ( z <= 0.0 )
	{
		return y;
	}

	z = 1.0 / z;
	x *= z;
	z *= y; /* It's not Z now, but y */
	z = 0.256 + (-0.184 + (-2.527 + 4.656 * x * x + 4.657 * z * z * z) * x)
		* z;
	y *= exp10 ( z ) * wy;

	return (y > 1.0 ? 1.0 : y);
}



#define RED_X		0.640
#define RED_Y		0.330
#define GREEN_X		0.300
#define GREEN_Y		0.600
#define BLUE_X		0.150
#define BLUE_Y		0.060

#if 1 // 6500K whitepoint
	#define WHITE_X		0.3127
	#define WHITE_Y		0.329
#else // 9300K whitepoint
	#define WHITE_X		0.2848
	#define WHITE_Y		0.2932
#endif

static double det (
	double		x1,
	double		y1,
	double		x2,
	double		y2,
	double		x3,
	double		y3
	)
{
	return ( (y1 - y2) * x3 + (y2 - y3) * x1 + (y3 - y1) * x2 );
}

static void make_rgb_xyz ( void )
{
	double		rgbdet,
			wr,
			wg,
			wb;


	rgbdet = det (RED_X, RED_Y, GREEN_X, GREEN_Y, BLUE_X, BLUE_Y) * WHITE_Y;
	wr = det ( WHITE_X, WHITE_Y, GREEN_X, GREEN_Y, BLUE_X, BLUE_Y) / rgbdet;
	wg = det ( RED_X, RED_Y, WHITE_X, WHITE_Y, BLUE_X, BLUE_Y ) / rgbdet,
	wb = det ( RED_X, RED_Y, GREEN_X, GREEN_Y, WHITE_X, WHITE_Y ) / rgbdet;

	rxyz[0] = RED_X * wr;
	rxyz[1] = RED_Y * wr;
	rxyz[2] = (1.0 - RED_X - RED_Y) * wr;
	gxyz[0] = GREEN_X * wg;
	gxyz[1] = GREEN_Y * wg;
	gxyz[2] = (1.0 - GREEN_X - GREEN_Y) * wg;
	bxyz[0] = BLUE_X * wb;
	bxyz[1] = BLUE_Y * wb;
	bxyz[2] = (1.0 - BLUE_X - BLUE_Y) * wb;
	xyz2XN ( wXN, WHITE_X / WHITE_Y, 1.0, ( 1 - WHITE_X - WHITE_Y ) /
		WHITE_Y );

	/* Normalize brightness of white */
	wy = 1.0 / exp10 ( 0.256 + (-0.184 + (-2.527 + 4.656 * WHITE_X *
		WHITE_X + 4.657 * WHITE_Y * WHITE_Y * WHITE_Y) * WHITE_X) *
		WHITE_Y );
}



#ifdef SRGB

	#define GAMMA_POW	2.4
	#define GAMMA_OFS	0.055

	#if 1	/* Standard */
		#define GAMMA_SPLIT	0.04045
		#define GAMMA_DIV	12.92
	#else	/* C1 continuous */
		#define GAMMA_SPLIT	0.03928
		#define GAMMA_DIV	12.92321
	#endif

	#define KGAMMA		800

#else	/* ITU-R 709 */

	#define GAMMA_POW	(1.0 / 0.45)
	#define GAMMA_OFS	0.099
	#define GAMMA_SPLIT	0.081
	#define GAMMA_DIV	4.5
	#define KGAMMA		300

#endif

static void make_gamma (
	double		* Gamma,
	int		cnt
	)
{
	int		i,
			k;
	double		mult = 1.0 / (double)(cnt - 1);


	k = (int)( GAMMA_SPLIT * (cnt - 1) ) + 1;

	for ( i = k; i < cnt; i++ )
	{
		Gamma[i] = pow ( ( (double)i * mult + GAMMA_OFS ) /
			( 1.0 + GAMMA_OFS ), GAMMA_POW );
	}

	mult /= GAMMA_DIV;

	for ( i = 0; i < k; i++ )
	{
		Gamma[i] = (double)i * mult;
	}
}



static double		kgamma256 = KGAMMA * 4,
			kgamma64 = KGAMMA;
static unsigned char	ungamma256[KGAMMA * 4 + 1],
			ungamma64[KGAMMA + 1];



static void make_ungamma (
	double		* Gamma,
	double		* Midgamma,
	unsigned char	* Ungamma,
	int		kgamma,
	int		cnt
	)
{
	int		i,
			j,
			k = 0;


	Midgamma[0] = 0.0;

	for ( i = 1; i < cnt; i++ )
	{
		Midgamma[i] = ( Gamma[i - 1] + Gamma[i] ) / 2.0;
		j = Midgamma[i] * kgamma;

		for ( ; k < j; k++ )
		{
			Ungamma[k] = i - 1;
		}
	}

	for ( ; k <= kgamma; k++ )
	{
		Ungamma[k] = cnt - 1;
	}
}

static void make_CIE ( void )
{
	int		i;


	for ( i = 0; i < CIENUM; i++ )
	{
		CIE[i] = CIEpow ( i * ( 1.0 / CIENUM ) ) * 116.0 - 16.0;
	}

	CIE[CIENUM] = CIE[CIENUM + 1] = 100.0;
}

static void make_EXP ( void )
{
	int		i;


	for ( i = 0; i < EXPNUM; i++ )
	{
		EXP[i] = exp ( M_LN10 * ( i * ( 0.5 / EXPNUM ) + EXPLOW ) );
	}
}

static void init_cols ( void )
{
	make_gamma ( gamma256, 256 );
	make_gamma ( gamma64, 64 );
	make_ungamma ( gamma256, midgamma256, ungamma256, kgamma256, 256 );
	make_ungamma ( gamma64, midgamma64, ungamma64, kgamma64, 64 );
	make_CIE ();
	make_EXP ();
	make_rgb_xyz ();
}



#define UNGAMMA64(X) (ungamma64[ (int)( (X) * kgamma64) ] - \
	( (X) < midgamma64[ ungamma64[ (int)( (X) * kgamma64) ]]) )
#define UNGAMMA256(X) (ungamma256[(int)((X) * kgamma256) ] - \
	( (X) < midgamma256[ ungamma256[ (int)( (X) * kgamma256) ]]) )



static double _ungamma256 (
	double		v
	)
{
	return UNGAMMA256 ( v );
}

static double _gamma256 (
	unsigned char	v
	)
{
	return gamma256[v];
}



static FT_Library	freetype_library;



int mtpixel_init ( void )
{
	FT_Error	error;


	error = FT_Init_FreeType ( &freetype_library );
	if ( error )
	{
		return MTPIXEL_ERRNO_FREETYPE;
	}

	init_cols ();

	return 0;
}

int mtpixel_quit ()
{
	FT_Done_FreeType ( freetype_library );

	return 0;
}


#include "font_8.xbm"
#include "font_16.xbm"



#define FIXED_FONT_WIDTH 8



typedef struct
{
	int		w,
			h,		// Geometry of the bitmap
			x,
			y,		// Offset to draw the bitmap
			advance		// Pixels to advance horizontally
			;
	unsigned char	* pixels;	// Bitmap of the character
} mtFontCache;

struct _mtFontFT
{
	FT_Face		face;		// Font face
	FT_Int32	load_flags;	// Set on creation of font
	int		size;		// Point size
	mtTree		* cache_tree;	// Key = Unicode ID #
					// Data = pointer to mtFontCache
};



static int mtpixel_font_cache_destroy (
	mtFontFT	* ft
	)
{
	if ( ft && ft->cache_tree )
	{
		mtkit_tree_destroy ( ft->cache_tree );
		ft->cache_tree = NULL;
	}

	return 0;
}

static int fcache_cmp (
	void		* k1,
	void		* k2
	)
{
	if ( k1 < k2 )
	{
		return -1;
	}

	if ( k1 > k2 )
	{
		return 1;
	}

	return 0;			// k1 == k2
}

static void fcache_del (
	mtTreeNode	* node
	)
{
	mtFontCache	* fc = node->data;


	if ( fc )
	{
		free ( fc->pixels );
		free ( fc );
	}
}

static int mtpixel_font_cache_new (
	mtFontFT	* ft
	)
{
	mtpixel_font_cache_destroy ( ft );

	ft->cache_tree = mtkit_tree_new ( fcache_cmp, fcache_del );

	if ( ! ft->cache_tree )
	{
		return MTPIXEL_ERRNO_MALLOC;
	}

	return 0;
}

mtFont * mtpixel_font_new (
	char		* filename,
	int		face_index
	)
{
	mtFont		* font;


	font = calloc ( sizeof ( mtFont ), 1 );
	if ( ! font )
	{
		return NULL;
	}

	if ( ! filename )
	{
		font->type = MTPIXEL_FONT_FIXED;

		if ( face_index == 16 )
		{
			font->height = 16;
		}
		else
		{
			font->height = 8;
		}

		return font;
	}
	else
	{
		FT_Error	error;
		mtFontFT	* mtft;


		mtft = calloc ( sizeof ( mtFontFT ), 1 );
		if ( ! mtft )
		{
			free ( font );

			return NULL;
		}

		font->ft = mtft;

		error = FT_New_Face ( freetype_library, filename,
			face_index, &mtft->face );

		if ( error )
		{
			free ( mtft );
			free ( font );

			return NULL;
		}

		if ( FT_IS_SCALABLE ( mtft->face ) )
		{
			font->type = MTPIXEL_FONT_FT_SCALABLE;
		}
		else
		{
			font->type = MTPIXEL_FONT_FT_FIXED;
		}

		// Needed to make dingbat fonts work
		FT_Set_Charmap ( mtft->face, mtft->face->charmaps[0] );

		// Default settings
		mtpixel_font_set_size ( font, 12 );
		mtft->load_flags = FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT;
	}

	return font;
}

int mtpixel_font_destroy (
	mtFont		* font
	)
{
	if ( ! font )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( font->ft )
	{
		FT_Done_Face ( font->ft->face );

		mtpixel_font_cache_destroy ( font->ft );
		free ( font->ft );
	}

	free ( font );

	return 0;
}

int mtpixel_font_set_size (
	mtFont		* font,
	int		size
	)
{
	if ( ! font || size < 1 )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( font->type == MTPIXEL_FONT_FIXED )
	{
		if ( size == 16 )
		{
			font->height = 16;
		}
		else
		{
			font->height = 8;
		}
	}
	else
	{
		mtFontFT	* mtft;


		mtft = font->ft;
		mtpixel_font_cache_new ( mtft );

		if ( font->type == MTPIXEL_FONT_FT_SCALABLE )
		{
			mtft->size = size;

			if ( FT_Set_Char_Size ( mtft->face, size * 64, 0, 0, 0 )
				)
			{
				return MTPIXEL_ERRNO_FREETYPE;
			}

			font->height = ( mtft->face->size->metrics.ascender -
				mtft->face->size->metrics.descender ) >> 6;
		}
		else
		{
			font->height = mtft->face->available_sizes[0].height;

			if ( FT_Set_Pixel_Sizes ( mtft->face,
				mtft->face->available_sizes[0].width,
				font->height )
				)
			{
				return MTPIXEL_ERRNO_FREETYPE;
			}
		}
	}

	return 0;
}

mtFontRender * mtpixel_font_render_new (
	mtFont		* font,
	char		* utf8,
	mtBrush		* brush,
	mtImageType	image_type
	)
{
	mtFontRender	* ren;


	ren = calloc ( sizeof ( mtFontRender ), 1 );
	if ( ren )
	{
		if (	image_type != MTPIXEL_IMAGE_INDEXED &&
			image_type != MTPIXEL_IMAGE_RGB
			)
		{
			image_type = MTPIXEL_IMAGE_RGB;
		}

		ren->font = font;
		ren->utf8 = utf8;
		ren->brush = brush;
		ren->image_type = image_type;
	}

	return ren;
}

int mtpixel_font_render_destroy (
	mtFontRender	* ren
	)
{
	if ( ! ren )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	free ( ren->posx );
	free ( ren );

	return 0;
}

static void fixed_draw_bitmap (
	unsigned char	* mem,
	unsigned char	ch,
	int		w,
	int		x,
	int		y,
	int		height
	)
{
	int		i,
			j;
	unsigned char	* src,
			* dest,
			bitfield;


	if ( height == 16 )
	{
		src = font_16_bits;
	}
	else
	{
		src = font_8_bits;
	}

	src += height * ch;

	for ( j = 0; j < height; j++ )
	{
		dest = mem + x + (y + j) * w;
		bitfield = src[j];

		for ( i = 0; i < FIXED_FONT_WIDTH; i++ )
		{
			if ( bitfield & ( 1 << i ) )
			{
				*dest++ = 255;
			}
			else
			{
				*dest++ = 0;
			}
		}
	}
}

static unsigned char * mt_font_render_fixed (
	mtFontRender	* ren,
	int		render,
	int		get_posx
	)
{
	unsigned char	* mem = NULL,
			* txt = NULL;
	int		n = 0,
			minx,
			miny,
			maxx,
			maxy,
			pass,
			unicode_chars,
			penx,
			peny,
			font_h,
			pass_tot = 1
			;
	unsigned int	unicode;


	font_h = ren->font->height;
	minx = miny = 0;
	maxx = maxy = -( 2 << 24 );
	ren->width = 0;
	ren->height = 0;

	if ( render )
	{
		pass_tot = 2;
	}

	for ( pass = 0; pass < pass_tot; pass ++ )
	{
		penx = 0;
		peny = 0;
		txt = (unsigned char *)ren->utf8;

		for ( n = 0; ; n++ )
		{
			if ( ren->max_chars && n >= ren->max_chars )
			{
				break;
			}

			if ( ! txt || txt[0] == 0 )
			{
				break;
			}

			if ( txt[0] == '\n' )
			{
				txt ++;
				break;
			}

			unicode_chars = 1;

			if ( ( txt[0] & 128 ) == 128 )
			{
				unicode_chars = mtkit_utf8_to_utf32 ( txt,
					&unicode );

				if ( unicode_chars < 0 )
				{
					// Invalid UTF-8 char so stop
					break;
				}
			}
			else
			{
				// Simple ASCII character, so avoid the above
				// overhead
				unicode = txt[0];
			}

			if (	ren->max_width &&
				(penx + FIXED_FONT_WIDTH) > ren->max_width )
			{
				// Width limit reached so stop
				break;
			}

			txt += unicode_chars;

			if ( pass == 0 )
			{
				if ( (penx + FIXED_FONT_WIDTH - 1) > maxx )
				{
					maxx = penx + FIXED_FONT_WIDTH - 1;
				}

				maxy = peny + font_h - 1;
			}
			else if ( pass == 1 )
			{
				if ( mem )
				{
					fixed_draw_bitmap ( mem,
						(unsigned char) unicode,
						ren->width, penx, peny,
						font_h );
				}

				if ( ren->posx )
				{
					ren->posx[n] = penx;
				}
			}

			penx += FIXED_FONT_WIDTH;
		}

		if ( pass == 0 )	// Set up new clipboard
		{
			ren->width = maxx - minx + 1;
			ren->height = maxy - miny + 1;

			if ( ren->width > 0 && ren->height > 0 )
			{
				if ( render )
				{
					mem = calloc ( ren->width * ren->height,
						1 );

					if ( ! mem )
					{
						// Allocation failed so bail out
						pass = 10;
					}
				}
			}
			else
			{
				ren->width = 0;
				ren->height = 0;
			}

			ren->chars = n;
			ren->utf8_next = (char *)txt;

			if ( get_posx && n > 0 )
			{
				ren->posx = calloc ( sizeof ( int ), n );

				if ( ! ren->posx )
				{
					goto error;
				}
			}
		}
	}

	return mem;

error:
	free ( mem );

	return 0;
}

static void ft_draw_cache_bitmap (
	unsigned char	* mem,
	int		w,
	mtFontCache	* fc,
	int		x,
	int		y
	)
{
	unsigned char	* dest,
			* src = fc->pixels;
	int		j,
			p,
			q;


	for (	j = y, q = 0;
		q < fc->h;
		j++, q++
		)
	{
		dest = mem + j * w + x;

		for ( p = 0; p < fc->w; p++ )
		{
			*dest++ |= *src++;
		}
	}
}

static mtFontCache * mt_font_cache_glyph (
	mtFont		* font,
	FT_Int32	unicode,
	mtFontFT	* mtft
	)
{
	FT_Error	error;
	FT_UInt		glyph_index;
	mtFontCache	* cp;
	mtTreeNode	* node;


	if ( unicode < 0 )
	{
		return NULL;
	}

	node = mtkit_tree_node_find ( font->ft->cache_tree,
		(void *)(intptr_t)unicode );

	if ( node )
	{
		return (mtFontCache *)node->data;
	}

	// Glyph hasn't been cached yet so cerate it
	cp = calloc ( sizeof ( mtFontCache ), 1 );
	if ( ! cp )
	{
		goto fail;
	}

	glyph_index = FT_Get_Char_Index ( mtft->face, unicode );

	error = FT_Load_Glyph ( mtft->face, glyph_index, mtft->load_flags );
	if ( error )
	{
		goto fail;
	}

	cp->w = mtft->face->glyph->bitmap.width;
	cp->h = mtft->face->glyph->bitmap.rows;
	cp->x = mtft->face->glyph->bitmap_left;
	cp->y = font->height - mtft->face->glyph->bitmap_top
		+ (mtft->face->size->metrics.descender >> 6);

	if ( cp->w < 1 || cp->h < 1 )
	{
		cp->x = 0;
		cp->y = 0;
	}

	cp->advance = mtft->face->glyph->advance.x >> 6;

	if ( cp->w > 0 && cp->h > 0 )
	{
		cp->pixels = calloc ( cp->w * cp->h, 1 );

		if ( ! cp->pixels )
		{
			goto fail;
		}
	}

	if ( cp->advance < 1 )
	{
		cp->advance = 1;
	}


	unsigned char	* dest,
			* src;
	int		p,
			q,
			ppb = 0;


	switch ( mtft->face->glyph->bitmap.pixel_mode )
	{
	case FT_PIXEL_MODE_GRAY:
		ppb = 1;
		break;

	case FT_PIXEL_MODE_MONO:
		ppb = 8;
		break;
	}

	dest = cp->pixels;

	if ( ppb == 1 )			// 8 bits per pixel greyscale
	{
		for ( q = 0; q < cp->h; q++ )
		{
			src = mtft->face->glyph->bitmap.buffer +
				q * mtft->face->glyph->bitmap.pitch;

			for ( p = 0; p < cp->w; p++ )
			{
				*dest++ = *src++;
			}
		}
	}

	if ( ppb == 8 )			// 1 bit per pixel mono
	{
		int		sh,
				bits;
		unsigned char	field;


		for ( q = 0; q < cp->h; q++ )
		{
			src = mtft->face->glyph->bitmap.buffer +
				q * mtft->face->glyph->bitmap.pitch;

			for ( p = 0; p < cp->w; p += 8 )
			{
				field = src[0];

				for (	bits = 128, sh = 0;
					sh<8;
					sh ++, bits >>= 1
					)
				{
					if ( (sh + p) >= cp->w )
					{
						break;
					}

					if ( field & bits )
					{
						*dest++ = 255;
					}
					else
					{
						*dest++ = 0;
					}
				}

				src ++;
			}
		}
	}

	// Add the glyph to the cache tree
	if ( mtkit_tree_node_add ( font->ft->cache_tree,
		(void *)(intptr_t)unicode, cp ) == 0
		)
	{
		goto fail;
	}

	return cp;			// Success

fail:
	free ( cp );

	return NULL;			// Failure
}


static unsigned char * mt_font_render_ft_cache (
	mtFontRender	* ren,
	int		render,
	mtFontFT	* mtft,
	int		get_posx
	)
{
	unsigned char	* mem = NULL,
			* txt = NULL;
	int		n = 0,
			minx,
			miny,
			maxx,
			maxy,
			bx,
			by,
			bw = 0,
			bh = 0,
			pass,
			penx,
			peny,
			unicode_chars,
			pass_tot = 1
			;
	unsigned int	unicode;
	mtFontCache	* cp;


	minx = miny = 0;
	maxx = maxy = -( 2 << 24 );
	ren->width = 0;
	ren->height = 0;

	if ( render )
	{
		pass_tot = 2;
	}

	for ( pass = 0; pass < pass_tot; pass ++ )
	{
		penx = 0;
		peny = 0;
		txt = (unsigned char *)ren->utf8;

		for ( n = 0; ; n++ )
		{
			if ( ren->max_chars && n >= ren->max_chars )
			{
				break;
			}

			if ( ! txt || txt[0] == 0 )
			{
				break;
			}

			if ( txt[0] == '\n' )
			{
				txt ++;
				break;
			}

			unicode_chars = 1;
			if ( (txt[0] & 128) == 128 )
			{
				unicode_chars = mtkit_utf8_to_utf32 ( txt,
					&unicode );

				if ( unicode_chars < 0 )
				{
					// Invalid UTF-8 char so stop
					break;
				}
			}
			else
			{
				// Simple ASCII character, so avoid the above
				// overhead
				unicode = txt[0];
			}

			cp = mt_font_cache_glyph ( ren->font, unicode, mtft );
			if ( ! cp )
			{
				// Problem so crash out
				break;
			}

			bx = penx + cp->x;
			by = peny + cp->y;
			bw = cp->w;
			bh = cp->h;

			if ( bw == 0 )
			{
				if ( ren->font->type == MTPIXEL_FONT_FT_SCALABLE
					)
				{
					bx = 1 + penx + cp->advance;
				}
				else
				{
					bx = penx + cp->advance;
				}
			}

			if (	ren->max_width &&
				(bx + bw - minx) > ren->max_width )
			{
				// Width limit reached so stop
				break;
			}

			txt += unicode_chars;

			if ( bw == 0 )
			{
				// Needed to correctly judge position with empty
				// bitmap

				bx = penx;
				bh = ren->font->height;
				bw = cp->advance;
			}

			// Calculate new max/min boundaries
			if ( pass == 0 && bw > 0 && bh > 0 )
			{
				// Only do this if there is a bitmap
				if ( bx < minx )
				{
					minx = bx;
				}

				if ( by < miny )
				{
					miny = by;
				}

				if ( bx + bw - 1 > maxx )
				{
					maxx = bx + bw - 1;
				}

				if ( by + bh - 1 > maxy )
				{
					maxy = by + bh - 1;
				}
			}

			penx += cp->advance;

			if ( mem )
			{
				// Draw bitmap onto clipboard memory in pass 1
				ft_draw_cache_bitmap ( mem, ren->width, cp,
					bx-minx, by-miny );

				if ( ren->posx )
				{
					ren->posx[n] = bx-minx;
				}
			}
		}

		if ( pass == 0 )	// Set up new clipboard
		{
			int		min_h = ren->font->height;


			if ( (maxy - miny) < min_h )
			{
				// Ensure that the last line is the full height
				maxy = miny + min_h - 1;
			}

			ren->width = maxx - minx + 1;
			ren->height = maxy - miny + 1;

			if ( ren->width > 0 && ren->height > 0 )
			{
				if ( render )
				{
					mem = calloc ( ren->width * ren->height,
						1 );

					if ( ! mem )
					{
						// Allocation failed so bail out
						pass = 10;
					}
				}
			}
			else
			{
				ren->width = 0;
				ren->height = 0;
			}

			ren->chars = n;
			ren->utf8_next = (char *)txt;

			if ( get_posx && n > 0 )
			{
				ren->posx = calloc ( sizeof ( int ), n );

				if ( ! ren->posx )
				{
					goto error;
				}
			}
		}
	}

	return mem;

error:
	free ( mem );

	return NULL;
}

static void clear_render_outputs (
	mtFontRender	* ren
	)
{
	ren->width = 0;
	ren->height = 0;
	ren->chars = 0;

	free ( ren->posx );
	ren->posx = NULL;
}

static int extent_render (
	mtFontRender	* ren,
	mtImage		** image,
	int		get_posx
	)
{
	int		type = 0,
			chans;
	unsigned char	* mem = NULL;


	if ( ! ren || ! ren->font )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( image )
	{
		type = 1;
	}

	clear_render_outputs ( ren );

	if ( ren->font->type == MTPIXEL_FONT_FIXED )
	{
		mem = mt_font_render_fixed ( ren, type, get_posx );
	}
	else
	{
		if ( ! ren->font->ft )
		{
			return MTPIXEL_ERRNO_ARGUMENT;
		}

		mem = mt_font_render_ft_cache ( ren, type, ren->font->ft,
			get_posx );
	}

	if ( image )
	{
		if ( mem )
		{
			if ( ren->brush )
			{
				chans = MTPIXEL_CHANNEL_MASK_IMAGE;
			}
			else
			{
				chans = 0;
			}

			image[0] = mtpixel_image_new ( ren->width, ren->height,
					ren->image_type, chans );

			if ( ! image[0] )
			{
				free ( mem );
				return MTPIXEL_ERRNO_MALLOC;
			}

			// Rendered text is alpha
			image[0]->channel[ MTPIXEL_CHANNEL_ALPHA ] = mem;

			if ( ren->brush )
			{
				int	t,
					s;


				t = ren->brush->chan_dis[ MTPIXEL_CHANNEL_ALPHA
					];
				s = ren->brush->size;


				// Don't touch the alpha channel + fill
				// rectangle
				ren->brush->chan_dis[MTPIXEL_CHANNEL_ALPHA] = 1;
				ren->brush->size = 0;

				mtpixel_brush_rectangle ( image[0], ren->brush,
					0, 0, ren->width, ren->height );

				ren->brush->chan_dis[MTPIXEL_CHANNEL_ALPHA] = t;
				ren->brush->size = s;
			}
		}
		else
		{
			image[0] = NULL;

			return MTPIXEL_ERRNO_MALLOC;
		}
	}
	else
	{
		// No image being created so lose mem if it exists
		free ( mem );
	}

	return 0;
}

mtImage * mtpixel_font_render (
	mtFontRender	* ren
	)
{
	mtImage * image = NULL;


	extent_render ( ren, &image, 0 );

	return image;
}

mtImage * mtpixel_font_render_posx (
	mtFontRender	* ren
	)
{
	mtImage		* image = NULL;


	extent_render ( ren, &image, 1 );

	return image;
}

int mtpixel_font_extent (
	mtFontRender	* ren
	)
{
	return extent_render ( ren, NULL, 0 );
}

mtPolygon * mtpixel_polygon_new ( void )
{
	mtPolygon	* new;


	new = calloc ( sizeof ( mtPolygon ), 1 );
	if ( ! new )
	{
		return NULL;
	}

	return new;
}

int mtpixel_polygon_add_point (
	mtPolygon	* polygon,
	int		x,
	int		y
	)
{
	int		i;


	if ( ! polygon )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	i = polygon->size;

	if ( i >= MTPIXEL_POLYGON_MAX )
	{
		return MTPIXEL_ERRNO_FULL;
	}

	polygon->point[i][0] = x;
	polygon->point[i][1] = y;
	polygon->size ++;
	polygon->mode &= ~MTPIXEL_POLYGON_READY;

	return 0;
}

int mtpixel_polygon_clear (
	mtPolygon	* polygon
	)
{
	if ( ! polygon )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	polygon->size = 0;
	polygon->mode = MTPIXEL_POLYGON_UNREADY;

	return 0;
}

mtPolygon * mtpixel_polygon_duplicate (
	mtPolygon	* polygon
	)
{
	mtPolygon	* new;


	if ( ! polygon )
	{
		return NULL;
	}

	new = mtpixel_polygon_new ();
	if ( ! new )
	{
		return NULL;
	}

	memcpy ( new, polygon, sizeof ( mtPolygon ) );

	return new;
}

int mtpixel_polygon_destroy (
	mtPolygon	* polygon
	)
{
	if ( ! polygon )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	free ( polygon );

	return 0;
}

static int mtpixel_poly_init (
	mtPolygon	* polygon
	)
{
	int		i;


	if ( polygon->mode & MTPIXEL_POLYGON_READY )
	{
		return 0;		// Already done
	}

	if ( polygon->size > 2 && polygon->size <= MTPIXEL_POLYGON_MAX )
	{
		polygon->min_x = polygon->point[0][0];
		polygon->max_x = polygon->point[0][0];
		polygon->min_y = polygon->point[0][1];
		polygon->max_y = polygon->point[0][1];

		for ( i = 1; i < polygon->size; i++ )
		{
			mtMIN ( polygon->min_x, polygon->min_x,
				polygon->point[i][0] )
			mtMAX ( polygon->max_x, polygon->max_x,
				polygon->point[i][0] )
			mtMIN ( polygon->min_y, polygon->min_y,
				polygon->point[i][1] )
			mtMAX ( polygon->max_y, polygon->max_y,
				polygon->point[i][1] )
		}

		polygon->mode |= MTPIXEL_POLYGON_READY;
	}
	else
	{
		return 1;		// Bogus polygon
	}

	return 0;			// Success
}



static mtPalette default_palette = {
{
{ 0, 0, 0 }, { 255, 0, 0 }, { 0, 255, 0 }, { 255, 255, 0 },
{ 0, 0, 255 }, { 255, 0, 255 }, { 0, 255, 255 }, { 255, 255, 255 },
{ 219, 219, 219 }, { 182, 182, 182 }, { 146, 146, 146 }, { 109, 109, 109 },
{ 73, 73, 73 }, { 36, 36, 36 }, { 219, 0, 0 }, { 182, 0, 0 },
{ 146, 0, 0 }, { 109, 0, 0 }, { 73, 0, 0 }, { 36, 0, 0 },
{ 0, 219, 0 }, { 0, 182, 0 }, { 0, 146, 0 }, { 0, 109, 0 },
{ 0, 73, 0 }, { 0, 36, 0 }, { 219, 219, 0 }, { 182, 182, 0 },
{ 146, 146, 0 }, { 109, 109, 0 }, { 73, 73, 0 }, { 36, 36, 0 },
{ 0, 0, 219 }, { 0, 0, 182 }, { 0, 0, 146 }, { 0, 0, 109 },
{ 0, 0, 73 }, { 0, 0, 36 }, { 219, 0, 219 }, { 182, 0, 182 },
{ 146, 0, 146 }, { 109, 0, 109 }, { 73, 0, 73 }, { 36, 0, 36 },
{ 0, 219, 219 }, { 0, 182, 182 }, { 0, 146, 146 }, { 0, 109, 109 },
{ 0, 73, 73 }, { 0, 36, 36 }, { 255, 219, 219 }, { 219, 182, 182 },
{ 182, 146, 146 }, { 146, 109, 109 }, { 109, 73, 73 }, { 73, 36, 36 },
{ 255, 182, 182 }, { 219, 146, 146 }, { 182, 109, 109 }, { 146, 73, 73 },
{ 109, 36, 36 }, { 255, 146, 146 }, { 219, 109, 109 }, { 182, 73, 73 },
{ 146, 36, 36 }, { 255, 109, 109 }, { 219, 73, 73 }, { 182, 36, 36 },
{ 255, 73, 73 }, { 219, 36, 36 }, { 255, 36, 36 }, { 219, 255, 219 },
{ 182, 219, 182 }, { 146, 182, 146 }, { 109, 146, 109 }, { 73, 109, 73 },
{ 36, 73, 36 }, { 182, 255, 182 }, { 146, 219, 146 }, { 109, 182, 109 },
{ 73, 146, 73 }, { 36, 109, 36 }, { 146, 255, 146 }, { 109, 219, 109 },
{ 73, 182, 73 }, { 36, 146, 36 }, { 109, 255, 109 }, { 73, 219, 73 },
{ 36, 182, 36 }, { 73, 255, 73 }, { 36, 219, 36 }, { 36, 255, 36 },
{ 219, 219, 255 }, { 182, 182, 219 }, { 146, 146, 182 }, { 109, 109, 146 },
{ 73, 73, 109 }, { 36, 36, 73 }, { 182, 182, 255 }, { 146, 146, 219 },
{ 109, 109, 182 }, { 73, 73, 146 }, { 36, 36, 109 }, { 146, 146, 255 },
{ 109, 109, 219 }, { 73, 73, 182 }, { 36, 36, 146 }, { 109, 109, 255 },
{ 73, 73, 219 }, { 36, 36, 182 }, { 73, 73, 255 }, { 36, 36, 219 },
{ 36, 36, 255 }, { 255, 255, 219 }, { 219, 219, 182 }, { 182, 182, 146 },
{ 146, 146, 109 }, { 109, 109, 73 }, { 73, 73, 36 }, { 255, 255, 182 },
{ 219, 219, 146 }, { 182, 182, 109 }, { 146, 146, 73 }, { 109, 109, 36 },
{ 255, 255, 146 }, { 219, 219, 109 }, { 182, 182, 73 }, { 146, 146, 36 },
{ 255, 255, 109 }, { 219, 219, 73 }, { 182, 182, 36 }, { 255, 255, 73 },
{ 219, 219, 36 }, { 255, 255, 36 }, { 255, 219, 255 }, { 219, 182, 219 },
{ 182, 146, 182 }, { 146, 109, 146 }, { 109, 73, 109 }, { 73, 36, 73 },
{ 255, 182, 255 }, { 219, 146, 219 }, { 182, 109, 182 }, { 146, 73, 146 },
{ 109, 36, 109 }, { 255, 146, 255 }, { 219, 109, 219 }, { 182, 73, 182 },
{ 146, 36, 146 }, { 255, 109, 255 }, { 219, 73, 219 }, { 182, 36, 182 },
{ 255, 73, 255 }, { 219, 36, 219 }, { 255, 36, 255 }, { 219, 255, 255 },
{ 182, 219, 219 }, { 146, 182, 182 }, { 109, 146, 146 }, { 73, 109, 109 },
{ 36, 73, 73 }, { 182, 255, 255 }, { 146, 219, 219 }, { 109, 182, 182 },
{ 73, 146, 146 }, { 36, 109, 109 }, { 146, 255, 255 }, { 109, 219, 219 },
{ 73, 182, 182 }, { 36, 146, 146 }, { 109, 255, 255 }, { 73, 219, 219 },
{ 36, 182, 182 }, { 73, 255, 255 }, { 36, 219, 219 }, { 36, 255, 255 },
{ 255, 219, 182 }, { 219, 182, 146 }, { 182, 146, 109 }, { 146, 109, 73 },
{ 109, 73, 36 }, { 73, 36, 0 }, { 255, 182, 146 }, { 219, 146, 109 },
{ 182, 109, 73 }, { 146, 73, 36 }, { 109, 36, 0 }, { 255, 182, 219 },
{ 219, 146, 182 }, { 182, 109, 146 }, { 146, 73, 109 }, { 109, 36, 73 },
{ 73, 0, 36 }, { 255, 146, 182 }, { 219, 109, 146 }, { 182, 73, 109 },
{ 146, 36, 73 }, { 109, 0, 36 }, { 219, 182, 255 }, { 182, 146, 219 },
{ 146, 109, 182 }, { 109, 73, 146 }, { 73, 36, 109 }, { 36, 0, 73 },
{ 182, 146, 255 }, { 146, 109, 219 }, { 109, 73, 182 }, { 73, 36, 146 },
{ 36, 0, 109 }, { 182, 219, 255 }, { 146, 182, 219 }, { 109, 146, 182 },
{ 73, 109, 146 }, { 36, 73, 109 }, { 0, 36, 73 }, { 146, 182, 255 },
{ 109, 146, 219 }, { 73, 109, 182 }, { 36, 73, 146 }, { 0, 36, 109 },
{ 182, 255, 219 }, { 146, 219, 182 }, { 109, 182, 146 }, { 73, 146, 109 },
{ 36, 109, 73 }, { 0, 73, 36 }, { 146, 255, 182 }, { 109, 219, 146 },
{ 73, 182, 109 }, { 36, 146, 73 }, { 0, 109, 36 }, { 219, 255, 182 },
{ 182, 219, 146 }, { 146, 182, 109 }, { 109, 146, 73 }, { 73, 109, 36 },
{ 36, 73, 0 }, { 182, 255, 146 }, { 146, 219, 109 }, { 109, 182, 73 },
{ 73, 146, 36 }, { 36, 109, 0 }, { 255, 182, 0 }, { 219, 146, 0 },
{ 182, 109, 0 }, { 146, 73, 0 }, { 255, 0, 182 }, { 219, 0, 146 },
{ 182, 0, 109 }, { 146, 0, 73 }, { 0, 182, 255 }, { 0, 146, 219 },
{ 0, 109, 182 }, { 0, 73, 146 }, { 0, 0, 0 }, { 0, 0, 0 }
},
	-1, 256

	};



mtImage * mtpixel_image_new (
	int		width,
	int		height,
	mtImageType	type,
	mtChannelMask	chan_mask
	)
{
	int		c,
			err;
	mtImage		* image;


	if ( width < 1 || height < 1 )
	{
		return NULL;
	}

	switch ( type )
	{
		case MTPIXEL_IMAGE_INDEXED:
		case MTPIXEL_IMAGE_RGB:
			break;

		default:
			return NULL;
	}

	image = calloc ( sizeof ( mtImage ), 1 );
	if ( ! image )
	{
		return NULL;
	}

	image->width = width;
	image->height = height;
	image->type = type;

	mtpixel_palette_copy ( &image->palette, &default_palette );

	err = 0;

	for ( c = 0; c < MTPIXEL_CHANNEL_TOTAL; c++ )
	{
		if ( chan_mask & (1 << c) )
		{
			err |= mtpixel_channel_new ( image, c );
		}
	}

	if ( err )
	{
		mtpixel_image_destroy ( image );
		return NULL;
	}

	return image;
}

mtImage * mtpixel_image_new_rgb (
	int		width,
	int		height
	)
{
	return mtpixel_image_new ( width, height, MTPIXEL_IMAGE_RGB,
		MTPIXEL_CHANNEL_MASK_IMAGE );
}

mtImage * mtpixel_image_new_indexed (
	int		width,
	int		height
	)
{
	return mtpixel_image_new ( width, height, MTPIXEL_IMAGE_INDEXED,
		MTPIXEL_CHANNEL_MASK_IMAGE );
}



typedef struct
{
	int		w,
			h,
			x_b,
			y_b,
			x_a,
			y_a;
} Coverlap;



// Image B is placed at coords x,y relative to image A - what is the overlap?
static void calc_overlap (
	mtImage		* a,
	mtImage		* b,
	int		x,
	int		y,
	Coverlap	* cov
	)
{
	cov->x_b = 0;
	cov->y_b = 0;
	cov->x_a = x;
	cov->y_a = y;
	cov->w = b->width;
	cov->h = b->height;

	if ( cov->x_a < 0 )
	{
		cov->w += cov->x_a;
		cov->x_b -= cov->x_a;
		cov->x_a = 0;
	}
	if ( cov->y_a < 0 )
	{
		cov->h += cov->y_a;
		cov->y_b = -cov->y_a;
		cov->y_a = 0;
	}

	if ( (cov->x_a + cov->w) > a->width )
	{
		cov->w = a->width - cov->x_a;
	}

	if ( (cov->y_a + cov->h) > a->height )
	{
		cov->h = a->height - cov->y_a;
	}
}

mtImage * mtpixel_image_new_from_image (
	mtImage		* image,
	int		x,
	int		y,
	int		width,
	int		height
	)
{
	unsigned char	* src,
			* dest;
	int		chan_mask,
			i,
			j,
			dest_stride,
			src_stride,
			bpp;
	mtImage		* new;
	Coverlap	cov;


	if ( ! image )
	{
		return NULL;
	}

	if ( width < 1 )
	{
		width = image->width;
	}

	if ( height < 1 )
	{
		height = image->height;
	}

	chan_mask = mtpixel_channel_get_mask ( image );

	new = mtpixel_image_new ( width, height, image->type, chan_mask );
	if ( ! new )
	{
		return NULL;
	}

	memcpy ( &new->palette, &image->palette, sizeof ( image->palette ) );

	calc_overlap ( image, new, x, y, &cov );

	if ( cov.w > 0 && cov.h > 0 )
	{
		for ( i = 0; i < MTPIXEL_CHANNEL_TOTAL; i++ )
		{
			if ( ! image->channel[i] || ! new->channel[i] )
			{
				continue;
			}

			if ( i == MTPIXEL_CHANNEL_IMAGE )
			{
				bpp = new->type;
			}
			else
			{
				bpp = 1;
			}

			dest_stride = width * bpp;
			src_stride = image->width * bpp;

			for ( j = 0; j < cov.h; j++ )
			{
				src = image->channel[i] +
					(j + cov.y_a) * src_stride +
					cov.x_a * bpp;

				dest = new->channel[i] +
					(j + cov.y_b) * dest_stride +
					cov.x_b * bpp;

				memcpy ( dest, src, cov.w * bpp );
			}
		}
	}

	return new;
}

mtImage * mtpixel_image_new_from_polygon (
	mtImage		* image,
	mtPolygon	* polygon
	)
{
	unsigned char	* dalpha;
	int		i;
	mtImage		* new;
	mtPolygon	* npoly = NULL;
	mtBrush		* nbrush = NULL;


	if ( ! image || ! polygon || mtpixel_poly_init ( polygon ) )
	{
		return NULL;
	}

	new = mtpixel_image_new_from_image ( image, polygon->min_x,
		polygon->min_y, polygon->max_x - polygon->min_x + 1,
		polygon->max_y - polygon->min_y + 1 );

	if ( ! new )
	{
		return NULL;
	}

	// Store old alpha & add alpha channel for polygon
	dalpha = new->channel[ MTPIXEL_CHANNEL_ALPHA ];
	new->channel[ MTPIXEL_CHANNEL_ALPHA ] = NULL;

	if ( mtpixel_channel_new ( new, MTPIXEL_CHANNEL_ALPHA ) )
	{
		goto fail;
	}

	npoly = mtpixel_polygon_duplicate ( polygon );
	if ( ! npoly )
	{
		goto fail;
	}

	// Change the coords to subtract the mins
	for ( i = 0; i < npoly->size; i++ )
	{
		npoly->point[i][0] -= npoly->min_x;
		npoly->point[i][1] -= npoly->min_y;
	}

	npoly->max_x -= npoly->min_x;
	npoly->max_y -= npoly->min_y;
	npoly->min_x = 0;
	npoly->min_y = 0;

	nbrush = mtpixel_brush_new ();
	if ( ! nbrush )
	{
		goto fail;
	}

	// Disable all channels except alpha to draw polygon onto new alpha
	// channel (value 255)

	mtpixel_brush_channel_mask_set ( nbrush, MTPIXEL_CHANNEL_MASK_ALL );
	mtpixel_brush_channel_mask_clear ( nbrush, MTPIXEL_CHANNEL_MASK_ALPHA );
	mtpixel_brush_set_style ( nbrush, MTPIXEL_BRUSH_SHAPE_SQUARE,
		MTPIXEL_BRUSH_MODE_PATTERN, 0, 0, 0, 255 );
	mtpixel_brush_set_index ( nbrush, NULL, 255, 255, MTPIXEL_CHANNEL_ALPHA
		);

	mtpixel_brush_polygon ( new, nbrush, npoly );

	mtpixel_brush_destroy ( nbrush );
	mtpixel_polygon_destroy ( npoly );

	if ( dalpha )
	{
		int		j;
		unsigned char	* src,
				* dest;


		j = new->width * new->height;
		src = dalpha;
		dest = new->channel[ MTPIXEL_CHANNEL_ALPHA ];

		// Multiply it out the old image alpha with the new alpha
		for ( i = 0; i < j; i++ )
		{
			dest[i] = src[i] * dest[i] / 255;
		}

		free ( dalpha );
	}

	return new;

fail:
	free ( dalpha );
	mtpixel_image_destroy ( new );
	mtpixel_brush_destroy ( nbrush );
	mtpixel_polygon_destroy ( npoly );

	return NULL;
}

mtImage * mtpixel_image_duplicate (
	mtImage		* image
	)
{
	int		i,
			bpp;
	mtImage		* new;


	if ( ! image )
	{
		return NULL;
	}

	new = mtpixel_image_new ( image->width, image->height, image->type,
			mtpixel_channel_get_mask ( image ) );

	if ( ! new )
	{
		return NULL;
	}

	// Copy palette
	memcpy ( &new->palette, &image->palette, sizeof ( image->palette ) );

	// Copy pixels
	for ( i = 0; i < MTPIXEL_CHANNEL_TOTAL; i++ )
	{
		if ( ! image->channel[i] )
		{
			continue;
		}

		if ( i == MTPIXEL_CHANNEL_IMAGE )
		{
			bpp = image->type;
		}
		else
		{
			bpp = 1;
		}

		memcpy ( new->channel[i], image->channel[i],
			image->width * image->height * bpp );
	}

	return new;
}

int mtpixel_image_destroy (
	mtImage		* image
	)
{
	int		c;


	if ( ! image )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	for ( c = 0; c < MTPIXEL_CHANNEL_TOTAL; c++ )
	{
		mtpixel_channel_destroy ( image, c );
	}

	free ( image );

	return 0;
}

// Paste pixel for pixel in this channel operation
static void verbatim_paste (
	int		chan,
	mtImage		* dest,
	mtImage		* src,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	unsigned char	* dmem,
			* smem;
	int		a1,
			a2,
			i,
			j,
			dest_stride,
			src_stride,
			bpp;
	Coverlap	cov;


	if ( chan == MTPIXEL_CHANNEL_IMAGE )
	{
		bpp = dest->type;
	}
	else
	{
		bpp = 1;
	}

	calc_overlap ( dest, src, x, y, &cov );

	// Copy pixels that overlap
	if ( cov.w > 0 && cov.h > 0 )
	{
		dest_stride = dest->width * bpp;
		src_stride = src->width * bpp;

		for ( j = 0; j < cov.h; j++ )
		{
			if ( bpp == 1 || brush->opacity == 255 )
			{
				smem = src->channel[chan] +
					(j + cov.y_b) * src_stride +
					cov.x_b * bpp;

				dmem = dest->channel[chan] +
					(j + cov.y_a) * dest_stride +
					cov.x_a * bpp;

				memcpy ( dmem, smem, cov.w * bpp );
			}
			else
			{
				smem = src->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_b) * src_stride +
					bpp * cov.x_b;

				dmem = dest->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_a) * dest_stride +
					bpp * cov.x_a;

				a1 = brush->opacity;
				a2 = 255 - a1;

				for ( i = 0; i < cov.w; i++, dmem += bpp,
					smem += bpp )
				{
					dmem[0] = (a1 * smem[0] + a2 * dmem[0])
						/ 255;
					dmem[1] = (a1 * smem[1] + a2 * dmem[1])
						/ 255;
					dmem[2] = (a1 * smem[2] + a2 * dmem[2])
						/ 255;
				}
			}
		}
	}
}

// Paste using alpha in this channel operation
static void alpha_paste (
	mtImage		* dest,
	mtImage		* src,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	unsigned char	* dmem,
			* smem,
			* amem;
	int		aa,
			a1,
			a2,
			i,
			j,
			dest_stride,
			src_stride,
			bpp;
	Coverlap	cov;


	bpp = dest->type;
	calc_overlap ( dest, src, x, y, &cov );

	// Copy pixels that overlap
	if ( cov.w > 0 && cov.h > 0 )
	{
		dest_stride = dest->width * bpp;
		src_stride = src->width * bpp;

		for ( j = 0; j < cov.h; j++ )
		{
			if ( bpp == MTPIXEL_IMAGE_INDEXED )
			{
				smem = src->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_b) * src_stride + cov.x_b;
				amem = src->channel[MTPIXEL_CHANNEL_ALPHA] +
					(j + cov.y_b) * src->width + cov.x_b;
				dmem = dest->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_a) * dest_stride + cov.x_a;

				for ( i = 0; i < cov.w; i++, amem ++, dmem ++,
					smem ++ )
				{
					if ( amem[0] > 127 )
					{
						dmem[0] = smem[0];
					}
				}
			}
			else if ( bpp == MTPIXEL_IMAGE_RGB )
			{
				smem = src->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_b) * src_stride +
					3 * cov.x_b;

				amem = src->channel[MTPIXEL_CHANNEL_ALPHA] +
					(j + cov.y_b) * src->width + cov.x_b;

				dmem = dest->channel[MTPIXEL_CHANNEL_IMAGE] +
					(j + cov.y_a) * dest_stride +
					3 * cov.x_a;

				aa = brush->opacity;

				for ( i = 0; i < cov.w; i++, amem ++,
					dmem += 3, smem += 3 )
				{
					a1 = amem[0] * aa / 255;
					a2 = 255 - a1;
					dmem[0] = (a1 * smem[0] + a2 * dmem[0])
						/ 255;
					dmem[1] = (a1 * smem[1] + a2 * dmem[1])
						/ 255;
					dmem[2] = (a1 * smem[2] + a2 * dmem[2])
						/ 255;
				}
			}
		}
	}
}

int mtpixel_image_paste (
	mtImage		* dest,
	mtImage		* src,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	int		c;


	if ( ! dest || ! src || ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( dest->type != src->type )
	{
		return MTPIXEL_ERRNO_IMAGE_TYPE;
	}

	for ( c = MTPIXEL_CHANNEL_IMAGE; c <= MTPIXEL_CHANNEL_ALPHA; c++ )
	{
		if (	! dest->channel[c] ||
			! src->channel[c] ||
			brush->chan_dis[c]
			)
		{
			continue;
		}

		if ( c == MTPIXEL_CHANNEL_IMAGE )
		{
			if (	! src->channel[MTPIXEL_CHANNEL_ALPHA] ||
				brush->chan_dis[MTPIXEL_CHANNEL_ALPHA] ||
				(src->channel[MTPIXEL_CHANNEL_ALPHA] &&
					dest->channel[MTPIXEL_CHANNEL_ALPHA]) )
			{
				verbatim_paste ( c, dest, src, brush, x, y );
			}
			else
			{
				alpha_paste ( dest, src, brush, x, y );
				c++;

				// Don't do the alpha channel as we have used
				// it here
			}
		}
		else
		{
			verbatim_paste ( c, dest, src, brush, x, y );
		}
	}

	return 0;
}

int mtpixel_image_merge_duplicates (
	mtImage		* image,
	int		* duplicates
	)
{
	unsigned char	dupes[256],
			* img;
	int		i,
			j,
			found = 0;
	mtPalette	* pal;


	if (	! image ||
		! image->channel[MTPIXEL_CHANNEL_IMAGE] ||
		image->type != MTPIXEL_IMAGE_INDEXED ||
		image->palette.size < 3
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	pal = &image->palette;
	dupes[0] = 0;

	for ( i = image->palette.size - 1; i > 0; i-- )
	{
		dupes[i] = i;		// Start with a clean sheet

		for ( j = 0; j < i; j++ )
		{
			if (	pal->col[i][0] == pal->col[j][0] &&
				pal->col[i][1] == pal->col[j][1] &&
				pal->col[i][2] == pal->col[j][2]
				)
			{
				found ++;

				// Point to first duplicate in the palette
				dupes[i] = j;

				break;
			}
		}
	}

	img = image->channel[MTPIXEL_CHANNEL_IMAGE];
	j = image->width * image->height;

	for ( i = 0; i < j; i++ )	// Scan canvas for duplicates
	{
		img[0] = dupes[ img[0] ];
		img ++;
	}

	if ( duplicates )
	{
		duplicates[0] = found;
	}

	return 0;
}



enum
{
	I_FLIP_HOR,
	I_FLIP_VER,
	I_FLIP_90,
	I_FLIP_180,
	I_FLIP_270
};

static mtImage * image_flipper (
	mtImage		* image,
	int		type
	)
{
	int		c,
			w,
			h,
			ow,
			oh,
			x,
			y,
			bpp,
			stride,
			stride_o;
	unsigned char	* src,
			* dest;
	mtImage		* new;


	if ( ! image )
	{
		return NULL;
	}

	new = mtpixel_image_duplicate ( image );
	if ( ! new )
	{
		return NULL;
	}

	if ( type == I_FLIP_90 || type == I_FLIP_270 )
	{
		new->width = image->height;
		new->height = image->width;
	}

	w = new->width;
	h = new->height;
	ow = image->width;
	oh = image->height;

	for ( c = 0; c < MTPIXEL_CHANNEL_TOTAL; c++ )
	{
		if ( ! new->channel[c] )
		{
			continue;
		}

		if ( c == MTPIXEL_CHANNEL_IMAGE )
		{
			bpp = image->type;
		}
		else
		{
			bpp = 1;
		}

		stride = w * bpp;
		stride_o = ow * bpp;

		for ( y = 0; y < oh; y++ )
		{
			src = image->channel[c] + stride_o * y;

			if ( type == I_FLIP_HOR || type == I_FLIP_180 )
			{
				if ( type == I_FLIP_HOR )
				{
					dest = new->channel[c] + stride * y +
						stride - bpp;
				}
				else	// I_FLIP_180
				{
					dest = new->channel[c] +
						stride * (h - y) - bpp;
				}

				if ( bpp == 1 )
				{
					for ( x = 0; x < ow; x++ )
					{
						*dest-- = *src++;
					}
				}
				else if ( bpp == 3 )
				{
					for ( x = 0; x < ow; x++ )
					{
						*dest++ = *src++;
						*dest++ = *src++;
						*dest++ = *src++;

						dest -= bpp * 2;
					}
				}
			}
			else if ( type == I_FLIP_VER )
			{
				dest = new->channel[c] + stride * (h - 1 - y);
				memcpy ( dest, src, stride );
			}
			else if ( type == I_FLIP_90 || type == I_FLIP_270 )
			{
				if ( type == I_FLIP_90 )
				{
					dest = new->channel[c] + bpp * (w - y) -
						bpp;
				}
				else	// I_FLIP_270
				{
					dest = new->channel[c] + bpp * y +
						(h - 1) * stride;
					stride = -stride;
				}

				if ( bpp == 1 )
				{
					for ( x = 0; x < ow; x++ )
					{
						dest[0] = *src++;
						dest += stride;
					}
				}
				else if ( bpp == 3 )
				{
					for ( x = 0; x < ow; x++ )
					{
						dest[0] = *src++;
						dest[1] = *src++;
						dest[2] = *src++;
						dest += stride;
					}
				}

				if ( type == I_FLIP_270 )
				{
					stride = -stride;
				}
			}
		}
	}

	return new;
}

mtImage * mtpixel_image_flip_vertical (
	mtImage		* image
	)
{
	return image_flipper ( image, I_FLIP_VER );
}

mtImage * mtpixel_image_flip_horizontal (
	mtImage		* image
	)
{
	return image_flipper ( image, I_FLIP_HOR );
}

static mtImage * new_rgb_indexed (
	mtImage		* image,
	mtImageType	new_type,
	mtImageType	old_type
	)
{
	int		i;
	mtImage		* new;


	if (	! image ||
		! image->channel[ MTPIXEL_CHANNEL_IMAGE ] ||
		image->type != old_type
		)
	{
		return NULL;
	}

	new = mtpixel_image_new ( image->width, image->height, new_type,
		mtpixel_channel_get_mask ( image ) );

	if ( ! new )
	{
		return NULL;
	}

	// Copy channels
	for ( i = 0; i < MTPIXEL_CHANNEL_TOTAL; i++ )
	{
		if ( ! image->channel[i] || i == MTPIXEL_CHANNEL_IMAGE )
		{
			continue;
		}

		memcpy ( new->channel[i], image->channel[i],
			image->width * image->height );
	}

	// Copy palette
	mtpixel_palette_copy ( &new->palette, &image->palette );

	return new;
}

mtImage * mtpixel_image_to_rgb (
	mtImage		* image
	)
{
	int		i,
			j;
	unsigned char	pix,
			* src,
			* dest;
	mtImage		* new;
	mtPalette	* pal;


	new = new_rgb_indexed ( image, MTPIXEL_IMAGE_RGB, MTPIXEL_IMAGE_INDEXED
		);

	if ( ! new )
	{
		return NULL;
	}

	// Create RGB pixels
	j = image->width * image->height;
	src = image->channel[MTPIXEL_CHANNEL_IMAGE];
	dest = new->channel[MTPIXEL_CHANNEL_IMAGE];
	pal = &image->palette;

	for ( i = 0; i < j; i++ )
	{
		pix = *src++;
		*dest++ = pal->col[pix][0];
		*dest++ = pal->col[pix][1];
		*dest++ = pal->col[pix][2];
	}

	return new;
}



/*
Dithering works with 6-bit colours, because hardware VGA palette is 6-bit,
and any kind of dithering is imprecise by definition anyway - WJ
*/

typedef struct {
	double		xyz256[768],
			gamma[256 * 2],
			lin[256 * 2];
	int		cspace,
			cdist,
			ncols;
	uint32_t	xcmap[64 * 64 * 2 + 128 * 2];	/* Cache bitmap */
	uint32_t	lcmap[64 * 64 * 2];		/* Extension bitmap */
	unsigned char	cmap[64 * 64 * 64 + 128 * 64];	/* Index cache */
} ctable;



static int lookup_srgb (
	ctable		* ctp,
	double		* srgb
	)
{
	int		i,
			j,
			k,
			n = 0,
			col[3];
	double		d,
			td,
			td2,
			tmp[3];


	/* Convert to 8-bit RGB coords */
	col[0] = _ungamma256 ( srgb[0] );
	col[1] = _ungamma256 ( srgb[1] );
	col[2] = _ungamma256 ( srgb[2] );

	/* Check if there is extended precision */
	k = ( ( col[0] & 0xFC ) << 10 ) +
		( ( col[1] & 0xFC ) << 4 ) + ( col[2] >> 2 );

	if ( ctp->lcmap[k >> 5] & (1 << (k & 31) ) )
	{
		k = 64 * 64 * 64 + ctp->cmap[k] * 64 +
			( ( col[0] & 3 ) << 4 ) +
			( ( col[1] & 3 ) << 2 ) + ( col[2] & 3 );
	}
	else
	{
		n = 256;	/* Use posterized values for 6-bit part */
	}

	/* Use colour cache if possible */
	if ( ctp->xcmap[k >> 5] & (1 << (k & 31) ) )
	{
		return ctp->cmap[k];
	}

	/* Prepare colour coords */
	switch ( ctp->cspace )
	{
	default:
	case 0: /* RGB */
		tmp[0] = ctp->lin[n + col[0]];
		tmp[1] = ctp->lin[n + col[1]];
		tmp[2] = ctp->lin[n + col[2]];
		break;

	case 1: /* sRGB */
		tmp[0] = ctp->gamma[n + col[0]];
		tmp[1] = ctp->gamma[n + col[1]];
		tmp[2] = ctp->gamma[n + col[2]];
		break;

	case 2: /* L*X*N* */
		_rgb2LXN ( tmp, ctp->gamma[ n + col[0] ],
			ctp->gamma[ n + col[1] ],
			ctp->gamma[ n + col[2] ] );
		break;
	}

	/* Find nearest colour */
	d = 1000000000.0;
	for ( i = j = 0; i < ctp->ncols; i++ )
	{
		switch ( ctp->cdist )
		{
		case 0: /* Largest absolute difference (Linf measure) */
			td = fabs ( tmp[0] - ctp->xyz256[i * 3] );
			td2 = fabs ( tmp[1] - ctp->xyz256[i * 3 + 1] );

			if ( td < td2 )
			{
				td = td2;
			}

			td2 = fabs ( tmp[2] - ctp->xyz256[i * 3 + 2]);

			if ( td < td2 )
			{
				td = td2;
			}

			break;

		case 1: /* Sum of absolute differences (L1 measure) */
			td = fabs ( tmp[0] - ctp->xyz256[i * 3] ) +
				fabs ( tmp[1] - ctp->xyz256[i * 3 + 1] ) +
				fabs ( tmp[2] - ctp->xyz256[i * 3 + 2] );
			break;

		default:
		case 2: /* Euclidean distance (L2 measure) */
			td = sqrt ( ( tmp[0] - ctp->xyz256[i * 3] ) *
				( tmp[0] - ctp->xyz256[i * 3] ) +
				( tmp[1] - ctp->xyz256[i * 3 + 1] ) *
				( tmp[1] - ctp->xyz256[i * 3 + 1] ) +
				( tmp[2] - ctp->xyz256[i * 3 + 2] ) *
				( tmp[2] - ctp->xyz256[i * 3 + 2] ) );
			break;
		}

		if ( td >= d )
		{
			continue;
		}

		j = i;
		d = td;
	}

	/* Store & return result */
	ctp->xcmap[k >> 5] |= 1 << ( k & 31 );
	ctp->cmap[k] = j;

	return j;
}


#define ALIGNTO( p, s )	( (void *)( p ) )


/*
Damping functions roughly resemble old GIMP's behaviour, but may need some
tuning because linear sRGB is just too different from normal RGB
*/

static int mem_dither (
	mtImage		* inew,
	mtImage		* iold,
	short		* dither,
	int		cspace,
	int		dist,
	int		limit,
	int		selc,
	int		serpent,
	int		rgb8b,
	double		emult
	)
{
	int		i,
			j,
			k,
			l,
			kk,
			j0,
			j1,
			dj,
			rlen,
			col0,
			col1,
			ncols,
			w,
			h;
	unsigned char	* ddata1,
			* ddata2,
			* src,
			* dest,
			* old,
			* new;
	double		* row0,
			* row1,
			* row2,
			* tmp,
			err,
			intd,
			extd,
			* gamma6,
			* lin6,
			tc0[3],
			tc1[3],
			color0[3],
			color1[3],
			fdiv = 0,
			gamut[6] = { 1, 1, 1, 0, 0, 0 };
	mtPalette	* pal;
	ctable		* ctp;


	old = iold->channel[MTPIXEL_CHANNEL_IMAGE];
	new = inew->channel[MTPIXEL_CHANNEL_IMAGE];
	pal = &inew->palette;
	ncols = inew->palette.size;
	w = inew->width;
	h = inew->height;

	/* Allocate working space */
	rlen = ( w + 4 ) * 3;
	k = ( rlen * 3 + 1 ) * sizeof ( double );
	ddata1 = calloc ( 1, k );
	ddata2 = calloc ( 1, sizeof ( ctable ) + sizeof ( double ) );

	if ( ! ddata1 || ! ddata2 )
	{
		free ( ddata1 );
		free ( ddata2 );

		return 1;
	}

	row0 = ALIGNTO ( ddata1, double );
	row1 = row0 + rlen;
	row2 = row1 + rlen;
	ctp = ALIGNTO ( ddata2, double );

	/* Preprocess palette to find whether to extend precision and where */
	for ( i = 0; i < ncols; i++ )
	{
		j = ( ( pal->col[i][0] & 0xFC ) << 10 ) +
			( ( pal->col[i][1] & 0xFC ) << 4 ) +
			( pal->col[i][2] >> 2);

		l = ctp->cmap[j];

		if ( ! l )
		{
			ctp->cmap[j] = l = i + 1;
			ctp->xcmap[l * 4 + 2] = j;
		}

		k = ( ( pal->col[i][0] & 3 ) << 4 ) +
			( ( pal->col[i][1] & 3 ) << 2 ) +
			( pal->col[i][2] & 3 );

		ctp->xcmap[ l * 4 + ( k & 1 ) ] |= 1 << ( k >> 1 );
	}

	memset ( ctp->cmap, 0, 64 * 64 * 64 );

	for ( k = 0 , i = 4; i < 256 * 4; i += 4 )
	{
		uint32_t	v = ctp->xcmap[i] | ctp->xcmap[i + 1];

		/* Are 2+ colors there somewhere? */

		if ( ! ( ( v & ( v - 1 ) ) |
			( ctp->xcmap[i] & ctp->xcmap[i + 1] ) )
			)
		{
			continue;
		}

		rgb8b = 1;		/* Force 8-bit precision */

		j = ctp->xcmap[i + 2];
		ctp->lcmap[j >> 5] |= 1 << ( j & 31 );
		ctp->cmap[j] = k++;
	}

	memset ( ctp->xcmap, 0, 257 * 4 * sizeof ( uint32_t ) );

	/* Prepare tables */
	for ( i = 0; i < 256; i++ )
	{
		j = ( i & 0xFC ) + ( i >> 6 );
		ctp->gamma[i] = _gamma256 ( i );
		ctp->gamma[i + 256] = _gamma256 ( j );
		ctp->lin[i] = i * ( 1.0 / 255.0 );
		ctp->lin[i + 256] = j * ( 1.0 / 255.0 );
	}

	/* Keep all 8 bits of input or posterize to 6 bits? */
	i = rgb8b ? 0 : 256;
	gamma6 = ctp->gamma + i;
	lin6 = ctp->lin + i;
	tmp = ctp->xyz256;

	for ( i = 0; i < ncols; i++ , tmp += 3 )
	{
		/* Update gamut limits */
		tmp[0] = gamma6[pal->col[i][0]];
		tmp[1] = gamma6[pal->col[i][1]];
		tmp[2] = gamma6[pal->col[i][2]];

		for ( j = 0; j < 3; j++ )
		{
			if ( tmp[j] < gamut[j] )
			{
				gamut[j] = tmp[j];
			}

			if ( tmp[j] > gamut[j + 3] )
			{
				gamut[j + 3] = tmp[j];
			}
		}

		/* Store colour coords */
		switch ( cspace )
		{
		default:
		case 0: /* RGB */
			tmp[0] = lin6[pal->col[i][0]];
			tmp[1] = lin6[pal->col[i][1]];
			tmp[2] = lin6[pal->col[i][2]];
			break;

		case 1: /* sRGB - done already */
			break;

		case 2: /* L*X*N* */
			_rgb2LXN ( tmp, tmp[0], tmp[1], tmp[2] );
			break;
		}
	}

	ctp->cspace = cspace;
	ctp->cdist = dist;
	ctp->ncols = ncols;
	serpent = serpent ? 0 : 2;

	if ( dither )
	{
		fdiv = 1.0 / *dither++;
	}

	/* Process image */
	for ( i = 0; i < h; i++ )
	{
		src = old + i * w * 3;
		dest = new + i * w;
		memset ( row2, 0, rlen * sizeof ( double ) );

		if ( serpent ^= 1 )
		{
			j0 = 0;
			j1 = w * 3;
			dj = 1;
		}
		else
		{
			j0 = ( w - 1 ) * 3;
			j1 = -3;
			dj = -1;
			dest += w - 1;
		}

		for ( j = j0; j != j1; j += dj * 3 )
		{
			for ( k = 0; k < 3; k++ )
			{
				/* Posterize to 6 bits as natural for palette */
				color0[k] = gamma6[ src[ j + k ] ];

				/* Add in error, maybe limiting it */
				err = row0[ j + k + 6 ];

				if ( limit == 1 ) /* To half of SRGB range */
				{
					err = err < -0.5 ? -0.5 :
						err > 0.5 ? 0.5 : err;
				}
				else if (limit == 2) /* To 1/4, with damping */
				{
					err = err < -0.1 ? (err < -0.4 ?
						-0.25 : 0.5 * err - 0.05) :
						err > 0.1 ? (err > 0.4 ?
						0.25 : 0.5 * err + 0.05) : err;
				}

				color1[k] = color0[k] + err;

				/* Limit result to palette gamut */
				if ( color1[k] < gamut[k] )
				{
					color1[k] = gamut[k];
				}

				if ( color1[k] > gamut[k + 3] )
				{
					color1[k] = gamut[k + 3];
				}
			}

			/* Output best colour */
			col1 = lookup_srgb ( ctp, color1 );
			dest[0] = col1;
			dest += dj;

			if ( ! dither )
			{
				continue;
			}

			/* Evaluate new error */
			tc1[0] = gamma6[pal->col[col1][0]];
			tc1[1] = gamma6[pal->col[col1][1]];
			tc1[2] = gamma6[pal->col[col1][2]];

			if ( selc ) /* Selective error damping */
			{
				col0 = lookup_srgb ( ctp, color0 );
				tc0[0] = gamma6[ pal->col[col0][0] ];
				tc0[1] = gamma6[ pal->col[col0][1] ];
				tc0[2] = gamma6[ pal->col[col0][2] ];

				/* Split error the obvious way */
				if ( ! ( selc & 1 ) && ( col0 == col1 ) )
				{
					color1[0] = (color1[0] - color0[0]) *
						emult + color0[0] - tc0[0];
					color1[1] = (color1[1] - color0[1]) *
						emult + color0[1] - tc0[1];
					color1[2] = (color1[2] - color0[2]) *
						emult + color0[2] - tc0[2];
				}
				/* Weigh component errors separately */
				else if ( selc < 3 )
				{
					for ( k = 0; k < 3; k++ )
					{
						intd = fabs ( color0[k] -
							tc0[k] );
						extd = fabs ( color0[k] -
							color1[k]);

						if ( intd + extd == 0.0 )
						{
							err = 1.0;
						}
						else
						{
							err = ( intd + emult *
								extd ) / ( intd
								+ extd );
						}

						color1[k] = err * ( color1[k] -
							tc1[k] );
					}
				}
				/* Weigh errors by vector length */
				else
				{
					intd = sqrt ( ( color0[0] - tc0[0] ) *
						( color0[0] - tc0[0] ) +
						( color0[1] - tc0[1] ) *
						( color0[1] - tc0[1] ) +
						( color0[2] - tc0[2] ) *
						( color0[2] - tc0[2] ) );

					extd = sqrt ( ( color0[0] - color1[0] )
						* ( color0[0] - color1[0] ) +
						( color0[1] - color1[1] ) *
						( color0[1] - color1[1] ) +
						( color0[2] - color1[2] ) *
						( color0[2] - color1[2] ) );

					if ( intd + extd == 0.0 )
					{
						err = 1.0;
					}
					else
					{
						err = ( intd + emult * extd ) /
							( intd + extd );
					}

					color1[0] = err * (color1[0] - tc1[0]);
					color1[1] = err * (color1[1] - tc1[1]);
					color1[2] = err * (color1[2] - tc1[2]);
				}
			}
			else /* Indiscriminate error damping */
			{
				color1[0] = ( color1[0] - tc1[0] ) * emult;
				color1[1] = ( color1[1] - tc1[1] ) * emult;
				color1[2] = ( color1[2] - tc1[2] ) * emult;
			}

			/* Distribute the error */
			color1[0] *= fdiv;
			color1[1] *= fdiv;
			color1[2] *= fdiv;

			for ( k = 0; k < 5; k++ )
			{
				kk = j + ( k - 2 ) * dj * 3 + 6;

				for ( l = 0; l < 3; l++ , kk++ )
				{
					row0[kk] += color1[l] * dither[k];
					row1[kk] += color1[l] * dither[k + 5];
					row2[kk] += color1[l] * dither[k + 10];
				}
			}
		}

		tmp = row0; row0 = row1; row1 = row2; row2 = tmp;
	}

	free ( ddata1 );
	free ( ddata2 );

	return 0;
}


/*
Dumb (but fast) Floyd-Steinberg dithering in RGB space, loosely based on
Dennis Lee's dithering implementation from dl3quant.c, in turn based on
dithering code from the IJG's jpeg library - WJ
*/

static int mem_dumb_dither (
	mtImage		* inew,
	mtImage		* iold,
	int		dither
	)
{
	unsigned short	cols[32768],
			sqrtb[512],
			* sqr;
	short		limtb[512],
			* lim,
			fr[3] = { 0, 0, 0 },
			* rows = NULL,
			* row0 = fr,
			* row1 = fr;
	unsigned char	clamp[768],
			* src,
			* dest,
			* old,
			* new;
	int		i,
			j,
			k,
			j0,
			dj,
			dj3,
			r,
			g,
			b,
			rlen,
			serpent = 2,
			width,
			height,
			ncols
			;
	mtPalette	* pal;


	width = inew->width;
	height = inew->height;
	ncols = inew->palette.size;
	pal = &inew->palette;

	/* Allocate working space */
	rlen = ( width + 2 ) * 3;

	if ( dither )
	{
		rows = calloc ( rlen * 2, sizeof ( short ) );

		if ( ! rows )
		{
			return 1;
		}

		serpent = 0;
	}

	/* Color cache, squares table, clamp table */
	memset ( cols, 0, sizeof ( cols ) );
	sqr = sqrtb + 256;

	for ( i = -255; i < 256; i++ )
	{
		sqr[i] = i * i;
	}

	memset ( clamp, 0, 256 );
	memset ( clamp + 512, 255, 256 );

	for ( i = 0; i < 256; i++ )
	{
		clamp[i + 256] = i;
	}

	/* Error limiter table, Dennis Lee's way */
#define ERR_LIM 20

	lim = limtb + 256;

	for ( i = 0; i < ERR_LIM; i++ )
	{
		lim[i] = i;
		lim[-i] = -i;
	}

	for ( ; i < 256; i++ )
	{
		lim[i] = ERR_LIM;
		lim[-i] = -ERR_LIM;
	}

#undef ERR_LIM

	old = iold->channel[ MTPIXEL_CHANNEL_IMAGE ];
	new = inew->channel[ MTPIXEL_CHANNEL_IMAGE ];

	/* Process image */
	for ( i = 0; i < height; i++ )
	{
		src = old + i * width * 3;
		dest = new + i * width;

		if ( serpent ^= 1 )
		{
			j0 = 0; dj = 1;
		}
		else
		{
			j0 = ( width - 1 ) * 3; dj = -1;
			dest += width - 1;
		}

		if ( dither )
		{
			row0 = row1 = rows + 3;
			*( serpent ? &row1 : &row0 ) += rlen;
			memset ( row1 - 3, 0, rlen * sizeof ( short ) );
			src += j0;
			row0 += j0;
			row1 += j0;
		}

		dj3 = dj * 3;

		for ( j = 0; j < width; j++ , src += dj3 , dest += dj )
		{
			r = clamp[src[0] + ( ( row0[0] + 0x1008 ) >> 4)];
			g = clamp[src[1] + ( ( row0[1] + 0x1008 ) >> 4)];
			b = clamp[src[2] + ( ( row0[2] + 0x1008 ) >> 4)];

			k =	( ( r & 0xF8 ) << 7) +
				( ( g & 0xF8 ) << 2) +
				( b >> 3 );

			if ( ! cols[k] ) /* Find nearest color in RGB */
			{
				int	i,
					j,
					n = 0,
					l = 1000000;

/*
Searching for color nearest to first color in cell, instead of to cell
itself, looks like a bug, but works like a feature - makes FS dither less
prone to patterning. This trick I learned from Dennis Lee's code - WJ
*/
				for ( i = 0; i < ncols; i++ )
				{
					j = sqr[r - pal->col[i][0]] +
						sqr[g - pal->col[i][1]] +
						sqr[b - pal->col[i][2]];

					if ( j >= l )
					{
						continue;
					}

					l = j;
					n = i;
				}

				cols[k] = n + 1;
			}

			dest[0] = k = cols[k] - 1;

			if ( ! dither )
			{
				continue;
			}

			r = lim[r - pal->col[k][0]];
			g = lim[g - pal->col[k][1]];
			b = lim[b - pal->col[k][2]];

			k = r + r;
			row1[0 + dj3] += r;
			row1[0 - dj3] += (r += k);
			row1[0 + 0  ] += (r += k);
			row0[0 + dj3] += r + k;

			k = g + g;
			row1[1 + dj3] += g;
			row1[1 - dj3] += (g += k);
			row1[1 + 0  ] += (g += k);
			row0[1 + dj3] += g + k;

			k = b + b;
			row1[2 + dj3] += b;
			row1[2 - dj3] += (b += k);
			row1[2 + 0  ] += (b += k);
			row0[2 + dj3] += b + k;

			row0 += dj3;
			row1 += dj3;
		}
	}

	free ( rows );

	return 0;
}

static void mem_quantize (
	mtImage		* new,
	mtImage		* old,
	int		type		// 1 = flat 2 = dither 3 = scatter
	)
{
	unsigned char	* new_img,
			* old_mem_image,
			pcol[3];
	int		i,
			j,
			k,
			w,
			h,
			closest[3][2],
			target_cols;
	mtPalette	* pal;


	new_img = new->channel[ MTPIXEL_CHANNEL_IMAGE ];
	old_mem_image = old->channel[ MTPIXEL_CHANNEL_IMAGE ];
	target_cols = old->palette.size;
	pal = &new->palette;
	w = old->width;
	h = old->height;

	for ( j = 0; j < h; j++ )		// Convert RGB to indexed
	{
		for ( i = 0; i < w; i++ )
		{
			pcol[0] = old_mem_image[ 3 * ( i + w * j ) ];
			pcol[1] = old_mem_image[ 1 + 3 * ( i + w * j ) ];
			pcol[2] = old_mem_image[ 2 + 3 * ( i + w * j ) ];

			closest[0][0] = 0; // 1st Closest palette item to pixel
			closest[1][0] = 100000000;
			closest[0][1] = 0; // 2nd Closest palette item to pixel
			closest[1][1] = 100000000;

			for ( k = 0; k < target_cols; k++ )
			{
				closest[2][0] = abs ( pcol[0] - pal->col[k][0] )
					+ abs ( pcol[1] - pal->col[k][1] ) +
					abs ( pcol[2] - pal->col[k][2] );

				if ( closest[2][0] < closest[1][0] )
				{
					closest[0][1] = closest[0][0];
					closest[1][1] = closest[1][0];
					closest[0][0] = k;
					closest[1][0] = closest[2][0];
				}
				else
				{
					if ( closest[2][0] < closest[1][1] )
					{
						closest[0][1] = k;
						closest[1][1] = closest[2][0];
					}
				}
			}
			if ( type == 1 )
			{
				// Flat conversion
				k = closest[0][0];
			}
			else
			{
				if ( closest[1][1] == 100000000 )
				{
					closest[1][0] = 0;
				}

				if ( closest[1][0] == 0 )
				{
					k = closest[0][0];
				}
				else
				{
				  if ( type == 2 )	// Dithered
				  {
				  	if ( closest[1][1] * .67 <
						( closest[1][1] -
							closest[1][0] )
						)
					{
						k = closest[0][0];
					}
					else
					{
					  	if ( closest[0][0] >
							closest[0][1]
							)
						{
							k = closest[0][
								( i + j ) % 2 ];
						}
						else
						{
							k = closest[0][
								( i + j + 1 ) %
								2 ];
						}
					}
				  }

				  if ( type == 3 )	// Scattered
				  {
				    if ( ( rand () % ( closest[1][1] +
				    	closest[1][0])) <= closest[1][1]
					)
					{
						k = closest[0][0];
					}
					else
					{
				    		k = closest[0][1];
					}
				  }
				}
			}

			*new_img++ = k;
		}
	}
}

mtImage * mtpixel_image_to_indexed (
	mtImage		* image,
	mtDitherType	type,
	mtDitherCS	cspace,
	mtDitherDiff	diff,
	mtDitherBL	bleed,
	int		serp,
	int		errp,
	mtDitherSep	sep,
	int		fullerr
	)
{
	/* Dithering filters */
	static short	fs_dither[16] =		/* Floyd-Steinberg dither */
			{ 16,  0, 0, 0, 7, 0,  0, 3, 5, 1, 0,  0, 0, 0, 0, 0 },
			s_dither[16] =		/* Stucki dither */
			{ 42,  0, 0, 0, 8, 4,  2, 4, 8, 4, 2,  1, 2, 4, 2, 1 };
	mtImage		* new;


	if (	! image					||
		image->palette.size < 2			||
		type < MTPIXEL_DITHER_TYPE_NONE		||
		type >= MTPIXEL_DITHER_TYPE_TOTAL	||
		cspace < MTPIXEL_DITHER_CS_NONE		||
		cspace >= MTPIXEL_DITHER_CS_TOTAL	||
		diff < MTPIXEL_DITHER_DIFF_NONE		||
		diff >= MTPIXEL_DITHER_DIFF_TOTAL	||
		bleed < MTPIXEL_DITHER_BL_NONE		||
		bleed >= MTPIXEL_DITHER_BL_TOTAL	||
		sep < MTPIXEL_DITHER_SEP_NONE		||
		sep >= MTPIXEL_DITHER_SEP_TOTAL		||
		serp < 0				||
		serp > 1				||
		errp < 0				||
		errp > 100				||
		fullerr < 0				||
		fullerr > 1
		)
	{
		return NULL;
	}

	new = new_rgb_indexed ( image, MTPIXEL_IMAGE_INDEXED, MTPIXEL_IMAGE_RGB
		);

	if ( ! new )
	{
		return NULL;
	}

	if (	type == 0	||
		cspace == 0	||
		diff == 0	||
		bleed == 0	||
		sep == 0
		)
	{
		unsigned char	* src,
				* end,
				* dest;
		int		i,
				best_i,
				best_err,
				err,
				r,
				g,
				b;


		// Force an ultra simple direct conversion
		src = image->channel[ MTPIXEL_CHANNEL_IMAGE ];
		dest = new->channel[ MTPIXEL_CHANNEL_IMAGE ];
		end = src + image->width * image->height * 3;

		while ( src < end )
		{
			for ( best_i = 0, best_err = 1 << 30, i = 0;
				i < new->palette.size; i++ )
			{
				r = new->palette.col[i][0] - src[0];
				g = new->palette.col[i][1] - src[1];
				b = new->palette.col[i][2] - src[2];
				err = r * r + g * g + b * b;

				if ( err < best_err )
				{
					best_i = i;
					best_err = err;

					if ( err == 0 )
					{
						break;
					}
				}
			}

			*dest++ = best_i;
			src += 3;
		}
	}
	else switch ( type )		// Dither pixels as requested
	{
	case MTPIXEL_DITHER_TYPE_NONE:
	case MTPIXEL_DITHER_TYPE_FLOYD1:
	case MTPIXEL_DITHER_TYPE_STUCKI:
		mem_dither ( new, image,
			type == MTPIXEL_DITHER_TYPE_NONE ? NULL :
				type == MTPIXEL_DITHER_TYPE_FLOYD1 ? fs_dither :
				s_dither,
			cspace, diff, bleed, sep,
			serp, fullerr, ( (double) errp ) / 100
			);
		break;

	case MTPIXEL_DITHER_TYPE_FLOYD2:
		mem_dumb_dither ( new, image, 1);
		break;

	case MTPIXEL_DITHER_TYPE_DITHER:
		mem_quantize ( new, image, 2 );
		break;

	case MTPIXEL_DITHER_TYPE_SCATTER:
		mem_quantize ( new, image, 3 );
		break;

	default:
		break;
	}

	return new;
}



///	Code for scaling contributed by Dmitry Groshev, January 2006

typedef struct
{
	int		idx;
	float		k;
} fstep;



static double Cubic (
	double		x,
	double		A
	)
{
	if ( x < -1.5 )
	{
		return 0.0;
	}
	else if ( x < -0.5 )
	{
		return ( A * ( -1.0 / 8.0 ) * ( ( ( x * 8.0 + 28.0 ) * x +
			30.0 ) * x + 9.0 ) );
	}
	else if ( x < 0.5 )
	{
		return ( 0.5 * ( ( ( -4.0 * A - 4.0 ) * x * x + A + 3.0 ) *
			x + 1.0 ) );
	}
	else if ( x < 1.5 )
	{
		return ( A * ( -1.0 / 8.0 ) * ( ( ( x * 8.0 - 28.0 ) * x +
			30.0 ) * x - 9.0 ) + 1.0 );
	}
	else
	{
		return 1.0;
	}
}

static double BH1 (
	double		x
	)
{
	if ( x < 1e-7 )
	{
		return 1.0;
	}

	return ( ( sin ( M_PI * x ) / ( M_PI * x ) ) * ( 0.42323 +
		0.49755 * cos ( x * ( M_PI * 2.0 / 6.0 ) ) +
		0.07922 * cos ( x * ( M_PI * 4.0 / 6.0 ) ) ) );
}

static double BH (
	double		x
	)
{
	double		y = 0.0,
			xx = fabs ( x );


	if ( xx < 2.5 )
	{
		y = BH1 ( xx + 0.5 );

		if ( xx < 1.5 )
		{
			y += BH1 ( xx + 1.5 );
		}

		if ( xx < 0.5 )
		{
			y += BH1 ( xx + 2.5 );
		}
	}

	return ( x > 0.0 ? 1.0 - y : y );
}

static fstep * make_filter (
	int		l0,
	int		l1,
	int		type,
	int		sharp
	)
{
	fstep		* res,
			* buf;
	double		Aarray[4] = { -0.5, -2.0 / 3.0, -0.75, -1.0 },
			x,
			y,
			basept,
			fwidth,
			delta,
			scale = (double)l1 / (double)l0,
			A = 0.0,
			kk = 1.0,
			sum;
	int		pic_tile = 0, /* Allow to enable tiling mode later */
			pic_skip = 0, /* Allow to enable skip mode later */
			i,
			j,
			k,
			ix;


	/* To correct scale-shift */
	delta = 0.5 / scale - 0.5;

	/* Untransformed bilinear is useless for reduction */
	if ( type == MTPIXEL_SCALE_BILINEAR )
	{
		sharp = 1;
	}

	if ( scale < 1.0 )
	{
		kk = scale;
	}
	else
	{
		sharp = 0;
	}

	switch ( type )
	{
	case MTPIXEL_SCALE_BILINEAR:
		fwidth = 2.0;
		break;

	case MTPIXEL_SCALE_BICUBIC:
	case MTPIXEL_SCALE_BICUBIC_EDGED:
	case MTPIXEL_SCALE_BICUBIC_BETTER:
	case MTPIXEL_SCALE_BICUBIC_SHARP:
		fwidth = 4.0;
		A = Aarray[type - 2];
		break;

	case MTPIXEL_SCALE_BLACKMAN_HARRIS:
		fwidth = 6.0;		/* Blackman-Harris windowed sinc */
		break;

	default:			/* Bug */
		fwidth = 0.0;
		break;
	}

	if ( sharp )
	{
		fwidth += scale - 1.0;
	}

	fwidth /= kk;

	i = (int)floor ( fwidth ) + 2;

	res = buf = calloc ( l1 * ( i + 1 ), sizeof ( fstep ) );
	if ( ! res )
	{
		return NULL;
	}

	fwidth *= 0.5;
	type = type * 2 + ( sharp ? 1 : 0 );

	for ( i = 0; i < l1; i++ )
	{
		basept = (double)i / scale + delta;
		k = (int)floor ( basept + fwidth );

		for ( j = (int)ceil ( basept - fwidth ); j <= k; j++ )
		{
			ix = j;

			if ( ( j < 0 ) || ( j >= l0 ) )
			{
				if ( pic_skip )
				{
					continue;
				}

				if ( pic_tile )
				{
					if ( ix < 0 )
					{
						ix = l0 - ( -ix % l0 );
					}

					ix %= l0;
				}
				else if ( l0 == 1 )
				{
					ix = 0;
				}
				else
				{
					ix = abs ( ix ) % ( l0 + l0 - 2 );
					if ( ix >= l0 )
					{
						ix = l0 + l0 - 2 - ix;
					}
				}
			}

			buf->idx = ix;
			x = fabs ( ( (double)j - basept ) * kk );
			y = 0;

			switch ( type )
			{
			case ( MTPIXEL_SCALE_BILINEAR * 2 ):
				y = 1.0 - x;
				break;

			case ( MTPIXEL_SCALE_BILINEAR * 2 + 1 ):
				if ( x <= 0.5 - scale / 2.0 )
				{
					y = 1.0;
				}
				else
				{
					y = 0.5 - (x - 0.5) / scale;
				}
				break;

			case ( MTPIXEL_SCALE_BICUBIC * 2 ):
			case ( MTPIXEL_SCALE_BICUBIC_EDGED * 2 ):
			case ( MTPIXEL_SCALE_BICUBIC_BETTER * 2 ):
			case ( MTPIXEL_SCALE_BICUBIC_SHARP * 2 ):
				if ( x < 1.0 )
				{
					y = ( ( A + 2.0 ) * x - ( A + 3 ) ) *
						x * x + 1.0;
				}
				else
				{
					y = A * ( ( ( x - 5.0 ) * x + 8.0 ) *
						x - 4.0 );
				}
				break;

			case ( MTPIXEL_SCALE_BICUBIC * 2 + 1 ):
			case ( MTPIXEL_SCALE_BICUBIC_EDGED * 2 + 1 ):
			case ( MTPIXEL_SCALE_BICUBIC_BETTER * 2 + 1 ):
			case ( MTPIXEL_SCALE_BICUBIC_SHARP * 2 + 1 ):
				y = Cubic ( x + scale * 0.5, A ) -
					Cubic ( x - scale * 0.5, A );
				break;

			case ( MTPIXEL_SCALE_BLACKMAN_HARRIS * 2 ):
				y = BH1 ( x );
				break;

			case ( MTPIXEL_SCALE_BLACKMAN_HARRIS * 2 + 1 ):
				y = BH ( x + scale * 0.5 ) -
					BH ( x - scale * 0.5 );
				break;

			default:	/* Bug */
				break;
			}

			buf->k = y * kk;

			if ( buf->k != 0.0 )
			{
				buf ++;
			}
		}

		buf->idx = -1;
		buf ++;
	}

	( buf - 1 )->idx = -2;

	/* Normalization pass */
	sum = 0.0;

	for ( buf = res, i = 0; ; i++ )
	{
		if ( buf[i].idx >= 0 )
		{
			sum += buf[i].k;
		}
		else
		{
			if ( ( sum != 0.0 ) && ( sum != 1.0 ) )
			{
				sum = 1.0 / sum;

				for ( j = 0; j < i; j++ )
				{
					buf[j].k *= sum;
				}
			}

			if ( buf[i].idx < -1 )
			{
				break;
			}

			sum = 0.0;
			buf += i + 1;
			i = -1;
		}
	}

	return res;
}



typedef struct
{
	char		* workarea;
	fstep		* hfilter,
			* vfilter;
} scale_context;



static void clear_scale (
	scale_context	* ctx
	)
{
	free ( ctx->workarea );
	free ( ctx->hfilter );
	free ( ctx->vfilter );
}

static int prepare_scale (
	mtImage		* image,
	scale_context	* ctx,
	int		ow,
	int		oh,
	int		nw,
	int		nh,
	int		type,
	int		sharp
	)
{
	ctx->workarea = NULL;
	ctx->hfilter = ctx->vfilter = NULL;

	if (	type == MTPIXEL_SCALE_NEAREST_NEIGHBOUR ||
		image->type == MTPIXEL_IMAGE_INDEXED
		)
	{
		return 1;
	}

	ctx->workarea = malloc ( ( 7 * ow + 1 ) * sizeof ( double ) );
	ctx->hfilter = make_filter ( ow, nw, type, sharp );
	ctx->vfilter = make_filter ( oh, nh, type, sharp );

	if ( ! ctx->workarea || ! ctx->hfilter || ! ctx->vfilter )
	{
		clear_scale ( ctx );
		return 0;
	}

	return 1;
}

static void do_scale (
	scale_context	* ctx,
	mtImage		* new_img,
	mtImage		* old_img,
	int		ow,
	int		oh,
	int		nw,
	int		nh,
	int		gcor
	)
{
	unsigned char	* src,
			* img,
			* imga;
	fstep		* tmp = NULL,
			* tmpx,
			* tmpy,
			* tmpp;
	double		* wrk,
			* wrk2,
			* wrka,
			* work_area;
	double		sum,
			sum1,
			sum2,
			kk,
			mult;
	int		i,
			j,
			cc,
			bpp,
			gc,
			mask_rgba;


	work_area = ALIGNTO ( ctx->workarea, double );

	if (	new_img->channel[ MTPIXEL_CHANNEL_IMAGE ] &&
		new_img->channel[ MTPIXEL_CHANNEL_ALPHA ]
		)
	{
		mask_rgba = MTPIXEL_CHANNEL_MASK_IMAGE |
			MTPIXEL_CHANNEL_MASK_ALPHA;
	}
	else
	{
		mask_rgba = 0;
	}

	/* For each destination line */
	tmpy = ctx->vfilter;

	for ( i = 0; i < nh; i++, tmpy ++ )
	{
		/* Process regular channels */
		for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc++ )
		{
			if ( ! new_img->channel[cc] )
			{
				continue;
			}

			/* Do RGBA separately */
			if ( mask_rgba & ( 1 << cc ) )
			{
				continue;
			}

			bpp = cc == MTPIXEL_CHANNEL_IMAGE ? 3 : 1;
			gc = cc == MTPIXEL_CHANNEL_IMAGE ? gcor : 0;
			memset ( work_area, 0, ow * bpp * sizeof ( double ) );
			src = old_img->channel[cc];

			/* Build one vertically-scaled row */
			for ( tmp = tmpy; tmp->idx >= 0; tmp ++ )
			{
				img = src + tmp->idx * ow * bpp;
				wrk = work_area;
				kk = tmp->k;

				if ( gc )	/* Gamma-correct */
				{
					for ( j = 0; j < ow * bpp; j++ )
					{
						*wrk++ += _gamma256 ( *img++ ) *
							kk;
					}
				}
				else		/* Leave as is */
				{
					for ( j = 0; j < ow * bpp; j++ )
					{
						*wrk++ += *img++ * kk;
					}
				}
			}

			/* Scale it horizontally */
			img = new_img->channel[cc] + i * nw * bpp;
			sum = sum1 = sum2 = 0.0;

			for ( tmpx = ctx->hfilter; ; tmpx ++ )
			{
				if ( tmpx->idx >= 0 )
				{
					wrk = work_area + tmpx->idx * bpp;
					kk = tmpx->k;
					sum += wrk[0] * kk;

					if ( bpp == 1 )
					{
						continue;
					}

					sum1 += wrk[1] * kk;
					sum2 += wrk[2] * kk;

					continue;
				}

				if ( gc )	/* Reverse gamma correction */
				{
					*img++ = sum < 0.0 ? 0 : sum > 1.0 ?
						0xFF : _ungamma256 ( sum );

					*img++ = sum1 < 0.0 ? 0 : sum1 > 1.0 ?
						0xFF : _ungamma256 ( sum1 );

					*img++ = sum2 < 0.0 ? 0 : sum2 > 1.0 ?
						0xFF : _ungamma256 ( sum2 );

					sum = sum1 = sum2 = 0.0;

					if ( tmpx->idx < -1 )
					{
						break;
					}

					continue;
				}

				if ( bpp > 1 )
				{
					j = (int)rint ( sum1 );
					img[1] = j < 0 ? 0 : j > 0xFF ? 0xFF :
						j;

					j = (int)rint ( sum2 );
					img[2] = j < 0 ? 0 : j > 0xFF ? 0xFF :
						j;

					sum1 = sum2 = 0.0;
				}

				j = (int)rint ( sum );
				img[0] = j < 0 ? 0 : j > 0xFF ? 0xFF : j;
				sum = 0.0;
				img += bpp;

				if ( tmpx->idx < -1 )
				{
					break;
				}
			}
		}

		/* Process RGBA */
		if ( mask_rgba != 0 )
		{
			memset ( work_area, 0, ow * 7 * sizeof ( double ) );

			/* Build one vertically-scaled row - with & w/o alpha */
			for ( tmp = tmpy; tmp->idx >= 0; tmp ++ )
			{
				img = old_img->channel[MTPIXEL_CHANNEL_IMAGE] +
					tmp->idx * ow * 3;
				imga = old_img->channel[MTPIXEL_CHANNEL_ALPHA] +
					tmp->idx * ow;

				wrk = work_area + ow;
				wrk2 = work_area + 4 * ow;

				if ( gcor )	/* Gamma-correct */
				{
					for ( j = 0; j < ow; j++ )
					{
						kk = imga[j] * tmp->k;
						work_area[j] += kk;

						wrk[0] += _gamma256 ( img[0] ) *
							tmp->k;
						wrk2[0] += _gamma256 ( img[0] )
							* kk;
						wrk[1] += _gamma256 ( img[1] ) *
							tmp->k;
						wrk2[1] += _gamma256 ( img[1] )
							* kk;
						wrk[2] += _gamma256 ( img[2] ) *
							tmp->k;
						wrk2[2] += _gamma256 ( img[2] )
							* kk;

						wrk += 3;
						wrk2 += 3;
						img += 3;
					}
				}
				else	/* Leave as is */
				{
					for ( j = 0; j < ow; j++ )
					{
						kk = imga[j] * tmp->k;
						work_area[j] += kk;

						wrk[0] += img[0] * tmp->k;
						wrk2[0] += img[0] * kk;
						wrk[1] += img[1] * tmp->k;
						wrk2[1] += img[1] * kk;
						wrk[2] += img[2] * tmp->k;
						wrk2[2] += img[2] * kk;

						wrk += 3;
						wrk2 += 3;
						img += 3;
					}
				}
			}

			/* Scale it horizontally */
			img = new_img->channel[ MTPIXEL_CHANNEL_IMAGE ] +
				i * nw * 3;
			imga = new_img->channel[ MTPIXEL_CHANNEL_ALPHA ] +
				i * nw;

			for (	tmpp = tmpx = ctx->hfilter;
				tmpp->idx >= -1;
				tmpx = tmpp + 1
				)
			{
				sum = 0.0;

				for ( tmpp = tmpx; tmpp->idx >= 0; tmpp++ )
				{
					sum += work_area[ tmpp->idx ] * tmpp->k;
				}

				j = (int)rint ( sum );
				*imga = j < 0 ? 0 : j > 0xFF ? 0xFF : j;
				wrk = work_area + ow;
				mult = 1.0;

				if ( *imga++ )
				{
					wrk = work_area + 4 * ow;
					mult /= sum;
				}

				sum = sum1 = sum2 = 0.0;

				for ( tmpp = tmpx; tmpp->idx >= 0; tmpp ++ )
				{
					wrka = wrk + tmpp->idx * 3;
					kk = tmpp->k;
					sum += wrka[0] * kk;
					sum1 += wrka[1] * kk;
					sum2 += wrka[2] * kk;
				}

				sum *= mult; sum1 *= mult; sum2 *= mult;

				if ( gcor )	/* Reverse gamma correction */
				{
					*img++ = sum < 0.0 ? 0 : sum > 1.0 ?
						0xFF : _ungamma256 ( sum );
					*img++ = sum1 < 0.0 ? 0 : sum1 > 1.0 ?
						0xFF : _ungamma256 ( sum1 );
					*img++ = sum2 < 0.0 ? 0 : sum2 > 1.0 ?
						0xFF : _ungamma256 ( sum2 );
				}
				else	/* Simply round to nearest */
				{
					j = (int)rint ( sum );
					*img++ = j < 0 ? 0 : j > 0xFF ? 0xFF :
						j;
					j = (int)rint ( sum1 );
					*img++ = j < 0 ? 0 : j > 0xFF ? 0xFF :
						j;
					j = (int)rint ( sum2 );
					*img++ = j < 0 ? 0 : j > 0xFF ? 0xFF :
						j;
				}
			}
		}

		if ( tmp->idx < -1 )
		{
			break;
		}

		tmpy = tmp;
	}

	clear_scale ( ctx );
}

static void do_scale_internal (
	scale_context	* ctx,
	mtImage		* new_img,
	mtImage		* old_img,
	int		type,
	int		gcor,
	int		ow,
	int		oh,
	int		nw,
	int		nh
	)
{
	unsigned char	* src,
			* dest;
	int		i,
			j,
			oi,
			oj,
			cc,
			bpp;
	double		scalex,
			scaley,
			deltax,
			deltay;


	if (	type != MTPIXEL_SCALE_NEAREST_NEIGHBOUR &&
		new_img->type == MTPIXEL_IMAGE_RGB
		)
	{
		do_scale ( ctx, new_img, old_img, ow, oh, nw, nh, gcor );
	}
	else
	{
		scalex = (double)ow / (double)nw;
		scaley = (double)oh / (double)nh;
		deltax = 0.5 * scalex - 0.5;
		deltay = 0.5 * scaley - 0.5;

		for ( j = 0; j < nh; j++ )
		{
			for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc++ )
			{
				if ( ! new_img->channel[cc] )
				{
					continue;
				}

				if ( cc == MTPIXEL_CHANNEL_IMAGE )
				{
					bpp = new_img->type;
				}
				else
				{
					bpp = 1;
				}

				dest = new_img->channel[cc] + nw * j * bpp;
				oj = rint ( scaley * j + deltay );
				src = old_img->channel[cc] + ow * oj * bpp;

				for ( i = 0; i < nw; i++ )
				{
					oi = (int)rint ( scalex * i + deltax ) *
						bpp;

					*dest++ = src[oi];

					if ( bpp == 1 )
					{
						continue;
					}

					*dest++ = src[oi + 1];
					*dest++ = src[oi + 2];
				}
			}
		}
	}
}

static mtImage * new_scale_rotate (
	mtImage		* image,
	int		nw,
	int		nh
	)
{
	mtImage		* new;


	if ( ! image || nw < 1 || nh < 1 )
	{
		return NULL;
	}

	new = mtpixel_image_new ( nw, nh, image->type,
		mtpixel_channel_get_mask ( image ) );

	if ( ! new )
	{
		return NULL;
	}

	// Copy palette
	mtpixel_palette_copy ( &new->palette, &image->palette );

	return new;
}

static mtImage * mt_scale (
	mtImage		* image,
	mtScaleType	type,
	int		width,
	int		height
	)
{
	mtImage		* new = NULL;
	unsigned char	* new_image,
			* mem_image;
	int		ow,
			oh,
			nw,
			nh,
			i,
			j,
			oi,
			oj,
			k,
			oi1,
			oj1,
			oi2,
			oj2,
			a[3] = { 0 },
			b[3] = { 0 },
			c[3] = { 0 },
			d[3] = { 0 },
			e[3] = { 0 },
			xlen,
			ylen,
			avi,
			avj;
	float		foi,
			foj,
			foi1,
			foj1,
			foi2,
			foj2,
			ab,
			ac,
			xf1,
			xf2,
			yf1,
			yf2,
			av_tot,
			pf1,
			pf2,
			ff[3];


	ow = image->width;
	oh = image->height;
	nw = width;
	nh = height;
	mem_image = image->channel[ MTPIXEL_CHANNEL_IMAGE ];

	if ( image->type == MTPIXEL_IMAGE_INDEXED )
	{
		new = mtpixel_image_new_indexed ( nw, nh );

		if ( ! new )
		{
			goto finish;
		}

		new_image = new->channel[ MTPIXEL_CHANNEL_IMAGE ];

		for ( j = 0; j < nh; j++ )
		{
			oj = oh * ( (float) j ) / nh;

			for ( i = 0; i < nw; i++ )
			{
				oi = ow * ( (float) i ) / nw;
				new_image[ i + nw * j ] =
					mem_image[ oi + ow * oj ];
			}
		}
	}
	else if ( image->type == MTPIXEL_IMAGE_RGB )
	{
		new = mtpixel_image_new_rgb ( width, height );

		if ( ! new )
		{
			goto finish;
		}

		new_image = new->channel[ MTPIXEL_CHANNEL_IMAGE ];

		if ( type == MTPIXEL_SCALE_MT_SMOOTH )
		{
		    for ( j = 0; j < nh; j++ )
		    {
			foj = ( (float) oh - 1 ) * ( (float) j ) / ( nh - 1 );

			for ( i = 0; i < nw; i++ )
			{
				foi = ( (float) ow - 1) * ( (float) i ) /
					( nw - 1 );

				oi  = foi;
				oj  = foj;

				if ( nw >= ow )	// Expanding X axis
				{
					oi2 = foi + 0.9999;
					ab = foi - ( (int) foi );

					for ( k = 0; k < 3; k++ )
					{
						a[k] = mem_image[ k + 3 *
							( oi + ow * oj ) ];
						b[k] = mem_image[ k + 3 *
							( oi2 + ow * oj ) ];
					}
				}
				else
				{
					ab = 0.0;
				}

				if ( nh >= oh )		// Expanding Y axis
				{
					oi2 = foi + 0.9999;
					oj2 = foj + 0.9999;
					ac = foj - ( (int) foj );

					for ( k = 0; k < 3; k++ )
					{
						c[k] = mem_image[ k + 3 * ( oi
							+ ow * oj2 ) ];
						d[k] = mem_image[ k + 3 * ( oi2
							+ ow * oj2 ) ];
					}
				}
				else
				{
					ac = 0.0;
				}

				if ( nw < ow || nh < oh )	// Shrinking
				{
				 av_tot = 0;
				 ff[0] = 0; ff[1] = 0; ff[2] = 0;

				 foi1 = ( (float) ow ) * ( (float) i ) / nw;
				 foj1 = ( (float) oh ) * ( (float) j ) / nh;
				 foi2 = ( (float) ow ) * ( (float) i + 1 ) / nw;
				 foj2 = ( (float) oh ) * ( (float) j + 1 ) / nh;
				 oi1 = foi1;
				 oj1 = foj1;
				 oi2 = foi2;
				 oj2 = foj2;

				 if ( ( foi2 - oi2 ) < 0.00001 )
				 {
				 	oi2 = oi2 - 1;
				 }

				 if ( ( foj2 - oj2 ) < 0.00001 )
				 {
				 	oj2 = oj2 - 1;
				 }

				 // X Fraction of 1st pixel to sample from
				 // original
				 xf1 = 1 - ( foi1 - oi1 );

				 // Y Fraction of 1st pixel to sample from
				 // original
				 yf1 = 1 - ( foj1 - oj1 );

				 // X Fraction of last pixel to sample from
				 // original
				 xf2 = foi2 - oi2;

				 // Y Fraction of last pixel to sample from
				 // original
				 yf2 = foj2 - oj2;

				 // X pixels to sample from original
				 xlen = oi2 - oi1;

				 // Y pixels to sample from original
				 ylen = oj2 - oj1;

				 if ( xlen == 0 ) xf1 = 1.0;
				 if ( ylen == 0 ) yf1 = 1.0;

				 for ( avj = 0; avj <= ylen; avj ++ )
				 {
				   pf1 = 1.0;

				   if ( avj == 0 )
				   {
				   	pf1 = yf1;
				   }
				   else if ( avj == ylen )
				   {
				   	pf1 = yf2;
				   }

				   for ( avi = 0; avi <= xlen; avi ++ )
				   {
				      pf2 = pf1;

				      if ( avi == 0 )
				      {
				      	pf2 = pf1 * xf1;
				      }
				      else if ( avi == xlen )
				      {
				      	pf2 = pf1 * xf2;
				      }

				      for ( k = 0; k < 3; k++ )
				      {
				        ff[k] = ff[k] + pf2 * (
						mem_image[ k + 3 * ( oi1 + avi +
						ow * ( oj1 + avj ) ) ] );
				      }

				      av_tot = av_tot + pf2;
				   }
				 }

				 // Average colour
				 for ( k = 0; k < 3; k++ )
				 {
					a[k] = 0.49999 + ff[k] / av_tot;
				 }
				}

				for ( k = 0; k < 3; k++ )
				{
					e[k] = ( 1 - ac) * ( ( 1 - ab ) * a[k] +
						ab * b[k] ) +
						ac * ( ( 1 - ab ) * c[k] +
						ab * d[k] );
				}

				new_image[ 3 * ( i + nw * j ) ] = e[0];
				new_image[ 1 + 3 * ( i + nw * j ) ] = e[1];
				new_image[ 2 + 3 * ( i + nw * j ) ] = e[2];
			}
		    }
		}
		else for ( j = 0; j < nh; j++ )	// Nearest neighbour RGB
		{
			oj = oh * ( (float) j ) / nh;

			for ( i = 0; i < nw; i++ )
			{
				oi = ow * ( (float) i ) / nw;

				new_image[ 3 * ( i + nw * j ) ] =
					mem_image[ 3 * ( oi + ow * oj ) ];

				new_image[ 1 + 3 * ( i + nw * j ) ] =
					mem_image[ 1 + 3 * ( oi + ow * oj ) ];

				new_image[ 2 + 3 * ( i + nw * j ) ] =
					mem_image[ 2 + 3 * ( oi + ow * oj ) ];
			}
		}
	}

	mtpixel_palette_copy ( &new->palette, &image->palette );

finish:
	return new;
}

mtImage * mtpixel_image_scale (
	mtImage		* image,
	mtScaleType	type,
	int		width,
	int		height,
	int		gamma_correction,
	int		sharper_reduce
	)
{
	scale_context	ctx;
	mtImage		* new;
	int		ow,
			oh;


	if (	! image				||
		type <= MTPIXEL_SCALE_NONE	||
		type >= MTPIXEL_SCALE_TOTAL	||
		width < 1			||
		height < 1
		)
	{
		return NULL;
	}

	if (	type == MTPIXEL_SCALE_MT	||
		type == MTPIXEL_SCALE_MT_SMOOTH
		)
	{
		return mt_scale ( image, type, width, height );
	}

	if ( image->type == MTPIXEL_IMAGE_INDEXED )
	{
		type = MTPIXEL_SCALE_NEAREST_NEIGHBOUR;
	}

	new = new_scale_rotate ( image, width, height );
	if ( ! new )
	{
		return NULL;
	}

	ow = image->width;
	oh = image->height;

	if ( ! prepare_scale ( image, &ctx, ow, oh, width, height, type,
		sharper_reduce ) )
	{
		mtpixel_image_destroy ( new );
		return NULL;
	}

	do_scale_internal ( &ctx, new, image, type, gamma_correction, ow, oh,
		width, height );

	return new;
}

static void mem_rotate_free_real (
	mtImage		* new_img,
	mtImage		* old_img,
	int		ow,
	int		oh,
	int		nw,
	int		nh,
	double		rangle,
	double		angle,
	int		mode,
	int		gcor
	)
{
	unsigned char	* src,
			* dest,
			* alpha,
			A_rgb[3],
			* pix1,
			* pix2,
			* pix3,
			* pix4;
	int		nx,
			ny,
			ox,
			oy,
			cc,
			bpp;
	double		s1,
			s2,
			c1,
			c2,		// Trig values
			cx0,
			cy0,
			cx1,
			cy1,
			x00,
			y00,
			x0y,
			y0y,		// Quick look up values
			fox,
			foy,
			k1,
			k2,
			k3,
			k4,		// Pixel weights
			aa1,
			aa2,
			aa3,
			aa4,
			aa,
			rr,
			gg,
			bb,
			tw,
			th,
			ta,
			ca,
			sa,
			sca,
			csa,
			Y00,
			Y0h,
			Yw0,
			Ywh,
			X00,
			Xwh;

	c2 = cos ( rangle );
	s2 = sin ( rangle );
	c1 = -s2;
	s1 = c2;

	/* Centerpoints, including half-pixel offsets */
	cx0 = ( ow - 1 ) / 2.0;
	cy0 = ( oh - 1 ) / 2.0;
	cx1 = ( nw - 1 ) / 2.0;
	cy1 = ( nh - 1 ) / 2.0;

	x00 = cx0 - cx1 * s1 - cy1 * s2;
	y00 = cy0 - cx1 * c1 - cy1 * c2;

	A_rgb[0] = 0;
	A_rgb[1] = 0;
	A_rgb[2] = 0;

	/* Prepare clipping rectangle */
	tw = 0.5 * ( ow + ( mode ? 1 : 0 ) );
	th = 0.5 * ( oh + ( mode ? 1 : 0 ) );
	ta = M_PI * ( angle / 180.0 - floor ( angle / 180.0 ) );
	ca = cos ( ta );
	sa = sin ( ta );
	sca = ca ? sa / ca : 0.0;
	csa = sa ? ca / sa : 0.0;

	Y00 = cy1 - th * ca - tw * sa;
	Y0h = cy1 + th * ca - tw * sa;
	Yw0 = cy1 - th * ca + tw * sa;
	Ywh = cy1 + th * ca + tw * sa;
	X00 = cx1 - tw * ca + th * sa;
	Xwh = cx1 + tw * ca - th * sa;

	bpp = new_img->type;

	for ( ny = 0; ny < nh; ny ++ )
	{
		int		xl,
				xm;


		/* Clip this row */
		if ( ny < Y0h )
		{
			xl = ceil ( X00 + ( Y00 - ny ) * sca );
		}
		else if ( ny < Ywh )
		{
			xl = ceil ( Xwh + ( ny - Ywh ) * csa );
		}
		else /* if (ny < Yw0) */
		{
			xl = ceil ( Xwh + ( Ywh - ny ) * sca );
		}

		if ( ny < Y00 )
		{
			xm = ceil ( X00 + ( Y00 - ny ) * sca );
		}
		else if ( ny < Yw0 )
		{
			xm = ceil ( X00 + ( ny - Y00 ) * csa );
		}
		else /* if (ny < Ywh) */
		{
			xm = ceil ( Xwh + ( Ywh - ny ) * sca );
		}

		if ( xl < 0 )
		{
			xl = 0;
		}

		if ( --xm >= nw )
		{
			xm = nw - 1;
		}

		x0y = ny * s2 + x00;
		y0y = ny * c2 + y00;

		for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc ++ )
		{
			if ( ! new_img->channel[cc] )
			{
				continue;
			}

			/* RGB nearest neighbour */
			if (	! mode &&
				cc == MTPIXEL_CHANNEL_IMAGE &&
				bpp == 3
				)
			{
				dest = new_img->channel[cc] + (ny * nw + xl) *
					3;

				for ( nx = xl; nx <= xm; nx++ , dest += 3)
				{
					ox = rint ( nx * s1 + x0y );
					oy = rint ( nx * c1 + y0y );
					src = old_img->channel[cc] +
						( oy * ow + ox ) * 3;

					dest[0] = src[0];
					dest[1] = src[1];
					dest[2] = src[2];
				}

				continue;
			}

			/* One-bpp nearest neighbour */
			if ( ! mode )
			{
				dest = new_img->channel[cc] + ny * nw + xl;

				for ( nx = xl; nx <= xm; nx ++ )
				{
					ox = rint ( nx * s1 + x0y );
					oy = rint ( nx * c1 + y0y );
					*dest++ = old_img->channel[cc][ oy * ow
						+ ox ];
				}

				continue;
			}

			/* RGB/RGBA bilinear */
			if ( cc == MTPIXEL_CHANNEL_IMAGE && ( bpp == 3 ) )
			{
				alpha = new_img->channel[MTPIXEL_CHANNEL_ALPHA];

				if ( alpha )
				{
					alpha += ny * nw + xl;
				}

				dest = new_img->channel[cc] + (ny * nw + xl) *
					3;

				for ( nx = xl; nx <= xm; nx++ , dest += 3 )
				{
					fox = nx * s1 + x0y;
					foy = nx * c1 + y0y;

					/* floor () is *SLOW* on Win32 -
						avoiding... */

					ox = (int)( fox + 2.0 ) - 2;
					oy = (int)( foy + 2.0 ) - 2;
					fox -= ox;
					foy -= oy;
					k4 = fox * foy;
					k3 = foy - k4;
					k2 = fox - k4;
					k1 = 1.0 - fox - foy + k4;
					pix1 = old_img->channel[cc] +
						(oy * ow + ox) * 3;
					pix2 = pix1 + 3;
					pix3 = pix1 + ow * 3;
					pix4 = pix3 + 3;

					if ( ox > ow - 2 )
					{
						pix2 = pix4 = A_rgb;
					}
					else if ( ox < 0 )
					{
						pix1 = pix3 = A_rgb;
					}

					if ( oy > oh - 2 )
					{
						pix3 = pix4 = A_rgb;
					}
					else if ( oy < 0 )
					{
						pix1 = pix2 = A_rgb;
					}

					if ( alpha )
					{
						aa1 = aa2 = aa3 = aa4 = 0.0;
						src = old_img->channel[
							MTPIXEL_CHANNEL_ALPHA ]
							+ oy * ow + ox;

						if ( pix1 != A_rgb )
						{
							aa1 = src[0] * k1;
						}

						if ( pix2 != A_rgb )
						{
							aa2 = src[1] * k2;
						}

						if ( pix3 != A_rgb )
						{
							aa3 = src[ow] * k3;
						}

						if ( pix4 != A_rgb )
						{
							aa4 = src[ow + 1] * k4;
						}

						aa = aa1 + aa2 + aa3 + aa4;

						if ( (*alpha++ = rint ( aa ) ) )
						{
							aa = 1.0 / aa;
							k1 = aa1 * aa;
							k2 = aa2 * aa;
							k3 = aa3 * aa;
							k4 = aa4 * aa;
						}
					}

					if ( gcor )	/* Gamma-correct */
					{
						rr = _gamma256 ( pix1[0] ) * k1
							+
							_gamma256 ( pix2[0] ) *
							k2 +
							_gamma256 ( pix3[0] ) *
							k3 +
							_gamma256 ( pix4[0] ) *
							k4;

						gg = _gamma256 ( pix1[1] ) * k1
							+
							_gamma256 ( pix2[1] ) *
							k2 +
							_gamma256 ( pix3[1] ) *
							k3 +
							_gamma256 ( pix4[1] ) *
							k4;

						bb = _gamma256 ( pix1[2] ) * k1
							+
							_gamma256 ( pix2[2] ) *
							k2 +
							_gamma256 ( pix3[2] ) *
							k3 +
							_gamma256 ( pix4[2] ) *
							k4;

						dest[0] = _ungamma256 ( rr );
						dest[1] = _ungamma256 ( gg );
						dest[2] = _ungamma256 ( bb );
					}
					else	/* Leave as is */
					{
						rr = pix1[0] * k1 +
							pix2[0] * k2 +
							pix3[0] * k3 +
							pix4[0] * k4;

						gg = pix1[1] * k1 +
							pix2[1] * k2 +
							pix3[1] * k3 +
							pix4[1] * k4;

						bb = pix1[2] * k1 +
							pix2[2] * k2 +
							pix3[2] * k3 +
							pix4[2] * k4;

						dest[0] = rint ( rr );
						dest[1] = rint ( gg );
						dest[2] = rint ( bb );
					}
				}

				continue;
			}

			/* Alpha channel already done... maybe */
			if ( cc == MTPIXEL_CHANNEL_ALPHA && ( bpp == 3 ) )
			{
				continue;
			}

			/* Utility channel bilinear */
			dest = new_img->channel[cc] + ny * nw + xl;
			for ( nx = xl; nx <= xm; nx ++ )
			{
				fox = nx * s1 + x0y;
				foy = nx * c1 + y0y;

				/* floor () is *SLOW* on Win32 - avoiding... */
				ox = (int)( fox + 2.0 ) - 2;
				oy = (int)( foy + 2.0 ) - 2;
				fox -= ox;
				foy -= oy;
				k4 = fox * foy;
				k3 = foy - k4;
				k2 = fox - k4;
				k1 = 1.0 - fox - foy + k4;
				src = old_img->channel[cc] + oy * ow + ox;
				aa1 = aa2 = aa3 = aa4 = 0.0;

				if ( ox < ow - 1 )
				{
					if ( oy < oh - 1 )
					{
						aa4 = src[ow + 1] * k4;
					}

					if ( oy >= 0 )
					{
						aa2 = src[1] * k2;
					}
				}
				if ( ox >= 0 )
				{
					if ( oy < oh - 1 )
					{
						aa3 = src[ow] * k3;
					}

					if ( oy >= 0 )
					{
						aa1 = src[0] * k1;
					}
				}

				*dest++ = rint ( aa1 + aa2 + aa3 + aa4 );
			}
		}
	}
}

mtImage * mtpixel_image_rotate (
	mtImage		* image,
	double		angle,
	int		smooth,
	int		gamma_correction
	)
{
	int		ow,
			oh,
			width,
			height,
			dx,
			dy;
	double		rangle,
			s2,
			c2;
	mtImage		* new;


	if ( ! image )
	{
		return NULL;
	}

	if ( angle == 90 )
	{
		return image_flipper ( image, I_FLIP_90 );
	}

	if ( angle == 180 )
	{
		return image_flipper ( image, I_FLIP_180 );
	}

	if ( angle == 270 )
	{
		return image_flipper ( image, I_FLIP_270 );
	}

	rangle = ( M_PI / 180.0 ) * angle;
	c2 = cos ( rangle );
	s2 = sin ( rangle );

	ow = image->width;
	oh = image->height;

	/* Preserve original centering */
	dx = ow & 1;
	dy = oh & 1;

	/* Exchange Y with X when rotated Y is nearer to old X */
	if ( ( dx ^ dy ) && ( fabs ( c2 ) < fabs ( s2 ) ) )
	{
		dx ^= 1;
		dy ^= 1;
	}

#define DD ( 127.0 / 128.0 ) /* Include all _visibly_ altered pixels */

	width = 2 * floor ( 0.5 * ( fabs ( ow * c2 ) + fabs ( oh * s2 ) - dx )
		+ DD ) + dx;

	height = 2 * floor ( 0.5 * ( fabs ( oh * c2 ) + fabs ( ow * s2 ) - dy )
		+ DD ) + dy;
#undef DD

	new = new_scale_rotate ( image, width, height );

	if ( ! new )
	{
		return NULL;
	}

	mem_rotate_free_real ( new, image, ow, oh, width, height, rangle, angle,
		smooth, gamma_correction );

	return new;
}

static void do_transform (
	int		start,
	int		step,
	int		cnt,
	unsigned char	* imgr,
	unsigned char	* img0,
	int		bcsp[6],
	int		allow[3]
	)
{
	static int	ixx[7] = { 0, 1, 2, 0, 1, 2, 0 },
			posm[9] = { 0, 0xFF00, 0x5500, 0x2480, 0x1100,
				0x0840, 0x0410, 0x0204, 0 },
			last_gamma,
			last_br,
			last_co;

	static unsigned char gamma_table[256],
			bc_table[256];
	int		do_gamma,
			do_bc,
			br,
			co,
			sa,
			ps,
			pmul,
			dH,
			sH,
			tH,
			ix0,
			ix1,
			ix2,
			c0,
			c1,
			c2,
			dc = 0,
			i,
			j,
			r,
			g,
			b,
			ofs3,
			opacity,
			op0 = 0,
			op1 = 0,
			op2 = 0,
			ops;
	double		w;
	unsigned char	rgb[3];


	cnt = start + step * cnt;

	if ( ! allow[0] )
	{
		op0 = 255;
	}

	if ( ! allow[1] )
	{
		op1 = 255;
	}

	if ( ! allow[2] )
	{
		op2 = 255;
	}

	ops = op0 + op1 + op2;

	br = bcsp[0] * 255;
	co = bcsp[1];

	if ( co > 0 )
	{
		co *= 3;
	}

	co += 100;
	co = ( 255 * co ) / 100;
	sa = ( 255 * bcsp[2] ) / 100;
	dH = sH = bcsp[5];
	ps = 8 - bcsp[3];
	pmul = posm[bcsp[3]];

	do_gamma = bcsp[4] - 100;
	do_bc = br | (co - 255);

	/* Prepare gamma table */
	if ( do_gamma && ( do_gamma != last_gamma ) )
	{
		last_gamma = do_gamma;
		w = 100.0 / (double)bcsp[4];

		for ( i = 0; i < 256; i++ )
		{
			gamma_table[i] = rint ( 255.0 * pow ( (double)i / 255.0,
				w ) );
		}
	}

	/* Prepare brightness-contrast table */
	if (	do_bc &&
		( ( br != last_br ) || ( co != last_co ) )
		)
	{
		last_br = br;
		last_co = co;

		for ( i = 0; i < 256; i++ )
		{
			j = ( ( i + i - 255 ) * co + ( 255 * 255 ) ) / 2 + br;
			j = j < 0 ? 0 : j > (255 * 255) ? (255 * 255) : j;

			bc_table[i] = ( j + ( j >> 8 ) + 1 ) >> 8;
		}
	}

	if ( dH )
	{
		if ( dH < 0 )
		{
			dH += 1530;
		}

		dc = ( dH / 510 ) * 2;
		dH -= dc * 255;

		if ( ( sH = dH > 255 ) )
		{
			dH = 510 - dH;
			dc = dc < 4 ? dc + 2 : 0;
		}
	}

	ix0 = ixx[dc];
	ix1 = ixx[dc + 1];
	ix2 = ixx[dc + 2];

	for ( i = start; i < cnt; i += step )
	{
		ofs3 = i * 3;
		rgb[0] = img0[ofs3 + 0];
		rgb[1] = img0[ofs3 + 1];
		rgb[2] = img0[ofs3 + 2];
		opacity = 0;

		/* If we do gamma transform */
		if ( do_gamma )
		{
			rgb[0] = gamma_table[ rgb[0] ];
			rgb[1] = gamma_table[ rgb[1] ];
			rgb[2] = gamma_table[ rgb[2] ];
		}

		/* If we do hue transform & colour has a hue */
		if (	dH &&
			( ( rgb[0] ^ rgb[1] ) | ( rgb[0] ^ rgb[2] ) )
			)
		{
			/* Min. component */
			c2 = dc;

			if ( rgb[ix2] < rgb[ix0] )
			{
				c2 ++;
			}

			if ( rgb[ ixx[c2] ] >= rgb[ ixx[c2 + 1] ] )
			{
				c2 ++;
			}

			/* Actual indices */
			c2 = ixx[c2];
			c0 = ixx[c2 + 1];
			c1 = ixx[c2 + 2];

			/* Max. component & edge dir */
			if ( ( tH = rgb[c0] <= rgb[c1] ) )
			{
				c0 = ixx[c2 + 2];
				c1 = ixx[c2 + 1];
			}

			/* Do adjustment */
			j = dH * ( rgb[c0] - rgb[c2] ) + 127; /* Round up (?) */
			j = ( j + ( j >> 8 ) + 1 ) >> 8;

			r = rgb[c0];
			g = rgb[c1];
			b = rgb[c2];

			if ( tH ^ sH )	/* Falling edge */
			{
				rgb[c1] = r = g > j + b ? g - j : b;
				rgb[c2] += j + r - g;
			}
			else		/* Rising edge */
			{
				rgb[c1] = b = g < r - j ? g + j : r;
				rgb[c0] -= j + g - b;
			}
		}

		r = rgb[ix0];
		g = rgb[ix1];
		b = rgb[ix2];

		/* If we do brightness/contrast transform */
		if ( do_bc )
		{
			r = bc_table[r];
			g = bc_table[g];
			b = bc_table[b];
		}

		/* If we do saturation transform */
		if (sa)
		{
			j = 0.299 * r + 0.587 * g + 0.114 * b;

			r = r * 255 + (r - j) * sa;
			r = r < 0 ? 0 : r > (255 * 255) ? (255 * 255) : r;
			r = (r + (r >> 8) + 1) >> 8;

			g = g * 255 + (g - j) * sa;
			g = g < 0 ? 0 : g > (255 * 255) ? (255 * 255) : g;
			g = (g + (g >> 8) + 1) >> 8;

			b = b * 255 + (b - j) * sa;
			b = b < 0 ? 0 : b > (255 * 255) ? (255 * 255) : b;
			b = (b + (b >> 8) + 1) >> 8;
		}

		/* If we do posterize transform */
		if ( ps )
		{
			r = ( (r >> ps) * pmul ) >> 8;
			g = ( (g >> ps) * pmul ) >> 8;
			b = ( (b >> ps) * pmul ) >> 8;
		}

		/* If we do partial masking */
		if (ops || opacity)
		{
			r = r * 255 + (img0[ofs3 + 0] - r) * (opacity | op0);
			r = ( r + ( r >> 8 ) + 1 ) >> 8;

			g = g * 255 + (img0[ofs3 + 1] - g) * (opacity | op1);
			g = ( g + ( g >> 8 ) + 1 ) >> 8;

			b = b * 255 + (img0[ofs3 + 2] - b) * (opacity | op2);
			b = ( b + ( b >> 8 ) + 1 ) >> 8;
		}

		imgr[ ofs3 + 0 ] = r;
		imgr[ ofs3 + 1 ] = g;
		imgr[ ofs3 + 2 ] = b;
	}
}

int mtpixel_image_color_transform (
	mtImage		* image,
	mtPalette	* palette,
	int		brightness,
	int		contrast,
	int		saturation,
	int		posterize,
	int		gamma,
	int		hue,
	int		red,
	int		green,
	int		blue
	)
{
	int		bcsp[6] = {
				brightness,
				contrast,
				saturation,
				posterize,
				gamma,
				hue
				},
			allow[3] = {
				red,
				green,
				blue
				};


	if (	image &&
		image->channel[ MTPIXEL_CHANNEL_IMAGE ] &&
		image->type == MTPIXEL_IMAGE_RGB
		)
	{
		do_transform ( 0, 1, image->width * image->height,
			image->channel[ MTPIXEL_CHANNEL_IMAGE ],
			image->channel[ MTPIXEL_CHANNEL_IMAGE ],
			bcsp, allow );
	}

	if ( palette )
	{
		do_transform ( 0, 1, 256,
			(unsigned char *)palette->col,
			(unsigned char *)palette->col,
			bcsp, allow );
	}

	return 0;
}

int mtpixel_image_invert (
	mtImage		* image,
	mtPalette	* palette,
	mtBrush		* brush
	)
{
	int		i,
			j,
			cc;


	if ( image )
	{
		unsigned char	* dest,
				pix;


		for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc++ )
		{
			if (	! image->channel[cc] ||
				( brush && brush->chan_dis[cc] )
				)
			{
				continue;
			}

			dest = image->channel[cc];
			j = image->width * image->height;

			if ( cc == MTPIXEL_CHANNEL_IMAGE )
			{
				j *= image->type;
			}

			for ( i = 0; i < j; i++ )
			{
				pix = dest[0];
				*dest++ = ~pix;
			}
		}
	}

	if ( palette )
	{
		for ( i = 0; i < 256; i++ )
		{
			palette->col[i][0] = ~palette->col[i][0];
			palette->col[i][1] = ~palette->col[i][1];
			palette->col[i][2] = ~palette->col[i][2];
		}
	}

	return 0;
}

static void do_grayscale (
	unsigned char	* mem,
	int		items
	)
{
	double		value;
	int		i,
			v;


	for ( i = 0; i < items; i++ )
	{
		value = _rgb2B ( _gamma256 ( mem[0] ),
				_gamma256 ( mem[1] ),
				_gamma256 ( mem[2] ) );

		v = _ungamma256 ( value );

		*mem++ = v;
		*mem++ = v;
		*mem++ = v;
	}
}

int mtpixel_image_grayscale (
	mtImage		* image,
	mtPalette	* palette
	)
{
	if ( image && image->type == MTPIXEL_IMAGE_RGB )
	{
		do_grayscale ( image->channel[ MTPIXEL_CHANNEL_IMAGE ],
			image->width * image->height );
	}

	if ( palette )
	{
		do_grayscale ( (unsigned char *)palette->col, 256 );
	}

	return 0;
}

static void do_posterize (
	unsigned char	* mem,
	int		bytes,
	int		bpp,
	int		level,
	int		red,
	int		green,
	int		blue
	)
{
	unsigned char	tab[256],
			pix;
	int		i,
			j;


	for ( i = 0; i < 256; i++ )
	{
		j = rint ( ( (double)i ) / 255 * ( level - 1 ) );
		tab[i] = j * 255 / ( level - 1 );
	}

	if ( bpp == 1 )
	{
		for ( i = 0; i < bytes; i++ )
		{
			pix = mem[0];
			*mem++ = tab[pix];
		}
	}
	else if ( bpp == 3 )
	{
		for ( i = 0; i < bytes; i += 3 )
		{
			if ( red )
			{
				mem[0] = tab[ mem[0] ];
			}

			if ( green )
			{
				mem[1] = tab[ mem[1] ];
			}

			if ( blue )
			{
				mem[2] = tab[ mem[2] ];
			}

			mem += 3;
		}
	}
}

int mtpixel_image_posterize (
	mtImage		* image,
	mtPalette	* palette,
	mtBrush		* brush,
	int		level,
	int		red,
	int		green,
	int		blue
	)
{
	if ( image )
	{
		int		bpp,
				cc;


		for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc ++ )
		{
			if (	! image->channel[cc] ||
				( brush && brush->chan_dis[cc] )
				)
			{
				continue;
			}

			if ( cc == MTPIXEL_CHANNEL_IMAGE )
			{
				bpp = image->type;
			}
			else
			{
				bpp = 1;
			}

			do_posterize ( image->channel[cc],
				image->width * image->height * bpp, bpp, level,
				red, green, blue );
		}
	}

	if ( palette )
	{
		do_posterize ( (unsigned char *)palette->col, 256 * 3, 3, level,
			red, green, blue );
	}

	return 0;
}



enum
{
	EFFECT_SHARPEN,
	EFFECT_SOFTEN,
	EFFECT_EDGE_DETECT,
	EFFECT_EMBOSS
};



static mtImage * do_effect (
	mtImage		* image,
	mtBrush		* brush,
	int		effect,
	int		val
	)
{
	unsigned char	* src,
			* dest;
	int		i,
			j,
			k = 0,
			ix,
			bpp,
			ll,
			dxp1,
			dxm1,
			dyp1,
			dym1,
			op,
			md,
			ms,
			cc;
	double		blur = (double)val / 200.0;
	mtImage		* new;


	new = mtpixel_image_duplicate ( image );
	if ( ! new )
	{
		return NULL;
	}

	for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc ++ )
	{
		if (	! new->channel[cc] ||
			( brush && brush->chan_dis[cc] )
			)
		{
			continue;
		}

		src = image->channel[cc];
		dest = new->channel[cc];

		if ( cc == MTPIXEL_CHANNEL_IMAGE )
		{
			bpp = new->type;
		}
		else
		{
			bpp = 1;
		}

		ll = new->width * bpp;
		ms = bpp == 3 ? 1 : 4;

		for ( ix = i = 0; i < new->height; i++ )
		{
			dyp1 = i < new->height - 1 ? ll : -ll;
			dym1 = i ? -ll : ll;

			for ( md = j = 0; j < ll; j++ , ix ++ )
			{
				op = 0;

				/* One step for 1 or 3 bytes */
				md += ms + ( md >> 1 );

				md &= 3;

				if ( op == 255 )
				{
					continue;
				}

				dxp1 = j < ll - bpp ? bpp : -bpp;
				dxm1 = j >= bpp ? -bpp : bpp;

				switch ( effect )
				{
				case EFFECT_EDGE_DETECT:
					k = src[ix];
					k = abs ( k - src[ix + dym1] ) +
						abs ( k - src[ix + dyp1] ) +
						abs ( k - src[ix + dxm1] ) +
						abs ( k - src[ix + dxp1] );
					break;

				case EFFECT_EMBOSS:
					k = src[ix + dym1] + src[ix + dxm1] +
						src[ix + dxm1 + dym1] +
						src[ix + dxp1 + dym1];

					k = k / 4 - src[ix] + 127;
					break;

				case EFFECT_SHARPEN:
					k = src[ix + dym1] + src[ix + dyp1] +
						src[ix + dxm1] +
						src[ix + dxp1] - 4 * src[ix];

					k = src[ix] - blur * k;
					break;

				case EFFECT_SOFTEN:
					k = src[ix + dym1] + src[ix + dyp1] +
						src[ix + dxm1] +
						src[ix + dxp1] - 4 * src[ix];

					k = src[ix] + ( 5 * k ) / ( 125 - val );
					break;
				}

				k = k < 0 ? 0 : k > 0xFF ? 0xFF : k;
				k = 255 * k + ( src[ix] - k ) * op;
				dest[ix] = ( k + ( k >> 8 ) + 1 ) >> 8;
			}
		}
	}

	return new;
}

mtImage * mtpixel_image_sharpen (
	mtImage		* image,
	mtBrush		* brush,
	int		value
	)
{
	return do_effect ( image, brush, EFFECT_SHARPEN, value );
}

mtImage * mtpixel_image_soften (
	mtImage		* image,
	mtBrush		* brush,
	int		value
	)
{
	return do_effect ( image, brush, EFFECT_SOFTEN, value );
}

mtImage * mtpixel_image_edge_detect (
	mtImage		* image,
	mtBrush		* brush
	)
{
	return do_effect ( image, brush, EFFECT_EDGE_DETECT, 0 );
}

mtImage * mtpixel_image_emboss (
	mtImage		* image,
	mtBrush		* brush
	)
{
	return do_effect ( image, brush, EFFECT_EMBOSS, 0 );
}



#include "patterns.xbm"



static int mt_round (			// Round a float to nearest whole number
	float		n
	)
{
	if ( n < 0 )
	{
		return ( (int) ( n - 0.4999999 ) );
	}
	else
	{
		return ( (int) ( n + 0.4999999 ) );
	}
}

mtBrush * mtpixel_brush_new ( void )
{
	mtBrush		* brush;

	brush = calloc ( sizeof ( mtBrush ), 1 );
	if ( ! brush )
	{
		return NULL;
	}

	brush->size = 1;
	brush->opacity = 255;
	brush->shape = MTPIXEL_BRUSH_SHAPE_SQUARE;
	brush->mode = MTPIXEL_BRUSH_MODE_PATTERN;

	return brush;
}

#define PATTERN_GRID_W 10

static void update_pattern (
	mtBrush		* brush
	)
{
	int		i,
			j,
			b,
			l,
			ii,
			k;


	/* Pattern bitmap starts here */
	l = brush->pattern * 8 - ( brush->pattern % PATTERN_GRID_W ) * 7;

	/* Set up pattern maps from XBM */
	for ( i = ii = 0; i < 8; i++ )
	{
		b = patterns_bits[l + i * PATTERN_GRID_W];

		for ( k = 0; k < 8; k++ , ii++ , b >>= 1 )
		{
			j = b & 1;
			brush->pattern_i[ii] =
				brush->col_i[ MTPIXEL_CHANNEL_IMAGE ][j];

			brush->pattern_rgb[ ii * 3 + 0 ] = brush->col_rgb[0][j];
			brush->pattern_rgb[ ii * 3 + 1 ] = brush->col_rgb[1][j];
			brush->pattern_rgb[ ii * 3 + 2 ] = brush->col_rgb[2][j];
		}
	}
}

int mtpixel_brush_set_RGB (
	mtBrush		* brush,
	int		col_a,
	int		col_b
	)
{
	if ( ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( col_a >= 0 )
	{
		brush->col_rgb[0][0] = MTPIXEL_INT_2_R ( col_a );
		brush->col_rgb[1][0] = MTPIXEL_INT_2_G ( col_a );
		brush->col_rgb[2][0] = MTPIXEL_INT_2_B ( col_a );
	}

	if ( col_b >= 0 )
	{
		brush->col_rgb[0][1] = MTPIXEL_INT_2_R ( col_b );
		brush->col_rgb[1][1] = MTPIXEL_INT_2_G ( col_b );
		brush->col_rgb[2][1] = MTPIXEL_INT_2_B ( col_b );
	}

	update_pattern ( brush );

	return 0;
}

int mtpixel_brush_set_index (
	mtBrush		* brush,
	mtPalette	* palette,
	int		index_a,
	int		index_b,
	mtChannelType	channel
	)
{
	if (	! brush ||
		(channel == MTPIXEL_CHANNEL_IMAGE && ! palette) ||
		channel < 0 ||
		channel >= MTPIXEL_CHANNEL_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( index_a >= 0 )
	{
		brush->col_i[channel][0] = index_a;
	}

	if ( index_b >= 0 )
	{
		brush->col_i[channel][1] = index_b;
	}

	if ( channel == MTPIXEL_CHANNEL_IMAGE )
	{
		if ( index_a >= 0 && index_a < MTPIXEL_PALETTE_ITEMS )
		{
			brush->col_rgb[0][0] = palette->col[ index_a ][0];
			brush->col_rgb[1][0] = palette->col[ index_a ][1];
			brush->col_rgb[2][0] = palette->col[ index_a ][2];
		}

		if ( index_b >= 0 && index_b < MTPIXEL_PALETTE_ITEMS )
		{
			brush->col_rgb[0][1] = palette->col[ index_b ][0];
			brush->col_rgb[1][1] = palette->col[ index_b ][1];
			brush->col_rgb[2][1] = palette->col[ index_b ][2];
		}

		update_pattern ( brush );
	}

	return 0;
}

int mtpixel_brush_set_style (
	mtBrush		* brush,
	mtBrushShape	shape,
	mtBrushMode	mode,
	int		pattern,
	int		size,
	int		flow,
	int		opacity
	)
{
	if ( ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( pattern >= 0 && pattern < MTPIXEL_BRUSH_PATTERN_TOTAL )
	{
		brush->pattern = pattern;
		update_pattern ( brush );
	}

	if (	shape>MTPIXEL_BRUSH_SHAPE_NONE &&
		shape<MTPIXEL_BRUSH_SHAPE_TOTAL
		)
	{
		brush->shape = shape;
	}

	if (	mode > MTPIXEL_BRUSH_MODE_NONE &&
		mode < MTPIXEL_BRUSH_MODE_TOTAL
		)
	{
		brush->mode = mode;
	}

	if ( size >= 0 )
	{
		brush->size = size;
	}

	if ( flow >= 0 )
	{
		brush->flow = flow;
	}

	if ( opacity >= 0 )
	{
		brush->opacity = opacity;
	}

	return 0;
}

int mtpixel_brush_set_size (
	mtBrush		* brush,
	int		size
	)
{
	if ( ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	brush->size = size;

	return 0;
}

mtBrush * mtpixel_brush_duplicate (
	mtBrush		* brush
	)
{
	mtBrush		* new;


	if ( ! brush )
	{
		return NULL;
	}

	new = mtpixel_brush_new ();
	if ( new )
	{
		memcpy ( new, brush, sizeof ( mtBrush ) );
	}

	return new;
}

int mtpixel_brush_destroy (
	mtBrush		* brush
	)
{
	if ( ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	free ( brush );

	return 0;
}

static int im_chan_setclear (
	mtBrush		* brush,
	mtChannelMask	mask,
	int		val
	)
{
	int		c;


	if ( ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	for ( c = 0; c < MTPIXEL_CHANNEL_TOTAL; c++ )
	{
		if ( mask & ( 1 << c ) )
		{
			brush->chan_dis[c] = val;
		}
	}

	return 0;
}

int mtpixel_brush_channel_mask_set (
	mtBrush		* brush,
	mtChannelMask	mask
	)
{
	return im_chan_setclear ( brush, mask, 1 );
}

int mtpixel_brush_channel_mask_clear (
	mtBrush		* brush,
	mtChannelMask	mask
	)
{
	return im_chan_setclear ( brush, mask, 0 );
}


#define GRAD_CUSTOM_DATA( X ) ( (X) ? MTPIXEL_GRAD_POINTS * ( (X) * 4 + 2) : 0 )
#define GRAD_CUSTOM_DMAP( X ) ( MTPIXEL_GRAD_POINTS * ( (X) * 4 + 3 ) )
#define GRAD_CUSTOM_OPAC( X ) ( MTPIXEL_GRAD_POINTS * ( (X) * 4 + 4 ) )
#define GRAD_CUSTOM_OMAP( X ) ( MTPIXEL_GRAD_POINTS * ( (X) * 4 + 5 ) )

#define GRAD_NONE	0
#define GRAD_START	1
#define GRAD_END	2
#define GRAD_DONE	3


/* Reevaluate gradient placement functions */
static void grad_update (
	mtGradInfo	* grad
	)
{
	double		len,
			len1,
			l2;


	/* Distance for gradient mode */
	grad->wmode = grad->gmode;
	len = grad->len;

	while ( 1 )
	{
		/* Stroke gradient */
		if (grad->status == GRAD_NONE)
		{
			if ( ! grad->len )
			{
				len = grad->rep + grad->ofs;
			}

			if ( len <= 0.0 )
			{
				grad->wmode = MTPIXEL_GRADIENT_MODE_NONE;
			}

			break;
		}

		/* Placement length */
		l2 = sqrt ( ( grad->x2 - grad->x1 ) * ( grad->x2 - grad->x1 ) +
			( grad->y2 - grad->y1 ) * ( grad->y2 - grad->y1 ) );

		if ( ! grad->len )
		{
			len = l2;
		}

		if ( l2 == 0.0 )
		{
			grad->wmode = MTPIXEL_GRADIENT_MODE_RADIAL;
			break;
		}

		grad->xv = ( grad->x2 - grad->x1 ) / l2;
		grad->yv = ( grad->y2 - grad->y1 ) / l2;

		break;
	}

	/* Base length (one repeat) */
	len1 = grad->rep > 0 ? grad->rep : len - grad->ofs;

	if ( len1 < 1.0 )
	{
		len1 = 1.0;
	}

	grad->wrep = len1;
	grad->wil1 = 1.0 / len1;

	/* Inverse period */
	l2 = 1.0;

	if ( grad->rmode == MTPIXEL_GRADIENT_BOUND_MIRROR )
	{
		/* Mirror repeat */
		l2 = len1 + len1;
	}
	else if ( grad->rmode == MTPIXEL_GRADIENT_BOUND_REPEAT )
	{
		/* Repeat */
		l2 = len1 + 1.0;
	}

	grad->wil2 = 1.0 / l2;
}

/* Setup gradient mapping */
static void gmap_setup (
	mtGradient	* grad,
	mtGradMap	* gmap,
	mtGradStore	gstore,
	int		slot
	)
{
	unsigned char	* data,
			* map;


	data = grad->grad_def + ( slot ? 8 + slot * 4 : 4 );
	map = grad->grad_def + 10 + slot * 4;
	gmap->vslen = 2;

	if ( gmap->gtype == MTPIXEL_GRADIENT_TYPE_CUSTOM )
	{
		gmap->vs = gstore + GRAD_CUSTOM_DATA ( slot );
		gmap->vsmap = gstore + GRAD_CUSTOM_DMAP ( slot );

		if ( gmap->cvslen > 1 )
		{
			gmap->vslen = gmap->cvslen;
		}
		else
		{
			memcpy ( gmap->vs, data, slot ? 2 : 6 );
			gmap->vsmap[0] = map[0];
		}
	}
	else
	{
		gmap->vs = data;
		gmap->vsmap = map;
		grad->grad_def[10 + slot * 4] = (unsigned char)gmap->gtype;
	}

	gmap->oplen = 2;

	if ( gmap->otype == MTPIXEL_GRADIENT_TYPE_CUSTOM )
	{
		gmap->op = gstore + GRAD_CUSTOM_OPAC ( slot );
		gmap->opmap = gstore + GRAD_CUSTOM_OMAP ( slot );

		if ( gmap->coplen > 1 )
		{
			gmap->oplen = gmap->coplen;
		}
		else
		{
			gmap->op[0] = grad->grad_def[0];
			gmap->op[1] = grad->grad_def[1];
			gmap->opmap[0] = grad->grad_def[2];
		}
	}
	else
	{
		gmap->op = grad->grad_def;
		gmap->opmap = grad->grad_def + 2;
		grad->grad_def[2] = gmap->otype;
	}
}

/* Store default gradient */
static void grad_def_update (
	mtBrush		* brush,
	mtChannelType	channel,
	mtImageType	bpp
	)
{
	int		ix;
	mtGradMap	* gradmap;


	/* Gradient slot (first RGB, then 1-bpp channels) */
	ix = channel + ( ( 0x81 + channel + channel - bpp ) >> 7 );
	gradmap = brush->gradient.graddata + ix;

	brush->gradient.grad_def[0] = brush->opacity;

	/* !!! As there's only 1 tool_opacity, use 0 for 2nd point */
	brush->gradient.grad_def[1] = 0;
	brush->gradient.grad_def[2] = gradmap->otype;

	brush->gradient.grad_def[10 + ix * 4] = gradmap->gtype;
	if ( ix )
	{
		brush->gradient.grad_def[8 + ix * 4] =
			brush->col_i[ ix - 1 ][0];

		brush->gradient.grad_def[9 + ix * 4] =
			brush->col_i[ ix - 1 ][1];

		brush->gradient.grad_def[12] =
			brush->col_i[ MTPIXEL_CHANNEL_IMAGE ][0];

		brush->gradient.grad_def[13] =
			brush->col_i[ MTPIXEL_CHANNEL_IMAGE ][1];
	}
	else
	{
		brush->gradient.grad_def[4] = brush->col_rgb[0][0];
		brush->gradient.grad_def[5] = brush->col_rgb[1][0];
		brush->gradient.grad_def[6] = brush->col_rgb[2][0];
		brush->gradient.grad_def[7] = brush->col_rgb[0][1];
		brush->gradient.grad_def[8] = brush->col_rgb[1][1];
		brush->gradient.grad_def[9] = brush->col_rgb[2][1];
	}

	gradmap = brush->gradient.graddata + MTPIXEL_CHANNEL_ALPHA + 1;
	brush->gradient.grad_def[ 12 + MTPIXEL_CHANNEL_ALPHA * 4 ] =
		brush->col_i[ MTPIXEL_CHANNEL_ALPHA ][0];
	brush->gradient.grad_def[ 13 + MTPIXEL_CHANNEL_ALPHA * 4 ] =
		brush->col_i[ MTPIXEL_CHANNEL_ALPHA ][1];
	brush->gradient.grad_def[ 14 + MTPIXEL_CHANNEL_ALPHA * 4 ] =
		gradmap->gtype;
}

static void update_gradients (
	mtBrush		* brush,
	mtChannelType	channel
	)
{
	int		i;


	for ( i = 0; i < MTPIXEL_CHANNEL_TOTAL; i++ )
	{
		grad_update ( brush->gradient.gradient + i );
	}

	for ( i = 0; i <= MTPIXEL_CHANNEL_TOTAL; i++ )
	{
		gmap_setup ( &brush->gradient, brush->gradient.graddata + i,
				brush->gradient.gradbytes, i );
	}

	grad_def_update ( brush, channel, MTPIXEL_IMAGE_INDEXED );
	grad_def_update ( brush, channel, MTPIXEL_IMAGE_RGB );
}

int mtpixel_brush_gradient_position (
	mtBrush		* brush,
	mtChannelType	channel,
	int		x1,
	int		y1,
	int		x2,
	int		y2
	)
{
	if ( ! brush || channel < 0 || channel >= MTPIXEL_CHANNEL_TOTAL )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	brush->gradient.gradient[ channel ].x1 = x1;
	brush->gradient.gradient[ channel ].y1 = y1;
	brush->gradient.gradient[ channel ].x2 = x2;
	brush->gradient.gradient[ channel ].y2 = y2;

	update_gradients ( brush, channel );

	return 0;
}

int mtpixel_brush_gradient (
	mtBrush		* brush,
	mtChannelType	channel,
	int		length,
	int		repeat_length,
	int		offset,
	mtGradientMode	grad_type,
	mtGradientBound	ext_type,
	mtGradientType	grad_col_type,
	int		grad_col_reverse,
	mtGradientType	grad_opac_type,		// Custom/Const
	int		grad_opac_reverse
	)
{
	int		idx = channel,
			b,
			bb[2] = { MTPIXEL_IMAGE_INDEXED, MTPIXEL_IMAGE_RGB };


	if ( ! brush || channel < 0 || channel >= MTPIXEL_CHANNEL_TOTAL )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	brush->gradient.gradient[ idx ].len = length;
	brush->gradient.gradient[ idx ].rep = repeat_length;
	brush->gradient.gradient[ idx ].ofs = offset;
	brush->gradient.gradient[ idx ].gmode = grad_type;
	brush->gradient.gradient[ idx ].rmode = ext_type;

	for ( b = 0; b < 2; b++ )
	{
		idx = channel + ( ( 0x81 + channel + channel - bb[b] ) >> 7 );

		brush->gradient.graddata[ idx ].gtype = grad_col_type;
		brush->gradient.graddata[ idx ].grev = grad_col_reverse;
		brush->gradient.graddata[ idx ].otype = grad_opac_type;
		brush->gradient.graddata[ idx ].orev = grad_opac_reverse;
		brush->gradient.gradient[ idx ].status = GRAD_DONE;
	}

	update_gradients ( brush, channel );

	return 0;
}

int mtpixel_brush_gradient_custom (
	mtBrush		* brush,
	mtChannelType	channel,
	int		points,
	int		table[256][2],
	int		opacity
	)
{
	unsigned char	* data,
			* map;
	int		i,
			idx,
			b,
			bb[2] = { MTPIXEL_IMAGE_INDEXED, MTPIXEL_IMAGE_RGB };


	if ( ! brush || channel < 0 || channel >= MTPIXEL_CHANNEL_TOTAL )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	for ( b = 0; b < 2; b++ )
	{
		idx = channel + ( ( 0x81 + channel + channel - bb[b]) >> 7 );

		if ( opacity )		// Populate opacity table
		{
			brush->gradient.graddata[ idx ].coplen = points;

			data = brush->gradient.gradbytes +
				GRAD_CUSTOM_OPAC ( idx );
			map  = brush->gradient.gradbytes +
				GRAD_CUSTOM_OMAP ( idx );

			for ( i = 0; i < points; i++ )
			{
				*data++ = table[i][0];
				*map++ = table[i][1];
			}
		}
		else			// Populate colour table
		{
			brush->gradient.graddata[ idx ].cvslen = points;
			map = brush->gradient.gradbytes +
				GRAD_CUSTOM_DMAP ( idx );
			data = brush->gradient.gradbytes +
				GRAD_CUSTOM_DATA ( idx );

			if ( idx )	// Indexed / channel
			{
				for ( i = 0; i < points; i++ )
				{
					*data++ = table[i][0];
					*map++  = table[i][1];
				}
			}
			else		// RGB
			{
				for ( i = 0; i < points; i++ )
				{
					*data++ = MTPIXEL_INT_2_R ( table[i][0]
						);
					*data++ = MTPIXEL_INT_2_G ( table[i][0]
						);
					*data++ = MTPIXEL_INT_2_B ( table[i][0]
						);
					*map++  = table[i][1];
				}
			}
		}
	}

	update_gradients ( brush, channel );

	return 0;
}

/*
	PAINTING FUNCTIONS
*/

/* Evaluate channel gradient at coordinate, return opacity */
/* Coordinate 0 is center of 1st pixel, 1 center of last */
static int grad_value (
	mtBrush		* brush,
	int		* dest,
	int		slot,
	double		x
	)
{
	int		i,
			k,
			i3,
			len,
			op;
	unsigned char	* gdata,
			* gmap;
	mtGradMap	* gradmap;
	double		xx,
			hsv[6];


	/* Gradient slot (first RGB, then 1-bpp channels) */
	gradmap = brush->gradient.graddata + slot;

	/* Get opacity */
	gdata = gradmap->op;
	gmap = gradmap->opmap;
	len = gradmap->oplen;
	xx = ( gradmap->orev ? 1.0 - x : x ) * ( len - 1 );
	i = xx;

	if ( i > len - 2 )
	{
		i = len - 2;
	}

	k = gmap[i] == MTPIXEL_GRADIENT_TYPE_CONST ? 0 : (int)( ( xx - i ) *
		0x10000 + 0.5);
	op = ( gdata[i] << 8 ) + ( ( k * ( gdata[i + 1] - gdata[i] ) + 127 )
		>> 8 );

	if ( ! op )
	{
		return 0;		/* Stop if zero opacity */
	}

	/* Get channel value */
	gdata = gradmap->vs; gmap = gradmap->vsmap; len = gradmap->vslen;
	xx = ( gradmap->grev ? 1.0 - x : x ) * ( len - 1 );
	i = xx;

	if ( i > len - 2 )
	{
		i = len - 2;
	}

	k = gmap[i] == MTPIXEL_GRADIENT_TYPE_CONST ? 0 : (int)( ( xx - i ) *
		0x10000 + 0.5 );

	if ( ! slot )			/* RGB */
	{
		i3 = i * 3;

		/* HSV interpolation */

		if (	gmap[i] == MTPIXEL_GRADIENT_TYPE_HSV ||
			gmap[i] == MTPIXEL_GRADIENT_TYPE_BK_HSV
			)
		{
			/* Convert */
			mtpixel_rgb2hsv ( gdata + i3 + 0, hsv + 0 );
			mtpixel_rgb2hsv ( gdata + i3 + 3, hsv + 3 );

			/* Grey has no hue */

			if ( hsv[1] == 0.0 )
			{
				hsv[0] = hsv[3];
			}

			if ( hsv[4] == 0.0 )
			{
				hsv[3] = hsv[0];
			}

			/* Prevent wraparound */
			i3 = gmap[i] == MTPIXEL_GRADIENT_TYPE_HSV ? 0 : 3;

			if ( hsv[i3] > hsv[i3 ^ 3] )
			{
				hsv[i3] -= 6.0;
			}

			/* Interpolate */
			hsv[0] += ( xx - i ) * ( hsv[3] - hsv[0] );
			hsv[1] += ( xx - i ) * ( hsv[4] - hsv[1] );
			hsv[2] += ( xx - i ) * ( hsv[5] - hsv[2] );

			/* Convert back */
			hsv[2] *= 512;
			hsv[1] = hsv[2] * ( 1.0 - hsv[1] );

			if ( hsv[0] < 0.0 )
			{
				hsv[0] += 6.0;
			}

			i3 = hsv[0];
			hsv[0] = ( hsv[0] - i3 ) * ( hsv[2] - hsv[1] );

			if ( i3 & 1 )
			{
				hsv[2] -= hsv[0];
				hsv[0] += hsv[2];
			}
			else
			{
				hsv[0] += hsv[1];
			}

			i3 >>= 1;
			dest[i3] = ( (int)hsv[2] + 1 ) >> 1;
			dest[ MOD3 ( i3 + 1 ) ] = ( (int)hsv[0] + 1 ) >> 1;
			dest[ MOD3 ( i3 + 2 ) ] = ( (int)hsv[1] + 1 ) >> 1;
		}
		else	/* RGB interpolation */
		{
			dest[0] = ( gdata[i3 + 0] << 8 ) +
				( ( k * ( gdata[i3 + 3] - gdata[i3 + 0] ) +
				127 ) >> 8 );

			dest[1] = ( gdata[i3 + 1] << 8 ) +
				( ( k * ( gdata[i3 + 4] - gdata[i3 + 1] ) +
				127 ) >> 8 );

			dest[2] = ( gdata[i3 + 2] << 8 ) +
				( ( k * ( gdata[i3 + 5] - gdata[i3 + 2]) +
				127 ) >> 8 );
		}
	}
	else if ( slot == MTPIXEL_CHANNEL_IMAGE + 1 ) /* Indexed */
	{
		dest[0] = gdata[i];
		dest[1] = gdata[ i + ( ( k + 0xFFFF ) >> 16 ) ];
		dest[ MTPIXEL_CHANNEL_IMAGE + 3 ] = ( k + 127 ) >> 8;
	}
	else	/* Utility */
	{
		dest[ slot + 2 ] = ( gdata[i] << 8 ) +
			( ( k * ( gdata[i + 1] - gdata[i] ) + 127 ) >> 8 );
	}

	return op;
}



/// Bayer ordered dithering

static const unsigned char bayer[16] =
	{
	0x00, 0x40, 0x10, 0x50, 0x04, 0x44, 0x14, 0x54,
	0x01, 0x41, 0x11, 0x51, 0x05, 0x45, 0x15, 0x55
	};

#define BAYER_MASK	15	/* Use 16x16 matrix */
#define BAYER( x, y )	( bayer[((x) ^ (y)) & BAYER_MASK] * 2 + bayer[(y) & BAYER_MASK])



/* Evaluate RGBA/indexed gradient at point, return opacity */
static int grad_pixel (
	int		channel,
	int		bpp,
	mtBrush		* brush,
	unsigned char	* dest,
	int		x,
	int		y
	)
{
	int		dither,
			op,
			slot,
			wrk[ MTPIXEL_CHANNEL_TOTAL + 3 ];
	mtGradInfo	* grad = brush->gradient.gradient + channel;
	double		dist,
			len1,
			l2;


	/* Disabled because of unusable settings? */
	if ( grad->wmode == MTPIXEL_GRADIENT_MODE_NONE )
	{
		return 0;
	}

	/* Distance for gradient mode */
	while ( 1 )
	{
		/* Linear/bilinear gradient */
		if ( grad->wmode <= MTPIXEL_GRADIENT_MODE_BILINEAR )
		{
			dist =	( x - grad->x1 ) * grad->xv +
				( y - grad->y1 ) * grad->yv;

			if ( grad->wmode == MTPIXEL_GRADIENT_MODE_LINEAR )
			{
				break;
			}

			dist = fabs ( dist );	/* Bilinear */

			break;
		}

		/* Radial gradient */
		if ( grad->wmode == MTPIXEL_GRADIENT_MODE_RADIAL )
		{
			dist = sqrt ( ( x - grad->x1 ) * ( x - grad->x1 ) +
				( y - grad->y1 ) * ( y - grad->y1 ) );
			break;
		}

		/* Square gradient */

		/*
		!!! Here is code duplication with linear/bilinear path - but
		merged paths actually LOSE in both time and space, at least
		with GCC - WJ
		*/

		dist = fabs ( ( x - grad->x1 ) * grad->xv + ( y - grad->y1 ) *
			grad->yv ) +
			fabs ( ( x - grad->x1 ) * grad->yv - ( y - grad->y1 ) *
			grad->xv );

		break;
	}

	dist -= grad->ofs;

	/* Apply repeat mode */
	len1 = grad->wrep;

	switch ( grad->rmode )
	{
	case MTPIXEL_GRADIENT_BOUND_MIRROR:	/* Mirror repeat */
		l2 = len1 + len1;
		dist -= l2 * (int)( dist * grad->wil2 );

		if ( dist < 0.0 )
		{
			dist += l2;
		}

		if ( dist > len1 )
		{
			dist = l2 - dist;
		}

		break;

	case MTPIXEL_GRADIENT_BOUND_REPEAT:	/* Repeat */
		l2 = len1 + 1.0;	/* Repeat period is 1 pixel longer */
		dist -= l2 * (int)( ( dist + 0.5 ) * grad->wil2 );

		if ( dist < -0.5 )
		{
			dist += l2;
		}

		break;

	case MTPIXEL_GRADIENT_BOUND_STOP:	/* Nothing is outside bounds */
		if (	dist < -0.5 ||
			dist >= len1 + 0.5
			)
		{
			return 0;
		}

	case MTPIXEL_GRADIENT_BOUND_LEVEL:	/* Constant extension */

	default:
		break;
	}

	/* Rescale to 0..1, enforce boundaries */
	dist = dist <= 0.0 ? 0.0 : dist >= len1 ? 1.0 : dist * grad->wil1;

	/* Value from Bayer dither matrix */
	dither = BAYER ( x, y );

	/* Get gradient */
	wrk[ MTPIXEL_CHANNEL_IMAGE + 3 ] = 0;
	slot = channel + ( ( 0x81 + channel + channel - bpp ) >> 7 );
	op = grad_value ( brush, wrk, slot, dist );

	if ( op + dither < 256 )
	{
		return 0;
	}

	if ( channel == MTPIXEL_CHANNEL_IMAGE )
	{
		if ( bpp == 3 )
		{
			dest[0] = ( wrk[0] + dither ) >> 8;
			dest[1] = ( wrk[1] + dither ) >> 8;
			dest[2] = ( wrk[2] + dither ) >> 8;
		}
		else
		{
			dest[ MTPIXEL_CHANNEL_IMAGE + 3 ] = (unsigned char)
				wrk[ ( wrk[ MTPIXEL_CHANNEL_IMAGE + 3 ] +
					dither ) >> 8];
		}
	}
	else
	{
		dest[ channel + 3 ] = ( wrk[ channel + 3 ] + dither ) >> 8;
	}

	return ( ( op + dither ) >> 8 );
}

static void put_pixel (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	int		cc,
			offset,
			o1,
			o2,
			pix,
			r,
			g,
			b;
	unsigned char	cset[ MTPIXEL_CHANNEL_TOTAL + 3 ];


	if (	x < 0			||
		y < 0			||
		x >= image->width	||
		y >= image->height
		)
	{
		return;			// Outside canvas
	}

	o1 = brush->opacity;
	o2 = 255 - brush->opacity;

	for ( cc = 0; cc < MTPIXEL_CHANNEL_TOTAL; cc ++ )
	{
		if ( ! image->channel[ cc ] || brush->chan_dis[ cc ] )
		{
			continue;
		}

		if ( brush->mode == MTPIXEL_BRUSH_MODE_GRADIENT )
		{
			o1 = grad_pixel ( cc, image->type, brush, cset, x, y );

			if ( ! o1 )
			{
				continue;
			}

			o2 = 255 - o1;

			if ( cc == MTPIXEL_CHANNEL_IMAGE )
			{
				offset = image->type * ( x + image->width * y );

				if ( image->type == MTPIXEL_IMAGE_RGB )
				{
					r = o2 * image->channel[cc][offset];
					g = o2 * image->channel[cc][offset + 1];
					b = o2 * image->channel[cc][offset + 2];

					image->channel[cc][offset] = ( r +
						o1 * ( cset[0] ) ) / 255;

					image->channel[cc][offset + 1] = ( g +
						o1 * ( cset[1] ) ) / 255;

					image->channel[cc][offset + 2] = ( b +
						o1 * ( cset[2] ) ) / 255;
				}
				else if ( image->type == MTPIXEL_IMAGE_INDEXED )
				{
					image->channel[ cc ][ offset ] =
						cset[ MTPIXEL_CHANNEL_IMAGE +
							3 ];
				}
			}
			else	// Non-image channel
			{
				offset = x + image->width * y;
				pix = o2 * image->channel[ cc ][ offset ];

				image->channel[ cc ][ offset ] = ( pix + o1 *
					cset[ cc + 3 ] ) / 255;
			}
		}
		else if ( brush->mode == MTPIXEL_BRUSH_MODE_PATTERN )
		{
			if ( cc == MTPIXEL_CHANNEL_IMAGE )
			{
				int		pato;
				unsigned char	* pat;


				offset = image->type * ( x + image->width * y );
				pato = ( ( x & 7 ) + 8 * ( y & 7 ) );

				if ( image->type == MTPIXEL_IMAGE_RGB )
				{
					pat = brush->pattern_rgb + 3 * pato;

					r = o2 * image->channel[cc][offset];
					g = o2 * image->channel[cc][offset + 1];
					b = o2 * image->channel[cc][offset + 2];

					image->channel[ cc ][ offset ] = ( r +
						o1 * ( *pat++ ) ) / 255;

					image->channel[ cc ][ offset + 1 ] =
						( g + o1 * ( *pat++ ) ) / 255;

					image->channel[ cc ][ offset + 2 ] =
						( b + o1 * ( *pat++ ) ) / 255;
				}
				else if ( image->type == MTPIXEL_IMAGE_INDEXED )
				{
					image->channel[ cc ][ offset ] =
						brush->pattern_i[ pato ];
				}
			}
			else	// Non-image channel
			{
				offset = x + image->width * y;
				pix = o2 * image->channel[ cc ][ offset ];

				image->channel[ cc ][ offset ] = ( pix +
					o1 * brush->col_i[ cc ][0] ) / 255;
			}
		}
	}
}

// Draw a filled rectangle
static void f_rectangle (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y,
	int		w,
	int		h
	)
{
	int		i,
			j;


	w += x;
	h += y;

	if ( x < 0 )
	{
		x = 0;
	}

	if ( y < 0 )
	{
		y = 0;
	}

	if ( w > image->width )
	{
		w = image->width;
	}

	if ( h > image->height )
	{
		h = image->height;
	}

	for ( i = y; i < h; i++ )
	{
		for ( j = x; j < w; j++ )
		{
			put_pixel ( image, brush, j, i );
		}
	}
}



/* Indices 0 and 1 are current X and Y, 2 is number of pixels remaining */
typedef int linedata[10];



static void line_init (
	linedata	line,
	int		x0,
	int		y0,
	int		x1,
	int		y1
	)
{
	line[0] = x0;
	line[1] = y0;
	line[6] = line[8] = x1 < x0 ? -1 : 1;
	line[7] = line[9] = y1 < y0 ? -1 : 1;
	line[4] = abs ( x1 - x0 );
	line[5] = abs ( y1 - y0 );

	if ( line[4] < line[5] )	/* More vertical */
	{
		line[2] = line[3] = line[5];
		line[4] *= 2;
		line[5] *= 2;
		line[6] = 0;
	}
	else	/* More horizontal */
	{
		line[2] = line[3] = line[4];
		line[4] = 2 * line[5];
		line[5] = 2 * line[2];
		line[7] = 0;
	}
}

static int line_step (
	linedata	line
	)
{
	line[3] -= line[4];

	if ( line[3] <= 0 )
	{
		line[3] += line[5];
		line[0] += line[8];
		line[1] += line[9];
	}
	else
	{
		line[0] += line[6];
		line[1] += line[7];
	}

	return ( --line[2] );
}

/* Produce a horizontal segment from two connected lines */
static void twoline_segment (
	int		* xx,
	linedata	line1,
	linedata	line2
	)
{
	xx[0] = xx[1] = line1[0];

	while ( 1 )
	{
		if ( ! line1[7] )	/* Segments longer than 1 pixel */
		{
			while ( line1[2] > 0		&&
				line1[3] > line1[4]
				)
			{
				line_step ( line1 );
			}
		}

		if ( xx[0] > line1[0] )
		{
			xx[0] = line1[0];
		}

		if ( xx[1] < line1[0] )
		{
			xx[1] = line1[0];
		}

		if ( line1[2] > 0 || line2[2] < 0 )
		{
			break;
		}

		memcpy ( line1, line2, sizeof ( linedata ) );
		line2[2] = -1;

		if ( xx[0] > line1[0] )
		{
			xx[0] = line1[0];
		}

		if ( xx[1] < line1[0] )
		{
			xx[1] = line1[0];
		}
	}
}

static void sline (
	mtImage		* image,
	mtBrush		* brush,
	int		x1,
	int		y1,
	int		x2,
	int		y2
	)
{
	int		i,
			xdo,
			ydo,
			px,
			py,
			todo;
	float		rat;


	xdo = x2 - x1;
	ydo = y2 - y1;
	mtMAX ( todo, abs ( xdo ), abs ( ydo ) )

	if ( todo == 0 )
	{
		todo = 1;
	}

	for ( i = 0; i <= todo; i++ )
	{
		rat = ( (float) i ) / todo;
		px = mt_round ( x1 + ( x2 - x1 ) * rat );
		py = mt_round ( y1 + ( y2 - y1 ) * rat );
		put_pixel ( image, brush, px, py );
	}
}

/* Draw whatever is bounded by two pairs of lines */
static void draw_quad (
	mtImage		* image,
	mtBrush		* brush,
	linedata	line1,
	linedata	line2,
	linedata	line3,
	linedata	line4
	)
{
	int		i,
			x1,
			x2,
			y1,
			xx[4];


	for ( ; line1[2] >= 0; line_step ( line1 ) , line_step ( line3 ) )
	{
		y1 = line1[1];
		twoline_segment ( xx + 0, line1, line2 );
		twoline_segment ( xx + 2, line3, line4 );

		if ( y1 < 0 || y1 >= image->height )
		{
			continue;
		}

		if ( xx[0] > xx[2] )
		{
			xx[0] = xx[2];
		}

		if ( xx[1] < xx[3] )
		{
			xx[1] = xx[3];
		}

		x1 = xx[0] < 0 ? 0 : xx[0];
		x2 = xx[1] >= image->width ? image->width - 1 : xx[1];

		for ( i = x1; i <= x2; i++ )
		{
			put_pixel ( image, brush, i, y1 );
		}
	}
}

/* Draw general parallelogram */
static void g_para (
	mtImage		* image,
	mtBrush		* brush,
	int		x1,
	int		y1,
	int		x2,
	int		y2,
	int		xv,
	int		yv
	)
{
	linedata	line1,
			line2,
			line3,
			line4;
	int		i,
			j,
			x[2] = { x1, x2 },
			y[2] = { y1, y2 };


	j = ( y1 < y2 ) ^ ( yv < 0 );
	i = j ^ 1;

	line_init ( line1, x[i], y[i], x[i] + xv, y[i] + yv );
	line_init ( line2, x[i] + xv, y[i] + yv, x[j] + xv, y[j] + yv );
	line_init ( line3, x[i], y[i], x[j], y[j] );
	line_init ( line4, x[j], y[j], x[j] + xv, y[j] + yv );

	draw_quad ( image, brush, line1, line2, line3, line4 );
}



static int	circ_r,
		circ_trace[128];

/*
This code uses midpoint ellipse algorithm modified for uncentered ellipses,
with floating-point arithmetics to prevent overflows. (C) Dmitry Groshev
*/

static int	xc2,
		yc2;



static void put4pix (
	mtImage		* image,
	mtBrush		* brush,
	int		dx,
	int		dy
	)
{
	int		x0 = xc2 - dx,
			x1 = xc2 + dx,
			y0 = yc2 - dy,
			y1 = yc2 + dy;


	if ( x1 < 0 || y1 < 0 )
	{
		return;
	}

	if ( x0 < 0 )
	{
		x0 = x1;
	}

	x0 >>= 1;
	x1 >>= 1;

	if ( y0 < 0 )
	{
		y0 = y1;
	}

	y0 >>= 1;
	y1 >>= 1;

	if (	x0 >= image->width ||
		y0 >= image->height
		)
	{
		return;
	}

	put_pixel ( image, brush, x0, y0 );

	if ( x0 != x1 )
	{
		put_pixel ( image, brush, x1, y0 );
	}

	if ( y0 == y1 )
	{
		return;
	}

	put_pixel ( image, brush, x0, y1 );

	if ( x0 != x1 )
	{
		put_pixel ( image, brush, x1, y1 );
	}
}

static void trace_ellipse (
	int		w,
	int		h,
	int		* left,
	int		* right
	)
{
	int		dx,
			dy;
	double		err,
			stx,
			sty,
			w2,
			h2;


	if ( left[0] > w )
	{
		left[0] = w;
	}

	if ( right[0] < w )
	{
		right[0] = w;
	}

	if ( h <= 1 )
	{
		return;			/* Too small */
	}

	h2 = h * h;
	w2 = w * w;
	dx = w & 1;
	dy = h;
	stx = h2 * dx;
	sty = w2 * dy;
	err = h2 * ( dx * 5 + 4 ) + w2 * ( 1 - h - h );

	while ( 1 )		/* Have to force first step */
	{
		if ( left[dy >> 1] > dx )
		{
			left[dy >> 1] = dx;
		}

		if ( right[dy >> 1] < dx )
		{
			right[dy >> 1] = dx;
		}

		if ( err >= 0.0 )
		{
			dy -= 2;
			sty -= w2 + w2;
			err -= 4.0 * sty;
		}

		dx += 2;
		stx += h2 + h2;
		err += 4.0 * (h2 + stx);

		if ( dy < 2 || stx >= sty )
		{
			break;
		}
	}

	err += 3.0 * ( w2 - h2 ) - 2.0 * ( stx + sty );

	while ( dy > 1 )
	{

		if ( left[dy >> 1] > dx )
		{
			left[dy >> 1] = dx;
		}

		if ( right[dy >> 1] < dx )
		{
			right[dy >> 1] = dx;
		}

		if ( err < 0.0 )
		{
			dx += 2;
			stx += h2 + h2;
			err += 4.0 * stx;
		}

		dy -= 2;
		sty -= w2 + w2;
		err += 4.0 * ( w2 - sty );
	}

	/* For too-flat ellipses */
	if ( left[1] > dx )
	{
		left[1] = dx;
	}

	if ( right[1] < w - 2 )
	{
		right[1] = w - 2;
	}
}

static void retrace_circle (
	int		r
	)
{
	int		sz,
			left[128];


	circ_r = r--;
	sz = ( ( r >> 1 ) + 1 ) * sizeof ( int );
	memset ( left, 0, sz );
	memset ( circ_trace, 0, sz );
	trace_ellipse ( r, r, left, circ_trace );
}

// Draw a filled circle
static void f_circle (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y,
	int		r
	)
{
	int		i,
			j,
			x0,
			x1,
			y0,
			y1,
			r1 = r - 1,
			half = r1 & 1;


	/* Prepare & cache circle contour */
	if ( circ_r != r )
	{
		retrace_circle ( r );
	}

	/* Draw result */
	for ( i = half; i <= r1; i += 2 )
	{
		y0 = y - ( ( i + half ) >> 1 );
		y1 = y + ( ( i - half ) >> 1 );

		if ( y0 >= image->height || y1 < 0 )
		{
			continue;
		}

		x0 = x - ( ( circ_trace[i >> 1] + half ) >> 1 );
		x1 = x + ( ( circ_trace[i >> 1] - half ) >> 1 );

		if ( x0 < 0 )
		{
			x0 = 0;
		}

		if ( x1 >= image->width )
		{
			x1 = image->width - 1;
		}

		for ( j = x0; j <= x1; j++ )
		{
			if ( y0 >= 0 )
			{
				put_pixel ( image, brush, j, y0 );
			}

			if ( y1 != y0 && y1 < image->height )
			{
				put_pixel ( image, brush, j, y1 );
			}
		}
	}
}

int mtpixel_brush_line (
	mtImage		* image,
	mtBrush		* brush,
	int		x1,
	int		y1,
	int		x2,
	int		y2
	)
{
	int		xv,
			yv,		// x & y vectors
			xv2,
			yv2;
	float		xuv,
			yuv,
			llen,		// x & y unit vectors, line length
			xv1,
			yv1;		// vector for shadow x & y


	if ( ! image || ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

// FIXME - paint first/last points as normal brush if circle/square else ignore

	if ( brush->size < 2 )
	{
		sline ( image, brush, x1, y1, x2, y2 );
	}
	else
	{
		xv = x2 - x1;
		yv = y2 - y1;
		llen = sqrt ( xv * xv + yv * yv );
		xuv = ( (float) xv ) / llen;
		yuv = ( (float) yv ) / llen;

		xv1 = -yuv * ( (float) brush->size - 0.5 );
		yv1 = xuv * ( (float) brush->size - 0.5 );

		xv2 = rint ( xv1 / 2 + 0.5 * ( ( brush->size + 1 ) % 2 ) );
		yv2 = rint ( yv1 / 2 + 0.5 * ( ( brush->size + 1 ) % 2 ) );

		xv1 = -yuv * ( (float) brush->size - 0.5 );
		yv1 = xuv * ( (float) brush->size - 0.5 );

		g_para ( image, brush, x1 - xv2, y1 - yv2, x2 - xv2, y2 - yv2,
			mt_round ( xv1 ), mt_round ( yv1 ) );
	}

	return 0;
}

int mtpixel_brush_rectangle (
	mtImage		* image,
	mtBrush		* brush,
	int		x1,
	int		y1,
	int		x2,
	int		y2
	)
{
	int		w,
			h,
			x,
			y,
			thick;


	if ( ! image || ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	x = x1 < x2 ? x1 : x2;
	y = y1 < y2 ? y1 : y2;
	w = abs ( x1 - x2 ) + 1;
	h = abs ( y1 - y2 ) + 1;
	thick = brush->size;

	if ( thick < 0 )
	{
		thick = 0;
	}

	if (	thick == 0	||
		2 * thick >= w	||
		2 * thick >= h
		)
	{
		f_rectangle ( image, brush, x, y, w, h );
	}
	else
	{
		f_rectangle ( image, brush, x, y, w, thick );
		f_rectangle ( image, brush, x, y + thick, thick,
			h - 2 * thick );
		f_rectangle ( image, brush, x, y + h - thick, w, thick );
		f_rectangle ( image, brush, x + w - thick, y + thick, thick,
			h - 2 * thick );
	}

	return 0;
}

static void wjellipse (
	mtImage		* image,
	mtBrush		* brush,
	int		xs,
	int		ys,
	int		w,
	int		h,
	int		type,
	int		thick
	)
{
	int		i,
			j,
			k,
			* left,
			* right;


	/* Prepare */
	yc2 = --h + ys + ys;
	xc2 = --w + xs + xs;
	k = type ? w + 1 : w & 1;
	j = h / 2 + 1;
	left = malloc ( 2 * j * sizeof ( int ) );

	if ( ! left )
	{
		return;
	}

	right = left + j;

	for ( i = 0; i < j; i++ )
	{
		left[i] = k;
		right[i] = 0;
	}

	/* Plot outer */
	trace_ellipse ( w, h, left, right );

	/* Plot inner */
	if ( type && ( thick > 1 ) )
	{
		/* Determine possible height */
		thick += thick - 2;

		for ( i = h; i >= 0; i -= 2 )
		{
			if ( left[i >> 1] > thick + 1 )
			{
				break;
			}
		}

		i = i >= h - thick ? h - thick : i + 2;

		/* Determine possible width */
		j = left[thick >> 1];

		if ( j > w - thick )
		{
			j = w - thick;
		}

		if ( j < 2 )
		{
			i = h & 1;
		}

		/* Do the plotting */
		for ( k = i >> 1; k <= h >> 1; k++ )
		{
			left[k] = w & 1;
		}

		if ( i > 1 )
		{
			trace_ellipse ( j, i, left, right );
		}
	}

	/* Draw result */
	for ( i = h & 1; i <= h; i += 2 )
	{
		for ( j = left[i >> 1]; j <= right[i >> 1]; j += 2 )
		{
			put4pix ( image, brush, j, i );
		}
	}

	free ( left );
}

int mtpixel_brush_ellipse (
	mtImage		* image,
	mtBrush		* brush,
	int		x1,
	int		y1,
	int		x2,
	int		y2
	)
{
	int		xs,
			ys,
			xl,
			yl,
			thick;


	if ( ! image || ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	xs = x1 < x2 ? x1 : x2;
	ys = y1 < y2 ? y1 : y2;
	xl = abs ( x2 - x1 ) + 1;
	yl = abs ( y2 - y1 ) + 1;
	thick = brush->size;

	if ( thick < 0 )
	{
		thick = 0;
	}

	/* Draw rectangle instead if too small */
	if ( xl <= 2 || yl <= 2 )
	{
		f_rectangle ( image, brush, xs, ys, xl, yl );
	}
	else
	{
		wjellipse ( image, brush, xs, ys, xl, yl,
			thick && (thick * 2 < xl) && (thick * 2 < yl), thick );
	}

	return 0;
}

static int cmp_floats (
	const void	* a,
	const void	* b
	)
{
	float		* fa = (float *)a,
			* fb = (float *)b;


	if ( fa[0] < fb[0] )
	{
		return -1;
	}

	if ( fa[0] > fb[0] )
	{
		return 1;
	}

	return 0;
}

static void poly_draw (
	mtImage		* image,
	mtBrush		* brush,
	mtPolygon	* polygon,
	int		type		// 0 = mask, 1 = indexed, 3 = RGB
					// +10 = outline
	)
{
	int		i,
			i2,
			i3,
			j,
			cuts,
			maxx = image->width,
			maxy = image->height,
			poly_lines[ MTPIXEL_POLYGON_MAX ][2][2],
			pminy,
			pmaxy;
	float		ratio,
			poly_cuts[ MTPIXEL_POLYGON_MAX ];


	// Populate poly_lines - smallest Y is first point
	for ( i = 0; i < polygon->size; i++ )
	{
		i2 = i + 1;

		if ( i2 >= polygon->size )
		{
			i2 = 0;		// Remember last point back to 1st point
		}

		if ( polygon->point[i][1] < polygon->point[i2][1] )
		{
			j = 0;
		}
		else
		{
			j = 1;
		}

		poly_lines[i][j][0] = polygon->point[i][0];
		poly_lines[i][j][1] = polygon->point[i][1];
		poly_lines[i][1-j][0] = polygon->point[i2][0];
		poly_lines[i][1-j][1] = polygon->point[i2][1];

		if ( type > 10 )
		{
			f_circle ( image, brush, polygon->point[i][0],
				polygon->point[i][1], brush->size );
			mtpixel_brush_line ( image, brush, polygon->point[i][0],
				polygon->point[i][1], polygon->point[i2][0],
				polygon->point[i2][1] );
		}

		// WARNING - This polygon code botches the edges when
		// opacity < 255 on RGB images

		if ( type == 1 || type == 3 )
		{
			sline ( image, brush, polygon->point[i][0],
				polygon->point[i][1], polygon->point[i2][0],
				polygon->point[i2][1] );
		}
	}

	if ( type > 10 )
	{
		return;			// If drawing outline only, finish now
	}

	pminy = polygon->min_y;
	pmaxy = polygon->max_y;

	if ( pminy < 0 )
	{
		pminy = 0;		// Vertical clipping
	}

	if ( pmaxy >= maxy )
	{
		pmaxy = maxy - 1;
	}

	for ( j = pminy; j <= pmaxy; j++ )	// Scanline
	{
		cuts = 0;

		// Count up line intersections - X value cuts
		for ( i = 0; i < polygon->size; i++ )
		{
			if (	j >= poly_lines[i][0][1] &&
				j <= poly_lines[i][1][1]
				)
			{
				if ( poly_lines[i][0][1] == poly_lines[i][1][1]
					)
				{
					// Line is horizontal so use each end
					// point as a cut

					poly_cuts[ cuts ++ ] =
						poly_lines[i][0][0];

					poly_cuts[ cuts ++ ] =
						poly_lines[i][1][0];
				}
				else
				{
					// Calculate cut X value - intersection
					// on y = j

					ratio = (float)( j - poly_lines[i][0][1]
						) / ( poly_lines[i][1][1] -
							poly_lines[i][0][1] );

					poly_cuts[ cuts ++ ] =
						poly_lines[i][0][0] + ratio *
						( poly_lines[i][1][0] -
							poly_lines[i][0][0] );

					if ( j == poly_lines[i][0][1] )
					{
						cuts --;
						// Don't count start point
					}
				}
			}
		}

		if ( cuts > 1 )
		{
			qsort ( poly_cuts, cuts, sizeof ( poly_cuts[0] ),
				cmp_floats );
		}

		// Paint from first X to 2nd, gap from 2-3, paint 3-4,
		// gap 4-5 ...

		for ( i = 0; i < (cuts - 1); i += 2 )
		{
			if ( poly_cuts[i] < maxx && poly_cuts[i + 1] >= 0 )
			{
				if ( poly_cuts[i] < 0 )
				{
					poly_cuts[i] = 0;
				}

				if ( poly_cuts[i] >= maxx )
				{
					poly_cuts[i] = maxx - 1;
				}

				if ( poly_cuts[i + 1] < 0 )
				{
					// Horizontal Clipping
					poly_cuts[i + 1] = 0;
				}

				if ( poly_cuts[i + 1] >= maxx )
				{
					poly_cuts[i + 1] = maxx - 1;
				}

				i3 = mt_round ( poly_cuts[i + 1] );

				for (	i2 = mt_round ( poly_cuts[i] );
					i2 <= i3;
					i2++
					)
				{
					put_pixel ( image, brush, i2, j );
				}
			}
		}
	}
}

int mtpixel_brush_polygon (
	mtImage		* image,
	mtBrush		* brush,
	mtPolygon	* polygon
	)
{
	if (	! image		||
		! brush		||
		! polygon	||
		mtpixel_poly_init ( polygon )
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( brush->size >= 1 )
	{
		poly_draw ( image, brush, polygon, image->type + 10 );
	}
	else
	{
		poly_draw ( image, brush, polygon, image->type );
	}

	return 0;
}

static int get_pixel (
	mtImage		* image,
	int		x,
	int		y
	)
{
	if ( image->type == 1 )
	{
		return image->channel[ MTPIXEL_CHANNEL_IMAGE ][ x +
			y * image->width ];
	}

	if ( image->type == 3 )
	{
		return MTPIXEL_RGB_2_INT (
			image->channel[ MTPIXEL_CHANNEL_IMAGE ][
				3 * ( x + y * image->width ) ],
			image->channel[ MTPIXEL_CHANNEL_IMAGE ][
				3 * ( x + y * image->width ) + 1 ],
			image->channel[ MTPIXEL_CHANNEL_IMAGE ][
				3 * ( x + y * image->width ) + 2 ]
				);
	}

	return 0;
}

/*
This flood fill algorithm processes image in quadtree order, and thus has
guaranteed upper bound on memory consumption, of order O ( width + height ).
(C) Dmitry Groshev
*/

#define QLEVELS 11
#define QMINSIZE 32
#define QMINLEVEL 5



static void wjfloodfill (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y,
	unsigned char	* bmap,
	int		lw
	)
{
	short		* nearq,
			* farq;
	int		col = 0,
			qtail[ QLEVELS + 1 ],
			ntail = 0,
			borders[4] = { 0, image->width, 0, image->height },
			corners[4],
			levels[4],
			coords[4],
			i,
			j,
			k,
			kk,
			lvl,
			tx,
			ty,
			bidx = 0,
			bbit = 0;
	char		* tmp = NULL;


	col = get_pixel ( image, x, y );

	i = ( ( image->width + image->height ) * 3 + QMINSIZE * QMINSIZE ) *
		2 * sizeof ( short );

	// Exact limit is less, but it's too complicated
	nearq = malloc ( i );

	if ( ! nearq )
	{
		return;
	}

	farq = nearq + QMINSIZE * QMINSIZE;
	memset ( qtail, 0, sizeof ( qtail ) );

	/* Start drawing */
	if ( bmap )
	{
		bmap[ y * lw + ( x >> 3 ) ] |= 1 << (x & 7);
	}
	else
	{
		put_pixel ( image, brush, x, y );

		if ( get_pixel ( image, x, y ) == col )
		{
			/* Can't draw */
			free ( nearq );

			return;
		}
	}

	while ( 1 )
	{
		/* Determine area */
		corners[0] = x & ~( QMINSIZE - 1 );
		corners[1] = corners[0] + QMINSIZE;
		corners[2] = y & ~( QMINSIZE - 1 );
		corners[3] = corners[2] + QMINSIZE;

		/* Determine queue levels */
		for ( i = 0; i < 4; i++ )
		{
			j = ( corners[i] & ~( corners[i] - 1 ) ) - 1;
			j = ( j & 0x5555) + ( ( j & 0xAAAA ) >> 1 );
			j = ( j & 0x3333) + ( ( j & 0xCCCC ) >> 2 );
			j = ( j & 0x0F0F) + ( ( j & 0xF0F0 ) >> 4 );
			levels[i] = ( j & 0xFF ) + ( j >> 8 ) - QMINLEVEL;
		}

		/* Process near points */
		while ( 1 )
		{
			coords[0] = x;
			coords[2] = y;

			for ( i = 0; i < 4; i++ )
			{
				coords[1] = x;
				coords[3] = y;
				coords[ ( i & 2 ) + 1] += ( ( i + i ) & 2 ) - 1;

				/* Is pixel valid? */
				if ( coords[i] == borders[i] )
				{
					continue;
				}

				tx = coords[1];
				ty = coords[3];

				if ( bmap )
				{
					bidx = ty * lw + ( tx >> 3 );
					bbit = 1 << ( tx & 7 );

					if ( bmap[ bidx ] & bbit )
					{
						continue;
					}
				}

				if ( get_pixel ( image, tx, ty ) != col )
				{
					continue;
				}

				/* Is pixel writable? */
				if ( bmap )
				{
					bmap[bidx] |= bbit;
				}
				else
				{
					put_pixel ( image, brush, tx, ty );

					if ( get_pixel ( image, tx, ty) == col )
					{
						continue;
					}
				}

				/* Near queue */
				if ( coords[i] != corners[i] )
				{
					nearq[ ntail ++ ] = tx;
					nearq[ ntail ++ ] = ty;

					continue;
				}

				/* Far queue */
				lvl = levels[i];

				for ( j = 0; j < lvl; j++ )
				{
					// Slide lower levels
					k = qtail[j];
					qtail[j] = k + 2;

					if ( k > qtail[j + 1] )
					{
						kk = qtail[j + 1];
						farq[k] = farq[kk];
						farq[k + 1] = farq[kk + 1];
					}
				}

				k = qtail[lvl];
				farq[k] = tx;
				farq[k + 1] = ty;
				qtail[lvl] = k + 2;
			}

			if ( ! ntail )
			{
				break;
			}

			y = nearq[--ntail];
			x = nearq[--ntail];
		}

		/* All done? */
		if ( ! qtail[0] )
		{
			break;
		}

		i = qtail[0] - 2;
		x = farq[i];
		y = farq[i + 1];
		qtail[0] = i;

		for ( j = 1; qtail[j] > i; j++ )
		{
			qtail[j] = i;
		}
	}

	free ( nearq );
	free ( tmp );
}

int mtpixel_brush_flood_fill (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	unsigned char	* pat,
			* temp;
	int		i,
			j,
			k,
			lw;


	if (	! image						||
		! image->channel[ MTPIXEL_CHANNEL_IMAGE ]	||
		! brush						||
		x < 0						||
		x >= image->width				||
		y >= image->height
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	/* Regular fill? */
	if ( brush->pattern == 0 && brush->opacity == 255 )
	{
		wjfloodfill ( image, brush, x, y, NULL, 0 );

		return 0;
	}

	lw = ( image->width + 7 ) >> 3;
	j = lw * image->height;
	pat = temp = calloc ( j, 1 );

	if ( ! pat )
	{
		return MTPIXEL_ERRNO_MALLOC;
	}

	wjfloodfill ( image, brush, x, y, pat, lw );

	for ( i = 0; i < image->height; i++ )
	{
		for ( j = 0; j < image->width; )
		{
			k = *temp++;

			if ( ! k )
			{
				j += 8;

				continue;
			}

			for ( ; k; k >>= 1 )
			{
				if ( k & 1 )
				{
					put_pixel ( image, brush, j, i );
				}

				j++;
			}

			j = ( j + 7 ) & ~( 7 );
		}
	}

	free ( pat );

	return 0;
}

int mtpixel_brush_plot (
	mtImage		* image,
	mtBrush		* brush,
	int		x,
	int		y
	)
{
	int		j,
			ts2,
			tr2,
			rx,
			ry;


	if ( ! image || ! brush )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	ts2 = brush->size >> 1;
	tr2 = brush->size - ts2 - 1;

	switch ( brush->shape )
	{
	case MTPIXEL_BRUSH_SHAPE_SQUARE:
		f_rectangle ( image, brush, x - ts2, y - ts2, brush->size,
			brush->size );
		break;

	case MTPIXEL_BRUSH_SHAPE_CIRCLE:
		f_circle ( image, brush, x, y, brush->size );
		break;

	case MTPIXEL_BRUSH_SHAPE_HORIZONTAL:
		sline ( image, brush, x - ts2, y, x + tr2, y );
		break;

	case MTPIXEL_BRUSH_SHAPE_VERTICAL:
		sline ( image, brush, x, y - ts2, x, y + tr2 );
		break;

	case MTPIXEL_BRUSH_SHAPE_SLASH:
		sline ( image, brush, x + tr2, y - ts2, x - ts2, y + tr2 );
		break;

	case MTPIXEL_BRUSH_SHAPE_BACKSLASH:
		sline ( image, brush, x - ts2, y - ts2, x + tr2, y + tr2 );
		break;

	case MTPIXEL_BRUSH_SHAPE_SPRAY:
		for ( j = 0; j < brush->flow; j++ )
		{
			rx = x - ts2 + rand () % brush->size;
			ry = y - ts2 + rand () % brush->size;

			if (	rx >= 0			&&
				ry >= 0			&&
				rx < image->width	&&
				ry < image->height
				)
			{
				put_pixel ( image, brush, rx, ry );
			}
		}
		break;

	default:
		break;
	}

	return 0;
}

int mtpixel_palette_load (
	mtPalette	* palette,
	char		* filename
	)
{
	int		rgb[3],
			i;
	FILE		* fp;
	char		input[128];


	if ( ! palette || ! filename )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	fp = fopen ( filename, "r" );
	if ( fp == NULL )
	{
		return MTPIXEL_ERRNO_FILE_OPEN;
	}

	if ( ! fgets ( input, 30, fp ) )
	{
		goto error;
	}

	if ( strncmp ( input, "GIMP Palette", 12 ) == 0 )
	{
		i = 0;

		while ( fgets ( input, 120, fp ) && ( i < 256 ) )
		{
			// Continue to read until EOF or new_mem_cols > 255
			// If line starts with a number or space assume its a
			// palette entry

			if (	input[0] == ' ' ||
				( input[0] >= '0' && input[0] <= '9')
				)
			{
				sscanf ( input, "%i %i %i", &rgb[0], &rgb[1],
					&rgb[2] );

				palette->col[i][0] = rgb[0];
				palette->col[i][1] = rgb[1];
				palette->col[i][2] = rgb[2];

				i++;
			}
		}

		palette->size = i;
	}
	else
	{
		goto error;
	}

	fclose ( fp );

	return 0;

error:
	fclose ( fp );

	return MTPIXEL_ERRNO_FILE_LOAD;
}

int mtpixel_palette_save (
	mtPalette	* palette,
	char		* filename
	)
{
	FILE		* fp;
	int		i;


	if ( ! palette || ! filename )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	fp = fopen ( filename, "w" );
	if ( fp == NULL )
	{
		return MTPIXEL_ERRNO_FILE_OPEN;
	}

	fprintf ( fp, "GIMP Palette\nName: mtPaint\nColumns: 16\n#\n" );

	for ( i = 0; i < palette->size; i++)
	{
		fprintf ( fp, "%3i %3i %3i\tUntitled\n",
			palette->col[i][0],
			palette->col[i][1],
			palette->col[i][2]
			);
	}

	fclose ( fp );

	return 0;
}

int mtpixel_palette_copy (
	mtPalette	* dest,
	mtPalette	* src
	)
{
	if ( ! dest || ! src )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	memcpy ( dest, src, sizeof ( mtPalette ) );

	return 0;
}

int mtpixel_palette_invert (
	mtPalette	* palette
	)
{
	int		i;


	if ( ! palette )
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	for ( i = 0; i < 256; i++ )
	{
		palette->col[i][0] = ~palette->col[i][0];
		palette->col[i][1] = ~palette->col[i][1];
		palette->col[i][2] = ~palette->col[i][2];
	}

	return 0;
}

// Calculate how many of each colour index is on the canvas
static void mem_get_histogram (
	mtImage		* image,
	int		mem_histogram[256]
	)
{
	int		i,
			j;
	unsigned char	* img;


	memset ( mem_histogram, 0, 256 * sizeof ( int ) );

	j = image->width * image->height;
	img = image->channel[ MTPIXEL_CHANNEL_IMAGE ];

	for ( i = 0; i < j; i++ )
	{
		mem_histogram[ *img++ ] ++;
	}
}

/* Nonclassical HSV: H is 0..6, S is 0..1, V is 0..255 */
void mtpixel_rgb2hsv (
	unsigned char	* rgb,
	double		* hsv
	)
{
	int		c0,
			c1,
			c2;


	if ( ! ( ( rgb[0] ^ rgb[1] ) | ( rgb[0] ^ rgb[2] ) ) )
	{
		hsv[0] = hsv[1] = 0.0;
		hsv[2] = rgb[0];

		return;
	}

	c2 = rgb[2] < rgb[0] ? 1 : 0;

	if ( rgb[c2] >= rgb[c2 + 1] )
	{
		c2++;
	}

	c0 = MOD3 ( c2 + 1 );
	c1 = ( c2 + c0 ) ^ 3;
	hsv[2] = rgb[c0] > rgb[c1] ? rgb[c0] : rgb[c1];
	hsv[1] = hsv[2] - rgb[c2];
	hsv[0] = c0 * 2 + 1 + ( rgb[c1] - rgb[c0] ) / hsv[1];
	hsv[1] /= hsv[2];
}

void mtpixel_hsv2rgb (
	unsigned char	* rgb,
	double		* hsv
	)
{
	double		h0,
			h1,
			h2;
	int		i;


	h2 = hsv[2] * 2;
	h1 = h2 * ( 1.0 - hsv[1] );
	i = hsv[0];
	h0 = ( hsv[0] - i ) * ( h2 - h1 );

	if ( i & 1 )
	{
		h2 -= h0;
		h0 += h2;
	}
	else
	{
		h0 += h1;
	}

	i >>= 1;
	rgb[i] = ( (int)h2 + 1 ) >> 1;
	rgb[ MOD3 ( i + 1 ) ] = ( (int)h0 + 1 ) >> 1;
	rgb[ MOD3 ( i + 2 ) ] = ( (int)h1 + 1 ) >> 1;
}

static double rgb_hsl (
	int		t,
	unsigned char	rgb[3]
	)
{
	double		hsv[3];


	if ( t == 2 )
	{
		return ( 0.299 * rgb[0] + 0.587 * rgb[1] + 0.114 * rgb[2] );
	}

	mtpixel_rgb2hsv ( rgb, hsv );

	return hsv[t];
}

int mtpixel_palette_sort (
	mtPalette	* palette,
	mtImage		* image,
	int		start,
	int		finish,
	mtPaletteSort	type,
	int		reverse
	)
{
	int		tab0[256],
			tab1[256],
			tmp,
			i,
			j,
			mem_histogram[256];
	mtPalette	old_pal;
	unsigned char	* img;


	if (	! palette ||
		start < 0 ||
		finish < 0 ||
		start >= 256 ||
		finish >= 256 ||
		( ! image && type == MTPIXEL_PALETTE_SORT_FREQUENCY ) ||
		finish == start ||
		start > palette->size ||
		finish > palette->size ||
		type <= MTPIXEL_PALETTE_SORT_NONE ||
		type >= MTPIXEL_PALETTE_SORT_TOTAL ||
		( image && ! image->channel[ MTPIXEL_CHANNEL_IMAGE ] )
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( finish < start )
	{
		i = start;
		start = finish;
		finish = i;
	}

	if ( type == MTPIXEL_PALETTE_SORT_FREQUENCY )
	{
		mem_get_histogram ( image, mem_histogram );
	}

	for ( i = 0; i < 256; i++ )
	{
		tab0[i] = i;
	}

	for ( i = start; i <= finish; i++ )
	{
		switch ( type )
		{
		case MTPIXEL_PALETTE_SORT_HUE:
			tab1[i] = rint ( 1000 * rgb_hsl (0, palette->col[i] ) );
			break;

		case MTPIXEL_PALETTE_SORT_SATURATION:
			tab1[i] = rint ( 1000 * rgb_hsl (1, palette->col[i] ) );
			break;

		case MTPIXEL_PALETTE_SORT_VALUE:
			tab1[i] = rint ( 1000 * rgb_hsl (2, palette->col[i] ) );
			break;

		case MTPIXEL_PALETTE_SORT_LUMINANCE:
			tab1[i] = _ungamma256 ( _rgb2B (
					_gamma256 ( palette->col[i][0] ),
					_gamma256 ( palette->col[i][1] ),
					_gamma256 ( palette->col[i][2] )
					) );
			break;

		case MTPIXEL_PALETTE_SORT_RED:
			tab1[i] = palette->col[i][0];
			break;

		case MTPIXEL_PALETTE_SORT_GREEN:
			tab1[i] = palette->col[i][1];
			break;

		case MTPIXEL_PALETTE_SORT_BLUE:
			tab1[i] = palette->col[i][2];
			break;

		case MTPIXEL_PALETTE_SORT_FREQUENCY:
			tab1[i] = mem_histogram[i];
			break;

		default:
			break;
		}
	}

	reverse = reverse ? 1 : 0;

	for ( j = finish; j > start; j-- )	// The venerable bubble sort
	{
		for ( i = start; i < j; i++ )
		{
			if ( tab1[i + 1 - reverse] < tab1[i + reverse] )
			{
				tmp = tab0[i];
				tab0[i] = tab0[i + 1];
				tab0[i + 1] = tmp;

				tmp = tab1[i];
				tab1[i] = tab1[i + 1];
				tab1[i + 1] = tmp;
			}
		}
	}

	mtpixel_palette_copy ( &old_pal, palette );

	for ( i = start; i <= finish; i++ )
	{
		palette->col[i][0] = old_pal.col[tab0[i]][0];
		palette->col[i][1] = old_pal.col[tab0[i]][1];
		palette->col[i][2] = old_pal.col[tab0[i]][2];
	}

	if ( image && image->type == MTPIXEL_IMAGE_INDEXED )
	{
		// Adjust canvas pixels if in indexed palette mode
		for ( i = 0; i < 256; i++ )
		{
			tab1[tab0[i]] = i;
		}

		img = image->channel[ MTPIXEL_CHANNEL_IMAGE ];
		j = image->width * image->height;

		for ( i = 0; i < j; i++ )
		{
			img[0] = tab1[ img[0] ];
			img ++;
		}
	}

	return 0;
}

static int mem_remove_unused_check (
	mtImage		* image,
	int		mem_histogram[256]
	)
{
	int		i,
			found = 0;


	mem_get_histogram ( image, mem_histogram );

	for ( i = 0; i < image->palette.size; i++ )
	{
		if ( ! mem_histogram[i] )
		{
			found ++;
		}
	}

	if ( ! found )
	{
		// All palette colours are used on the canvas
		return 0;
	}

	if ( image->palette.size - found < 2 )
	{
		// Canvas is all one colour
		return -1;
	}

	return found;
}

int mtpixel_palette_remove_unused (
	mtImage		* image,
	int		* removed
	)
{
	unsigned char	conv[256],
			* img;
	int		i,
			j,
			found,
			mem_histogram[256];


	if (	! image ||
		image->type != MTPIXEL_IMAGE_INDEXED ||
		! image->channel[ MTPIXEL_CHANNEL_IMAGE ]
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	found = mem_remove_unused_check ( image, mem_histogram );
	if ( found <= 0 )
	{
		goto done;
	}

	for ( i = j = 0; i < 256; i++ )	// Create conversion table
	{
		if ( mem_histogram[i] )
		{
			image->palette.col[j][0] = image->palette.col[i][0];
			image->palette.col[j][1] = image->palette.col[i][1];
			image->palette.col[j][2] = image->palette.col[i][2];
			conv[i] = j++;
		}
	}

	// Re-adjust transparent colour index if it exists
	if ( image->palette.trans >= 0 )
	{
		if ( mem_histogram[ image->palette.trans ] == 0 )
		{
			// No transparent pixels exist so remove reference
			image->palette.trans = -1;
		}
		else
		{
			// New transparency colour position
			image->palette.trans = conv[image->palette.trans];
		}
	}

	j = image->width * image->height;
	img = image->channel[ MTPIXEL_CHANNEL_IMAGE ];

	for ( i = 0; i < j; i++ )	// Convert canvas pixels as required
	{
		img[0] = conv[ img[0] ];
		img ++;
	}

	image->palette.size -= found;

done:
	if ( removed )
	{
		removed[0] = found;
	}

	return 0;
}

int mtpixel_palette_shift (
	mtPalette	* palette,
	int		start,
	int		finish,
	int		shift
	)
{
	mtPalette	old_pal;
	int		dir,
			minab,
			maxab,
			i,
			j;


	if (	! palette	||
		start == finish	||
		start < 0	||
		finish < 1	||
		start > 254	||
		finish > 255
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	if ( start > finish )
	{
		dir = -1;
	}
	else
	{
		dir = 1;
	}

	mtpixel_palette_copy ( &old_pal, palette );

	mtMIN ( minab, start, finish )
	mtMAX ( maxab, start, finish )

	shift = shift % ( maxab-minab + 1 );
	j = minab + dir * shift;

	while ( j > maxab )
	{
		j = j - ( maxab-minab + 1 );
	}

	while ( j < minab )
	{
		j = j + ( maxab-minab + 1 );
	}

	for ( i = minab; i <= maxab; i++ )
	{
		palette->col[i][0] = old_pal.col[j][0];
		palette->col[i][1] = old_pal.col[j][1];
		palette->col[i][2] = old_pal.col[j][2];
		j++;

		if ( j > maxab )
		{
			j = minab;
		}
	}

	return 0;
}

int mtpixel_palette_move (
	mtPalette	* palette,
	mtImage		* image,
	int		c1,
	int		c2
	)
{
	unsigned char	* img = NULL,
			rgb[3];
	int		i,
			j;


	if (	! palette	||
		c1 == c2	||
		c1 < 0		||
		c2 < 0		||
		c1 >= 256	||
		c2 >= 256
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	rgb[0] = palette->col[c1][0];
	rgb[1] = palette->col[c1][1];
	rgb[2] = palette->col[c1][2];

	j = c1 < c2 ? 1 : -1;

	for ( i = c1; i != c2; i += j )
	{
		palette->col[i][0] = palette->col[i + j][0];
		palette->col[i][1] = palette->col[i + j][1];
		palette->col[i][2] = palette->col[i + j][2];
	}

	palette->col[c2][0] = rgb[0];
	palette->col[c2][1] = rgb[1];
	palette->col[c2][2] = rgb[2];

	if ( image )
	{
		img = image->channel[ MTPIXEL_CHANNEL_IMAGE ];
	}

	if ( img && image->type == MTPIXEL_IMAGE_INDEXED )
	{
		unsigned char	table[256];


		for ( i = 0; i < 256; i++ )
		{
			table[i] = i + ( i > c2 ) - ( i > c1 );
		}

		table[c1] = c2;
		table[c2] += (c1 > c2);

		j = image->width * image->height;

		// Change pixel index to new palette
		for ( i = 0; i < j; i++ )
		{
			img[0] = table[ img[0] ];
			img ++;
		}
	}

	return 0;
}

int mtpixel_image_count_colors (
	mtImage		* image
	)
{
	if ( ! image || ! image->channel[ MTPIXEL_CHANNEL_IMAGE ] )
	{
		return 0;
	}

	if ( image->type == MTPIXEL_IMAGE_INDEXED )
	{
		int		i,
				count,
				mem_histogram[256];


		mem_get_histogram ( image, mem_histogram );
		count = 0;

		for ( i = 0; i < 256; i++ )
		{
			if ( mem_histogram[i] )
			{
				count ++;
			}
		}

		return count;
	}

	if ( image->type == MTPIXEL_IMAGE_RGB )
	{
		uint32_t	* tab,
				v;
		int		i,
				j,
				k,
				ix,
				w,
				h;
		unsigned char	* im;


		w = image->width;
		h = image->height;
		im = image->channel[ MTPIXEL_CHANNEL_IMAGE ];
		j = 0x80000;

		// HUGE colour cube
		tab = calloc ( j, sizeof ( uint32_t ) );

		if ( ! tab )
		{
			// Not enough memory Mr Greedy ;-)
			return 0;
		}

		k = w * h;
		for ( i = 0; i < k; i++ )	// Scan each pixel
		{
			ix = ( im[0] >> 5 ) + ( im[1] << 3 ) + ( im[2] << 11 );
			tab[ix] |= 1 << ( im[0] & 31 );
			im += 3;
		}

		for ( i = k = 0; i < j; i++ )	// Count each colour
		{
			v = tab[i];
			v = ( v & 0x55555555 ) + ( ( v >> 1 ) & 0x55555555 );
			v = ( v & 0x33333333 ) + ( ( v >> 2 ) & 0x33333333 );
			v = ( v & 0x0F0F0F0F ) + ( ( v >> 4 ) & 0x0F0F0F0F );
			v = ( v & 0x00FF00FF ) + ( ( v >> 8 ) & 0x00FF00FF );
			k += ( v & 0xFFFF ) + ( v >> 16 );
		}

		free ( tab );

		return k;
	}

	// Unknown image type
	return 0;
}

int mtpixel_palette_scale (
	mtPalette	* palette,
	int		i1,
	int		r1,
	int		g1,
	int		b1,
	int		i2,
	int		r2,
	int		g2,
	int		b2
	)
{
	double		r0,
			g0,
			b0,
			dr,
			dg,
			db,
			d = i2 - i1;
	int		i,
			step = i2 > i1 ? 1 : -1;


	if (	! palette	||
		i1 < 0		||
		i1 > 255	||
		i2 < 0		||
		i2 > 255	||
		i1 == i2
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	dr = ( r2 - r1 ) / d;
	r0 = r1 - dr * i1;
	dg = ( g2 - g1 ) / d;
	g0 = g1 - dg * i1;
	db = ( b2 - b1 ) / d;
	b0 = b1 - db * i1;

	for ( i = i1; i != i2 + step; i += step )
	{
		palette->col[i][0] = rint ( r0 + dr * i );
		palette->col[i][1] = rint ( g0 + dg * i );
		palette->col[i][2] = rint ( b0 + db * i );
	}

	return 0;
}

int mtpixel_channel_new (
	mtImage		* image,
	mtChannelType	channel
	)
{
	int		bpp = 1;


	if (	! image		||
		channel < 0	||
		channel >= MTPIXEL_CHANNEL_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	mtpixel_channel_destroy ( image, channel );	// Remove old channel

	if (	channel == MTPIXEL_CHANNEL_IMAGE &&
		image->type == MTPIXEL_IMAGE_RGB
		)
	{
		bpp = 3;
	}

	image->channel[ channel ] = calloc ( image->width * image->height * bpp,
		1 );

	if ( ! image->channel[ channel ] )
	{
		return MTPIXEL_ERRNO_MALLOC;
	}

	return 0;
}

static int channel_movecopy (
	mtImage		* image,
	mtChannelType	dest,
	mtChannelType	src,
	int		movecopy
	)
{
	if (	! image		||
		dest == src	||
		dest < 0	||
		src < 0		||
		src >= MTPIXEL_CHANNEL_TOTAL ||
		dest >= MTPIXEL_CHANNEL_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	// Destroy current channel
	mtpixel_channel_destroy ( image, dest );

	if ( movecopy == 0 )		// COPY
	{
		// Only duplicate if src contains something
		if ( image->channel[ src ] )
		{
			if ( dest == MTPIXEL_CHANNEL_IMAGE )
			{
				image->type = MTPIXEL_IMAGE_INDEXED;
			}

			image->channel[ dest ] =
				malloc ( image->width * image->height );

			if ( image->channel[ dest ] )
			{
				memcpy ( image->channel[ dest ],
					image->channel[ src ],
					image->width * image->height );
			}
		}
	}
	else	// MOVE
	{
		image->channel[ dest ] = image->channel[ src ];
		image->channel[ src ] = NULL;
	}

	return 0;
}

int mtpixel_channel_copy (
	mtImage		* image,
	mtChannelType	dest,
	mtChannelType	src
	)
{
	return channel_movecopy ( image, dest, src, 0 );
}

int mtpixel_channel_move (
	mtImage		* image,
	mtChannelType	dest,
	mtChannelType	src
	)
{
	return channel_movecopy ( image, dest, src, 1 );
}

int mtpixel_channel_destroy (
	mtImage		* image,
	mtChannelType	channel
	)
{
	if (	! image		||
		channel < 0	||
		channel >= MTPIXEL_CHANNEL_TOTAL
		)
	{
		return MTPIXEL_ERRNO_ARGUMENT;
	}

	free ( image->channel[ channel ] );
	image->channel[ channel ] = NULL;

	return 0;
}

mtChannelMask mtpixel_channel_get_mask (
	mtImage		* image
	)
{
	int		i;
	mtChannelMask	m;


	if ( ! image )
	{
		return 0;
	}

	m = MTPIXEL_CHANNEL_MASK_NONE;

	for ( i = 0; i < MTPIXEL_CHANNEL_TOTAL; i++ )
	{
		if ( image->channel[i] )
		{
			m |= ( 1 << i );
		}
	}

	return m;
}

